//-----------------------------------------------------------------------------
//
// A library which sits over Steinberg VST SDK and provides cleaner
// programming interface and some additional functionality.
//
//-----------------------------------------------------------------------------

/*
	VT Framework License (BSL 1.0)

	Copyright (C) 2007 Sergey A. Galin 
	E-mail: sergey.galin@gmail.com

	Permission is hereby granted, free of charge, to any person or organization
	obtaining a copy of the software and accompanying documentation covered by
	this license (the "Software") to use, reproduce, display, distribute,
	execute, and transmit the Software, and to prepare derivative works of the
	Software, and to permit third-parties to whom the Software is furnished to
	do so, all subject to the following:

	The copyright notices in the Software and this entire statement, including
	the above license grant, this restriction and the following disclaimer,
	must be included in all copies of the Software, in whole or in part, and
	all derivative works of the Software, unless such copies or derivative
	works are solely in the form of machine-executable object code generated by
	a source language processor.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
	SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
	FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
	ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
	DEALINGS IN THE SOFTWARE.
*/

#include <math.h>
#include <ctype.h>
#include <stdio.h>
#include <string.h>

#include "vt_vstaddons.hpp"

using namespace VT;

float VTTestValue=0.0f;


CVTAudioEffectX::CVTAudioEffectX (audioMasterCallback audioMaster, 
								  VstInt32 numPrograms, VstInt32 numParams,
								  bool hasDoublePrecision ):
	AudioEffectX( audioMaster, numPrograms, numParams ),
	vtParamCount(0),
	vtMaxParams(numParams),
	vtEditorOpenCount(0),
	bReloading(false)
{
	if( vtMaxParams ){
		vtParams = new Params::CParameter* [ vtMaxParams ];
		vtParamShouldBeDeleted = new bool[ vtMaxParams ];
		for( VstInt32 i=0; i<vtMaxParams; i++ )
				vtParamShouldBeDeleted[i] = true;	
	}else{
		vtParams = NULL;
		vtParamShouldBeDeleted = NULL;
	}

	setProgramName("Default");
	canProcessReplacing ();
	
	if( hasDoublePrecision )
		canDoubleReplacing ();
	
	#if defined(VT_USE_CHUNKS_SAVING)
		programsAreChunks( true );
	#endif
}


CVTAudioEffectX::~CVTAudioEffectX ()
{
	if( vtParams!=NULL ){
		for(int i=0; i<vtParamCount; i++)
			if( vtParamShouldBeDeleted[i] )
				delete vtParams[i];
		delete vtParams;
	}
	delete vtParamShouldBeDeleted;
}

int CVTAudioEffectX::VTAddParam(Params::CParameter *newparam, bool deleteIt)
{
	if( vtParams && vtParamCount<vtMaxParams ){
		vtParams[ vtParamCount ] = newparam;
		vtParamShouldBeDeleted[ vtParamCount ] = deleteIt;
		newparam -> paramPos = vtParamCount;
		return vtParamCount++;
	}else{
		VTDEBUGMSGT(_T("Failed to add more effect params!"));
		return -1;
	}
}

bool CVTAudioEffectX::string2parameter(VstInt32 index, char *text)
{
	if( text ){
		float tmp;
		VTParam(index)->visibleText2Visible(text, tmp);
		VTParam(index)->setFromVisibleValue( tmp );
	}	
	return true;
}


void CVTAudioEffectX::setParameter(const char *name, float value)
{
	int idx = VTParamIdx( name );
	if( idx >= 0 )
		setParameter( idx, value);
}




#if defined(VT_USE_CHUNKS_SAVING)

struct TVTSaveEffectParam{
	char name[ kVstMaxParamStrLen + 1 ];
	float value;
};

//
// Create a chunk where all parameters are saved.
//
VstInt32 CVTAudioEffectX::getChunk (void **data, bool isPreset) 
{
	// isPreset - true when saving a single program, false for all programs
	// Returns the size in bytes of the chunk (plug-in allocates the data array).
	(*data) = new TVTSaveEffectParam[ VTParamCount() ];
	for(int i=0; i<VTParamCount(); i++)
	{
		vst_strncpy(((TVTSaveEffectParam*)(*data))[i].name, 
			VTParam(i)->getCodeName(), kVstMaxParamStrLen);
		//((TVTSaveEffectParam*)(*data))[i].value = getParameter(i);
		((TVTSaveEffectParam*)(*data))[i].value = VTParam(i)->getSaveValue();
	}
	return sizeof(TVTSaveEffectParam) * VTParamCount();
}

//
// Create a chunk where all parameters are loaded.
//
VstInt32 CVTAudioEffectX::setChunk (void *data, VstInt32 byteSize, bool isPreset)
{
	// isPreset 	true when restoring a single program, false for all programs
	bReloading = true;

	int count = byteSize / sizeof(TVTSaveEffectParam);

	for(int i=0; i<count; i++){
		//setParameter( ((TVTSaveEffectParam*)data)[i].name, ((TVTSaveEffectParam*)data)[i].value );
		int idx = VTParamIdx( ((TVTSaveEffectParam*)data)[i].name, true ); // Use codeName of the param
		// VTASSERT2( idx>=0, ((TVTSaveEffectParam*)data)[i].name );
		if( idx >= 0 )
			VTParam(idx)->setFromSaveValue( ((TVTSaveEffectParam*)data)[i].value, true );
	}

	updateDisplay(); 
	
	bReloading = false;

	return byteSize; // !!!!!!!!!!!??????????????
}

#endif


void CVTAudioEffectX::resetToDefaults()
{
	bReloading = true;
	for(int i=0; i<VTParamCount(); i++)
		VTParam(i)->resetToDefault();
	updateDisplay();
	bReloading = false;
}

VstInt32 CVTAudioEffectX::VTParamIdx( const char *name, bool bUseCodeName ) const
{
	if( !bUseCodeName ){
		for(int i=0; i<VTParamCount(); i++)
			if( !strncmp(VTParam(i)->getName(), name, kMaxParamStrLen) )
				return i;
	}else{
		for(int i=0; i<VTParamCount(); i++)
			if( !strncmp(VTParam(i)->getCodeName(), name, kMaxParamStrLen) )
				return i;
	}
	return -1;
}

CParameter* CVTAudioEffectX::VTParam( const char *name ) const
{
	int idx = VTParamIdx( name );
	if( idx >= 0 )
		return VTParam( idx );
	else
		return 0;
}


