//-----------------------------------------------------------------------------
//
// A number of classes which expand functionality of VST GUI and
// provide a cleaner programming interface.
//
//-----------------------------------------------------------------------------

/*
	VT Framework License (BSL 1.0)

	Copyright (C) 2007 Sergey A. Galin 
	E-mail: sergey.galin@gmail.com

	Permission is hereby granted, free of charge, to any person or organization
	obtaining a copy of the software and accompanying documentation covered by
	this license (the "Software") to use, reproduce, display, distribute,
	execute, and transmit the Software, and to prepare derivative works of the
	Software, and to permit third-parties to whom the Software is furnished to
	do so, all subject to the following:

	The copyright notices in the Software and this entire statement, including
	the above license grant, this restriction and the following disclaimer,
	must be included in all copies of the Software, in whole or in part, and
	all derivative works of the Software, unless such copies or derivative
	works are solely in the form of machine-executable object code generated by
	a source language processor.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
	SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
	FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
	ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
	DEALINGS IN THE SOFTWARE.
*/

#ifndef VT_VSTGUIADDONS_HPP
#define VT_VSTGUIADDONS_HPP

#include <string.h>
#include "audioeffectx.h"
#include "vstcontrols.h"
#include "aeffguieditor.h"
#include "vt_vstaddons.hpp"

// This an experimental feature, doesn't work properly yet.
//#define VT_VST_REDUCEUPDATES

#if defined(VT_VST_REDUCEUPDATES)
	#ifdef _MSC_VER
		#pragma message("VT_VST_REDUCEUPDATES is enabled (this is an experimental feature!)")
	#else
		#warning VT_VST_REDUCEUPDATES is enabled (this is an experimental feature!)
	#endif
#else
	#ifdef _MSC_VER
		#pragma message("VT_VST_REDUCEUPDATES is disabled.")
	#else
		#warning VT_VST_REDUCEUPDATES is disabled.
	#endif
#endif


// Tag value ranges for controls.
// Controls of one kind should be enumerated using their tag field 
// starting from the constant which identifies the kind.
const long
								   // Negative tags are not handled in any way.
	cVTNoTagShift = 0,             // Normal controls: control value <=> effect param value
	cVTInternalCtlsShift = 1000,   // Internal GUI's controls (not related to effect)
	cVT01TagShift = 10000,         // Control value <=> effect param value as01()/from01()
	cVTTextTagShift = 20000,       // Text edits and combo boxes
	cVTSplashScreenTag = 99999,    // Any splash screens have this tag
	cVTNotEditorTagShift = 100000; // Controls which don't change effect's settings
	
	


// Don't forget that dB's are usually negative ;-)
inline float VT01toVUMeter( double level, double leftdb, double rightdb )
{
	double db = Clip( Convert_01TodB_NoClip( level ), leftdb, rightdb );
	return Clip( float((db-leftdb) / (rightdb-leftdb)), 0.0f, 0.9975f );
}


// This is for a meter which shows reduction level. The value passed is max(1-gain).
// Decibel values here are positive, e.g.: lefdb = 0 dB, rightdb = 30 dB.
inline float VT01toReductionVUMeter( double level, double leftdb, double rightdb )
{	
	double db = Clip( -Convert_01TodB_NoClip( 1.0-level ), leftdb, rightdb );
	return Clip( float((db-leftdb) / (rightdb-leftdb)), 0.0f, 0.9975f );
}







class CVTStdTextEdit;

#define VTCONSTRUCTIONTAG(effct) (effct->paramPos+getTagShift())


// Common interface expander for all our GUI controls.
// Be sure to initialize 'vtparam' with 'this' in all descendant classes constructors!
class CVTVSTGUIControl
{
public: 
	virtual int getTagShift(){ return cVTNoTagShift; }

	CVTVSTGUIControl(CControl *owner, Params::CParameter *prm):
		control( owner ),
		vtparam( prm ),
		m_uParamRangeChange( prm->GetRangeChange() )
	{
		// VTSetProperTag(); -- tag should be set by descendant constructor;
		// at the moment of construction of this object, the control object is 
		// probably not created yet, so who knows what may happen!
		// For the same reason, do not use VTGetProperTag() in descendant's
		// constructor; instead, pass a tag created via VTCONSTRUCTIONTAG
		// to control's constructor.
	}

	CControl *getControl() const { return control; }	
	Params::CParameter *getParam() const { return vtparam; }
	void setParam(Params::CParameter *newparam);
	

	// Reload value/default/min/max from the effect.
	// Returns true if value is changed.
	virtual bool VTUpdateFromEffect()=0; 
	
	// Calculate new value... Created especially for the stupid VST GUI edit boxes.
	virtual float VTNewInControlEffectValue()=0; 	
	virtual float VTNewEffectValue()=0; // New value, which should be set after editing
	virtual void VTUpdateEffect();
private:
	unsigned long m_uParamRangeChange;
	Params::CParameter *vtparam;
	CControl *control;	
protected:

	bool ParamRangeChanged() const { return m_uParamRangeChange!=vtparam->GetRangeChange(); }
	void RangeCorrected(){ m_uParamRangeChange = vtparam->GetRangeChange(); }

	void setControl(CControl *ctl){ control=ctl; }	
	virtual long VTGetTag()=0; // Plug into getTag() of the CControl (see VTCONTROLSTDFUNCIONS)
	virtual void VTSetTag(long newtag)=0; // Plug into setTag() of the CControl (see VTCONTROLSTDFUNCIONS)
	void VTSetProperTag(){ VTSetTag(VTCONSTRUCTIONTAG(vtparam)); }

	void setRangeFromParamValue();
	void setRangeFromParamVisibleValue();
};


#define VTCONTROLSTDFUNCIONS \
	protected: \
		virtual long VTGetTag(){ return getTag(); } \
		virtual void VTSetTag(long newtag){ setTag(newtag); }

#define VTCONTROLSCONSTRUCTION \
	VTSetProperTag(); \
	setControl((CControl*)this);





// Slider control with automatic update from/to effect parameter
class CVTStdSlider: public CSlider, public CVTVSTGUIControl
{
public: 
	virtual int getTagShift(){ return cVT01TagShift; }

	CVTStdSlider(int x1, int y1, CControlListener *listener, Params::CParameter *prm,
		CBitmap *back, CBitmap *handle,
		long style = kLeft|kHorizontal);

	// Uses fragment of window background (x1,y1,x2,y2) to create slider's background
	CVTStdSlider(int x1, int y1, int x2, int y2,
		CControlListener *listener, Params::CParameter *prm,
		CBitmap *windowBack, CBitmap *handle,
		long style = kLeft|kHorizontal);

	// Same, but with CRect
	CVTStdSlider(const CRect& pos, CControlListener *listener, Params::CParameter *prm,
		CBitmap *windowBack, CBitmap *handle,
		long style = kLeft|kHorizontal);

	virtual void setValue( float val ){ if( val!=getValue() ) CSlider::setValue(val); }
	virtual bool VTUpdateFromEffect();
	virtual float VTNewEffectValue(){ return getParam()->from01( Limit01(getValue()) ); }
	virtual float VTNewInControlEffectValue(){ return Limit01(getValue()); }

	VTCONTROLSTDFUNCIONS
};


class CVTKnob: public CKnob, public CVTVSTGUIControl
{
public:
	virtual int getTagShift(){ return cVT01TagShift; }
	CVTKnob(const CRect &size, CControlListener *listener, Params::CParameter *prm, CBitmap *background, CBitmap *handle);
	virtual void setValue( float val ){ if( val!=getValue() ) CKnob::setValue(val); }
	virtual bool VTUpdateFromEffect();
	virtual float VTNewEffectValue(){ return getParam()->from01( Limit01(getValue()) ); }
	virtual float VTNewInControlEffectValue(){ return Limit01(getValue()); }	
	VTCONTROLSTDFUNCIONS
};


// Two-state sticky button with automatic update from/to effect parameter
class CVTOnOffButton: public COnOffButton, public CVTVSTGUIControl
{
public:
	CVTOnOffButton(const CRect &size, CControlListener *listener, Params::CParameter *prm,
		CBitmap *background);
	virtual void setValue( float val ){ if( val!=getValue() ) COnOffButton::setValue(val); }
	virtual bool VTUpdateFromEffect();
	virtual float VTNewEffectValue(){ return getValue(); }
	virtual float VTNewInControlEffectValue(){ return float((getValue()>=0.5)?1.0:0.0); }

	VTCONTROLSTDFUNCIONS
};


// A label which has our standard appearance.
class CVTStdTextLabel: public CTextLabel
{
public:
	CVTStdTextLabel(const CRect &size, CBitmap *background=0, const char *txt="");

	virtual void setText (const char *txt){	
		if( strcmp(txt, getText()) ){	
			CTextLabel::setText(txt); setDirty(); 
		}
	}
};


// A label associated with effect parameter, which can update automatically
// using parameter's value and formatting preferences.
class CVTStdParamDisplayLabel: public CVTStdTextLabel, public CVTVSTGUIControl
{
public:
	virtual int getTagShift(){ return cVTNotEditorTagShift; }

	CVTStdParamDisplayLabel (const CRect &size, Params::CParameter *prm, CBitmap *background=0);	
	virtual bool VTUpdateFromEffect();
	virtual float VTNewEffectValue(){ return getValue(); }
	virtual float VTNewInControlEffectValue(){ return getValue(); }
	virtual void VTUpdateEffect(){ ; }
	
	VTCONTROLSTDFUNCIONS
};

// Peak indicator (in dB), useful to fit after VU-meters.
// Do updateValue() in onIdle() to have it updated.
// When peak level is over fAlarmThreshold, the peak is switched into
// "alarm mode" (i.e. colored in red and etc.)
class CVTPeakDisplay: public CTextLabel
{
public:
	// Font colors for normal mode and warning mode.
	CColor normalColor, alarmColor;

	// 'value' points to peak value variable. Peaks are read from there.
	// When the indicator is cleared via mouse click, *value is reset to 0.
	CVTPeakDisplay(const CRect &size, CBitmap *bgNormal, CBitmap *bgAlarm, float *value, float thresh=1.0);
	~CVTPeakDisplay();
	virtual void setText (const char *txt);
	void setState(float val, bool forceUpdate=false);
	void setValue(float value, bool forceUpdate=false);
	void updateValue(bool forceUpdate=false){ setValue(*pValue, forceUpdate); }
	void clear(){ setValue((*pValue)=resetValue); } 
	virtual void mouse (CDrawContext *pContext, CPoint &where, long buttons=-1){
		CTextLabel::mouse(pContext, where, buttons);
		clear();
	}
	void setResetValue(float x){ resetValue=x; }
	float getResetValue() const { return resetValue; }
protected:
	float *pValue;
	CBitmap *normalBg, *alarmBg;
	enum TVTPeakState { kVTPeakNormal, kVTPeakAlarm } alarmState;
	float fAlarmThreshold;
	float resetValue;
	void setAlarmState(bool forceUpdate=false);
	void setNormalState(bool forceUpdate=false);
};



// Text converter functions used by CVTStdTextEdit.
void VTStdTextConvert(char *input, char *string, void *userDta);
void VTStdString2FloatConvert(char *string, float &output);
void VTStdStringConvert(float value, char *string, void *userDta);

// Text edit control with automatic update from/to effect parameter.
class CVTStdTextEdit: public CTextEdit, public CVTVSTGUIControl
{
public:
	virtual int getTagShift(){ return cVTTextTagShift; }
	CVTStdTextEdit(const CRect &sz, CControlListener *listener, Params::CParameter *prm, CBitmap *background);
	virtual void setValue( float val ){ if( val!=getValue() ) CTextEdit::setValue(val); }
	virtual bool VTUpdateFromEffect();
	virtual float VTNewInControlEffectValue();
	virtual float VTNewEffectValue(){ return getParam()->fromVisibleValue( VTNewInControlEffectValue() ); }

	VTCONTROLSTDFUNCIONS
};



//
// Drop-down list control with automatic update from/to effect parameter.
// This is in fact not a "combo", because it doesn't support text editing,
// like edit box.
//
class CVTStdCombobox: public COptionMenu, public CVTVSTGUIControl
{
public:
	virtual int getTagShift(){ return cVTTextTagShift; }
	CVTStdCombobox(const CRect &sz, CControlListener *listener, 
		Params::CParameter *prm, CBitmap *background,
		CBitmap *bgWhenClick=0, const long style=0);
	
	virtual bool VTUpdateFromEffect();
	virtual void setValue( float val );	
	virtual float VTNewInControlEffectValue();	
	virtual float VTNewEffectValue();

	VTCONTROLSTDFUNCIONS

protected:
	virtual void UpdateOptionList();
};




// A VU meter which shows both RMS and amplitude indicators (in one "LED row")
class CVTRMSVUMeter: public CVuMeter
{
public:
	CVTRMSVUMeter(const CRect &size, CBitmap *onBitmap, CBitmap *rmsBitmp, CBitmap *offBitmap,
		long nbLed, const long style=kVertical);
	~CVTRMSVUMeter(){ if(rmsBitmap) rmsBitmap->forget(); }
	virtual void setValue(float val);
	virtual void setValue(float ampval, float rmsval);
	virtual bool attached (CView *parent);
	virtual void draw (CDrawContext *_pContext);
protected:
	float rmsOldValue, rmsValue;
	CBitmap *rmsBitmap;
	CRect rectRMS;
};











// Object which contains all actual CUI controls and listens to them.
class CVTStdContainer: public CViewContainer, public CControlListener
{
public:

	static const int 
		kDefaultMaxControls = 256,
		kDefaultMaxSplashes = 1;

	CVTStdContainer(const CRect &size, CFrame *pParent, CBitmap *pBackground, 
		CVTAudioEffectX *effct, int maxCtls = kDefaultMaxControls, int maxSplashes = kDefaultMaxSplashes);
	~CVTStdContainer();

	CVTAudioEffectX *VTEffect(){ return effect; }
	int VTNumControls(){ return iNumControls; }
	int VTMaxControls(){ return iMaxControls; }
	CVTVSTGUIControl* VTControl(int i){ return pControls[ i ]; }
	
	void VTAddControl(CVTVSTGUIControl* ctl){
		if( iNumControls<iMaxControls-1 ){
			pControls[ iNumControls++ ] = ctl; 
		}else{
			VTDEBUGMSGT( "Attempt to add more controls than declared into CVTStdContainer");
		}
	}	
	
	void VTAddViewAndControl(CVTVSTGUIControl* ctl){ VTAddControl(ctl); addView(ctl->getControl()); }	
	
	int tagToControlIdx(long tag);	
	virtual bool VTUpdateAllFromEffect(); // true if any control is updated
	virtual void valueChanged (CDrawContext *pContext, CControl *pControl);
	
	virtual void onIdle(){ }

	virtual void update (CDrawContext *pContext){ 
		bool bUpdated = VTUpdateAllFromEffect(); 
		// if( !pSplashScreen )
		if( !iNumSplashScreens )
			CViewContainer::update(pContext); 
		else{
			bool noSplashesOn = true;
			for( int i=0; i<iNumSplashScreens; i++ )
				if( pSplashScreens[i]->getValue()>=0.1f ){
					noSplashesOn = false;
					break;
				}

			// A workaround. Otherwise, the splash won't show up.
			// VST GUI is bad, m'kay!
			// if( pSplashScreen->getValue()<0.1f ) // Splash is not visible
			if( noSplashesOn )
				CViewContainer::update(pContext); 
		}
		#if defined(VT_VST_REDUCEUPDATES)
			if( bUpdated ) 
				setDirty();
		#else
			setDirty();
		#endif			
	}

private:
	CVTAudioEffectX *effect;
	int iNumControls, iMaxControls, iNumSplashScreens, iMaxSplashScreens;
	CVTVSTGUIControl **pControls; 	
	CSplashScreen **pSplashScreens;

protected:
	// Create splash (help) screen displayed by button.
	// Note that only ONE splash screen can be created so far using this function.
	// The splash & the button should be in their own images.
	virtual CSplashScreen* CreateSplashScreen(const CPoint& clickPos, CBitmap *pButtonImg, 
		const CPoint& splashPos, CBitmap *pSplashImg);
	// Same, but automatically loads images from resources.
	virtual CSplashScreen* CreateSplashScreen(const CPoint& clickPos, long iButtonResId, 
		const CPoint& splashPos, long iSplashResId);
};


// GUI editor
class CVTStdAEffGUIEditor: public AEffGUIEditor
{
public:
	CVTStdAEffGUIEditor(CVTAudioEffectX* effct);
	CVTStdAEffGUIEditor(CVTAudioEffectX* effct, int bgbitmapid);
	~CVTStdAEffGUIEditor();

	CVTAudioEffectX* VTEffect(){ return (CVTAudioEffectX*) effect; }

	void VTSetBackground(CBitmap *bmp);
	void VTSetBackground(int bgbitmapid);	
	
	void idle();

protected:
	volatile bool vtInIdle;
	CBitmap *backgroundBitmap;
	CVTStdContainer *objectContainer;
	virtual void close ();	
	virtual bool open (void *ptr);
	void setObjectContainer(CVTStdContainer *c){ frame = (CFrame*)c; objectContainer = c; }
	virtual CVTStdContainer* CreateContainer(const CRect& size, CFrame *frm)=0;
};


class CVTBitmap: public CBitmap
{
public:
	CVTBitmap( LPCVTCHAR path );
};



#endif

