//-----------------------------------------------------------------------------
//
// Arrays suitable for storing all types of m_pData: 
// objects, pointers, simple types.
//
//-----------------------------------------------------------------------------

/*
	VT Framework License (BSL 1.0)

	Copyright (C) 2006 Sergey A. Galin 
	E-mail: sergey.galin@gmail.com

	Permission is hereby granted, free of charge, to any person or organization
	obtaining a copy of the software and accompanying documentation covered by
	this license (the "Software") to use, reproduce, display, distribute,
	execute, and transmit the Software, and to prepare derivative works of the
	Software, and to permit third-parties to whom the Software is furnished to
	do so, all subject to the following:

	The copyright notices in the Software and this entire statement, including
	the above license grant, this restriction and the following disclaimer,
	must be included in all copies of the Software, in whole or in part, and
	all derivative works of the Software, unless such copies or derivative
	works are solely in the form of machine-executable object code generated by
	a source language processor.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
	SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
	FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
	ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
	DEALINGS IN THE SOFTWARE.
*/


#ifndef VT_ARRAYS_H
#define VT_ARRAYS_H

#include <stdlib.h>
#include "vt_assert.hpp"
#include "vt_audiomath.hpp"
#include "vt_localization.hpp"

namespace VT {

	//
	// Portable simple array class.
	//
	// Can be used to construct arrays of objects.
	//
	// This type of array can be used as a base for strings and pointer arrays.
	// The memory is never reallocated unless vector is resized over m_uAllocSize.
	// Some memory buffer is always allocated, which saves us from
	// enormous checks for NULL pointers.
	//
	// The memory is never reallocated unless vector is resized over m_uAllocSize.
	// This default behaviour should never be changed, because the objects are 
	// used in sound processing routines which should not do any OS calls.
	// For example, VST's processXXX under Mac OS X VST's can't do that.
	// shrink() can be used to reduce memory size occupied by the vector.
	//
	template<class T> class CArray
	{
	public:
		static const size_t kMinAllocSize = 16;
		virtual size_t GetMinAllocSize() const { return kMinAllocSize; }
		

		virtual LPCVTCHAR GetClassName() const { return _T("CArray"); }

		// Default constructor
		// Note that size can be 0.
		CArray<T>(size_t size=0, size_t alloc=kMinAllocSize): 
			m_uDataSize( size ),
			m_uGrowSize( kMinAllocSize )
		{
			m_uAllocSize = Max(Max(alloc, size), kMinAllocSize);
			m_pData = new T[ m_uAllocSize ];
		}

		virtual size_t GetGrowSize() const { return m_uGrowSize; } 
		virtual void SetGrowSize( size_t sz ){ m_uGrowSize = sz; }

		// Copy constructor
		template<class U> CArray<T>(const CArray<U> &src): 
			m_uDataSize(src.getSize()), m_uAllocSize(src.getSize())
		{
			m_pData = new T[ m_uAllocSize ]; copyFrom( src );
		}

		CArray& operator=( const CArray<T> &src ){
			copyFrom( src );
			return *this;
		}

		~CArray(){ if(m_pData!=NULL && m_uAllocSize){ delete[] m_pData; m_pData=NULL; } }

		// Resize array. If array is upsized and savedata==true, then
		// values above the old vector size are undefined.
		// If array is downsized, it is not reallocated and old m_pData
		// always stays there.
		virtual void resize(size_t newDataSize, bool saveData=false, size_t newAllocSize=0);

		// Shrink array for minimal memory usage, but not smaller than GetMinAllocSize().
		// Note that default value for saveData is different from one of resize().
		virtual void shrink( bool saveData=true, size_t minAlloc=0 );

		void fill(const T& f);
		
		VTINLINE void set(size_t x, T value){ m_pData[x]=value; } // Assign to item in array
		const T& getConst(size_t idx) const { return m_pData[idx]; } // Get direct const access to item in array
		T& get(size_t idx){ return m_pData[idx]; } // Get direct access to item in array
		T& operator[](size_t idx){ return m_pData[idx]; } // Get direct access to item in array
		VTINLINE size_t getSize() const { return m_uDataSize; }

		// Copy size & m_pData from the 'src' vector; note that 'src' may 
		// have different m_pData type; in this case, C++ type conversion 
		// occurs upon conversion of array elements.
		template<class U> void copyFrom(const CArray<U>& src);
		template<class U> CArray<T> operator=(const CArray<U>& src){ copyFrom(src); return *this; }
		template<class U> void copyFrom(const U *src, size_t nElements);
		
		// Low-level access ;-)
		VTINLINE T* getDataPtr(){ return m_pData; }
		VTINLINE const T* getRODataPtr() const { return m_pData; }


		// Copy without element conversion (it's faster, we just use memcpy()).
		// These functions don't override CopyFrom() to avoid compiler confusion.
		// Do not use these functions on arrays of objects!
		void rawCopyFromSameType(const CArray<T>& src){ 
			this->resize( src.getSize() );
			memcpy( m_pData, src.getRODataPtr(), sizeof(T)*src.getSize() );
		}	
		void rawCopyFromSameType(const T* src, size_t nElements){ 
			this->resize( nElements );
			memcpy( m_pData, src, sizeof(T)*nElements );
		}	

		VTINLINE void ZeroMemoryBytes(){ 
			if( m_pData!=NULL && m_uDataSize>0 )
				memset( m_pData, 0, m_uDataSize*sizeof(T) ); 
		}

		// Fast content swapping
		VTINLINE void Swap( CArray<T> *other ){
			Math::Swap<T*>( &m_pData, &(other->m_pData) );
			Math::Swap<size_t>( &m_uDataSize, &(other->m_uDataSize) );
			Math::Swap<size_t>( &m_uAllocSize, &(other->m_uAllocSize) );
		}

	protected:
		T *m_pData;
		size_t m_uDataSize, m_uAllocSize, m_uGrowSize;
	};




	//////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////
	//
	// Implementation of the template class functions
	//


	//
	// Note that this function doesn't use memcpy(), realloc() and etc.
	// This makes it slower for scalar types, but compatible with object-type T's.
	//
	template<class T> void CArray<T>::resize(size_t newDataSize, bool saveData, size_t newAllocSize)
	{
		VTASSERT( this!=NULL );		

		// If we dont't have enough space for new m_pData...
		if( newDataSize>m_uAllocSize || m_pData==NULL || newAllocSize>m_uAllocSize ){
			
			// *** Caclulate new array size ***
			if( !newAllocSize ) // Then calculate it using GetGrowSize()
				newAllocSize = newDataSize+GetGrowSize(); // > m_uAllocSize
			if( newAllocSize<m_uAllocSize ) 
				newAllocSize = m_uAllocSize; // We won't shrink array here!
			if( newAllocSize<newDataSize ) 
				newAllocSize = newDataSize; // Enlarging to fit m_pData

			// If we don't have any data to save, drop the flag, to have
			// memory released before allocating new block.
			if( !m_uDataSize || m_pData==NULL )
				saveData = false;
			
			// *** Reallocate memory ***
			if( saveData && m_pData!=NULL && m_uDataSize>0 && m_uAllocSize>0 ){ 
				// Expand the array and save all existing values.
				T *saveDataPtr = m_pData; // Save pointer to old m_pData
				m_pData = new T [ newAllocSize ];
				if( saveDataPtr!=NULL ){
					// printf("Copying %s m_pData: from %d/%d to %d/%d elements.\n", GetClassName(), 
					//	int(m_uDataSize), int(m_uAllocSize), int(newDataSize), int(newAllocSize));
					for(size_t i=0; i<m_uDataSize; i++) // Copy old elements to new place
						m_pData[i] = saveDataPtr[i];
					delete[] saveDataPtr;
				}
			}else{ // Don't care about the old m_pData
				if( m_pData!=NULL )
					delete[] m_pData;
				m_pData = new T[ newAllocSize ];
			}
			m_uAllocSize = newAllocSize;
		}
		m_uDataSize = newDataSize;
		VTASSERT( m_pData!=NULL );
		VTASSERT( m_uDataSize<=m_uAllocSize );
		VTASSERT( m_uAllocSize>=GetMinAllocSize() );
	}

	template<class T> void CArray<T>::shrink( bool saveData, size_t minAlloc )
	{
		if( m_uDataSize<m_uAllocSize ){
			size_t newAllocSize = Max( m_uDataSize, minAlloc );
			if( newAllocSize<GetMinAllocSize() )
				newAllocSize = GetMinAllocSize();
			T *saveDataPtr = m_pData; // Save pointer to old m_pData
			// Destroy old m_pData before allocating new memory, if possible
			if( saveDataPtr!=NULL ){
				if( !saveData || m_uDataSize==0 ){
					delete[] saveDataPtr; 
					saveDataPtr = NULL;
				}
			}
			// Allocate new m_pData
			m_pData = new T[ newAllocSize ]; 
			// Copy old m_pData to new place
			if( saveDataPtr!=NULL ){ 
				if( saveData ){
					for(size_t i=0; i<m_uDataSize; i++) // Copy old elements to new place
						m_pData[i] = saveDataPtr[i];
					delete[] saveDataPtr; // Destroy old m_pData
				}
			}
			m_uAllocSize = newAllocSize;
		}
	}

	template<class T> void CArray<T>::fill(const T& f)
	{
		for(size_t i=0; i<getSize(); i++)
			m_pData[i] = f;
	}

	template<class T> template<class U> void CArray<T>::copyFrom(const CArray<U>& src)
	{
		resize(src.getSize());
		for(size_t i=0; i<getSize(); i++)
			// Important!!! Do it like that (through const& and without explicit type cast) 
			// to have proper operator= used.
			// For precision-reducing conversions this may cause an MSC warning C4244.
			// But we cannot use type casting (T(src.getConst(i))) because it may cause
			// construction of temporary T obejct from U and then doing an assignment, 
			// which is not only inefficient but also done in a wrong way (at least in MSC).
			#if defined(_MSC_VER)
				#pragma warning( suppress: 4244 )
			#endif
			m_pData[i]=src.getConst(i); 
	}

	template<class T> template<class U> void CArray<T>::copyFrom(const U *src, size_t nElements)
	{
		resize(nElements);
		for(size_t i=0; i<getSize(); i++)
			// Important!!! Do it like that (without explicit type cast) to have proper 
			// operator= used.
			// For precision-reducing conversions this may cause an MSC warning C4244.
			// But we cannot use type casting (T(src.getConst(i))) because it may cause
			// construction of temporary T from U and then doing an assignment, which is 
			// not only inefficient but also done in a wrong way (at least in MSC).
			#if defined(_MSC_VER)
				#pragma warning( suppress: 4244 )
			#endif
			m_pData[i]=src[i];
	}

} // namespace

#endif





