//-----------------------------------------------------------------------------
//
// Misc math stuff mostly for use in sound applications
//
//-----------------------------------------------------------------------------

/*
	VT Framework License (BSL 1.0)

	Copyright (C) 2005 Sergey A. Galin 
	E-mail: sergey.galin@gmail.com

	Permission is hereby granted, free of charge, to any person or organization
	obtaining a copy of the software and accompanying documentation covered by
	this license (the "Software") to use, reproduce, display, distribute,
	execute, and transmit the Software, and to prepare derivative works of the
	Software, and to permit third-parties to whom the Software is furnished to
	do so, all subject to the following:

	The copyright notices in the Software and this entire statement, including
	the above license grant, this restriction and the following disclaimer,
	must be included in all copies of the Software, in whole or in part, and
	all derivative works of the Software, unless such copies or derivative
	works are solely in the form of machine-executable object code generated by
	a source language processor.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
	SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
	FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
	ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
	DEALINGS IN THE SOFTWARE.
*/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "vt_assert.hpp"
#include "vt_audiomath.hpp"

#ifndef _CRT_SECURE_NO_DEPRECATE
	#define _CRT_SECURE_NO_DEPRECATE
#endif

#if defined(_DEBUG)
	#include "vt_messageboxes.hpp"
#else
	#if !defined(VTASSERT)
		#define VTASSERT(x) {}
	#endif
#endif

using namespace VT;
using namespace VT::Math;

void Math::dB2string (float value, VTCHAR* text, size_t maxLen, bool showPlusSign)
{
	if ( value <= 0 ){
		VTSTRNCPY(text, kNegativeInfinitySymbol, maxLen);
	}else{
		float val = (float)(20. * log10 (value));
		if(val>0){
			if( showPlusSign ){
				text[0]=_T('+');
				float2string (val, text+1, maxLen-1);
			}else
				float2string (val, text, maxLen);
		}else{
			if( val<kMin_dB )
				VTSTRNCPY (text, kNegativeInfinitySymbol, maxLen);
			else
				float2string (val, text, maxLen);
		}
	}
}

void Math::float2string (float value, VTCHAR* text, size_t maxLen, bool showPlusSign)
{
	LPCVTCHAR format = (showPlusSign)?_T("%+0.02f"): _T("%0.02f");
	#ifndef UNICODE
		VTSNPRINTFC(text, maxLen, format, value);
	#else
		VTCHAR buffer[64];
		VTSPRINTF(buffer, format, value);
		VTSTRNCPY(text, buffer, maxLen);
	#endif
}

void Math::dB2stringNC (float value, VTCHAR* text, size_t maxLen, bool showPlusSign)
{
	if ( value <= kMin_dB ){
		VTSTRNCPY(text, kNegativeInfinitySymbol, maxLen);
	}else{
		if(value>0){
			if( showPlusSign ){
				text[0]=_T('+');
				float2string (value, text+1, maxLen-1);	
			}else
				float2string (value, text, maxLen);	
		}else
			float2string (value, text, maxLen);	
	}
}

void Math::int2string (float value, VTCHAR* text, size_t maxLen, bool showPlusSign)
{
	VTCHAR buf[64];
	VTSPRINTF(buf, (showPlusSign)?_T("%+ld"):_T("%ld"), long(SRound(value)));
	VTSTRNCPY( text, buf, maxLen );
}


double Math::Convert_01TodB(double value, double max){ 
	if( value < kMin_dB_01 ) // Prevent logarithm overflow
		return kMin_dB; 
	else{
		double ret=float(20.0 * log10 (value)); 
		return (ret<kMin_dB)? kMin_dB: (ret>max)?max: ret;
	}
}

double Math::Convert_dBTo01(double db, double max){ 	
	if( db<kMin_dB ) 
		return 0.0;
	if( db>kMax_dB )
		return kMax_dB_01;
	float ret=float(pow(10.0, db/20.0));
	if(ret>=max) 
		return max;
	else
		return ret;
}

void Math::VisibleTextToVisible_dB(VTCHAR *string, float &output)
{
	int parsed = VTSSCANF(string, _T("%f"), &output);
	if( !parsed ){
		// If failed to parse the string, then have to
		// return kMin_dB if string starts with "-" sign.
		for(int i=0; string[i]; i++)
			if(!VTISSPACE(string[i]))
				if(string[i]==_T('-')){
					output=kMin_dB;
					break;
				}
	}
}


// Calculate mod value which gives ControlHandle_01_Log(x, mod) => y
// The function uses bisection algorithm and smin and smax determine area 
// of search for result.
double Math::CalculateLogMod(double x, double y, double precision, double smin, double smax)
{
	// Calculate mod value which gives ControlHandle_01_Log(x) => y
	// using bisection algorithm.
	double 
		left=smin, 
		right=smax, 
		middle=(left+right)/2.0,
		leftval = ControlHandle_01_Log(x, left),
		rightval = ControlHandle_01_Log(x, right),
		middleval = ControlHandle_01_Log(x, middle);
	
	const int kMaxIterations = 1000;
	int iteration = 0;	

	while(
		right-left > precision || 
		Abs(middleval-y) > precision ||
		Abs(leftval-rightval) > precision ) {
			if( middleval < y ){
			left = middle;
			leftval = middleval;
		}else{
			right = middle;
			rightval = middleval;
		}
		middle = (left + right) / 2.0;
		middleval = ControlHandle_01_Log(x, middle);

		if( iteration++ > kMaxIterations ){
			#if defined(_DEBUG)
				VTCHAR msg[256];
				VTSPRINTF( msg, 
					_T( "Failed to find a result in a reasonable number of iterations.") VTENDLT
					_T("X=%lf Y=%lf precision=%lf smin=%lf smax=%lf"),
					x, y, precision, smin, smax );
				VTDEBUGMSG( msg );
			#endif			
			return kDefaultLogMod;
		}

		if( left==right || left==middle || right==middle ){
			#if defined(_DEBUG)
				VTCHAR msg[256];
				VTSPRINTF( msg, 
					_T( "Search ranges collision. Not enough precision?") VTENDLT
					_T("X=%lf Y=%lf precision=%lf smin=%lf smax=%lf"),
					x, y, precision, smin, smax );
				VTDEBUGMSG( msg );
			#endif			
			return kDefaultLogMod;
		}
	};	
	
	#if defined(_DEBUG)
		// Verifying precision of the result
		VTCHAR msg[256];
		VTSPRINTF( msg, _T( "X=%lf Y=%lf precision=%lf smin=%lf smax=%lf" ), 
			x, y, precision, smin, smax );
		VTASSERT2( right-left <= precision, msg );
		VTASSERT2( Abs(middleval-y) <= precision, msg );
		VTASSERT2( Abs(leftval-rightval) <= precision, msg );
		double logresult = ControlHandle_01_Log(x, middle);
		VTASSERT( Abs( logresult-y )<precision );
	#endif

	return middle;
}

// Greatest Common Divisor
SDWORD Math::GCD(SDWORD a, SDWORD b)
{
	if( a<=0 || b<=0 )
		return 0;
	while( a!=b ){
		if( a>b )
			a-= b;
		else
			b-= a;
	}
	return a;
}

// Least Common Multiple and Greatest Common Divisor
bool Math::LCM_and_GCD(SDWORD a, SDWORD b, SDWORD* lcm, SDWORD* gcd)
{
	//#if defined(VT_CRTDEBUG)
	//	printf("*** LCM_and_GCD(%d, %d)\n", a, b);
	//#endif
	bool bRet = true;
	if( a<=0 || b<=0 )
		return false;
	SDWORD divider = GCD(a, b);
	if( divider<=0 )
		return false;
	//#if defined(VT_CRTDEBUG)
	//	printf("*** LCM_and_GCD(): a=%d b=%d d=%d\n", a, b, divider);
	//#endif
	if( gcd!=NULL )
		(*gcd) = divider;
	if( lcm!=NULL ){
		// Overflow check. "*0.999" is to avoid round-up errors.
		if( double(a)*double(b)/double(divider)>double(kInt32Max)*0.999 ){
			#if defined(VT_CRTDEBUG)
				printf("*** LCM_and_GCD(): Overflow! a=%d b=%d d=%d max=%d\n",
					a, b, divider, kInt32Max);
			#endif
			(*lcm) = 0;
			bRet = false;
		}else{
			(*lcm) = (a/divider)*b; // Note that the division goes first!
		}
	}
	return bRet;
}




