//-----------------------------------------------------------------------------
//
// Misc math stuff mostly for use in sound applications
//
//-----------------------------------------------------------------------------

/*
	VT Framework License (BSL 1.0)

	Copyright (C) 2005 Sergey A. Galin 
	E-mail: sergey.galin@gmail.com

	Permission is hereby granted, free of charge, to any person or organization
	obtaining a copy of the software and accompanying documentation covered by
	this license (the "Software") to use, reproduce, display, distribute,
	execute, and transmit the Software, and to prepare derivative works of the
	Software, and to permit third-parties to whom the Software is furnished to
	do so, all subject to the following:

	The copyright notices in the Software and this entire statement, including
	the above license grant, this restriction and the following disclaimer,
	must be included in all copies of the Software, in whole or in part, and
	all derivative works of the Software, unless such copies or derivative
	works are solely in the form of machine-executable object code generated by
	a source language processor.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
	SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
	FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
	ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
	DEALINGS IN THE SOFTWARE.
*/

#ifndef VT_AUDIOMATH_HPP
#define VT_AUDIOMATH_HPP

#include <math.h>
#include "vt_portable.hpp"
#include "vt_assert.hpp"

#if defined(_MSC_VER)
	#define MSCFASTMATH
#endif

namespace VT {

	namespace Math {

		typedef long double TLongDouble;
		typedef VTFSIZE TAudioStreamSize;


		// *** Math Constants ***
		const long double
			kPi			= 3.1415926535897932384626433832795,
			k2Pi		= 6.283185307179586476925286766559,
			kPi2        = 1.5707963267948966192313216916398,
			kSqrt2		= 1.4142135623730950488016887242097,
			kLog10_2	= 0.301029995663981195213738894724,
			kLog_2		= 0.69314718055994530941723212145818;

		const double kRounder = 0.499999999;

		static const VTCHAR 
			*kInfinitySymbol			= _T("oo"),
			*kPositiveInfinitySymbol	= _T("+oo"),
			*kNegativeInfinitySymbol	= _T("-oo");

		const SDWORD kInt32Max = SDWORD(2147483647);
		const SDWORD kDefMaxParamStrLen = 8; // Do not change, this is the VST SDK limit

		const float 	
			kMin_dB = -150.0, // All decibel values <= this threshold can be treated as -oo
			kMin_dB_01 = float(1e-10), // All [0..1] values <= this threshold can be treated as -oo
			kMax_dB = 150.0, // All decibel values <= this threshold can be treated as +oo
			kMax_dB_01 = float(1e10); // All [0..1] values <= this threshold can be treated as +oo


		inline double Log2(double x){ return log(x)/kLog_2; }
		inline double Log(double x, double base){ return log(x)/log(base); }

		void dB2string (float value, VTCHAR* text, size_t maxLen = kDefMaxParamStrLen, bool showPlusSign=true);
		void float2string (float value, VTCHAR* text, size_t maxLen = kDefMaxParamStrLen, bool showPlusSign=false);
		void int2string (float value, VTCHAR* text, size_t maxLen = kDefMaxParamStrLen, bool showPlusSign=false);
		void dB2stringNC (float value, VTCHAR* text, size_t maxLen = kDefMaxParamStrLen, bool showPlusSign=true);

		VTINLINE double Convert_01TodB_NoClip(double value);
		VTINLINE double Convert_dBTo01_NoClip(double db);

		// These 2 functions also clip input and output values into the acceptable ranges:
		// [0..max] and [ kMin_dB..max ] (for dB values)
		double Convert_01TodB(double value, double max=0.0);
		double Convert_dBTo01(double db, double max=1.0);

		template<class T> VTINLINE T Limit01( T x ){ return (x<T(0))? T(0): ((x>T(1))?T(1):x); }
		template<class T> VTINLINE T Clip( T x, T min, T max ){ return (x<min)?min: ((x>max)?max: x); }
		template<class T> VTINLINE T Abs( T x ){ return (x<T(0))?-x:x; }
		template<class T> VTINLINE T Max( T x, T y ){ return (x>y)?x:y; }
		template<class T> VTINLINE T Min( T x, T y ){ return (x<y)?x:y; }
		template<class T> VTINLINE void Swap( T *a, T *b ){ T t=*a; *a=*b; *b=t; }

		// Round positive floating-point to integer
		template<class T> VTINLINE SDWORD URound( T x ){ return SDWORD(x+T(kRounder)); }

		template<class T> VTINLINE SDWORD SRound( T x ){ 
			return (x>=0)?SDWORD(x+T(kRounder)) : SDWORD(x-T(kRounder)); }

		// Remember to call it as SRoundT<T, INTT> because result type 
		// cannot be detected automatically by compiler.
		template<class T, class INTT> VTINLINE INTT SRoundT( T x ){ 
			return (x>=0)?INTT(x+T(kRounder)) : INTT(x-T(kRounder)); }

		// Round floating-point value and clip it within [min, max]
		// This function is more integer-overflow-safe than either doing clip and then round
		// or vice versa.
		template<class T, class INTT> VTINLINE INTT ClipNRound( T x, INTT min, INTT max ){ 
			if( x-T(kRounder)<=T(min) )
				return min;
			if( x+T(kRounder)>=T(max) )
				return max;
			register INTT ret = (x>=0)?INTT(x+T(kRounder)) : INTT(x-T(kRounder)); 
			// The reason for double-clipping is handling of weird numbers, 
			// like NaN, INF, and etc. They somehow pass the first range check 
			// (at least in MSC) but they turn into -INT_MAX or INT_MAX upon
			// conversion to integer.
			return Clip( ret, min, max );
		}

		// Safe counter addition.
		// Adds 'toadd' to 'counter', if 'counter' is less than 'maxValue'.
		// Also does overflow check, so if T overflows after (counter+toadd) then
		// value of 'counter' is not reduced, but changed to maxValue.
		// Note that 'counter' may become greater than maxValue after this function! 
		// 'toadd' must be non-negative!
		template<class T> VTINLINE void AddToCounter( T& counter, T toadd, T maxValue = T(1000000000) ){
			VTASSERT( toadd>=T(0) );
			if( counter<maxValue ){
				register T tmp = counter + toadd;
				if( tmp>=counter ) // If toadd==0, then this branch should be executed.
					counter = tmp;
				else // tmp became less than counter => overflow
					counter = maxValue;
			} // else maximum is already reached => do nothing
		}

		// Float to float round-up, unsigned
		template<class T> VTINLINE T UFRound( T x ){ return floor(x+T(kRounder)); }

		// Milliseconds to samples
		VTINLINE TAudioStreamSize ms2samples(double msec, double srate){ 
			return TAudioStreamSize(ceil(srate * (msec / 1000.0))); }

		// Calculate Least Common Multiple and Greatest Common Divisor of two
		// positive numbers.
		// The results are written to the buffers pointed by lcm & gcd. 
		// The pointers can be NULL, if you don't need either LCM or GCD.
		// In case the result cannot be computed, the function returns false. That
		// may happen if a or b is 0 or if LCM is too big for 32 bit integer (and
		// lcm is not NULL).
		bool LCM_and_GCD(SDWORD a, SDWORD b, SDWORD* lcm, SDWORD* gcd);
		// Find Greatest Common Divisor of a and b. Returns 0 on error.
		SDWORD GCD(SDWORD a, SDWORD b);


		// Sinc function. There are two different definitions of Sinc, one with
		// multiplication of argument by Pi (commonly used in DSP), so called
		// "Normalized Sinc", and one without (as in math, "Unnormalized"). 

		template<class T> VTINLINE T UnnormSinc( T x, T eps ){ return (x>=-eps && x<=eps)? T(1.0): T(sin(x)/(x)); }
		template<class T> VTINLINE T UnnormSinc( T x ){ return UnnormSinc( x, T(1e-20) ); }
		
		// Specialized templates. Optimal values 'eps' constants can be calculated using 
		// SincEps program located in Utilites directory. There is a very tiny effect on
		// speed from these constants, however.
		// !!!!! Currently the constants are for float == 32-bit IEEE, 
		// double and long double == 64-bit IEEE. !!!!!!
		template<> VTINLINE float UnnormSinc<float>( float x ){ 
			return UnnormSinc( x, 9.999989742937e-6f ); }
		template<> VTINLINE double UnnormSinc<double>( double x ){ 
			return UnnormSinc( x, 2.1491173166727281e-8 ); }
		template<> VTINLINE TLongDouble UnnormSinc<TLongDouble>( TLongDouble x ){ 
			return UnnormSinc( x, TLongDouble(2.1491173166727281e-8) ); }	
		
		template<class T> VTINLINE T NormSinc( T x, T eps ){ return UnnormSinc( T(x*kPi), eps ); }
		template<class T> VTINLINE T NormSinc( T x ){ return UnnormSinc( T(x*kPi) ); }

		//
		// Ustinov's volume calculator has 2 parameters: input volume controller,
		// ranging from 0 to 1, and sensitivity of volume to the parameter.
		// sensitivity01 == 1 => return value = controller01 (volume fully depends on the controller).
		// sensitivity01 == 0 => return value is 1 (volume is not changed by the controller).
		//
		template<class T> VTINLINE T UstinovVolumeCalculator( T controller01, T sensitivity01 ){
			// CC' := CC * k + (1-k)
			//     k := 1 => CC' == CC (changes from 0 to 1)
			//     k := 0 => CC' == 1 
			//     k := 0.5 => CC' == 0.5 * CC + 0.5 (changes from 0.5 to 1)
			return controller01 * sensitivity01 + (T(1.0) - sensitivity01);
		}


		// For logarithmic handlers, convert [0..1] <-> [0..1] using log scale.
		// The scale can be adjusted via 'mod' parameter. kDefaultLogMod seems to be a good all-rounder.
		// You can use CalculateLogMod() to calculate 'mod' which will give a better handling.

		const double kDefaultLogMod = kSqrt2;

		VTINLINE double ControlHandleLog( double x ){ return log10(x); }
		VTINLINE double ControlHandlePow( double x ){ return pow(10.0, x); }

		VTINLINE double ControlHandle_01_Log(double x01, double mod=kDefaultLogMod){ 
			return Limit01( ControlHandleLog((x01 * (ControlHandlePow(mod)-1.0))+1.0) / mod ); }
		VTINLINE double ControlHandle_Log_01(double log01, double mod=kDefaultLogMod){ 
			return Limit01( (ControlHandlePow(log01 * mod)-1.0) / (ControlHandlePow(mod)-1.0) ); }

		// Calculate mod value which gives ControlHandle_01_Log(x, mod) => y
		// The function uses bisection algorithm and smin and smax determine area 
		// of search for result.
		double CalculateLogMod(double x, double y, double precision=0.01, double smin=0.1, double smax=10.0);

		// Round x from [0..1] to nearest value of gridcount element grid
		template<class T> VTINLINE T GridRound01(T x, T gridcount){ return UFRound(x*gridcount)/gridcount; }

		// Converts string to numeric value and writes to 'output'; 
		// if failed to parse a number from the string, then if first 
		// nonspace character of the string is minus ("-") 'output' 
		// is set to -kMin_dB; otherwise (if string was not understood 
		// at all), 'output' is not changed.
		// This function should not be changed for VST SDK compatibility.
		void VisibleTextToVisible_dB(VTCHAR *string, float &output);

		// This class provides a calculator for frequency response graphs,
		// which can convert pixels to frequency scale in Hertz and vice versa, 
		// using logarithmic mapping.
		class CPix2Hz
		{
		public:
			SDWORD minHz, maxHz, minPix, maxPix;	

			CPix2Hz(): logMod(kSqrt2) {}
			CPix2Hz(SDWORD mih, SDWORD mah, SDWORD mip, SDWORD map): 
				logMod(kSqrt2) { init(mih, mah, mip, map); }
			void init(SDWORD mih, SDWORD mah, SDWORD mip, SDWORD map){ 
				minHz=mih; maxHz=mah; minPix=mip; maxPix=map; }

			// Adjust logarithmic scale, so 'hz' frequency is located at 
			// pixel row 'pix' ('pix' is between minPix and maxPix).
			void adjustLogMod(double hz, double pix){
				logMod = CalculateLogMod(hzhandle(hz), pixhandle(pix));
			}
			
			// It may look strange that pixel values can be floating point...
			// Don't be scared, these are used for interpolation stuff :-)

			VTINLINE double hz2pix_f(double hz){
				double pix = ControlHandle_01_Log( hzhandle(hz), logMod );
				return pix*(maxPix-minPix)+minPix;
			}

			VTINLINE SDWORD hz2pix(double hz){ return URound(hz2pix_f(hz)); }

			VTINLINE double pix2hz(double pix){		
				double hz = ControlHandle_Log_01( pixhandle(pix), logMod );
				return hz*(maxHz-minHz)+minHz;		
			}
		private:
			double logMod;

			// Convert hz into [0..1] scale
			VTINLINE double hzhandle(double hz){
				//return (Clip(hz, double(minHz), double(maxHz))-double(minHz))/double(maxHz-minHz+1);
				return Limit01( (hz-double(minHz)) / (maxHz-minHz) );
			}

			// Convert pixels into [0..1] scale
			VTINLINE double pixhandle(double pix){
				//return (Clip(pix, double(minPix), double(maxPix))-double(minPix))/double(maxPix-minPix+1);
				return Limit01( (pix-double(minPix)) / (maxPix-minPix) );
			}
		};









		//////////////////////////////////////////////////////////////////////////////
		//////////////////////////////////////////////////////////////////////////////
		//
		// Implementation of template & inline functions
		//


		double Convert_01TodB_NoClip(double value){ 
			if( value < kMin_dB_01 ) // Prevent logarithm overflow (when entered 0)
				return kMin_dB; 
			else{
				return double(20.0 * log10 (value)); 	
			}
		}

		double Convert_dBTo01_NoClip(double db){ 
			if( db>kMax_dB )
				return kMax_dB_01;
			else
				return double(pow(10.0, db/20.0)); 
		}
	
	} // namespace Math
} // namespace

#if !defined(VT_STRICT_NAMESPACES)
	using namespace VT::Math;
#endif


#endif


