//-----------------------------------------------------------------------------
//
// Basic Audio Types and Conversion Routines
//
//-----------------------------------------------------------------------------

/*
	VT Framework License (BSL 1.0)

	Copyright (C) 2006 Sergey A. Galin 
	E-mail: sergey.galin@gmail.com

	Permission is hereby granted, free of charge, to any person or organization
	obtaining a copy of the software and accompanying documentation covered by
	this license (the "Software") to use, reproduce, display, distribute,
	execute, and transmit the Software, and to prepare derivative works of the
	Software, and to permit third-parties to whom the Software is furnished to
	do so, all subject to the following:

	The copyright notices in the Software and this entire statement, including
	the above license grant, this restriction and the following disclaimer,
	must be included in all copies of the Software, in whole or in part, and
	all derivative works of the Software, unless such copies or derivative
	works are solely in the form of machine-executable object code generated by
	a source language processor.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
	SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
	FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
	ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
	DEALINGS IN THE SOFTWARE.
*/

#if !defined(VT_AUDIOTYPES_HPP)
#define VT_AUDIOTYPES_HPP

#include "vt_audiomath.hpp"
#include "vt_arrays.hpp"
#include "vt_dithering.hpp"

namespace VT {

	const int kMaxAudioSampleRate	= 192000;
	const int kMaxAudioChannels		= 8;



	// Audio sample types definition.
	// Meaning of the suffixes:
	//   F - floating point
	//   S - signed fixed point
	//   U - unsigned fixed point
	typedef SWORD			TSample16S;
	typedef SDWORD			TSample32S;
	typedef float			TSample32F;
	typedef double			TSample64F;
	typedef signed char		TSample8S;
	typedef unsigned char	TSample8U;	
	typedef TSample32S		TSample20S; // 24-bit data can be stored in DWORD's
	typedef TSample32S		TSample24S; // 20-bit data can be stored in DWORD's

	enum TSampleTypeIdx {
		kInt8S,
		kInt8U,
		kInt16S,
		kInt20S,
		kInt24S,
		kInt32S,
		kFP32,
		kFP64,
		kUnknownSampleType
	};

	const SWORD			kInt16SMax = 32767;
	const SDWORD		kInt20SMax = 524287;
	const SDWORD		kInt24SMax = 8388607;
	const SDWORD		kInt32SMax = 2147483647;
	const TSample32F	kFP32Max = 1.0f;
	const TSample64F	kFP64Max = 1.0;

	size_t GetAudioTypeSize( TSampleTypeIdx idx );
	inline bool IsFPFormat( TSampleTypeIdx idx ){ return idx==kFP32 || idx==kFP64; }
	inline bool IsIntFormat( TSampleTypeIdx idx ){ return !IsFPFormat(idx); }
	

	// Convert 24 bit sample (as byte sequence) to 32 bit integer.
	VTINLINE SDWORD ConvertSample24to32( const unsigned char *data ){
		register SDWORD s = SDWORD( *(data++) )<<8;
		s |= SDWORD( *(data++) )<<16;
		s |= SDWORD( *(data++) )<<24;
		return s;
	}

	VTINLINE SDWORD ConvertSample20to32( const unsigned char *data ){ return ConvertSample24to32( data ); }

	// Convert 32 bit integer sample to 24 bit sequence without dithering.
	// For dithering, include vt_dithering.hpp, convert 32 bits to double
	// and then to 24 bits using VT_64F_24S_Ditherer, or implement
	// integer dithering... whatever.
	VTINLINE void ConvertSample32to24( TSample32S sample, unsigned char *out ){
		sample >>= 8; *(out++) = (unsigned char)sample;
		sample >>= 8; *(out++) = (unsigned char)sample;
		sample >>= 8; *out = (unsigned char)sample;
	}

	// Convert audio sample: 32 bit to 20 bit (as 3-byte sequence).
	VTINLINE void ConvertSample32to20( TSample32S sample, unsigned char *out ){ ConvertSample32to24( sample, out ); }

	// Convert audio sample: 32 bit to 24 bit sequence dropping higher byte.
	VTINLINE void ConvertSample32to24NoScaling( TSample32S sample, unsigned char *out ){
		*(out++) = (unsigned char)sample;
		sample >>= 8; *(out++) = (unsigned char)sample;
		sample >>= 8; *out = (unsigned char)sample;
	}

	// Convert audio sample: 32 bit to 20 bit sequence dropping higher byte.
	VTINLINE void ConvertSample32to20NoScaling( TSample32S sample, unsigned char *out ){
		ConvertSample32to24NoScaling( sample, out ); }

	// Unsigned 8 bit sample to signed and vice versa. This is useful
	// because 8-bit WAV files are unsigned, while everything good is signed.
	VTINLINE TSample8S ConvertSample8Uto8S( TSample8U src ){ return (TSample8S)(SWORD(src)-128); }
	VTINLINE TSample8U ConvertSample8Sto8U( TSample8S src ){ return (TSample8U)(SWORD(src)+128); }

	// Conversion of SIGNED integer samples to floating-point.
	// For quality reverse conversion, use CDithering.
	template<class T> VTINLINE T ConvertSample16toFP( TSample16S s ){ return T(s)/T(32768.0); }
	template<class T> VTINLINE T ConvertSample20toFP( TSample20S s ){ return T(s)/T(524288.0); }
	template<class T> VTINLINE T ConvertSample24toFP( TSample24S s ){ return T(s)/T(8388608.0); }
	template<class T> VTINLINE T ConvertSample32toFP( TSample32S s ){ return T(s)/T(2147483648.0); }
	template<class T> VTINLINE T ConvertSample8toFP( TSample8U s ){ return T(ConvertSample8Uto8S(s))/T(128.0); }

	
	//
	// Defining dithering classes for all standard audio types
	//
	typedef CDitherer<TSample32F, TSample16S, -32767, 32767> VT_32F_16S_Ditherer;
	typedef CDitherer<TSample64F, TSample16S, -32767, 32767> VT_64F_16S_Ditherer;
	typedef CDitherer<TSample32F, TSample32S, -2147483647, 2147483647> VT_32F_32S_Ditherer;
	typedef CDitherer<TSample64F, TSample32S, -2147483647, 2147483647> VT_64F_32S_Ditherer;
	typedef CDitherer<TSample32F, TSample8S, -127, 127> VT_32F_8S_Ditherer;
	typedef CDitherer<TSample64F, TSample8S, -127, 127> VT_64F_8S_Ditherer;

	// For 20 and 24 bits, use 32 bit integer for storage; then the highest bit
	// should be dropped if necessary.
	typedef CDitherer<TSample32F, TSample32S, -524287, 524287> VT_32F_20S_Ditherer;
	typedef CDitherer<TSample64F, TSample32S, -524287, 524287> VT_64F_20S_Ditherer;
	typedef CDitherer<TSample32F, TSample32S, -8388607, 8388607> VT_32F_24S_Ditherer;
	typedef CDitherer<TSample64F, TSample32S, -8388607, 8388607> VT_64F_24S_Ditherer;


	//
	// Block audio data converters
	//
	typedef void(*TPAudioConverterFunc)( const void*, void*, size_t );

	// Anything to 32-bit float
	void ConvertAudio8U_32F( const void* src, void *dest, size_t nSamples );
	void ConvertAudio8S_32F( const void* src, void *dest, size_t nSamples );
	void ConvertAudio16S_32F( const void* src, void *dest, size_t nSamples );
	void ConvertAudio20S_32F( const void* src, void *dest, size_t nSamples );
	void ConvertAudio24S_32F( const void* src, void *dest, size_t nSamples );
	void ConvertAudio32S_32F( const void* src, void *dest, size_t nSamples );
	void ConvertAudio64F_32F( const void* src, void *dest, size_t nSamples );

	// Anything to 64-bit float
	void ConvertAudio8U_64F( const void* src, void *dest, size_t nSamples );
	void ConvertAudio8S_64F( const void* src, void *dest, size_t nSamples );
	void ConvertAudio16S_64F( const void* src, void *dest, size_t nSamples );
	void ConvertAudio20S_64F( const void* src, void *dest, size_t nSamples );
	void ConvertAudio24S_64F( const void* src, void *dest, size_t nSamples );
	void ConvertAudio32S_64F( const void* src, void *dest, size_t nSamples );
	void ConvertAudio32F_64F( const void* src, void *dest, size_t nSamples );

	// 32-bit float to integer (no dithering!)
	// For dithered conversion, use CDitherer template.
	void ConvertAudio32F_8U( const void* src, void *dest, size_t nSamples );
	void ConvertAudio32F_8S( const void* src, void *dest, size_t nSamples );
	void ConvertAudio32F_16S( const void* src, void *dest, size_t nSamples );
	void ConvertAudio32F_20S( const void* src, void *dest, size_t nSamples );
	void ConvertAudio32F_24S( const void* src, void *dest, size_t nSamples );
	void ConvertAudio32F_32S( const void* src, void *dest, size_t nSamples );

	// 64-bit float to integer (no dithering!)
	// For dithered conversion, use CDitherer template.
	void ConvertAudio64F_8U( const void* src, void *dest, size_t nSamples );
	void ConvertAudio64F_8S( const void* src, void *dest, size_t nSamples );
	void ConvertAudio64F_16S( const void* src, void *dest, size_t nSamples );
	void ConvertAudio64F_20S( const void* src, void *dest, size_t nSamples );
	void ConvertAudio64F_24S( const void* src, void *dest, size_t nSamples );
	void ConvertAudio64F_32S( const void* src, void *dest, size_t nSamples );

	// Convert anything to floating point
	void ConvertAudioTo32F( const void* src, TSampleTypeIdx srcType, void *dest, size_t nSamples );
	void ConvertAudioTo64F( const void* src, TSampleTypeIdx srcType, void *dest, size_t nSamples );
	
	// Stub functions (do memcpy).
	void ConvertAudio8BitStub( const void* src, void *dest, size_t nSamples );
	void ConvertAudio16BitStub( const void* src, void *dest, size_t nSamples );
	void ConvertAudio20BitStub( const void* src, void *dest, size_t nSamples );
	void ConvertAudio24BitStub( const void* src, void *dest, size_t nSamples );
	void ConvertAudio32BitStub( const void* src, void *dest, size_t nSamples );
	void ConvertAudio64BitStub( const void* src, void *dest, size_t nSamples );

	// Get pointer to sample conversion function.
	// This can be used to avoid switch & function call in ConvertAudioToXXX() 
	// and speed up conversion (very little though, unless audio blocks are
	// very small).	
	TPAudioConverterFunc GetAudioConverterTo32F( TSampleTypeIdx srcType );
	TPAudioConverterFunc GetAudioConverterTo64F( TSampleTypeIdx srcType );
	TPAudioConverterFunc GetAudioConverterFrom32F( TSampleTypeIdx destType );
	TPAudioConverterFunc GetAudioConverterFrom64F( TSampleTypeIdx destType );
	TPAudioConverterFunc GetAudioConverterStub( TSampleTypeIdx thetype );

	//
	// A handy class which converts any sample type to any, 
	// maintaining internal buffer for intermediate floating-point data.
	//
	class CAudioBitConverter
	{
	public:
		CAudioBitConverter();
		CAudioBitConverter( TSampleTypeIdx srcType, TSampleTypeIdx destType, size_t bufferPrealloc = 0,
			TDitheringMethod dithering = kDitheringNone, unsigned nDitheringChannels = 1 );
		
		// Initialize source and destination conversion types.
		// 'dither' can be set to enable dithering (valid only if output 
		// format is integer).
		void Init( TSampleTypeIdx srcType, TSampleTypeIdx destType, 
			TDitheringMethod dither = kDitheringNone, unsigned nDitheringChannels = 1 );

		// If converter is working with dithering for multiple audio files, 
		// it should be restarted between different files.
		void Restart();
		
		// Pre-allocate buffers for nSamples block size. Do it after Init().
		void PreallocBuffers( size_t nSamples );		
		
		// Convert block
		void Convert( const void *src, void *dest, size_t nSamples );
		
		// Do we use 64 bit for conversion? That makes no sense if
		// input and output formats are low-precision.
		// 4-byte IEEE float has 8-bit exponent and 23-bit mantissa,
		// so calculation which involve less than 24 bits (8, 16, 20)
		// can be performed in 32 bit float.
		bool Uses64Bit() const { 
			return 
				m_SrcType==kFP64 || m_DestType==kFP64 || 
				m_SrcType==kInt32S || m_DestType==kInt32S || 
				m_SrcType==kInt24S || m_DestType==kInt24S; }

		bool UsesDirectConversion() const;
		bool UsesDithering() const { 
			return m_DitheringMethod!=kDitheringNone && IsIntFormat(m_DestType) && !UsesDirectConversion(); }

		bool IsInitialized() const { return m_pfInConverter!=NULL; }		
	private:
		TSampleTypeIdx m_SrcType, m_DestType;
		TPAudioConverterFunc m_pfInConverter;
		TPAudioConverterFunc m_pfOutConverter;
		CArray<TSample32F> m_Buffer32;
		CArray<TSample64F> m_Buffer64;
		unsigned m_uChannels;
		TDitheringMethod m_DitheringMethod;

		VT_32F_16S_Ditherer		m_Dither32F_16S[kMaxAudioChannels];
		VT_64F_16S_Ditherer		m_Dither64F_16S[kMaxAudioChannels];
		VT_32F_32S_Ditherer		m_Dither32F_32S[kMaxAudioChannels];
		VT_64F_32S_Ditherer		m_Dither64F_32S[kMaxAudioChannels];
		VT_32F_8S_Ditherer		m_Dither32F_8S[kMaxAudioChannels];
		VT_64F_8S_Ditherer		m_Dither64F_8S[kMaxAudioChannels];
		VT_32F_20S_Ditherer		m_Dither32F_20S[kMaxAudioChannels];
		VT_64F_20S_Ditherer		m_Dither64F_20S[kMaxAudioChannels];
		VT_32F_24S_Ditherer		m_Dither32F_24S[kMaxAudioChannels];
		VT_64F_24S_Ditherer		m_Dither64F_24S[kMaxAudioChannels];
	};




	// 
	// Audio frame template 
	//
	// T - sample type
	// NCH - number of channels
	template<class T, int NCH> struct TAudioFrame
	{
		T channels[NCH];	
		
		VTINLINE TAudioFrame(){} // Default constructor does nothing
		
		VTINLINE TAudioFrame(T x){ Set(x); }	

		template<class U> VTINLINE void Zero(){ 
			for(register int i=0; i<NCH; i++) channels[i]=T(0); }
		
		template<class U> VTINLINE void Set(U x){ 
			for(register int i=0; i<NCH; i++) channels[i]=T(x); }
		
		template<class U> VTINLINE void Add( const TAudioFrame<U, NCH>& add ){
			for(register int i=0; i<NCH; i++) channels[i] = T( channels[i]+add.channels[i] ); }

		VTINLINE void AddScalar( T add ){
			for(register int i=0; i<NCH; i++) channels[i] += add; }
		
		template<class U> VTINLINE TAudioFrame<T, NCH>& operator+=( const TAudioFrame<U, NCH>& add ){ 
			Add( add ); return *this; }
		
		template<class U> VTINLINE void Mul( U mul ){ 
			for(register int i=0; i<NCH; i++) channels[i] = T( channels[i]*mul ); }

		template<class U> VTINLINE void Mul( const TAudioFrame<U, NCH> mul ){
			for(register int i=0; i<NCH; i++) channels[i] = T( channels[i]*mul.channels[i] ); }
		
		template<class U> VTINLINE TAudioFrame<T, NCH>& operator*=( U mul ){  
			Mul( mul ); return *this; }

		template<class U, class R> VTINLINE void AddMultiplied( const TAudioFrame<U, NCH>& add, R k ){
			for(register int i=0; i<NCH; i++) channels[i]+=T(add.channels[i]*k); }
		
		template<class U, class R> VTINLINE void SetMultiplied( const TAudioFrame<U, NCH>& add, R k ){
			for(register int i=0; i<NCH; i++) channels[i]=T(add.channels[i]*k); }

		template<class U, class R> VTINLINE void MultiplyAndAddMultiplied( R kmine, const TAudioFrame<U, NCH>& add, R k ){
			for(register int i=0; i<NCH; i++) 
				channels[i] = T( channels[i]*kmine + add.channels[i]*k ); }

		template<class U, class R> VTINLINE void MultiplyAndAdd( U kmine, const R add ){
			for(register int i=0; i<NCH; i++) channels[i] = T( channels[i]*kmine + add ); }

		bool operator==(const TAudioFrame<T, NCH>& other){
			for(register int i=0; i<NCH; i++) 
				if( channels[i] != other.channels[i] )
					return false;
			return true;
		}
	};




	


} // namespace

#endif


