//-----------------------------------------------------------------------------
//
// Encryption wrappers, currently supporting Crypto++.
//
//-----------------------------------------------------------------------------

/*
	VT Framework License (BSL 1.0)

	Copyright (C) 2006 Sergey A. Galin 
	E-mail: sergey.galin@gmail.com

	Permission is hereby granted, free of charge, to any person or organization
	obtaining a copy of the software and accompanying documentation covered by
	this license (the "Software") to use, reproduce, display, distribute,
	execute, and transmit the Software, and to prepare derivative works of the
	Software, and to permit third-parties to whom the Software is furnished to
	do so, all subject to the following:

	The copyright notices in the Software and this entire statement, including
	the above license grant, this restriction and the following disclaimer,
	must be included in all copies of the Software, in whole or in part, and
	all derivative works of the Software, unless such copies or derivative
	works are solely in the form of machine-executable object code generated by
	a source language processor.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
	SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
	FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
	ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
	DEALINGS IN THE SOFTWARE.
*/

#ifndef VT_CRYPTO_HPP
#define VT_CRYPTO_HPP

#include "vt_portable.hpp"

//#define USE_CRYPTOPP551_STATIC

//
// Select which encryption library to use (we support Crypto++ 5.5.1 so far ;-))
//
#if defined(USE_CRYPTOPP551_STATIC) && !defined(USE_CRYPTOPP551)
	#define USE_CRYPTOPP551
#endif

// If no library specified, use Crypto++ 
#if !defined(USE_CRYPTOPP551) // && !defined(...)...
	#define USE_CRYPTOPP551
#endif


// Define LSString functions?
#if !defined(USE_LSSTRING) && defined(USE_LSGUI)
	#define USE_LSSTRING
#endif

#if defined(USE_LSSTRING)
	#include "LSString.h"
#endif

// -------------------------------------------------------------------------
// Prototypes
// -------------------------------------------------------------------------

// The default hash function
#define VTCRYPT_HASH_FN SHA1

#if defined(USE_CRYPTOPP551)
	#if defined(USE_CRYPTOPP551_STATIC)
		#define VTCRYPT_WEAK_ALGORITHMS_ENABLED
	#endif
#endif

namespace VT {
	namespace Crypto {
		// Something which is OK for all used hashes.
		const int 
			kHashStringSize =   132, // Max length of Base16 encoding of a hash, 
									 // plus added something for '\0'
			kDigestBufferSize =  64; // The biggest is 512 bits / 64 bytes so far

		// void RandomFill( BYTE *out, size_t size ); unimplemented

		// Increase memory allocation size 'sz' to an integer number of blocks.
		size_t AlignToBlocks( size_t sz, size_t blocksz );

		// out must have size of at least inbytes*2+1
		void Base16(const unsigned char *data, size_t inbytes, VTCHAR *out);

		// Decode BASE16 string to binary data. The input string may 
		// contain space characters.
		bool Base16Decode( LPCVTCHAR src, size_t inchars, unsigned char *out );

		// Calculate hash (in binary form). Returns real size of digest in bytes.
		// Output buffer must be at least kDigestBufferSize bytes size.
		
		static const size_t kSHA1DigestLength = 20; 
		size_t SHA1Digest(const char *data, size_t dataSize, BYTE *digestOut);
		void SHA1(const char *data, size_t dataSize, VTCHAR *out);
		
		// SHA1 produces 512 bit / 64 byte digest
		static const size_t kSHA512DigestLength = 512/8;
		size_t SHA512Digest(const char *data, size_t dataSize, BYTE *digestOut);
		void SHA512(const char *data, size_t dataSize, VTCHAR *out);

		#if defined(VTCRYPT_WEAK_ALGORITHMS_ENABLED)
			// 16 byte digest
			size_t MD5Digest(const char *data, size_t dataSize, BYTE *digestOut);
			void MD5(const char *data, size_t dataSize, VTCHAR *out);
		#endif

		namespace AES
		{
			// Key size = 256, block size = 128
			const int
				kStdKeySizeBytes		= 256/8, 
				kDefaultKeySizeBytes	= 128/8,
				kBlockSizeBytes			= 128/8;

			// Get size of a buffer needed to hold encrypted sz bytes.
			inline size_t EncryptedSize( size_t sz ){ 
				size_t al = AlignToBlocks( sz, kBlockSizeBytes ); 
				if( al == sz )
					al += kBlockSizeBytes;
				return al;
			}

			// If 'out' is NULL, then memory is allocated via 'operator new'. 
			// If 'out' is not NULL, also pass its size into out_size, and make sure
			// it has at least AlignToBlocks( insz ) bytes.
			// Returns NULL on error.
			// Upon return, *out_size contains number of bytes allocated / used in out.
			BYTE *Encrypt(const BYTE *src, size_t insz, BYTE *out, size_t *out_size, 
				const BYTE *key, int keySize = kStdKeySizeBytes );
			
			// If 'out' is NULL, then memory is allocated via 'operator new'. 
			// If 'out' is not NULL, also pass its size into out_size.
			// Returns NULL on error.
			// Upon return, *out_size contains number of bytes allocated / used in out.
			BYTE *Decrypt(const BYTE *src, size_t insz, BYTE *out, size_t *out_size, 
				const BYTE *key, int keySize = kStdKeySizeBytes );
		
		} // namespace

		// Hash something without bothering about the algorithm used.
		void Hash(const char *data, size_t dataSize, VTCHAR *out, int trim=-1);

		#if defined(USE_LSSTRING)
			LString Base16(const unsigned char *data, size_t inbytes);
			LString SHA512(const char *data, size_t dataSize);
			LString SHA512(const LString& str);
			LString SHA1(const char *data, size_t dataSize);
			LString SHA1(const LString& str);
			// Hash something without bothering about the algorithm used.
			LString Hash(const char *data, size_t dataSize);
			LString Hash(const LString& str);
		#endif

	} // namespace
} // namespace


#endif

