//-----------------------------------------------------------------------------
//
// Dithering math
//
// This file provides tools to implement audio dithering using various 
// algorithms or noise shaping for better quality of 
// floating point -> fixed point conversion.
//
//-----------------------------------------------------------------------------

/*
	VT Framework License (BSL 1.0)

	Copyright (C) 2006 Sergey A. Galin 
	E-mail: sergey.galin@gmail.com

	Permission is hereby granted, free of charge, to any person or organization
	obtaining a copy of the software and accompanying documentation covered by
	this license (the "Software") to use, reproduce, display, distribute,
	execute, and transmit the Software, and to prepare derivative works of the
	Software, and to permit third-parties to whom the Software is furnished to
	do so, all subject to the following:

	The copyright notices in the Software and this entire statement, including
	the above license grant, this restriction and the following disclaimer,
	must be included in all copies of the Software, in whole or in part, and
	all derivative works of the Software, unless such copies or derivative
	works are solely in the form of machine-executable object code generated by
	a source language processor.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
	SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
	FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
	ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
	DEALINGS IN THE SOFTWARE.
*/

#ifndef VT_DITHERING_H
#define VT_DITHERING_H

#include <math.h>
#include <stdlib.h>
#include "vt_audiomath.hpp"
#include "vt_localization.hpp"
#include "vt_assert.hpp"

namespace VT {

	enum TDitheringMethod
	{
		kDitheringNone       = 0, // No dithering - just clip and round
		kDitheringTriangular = 1, // Triangular dithering
		kDitheringShaping    = 2  // Noise shaping
	};	

	const TDitheringMethod kDitheringDefaultMethod = kDitheringShaping;

	// This class can be used to convert blocks of samples from some floating-point type
	// to some signed fixed-point type, optionally using some method of dithering.
	// Please note that noise-shaping method uses a 1-sample feedback buffer, so if
	// you want to have this method working correctly, use different resampler objects
	// for each audio stream or channel.
	// All input samples are also clipped to fit [MIX..MAX] range.
	//
	// Example of usage:
	//   
	//   VT_32F_16S_Ditherer ditherer;
	//   ditherer.SetMethod( VT_32F_16S_Ditherer::kDitheringTriangular );
	//   float inSamples[1024]; // Input buffer
	//   signed short int outSamples[1024]; // Output buffer
	//   while( /* input available */ ){
	//     int samplesRead = ... // load next block of 1..1024 samples into inSamples[]
	//     ditherer.Dither( inSamples, outSamples, samplesRead, 32767.0f );
	//     .. // write samplesRead samples from outBuffer
	//   }
	//   ditherer.Restart(); // Optionally Restart() if you want to re-use the object
	//                       // to resample another audio stream.   
	//
	template<class TFP, class TINT, long MIN, long MAX> class CDitherer
	{
	public:		
		static const int kNumberOfMethods = 3;	
	protected:
		TDitheringMethod mMethod;
		TFP mBuffer; // Buffer used for noise shaping mode only
	public: 
		CDitherer(): mMethod( kDitheringDefaultMethod ), mBuffer( 0 ){}

		// Notify that another stream is started.
		// Calling this function is recommended, but not critical.
		VTINLINE void Restart(){ mBuffer = 0; }

		VTINLINE void SetMethod( TDitheringMethod method );
		// Ignores bad method setting (replaces it with default)
		VTINLINE void SetMethodSafe( TDitheringMethod method ); 
		VTINLINE TDitheringMethod GetMethod() const { return mMethod; }

		// Process source block of floating-point samples to dest block
		// of fixed-point samples. Source samples are also multiplied by scale,
		// dithered and clipped to be within [MIX..MAX]. 
		void Dither( const TFP* source, 
			TINT* dest, size_t nSamples, TFP scale );

		// One-sample dither (note that you have to do multiplication before 
		// calling this function).
		// !!!WARNING!!!: Don't forget to scale source!
		VTINLINE TINT Dither( TFP source );

		static LPCVTCHAR GetMethodName( TDitheringMethod method );
		LPCVTCHAR GetMethodName() const { return this->GetMethodName( mMethod ); }
	};











	//////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////
	//
	// Implementation of the template class functions
	//




	//
	// IMPORTANT! Clipping must be done before conversion to target type (to avoid integer overflow).
	//
	//
	template<class TFP, class TINT, long MIN, long MAX> 
		void CDitherer<TFP, TINT, MIN, MAX>::Dither( const TFP* source, TINT* dest, size_t nSamples, TFP scale )
	{
		switch( mMethod ){
			case kDitheringNone: 
				for( register size_t i = 0; i<nSamples; i++ )              
				  *(dest++) = SRoundT<TFP, TINT>( Clip( (*(source++)) * scale, TFP(MIN), TFP(MAX)) );
				return;

			case kDitheringTriangular: 
				for( register size_t i = 0; i<nSamples; i++ ){
					register TFP value = (*(source++)) * scale;
					register TFP fract = fmod( value, TFP(1.0) );
					if( value>=TFP(0.0) ){ // Sample (and fract) are positive
						value -= fract; // Leave only integer part of value
						if( rand() < int(fract*RAND_MAX) ) // fract is "big" =>
							value += TFP(1.0); // Rounding positive value to right
					}else{ // Sample (and tail) are negative
						VTASSERT( fract<=TFP(0.0) );
						value -= fract; // Leave only integer part of value					
						if( rand() < int((-fract)*RAND_MAX) ) // // fract is "big" (by absolute value) =>
							value -= TFP(1.0); // Rounding negative value to _left_
					}
					*(dest++) = SRoundT<TFP, TINT>( Clip( value, TFP(MIN), TFP(MAX) ));
				}
				return;

			case kDitheringShaping: 
				// Noise Shaping adds error of previous rounding from next sample to the current sample.
				// Previous sample error is stored in mBuffer to be saved between processing blocks.
				for( register size_t i = 0; i<nSamples; i++ ){				
					VTASSERT( mBuffer>=TFP(-1.0) && mBuffer<=TFP(1.0) ); // Check that error doesn't grow
					register TFP value = Clip(( (*(source++)) * scale ) + mBuffer, TFP(MIN), TFP(MAX));
					register TINT rounded = SRoundT<TFP, TINT>( value );
					mBuffer = value - TFP(rounded);
					*(dest++) = rounded;
				}
				return;
		}
		throw _T("Invalid dithering mode");
	}

	template<class TFP, class TINT, long MIN, long MAX> 
		VTINLINE TINT CDitherer<TFP, TINT, MIN, MAX>::Dither( TFP source )
	{
		switch( mMethod ){
			// case kDitheringNone: please see after the switch.
			case kDitheringTriangular: {
				// This check is solely to handle bad numbers, like +-INF, IND, NaN...
				if( source < TFP(MIN-1) ) 
					return MIN;
				if( source > TFP(MAX)+1 ) 
					return MAX;
				if( !(source>0) && !(source<0) ) // NaN
					return 0;

				register TFP fract = fmod( source, TFP(1.0) );
				register TFP value = source;
				if( value>=TFP(0.0) ){ // Sample (and fract) are positive
					value -= fract; // Leave only integer part of value
					if( rand() < int(fract*RAND_MAX) ) // fract is "big" =>
						value += TFP(1.0); // Rounding positive value to right
				}else{ // Sample (and tail) are negative
					VTASSERT( fract<=TFP(0.0) );
					value -= fract; // Leave only integer part of value					
					if( rand() < int((-fract)*RAND_MAX) ) // // fract is "big" (by absolute value) =>
						value -= TFP(1.0); // Rounding negative value to _left_
				}			
				register TINT ret = ClipNRound<TFP, TINT>( value, MIN, MAX );
				VTASSERT( ret>=MIN && ret<=MAX );
				VTASSERT( (ret>=0 && source>=0) || (ret<=0 && source<=0) );
				return ret;
			}
			case kDitheringShaping: {			
				register TFP value = Clip(source + mBuffer, TFP(MIN), TFP(MAX));
				register TINT rounded = ClipNRound<TFP, TINT>( value, MIN, MAX );			
				#if defined(_DEBUG)
				/*	if( Abs(value - TFP(rounded)) > TFP(1.0) ){
						OADEBUGMSG(LString(_T("Abs(value - TFP(rounded)) > TFP(1.0)\r\n"))<<
							_T("Source: ")<<double(source)<<_T("\r\n")<<
							_T("Value: ")<<double(value)<<_T("\r\n")<<
							_T("Rounded: ")<<int(rounded)<<_T("\r\n")<<
							_T("mBuffer: ")<<double(this->mBuffer));
						mBuffer = TFP(0.0);
					} */
				#endif
				mBuffer = value - TFP(rounded);
				// Check that error doesn't grow and that the rounding works correctly
				if( mBuffer<TFP(-1.0) || mBuffer>TFP(1.0) ){
					/* OADEBUGMSG(LString(_T("mBuffer<TFP(-1.0) || mBuffer>TFP(1.0)\r\n"))<<
						_T("Source: ")<<double(source)<<_T("\r\n")<<
						_T("Rounded: ")<<int(rounded)<<_T("\r\n")<<
						_T("mBuffer: ")<<double(this->mBuffer)); */
					mBuffer = TFP(0.0);
				}
				return rounded;
			}		
						  
		}
		// Default: no dithering
		return SRoundT<TFP, TINT>( Clip( source, TFP(MIN), TFP(MAX)) );
	}


	template<class TFP, class TINT, long MIN, long MAX> 
		LPCVTCHAR CDitherer<TFP, TINT, MIN, MAX>::GetMethodName( TDitheringMethod method )
	{
		switch( method ){
			case kDitheringNone: return LCT("No Dithering");
			case kDitheringTriangular: return LCT("Triangular");
			case kDitheringShaping: return LCT("Noise Shaping");
		}
		throw _T("Invalid dithering mode");
	}

	template<class TFP, class TINT, long MIN, long MAX> 
		void CDitherer<TFP, TINT, MIN, MAX>::SetMethod( TDitheringMethod method )
	{
		if( method>=kNumberOfMethods || method<0 )
			throw _T("Invalid dithering mode");
		mMethod = method;
		this->Restart();
	}

	template<class TFP, class TINT, long MIN, long MAX> 
		void CDitherer<TFP, TINT, MIN, MAX>::SetMethodSafe( TDitheringMethod method )
	{
		if( method>=kNumberOfMethods || method<0 )
			method = kDitheringDefaultMethod;
		mMethod = method;
		this->Restart();
	}

} // namespace

#endif
