//-----------------------------------------------------------------------------
//
// Filesystem abstraction layer.
//
//-----------------------------------------------------------------------------

/*
	VT Framework License (BSL 1.0)

	Copyright (C) 2006 Sergey A. Galin 
	E-mail: sergey.galin@gmail.com

	Permission is hereby granted, free of charge, to any person or organization
	obtaining a copy of the software and accompanying documentation covered by
	this license (the "Software") to use, reproduce, display, distribute,
	execute, and transmit the Software, and to prepare derivative works of the
	Software, and to permit third-parties to whom the Software is furnished to
	do so, all subject to the following:

	The copyright notices in the Software and this entire statement, including
	the above license grant, this restriction and the following disclaimer,
	must be included in all copies of the Software, in whole or in part, and
	all derivative works of the Software, unless such copies or derivative
	works are solely in the form of machine-executable object code generated by
	a source language processor.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
	SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
	FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
	ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
	DEALINGS IN THE SOFTWARE.
*/


#include "vt_file.hpp"
#include "vt_fileutils.hpp"

using namespace VT;

#if defined(VT_WINDOWS)
	inline size_t GetResourceSize(HRSRC hRes, HMODULE hModule)
	{
		return size_t( SizeofResource(hModule, hRes) );
	}
#endif

size_t CReseekingFileRedirector::read( void* ptr, size_t size, size_t n )
{ 
	if( !IsOK() )
		return 0;
	VTFSIZE savePos = m_pFile->tell();
	if( savePos!=m_iPosition ){
		if( m_pFile->seek( m_iPosition )!=0 ){
			m_pFile->seek( savePos );
			return false;
		}
	}
	size_t ret = m_pFile->read( ptr, size, n ); 
	m_iPosition = m_pFile->tell();
	m_pFile->seek( savePos );
	return ret;
}

size_t CReseekingFileRedirector::write( const void* ptr, size_t size, size_t n )
{ 
	if( !IsOK() )
		return 0;
	VTFSIZE savePos = m_pFile->tell();
	if( savePos!=m_iPosition ){
		if( m_pFile->seek( m_iPosition )!=0 ){
			m_pFile->seek( savePos );
			return false;
		}
	}
	size_t ret = m_pFile->write( ptr, size, n ); 
	m_iPosition = m_pFile->tell();
	m_pFile->seek( savePos );
	return ret;
}

int CReseekingFileRedirector::seek( VTFSIZE offset, int origin )
{ 
	if( !IsOK() )
		return 1;
	VTFSIZE savePos = m_pFile->tell();
	if( origin==SEEK_CUR ){
		if( m_pFile->seek( m_iPosition )!=0 ){
			m_pFile->seek( savePos );
			return 1;
		}
	}
	bool ret = m_pFile->seek( offset, origin )==0;
	if( ret )
		m_iPosition = m_pFile->tell();
	m_pFile->seek( savePos );
	return 0;
}

bool CReseekingFileRedirector::eof() const 
{ 
	if( !IsOK() )
		return true;
	VTFSIZE savePos = m_pFile->tell();
	if( m_pFile->seek( m_iPosition )!=0 ){
		m_pFile->seek( savePos );
		return true;
	}
	bool ret = m_pFile->eof(); 
	m_pFile->seek( savePos );
	return ret;
}







CMemoryFile::CMemoryFile(size_t sz, size_t alloc, size_t growSize):
	m_uGrowSize( growSize ),
	CArray<BYTE>( sz, alloc ),
	m_uPointer( 0 )
{
	ZeroMemoryBytes();
}


CMemoryFile::CMemoryFile( const CMemoryFile& srcFile ):
	CArray<BYTE>( srcFile.getMFSize() ),
	m_uPointer( 0 )
{
	// Copy array
	copyFrom( srcFile );
	m_uGrowSize = srcFile.m_uGrowSize;
}

size_t CMemoryFile::read( void* ptr, size_t size, size_t n )
{ 
	if( m_uPointer>getMFSize() )
		m_uPointer = getMFSize();
	size_t nbytes = size*n;
	// If we can't read as many as requested, 
	// then reduce number of records to read.
	if( m_uPointer + nbytes > getMFSize() ){
		n = (getMFSize() - m_uPointer) / size; 
		nbytes = size*n; // Updating read size in bytes
	}
	if( nbytes ){
		memcpy( ptr, getRODataPtr() + m_uPointer, nbytes );
		m_uPointer += nbytes;
	}
	return n;
}

size_t CMemoryFile::write( const void* ptr, size_t size, size_t n )
{ 
	if( m_uPointer>getSize() )
		m_uPointer = getMFSize();
	if( size==0 || n==0 )
		return 0;
	size_t nbytes = size*n;
	size_t end = m_uPointer + nbytes ;
	if( end>getMFSize() )
		resize( end, true );
	memcpy( getDataPtr() + m_uPointer, ptr, nbytes );
	m_uPointer += nbytes;
	return n;
}

// returns 0 on success; origin = SEEK_SET/SEEK_CUR/SEEK_END
int CMemoryFile::seek( VTFSIZE offset, int origin )
{ 
	VTFSIZE newPos;
	switch( origin ){
	case SEEK_SET:
		newPos = offset;
		break;
	case SEEK_CUR:
		newPos = offset + VTFSIZE(m_uPointer);
		break;
	case SEEK_END:
		newPos = getSize() + VTFSIZE(m_uPointer);
		break;
	default:
		return 1; // Error: invalid origin
	}
	if( newPos<0 || newPos>getSize() )
		return 1; // Invalid resulting position
	m_uPointer = size_t( newPos );
	return 0;
}

bool CMemoryFile::load( CFile *file, VTFSIZE size )
{
	if( size==0 ){
		Truncate();
		return true;
	}
	// If size<0, then whole rest of the file is loaded
	if( size<0 ){
		size = file->getSize() - file->tell();
		if( size<0 ){
			Truncate();
			return false;
		}
	}
	resize( size_t( size ), false );
	return file->read( getDataPtr(), getMFSize() );
}

bool CMemoryFile::save( CFile *file ) const
{
	return file->write( getRODataPtr(), getMFSize() );
}

bool CMemoryFile::load( LPCVTCHAR fn )
{
	CDiskFile file( fn, _T("rb") );
	if( !file.IsOK() )
		return false;
	return load( &file );
}

bool CMemoryFile::save( LPCVTCHAR fn ) const
{
	CDiskFile file( fn, _T("wb") );
	if( !file.IsOK() )
		return false;
	return save( &file );
}

void CMemoryFile::Resize( VTFSIZE newLengthSz )
{
	size_t newLength = size_t( newLengthSz ), oldSize = getMFSize();
	resize( newLength, true );
	if( newLength>oldSize )
		memset( getDataPtr()+oldSize, 0, newLength-oldSize );
	if( m_uPointer > newLength )
		m_uPointer = newLength;
}

void CMemoryFile::Fill( BYTE fill )
{
	size_t sz = getMFSize();
	if( sz )
		memset( getDataPtr(), fill, sz );
}


#if defined(VT_WINDOWS)



bool CMemoryFile::LoadResource(UINT iRes, LPCTSTR lpType, HMODULE hModule)
{
	Truncate();
	LPCTSTR lpszResourceName = MAKEINTRESOURCE (iRes);
	if( lpszResourceName == NULL ){
		VTDEBUGMSGT("lpszResourceName == NULL, bad iRes passed to LoadResource()?");
		return false;
	}
	HRSRC hRsrc = ::FindResource(hModule, lpszResourceName, lpType);
	return LoadResource(hRsrc, hModule);
}

bool CMemoryFile::LoadResource(HRSRC hRes, HMODULE hModule)
{
	Truncate();
	size_t rsize = GetResourceSize(hRes, hModule);
	if( !rsize ){
		VTDEBUGMSGT("Zero resource size!");
		return false;
	}
	HGLOBAL hMem = ::LoadResource(hModule, hRes);
	if( hMem ){
		void* resPtr = LockResource(hMem);
		if( resPtr ){
			resize( rsize );
			memcpy( getDataPtr(), resPtr, rsize );			
		}
		FreeResource( hMem );
		return true;
	}else{
		VTDEBUGMSGT("LoadResource() failed!");
		return false;
	}
	seek( 0 );
	return true;
}

#endif






bool CDirectory:: IsCaseSensitive()
{
	#if defined(VT_FS_CASEINSENSITIVE)
		return false;
	#else
		return true;
	#endif
}


CFilesystemDirectory::CFilesystemDirectory( LPCVTCHAR path )
{
	Open( path );
}

bool CFilesystemDirectory::Open( LPCVTCHAR path )
{
	if( path==NULL )
		m_Path[0]=0;
	else{
		VTSTRNCPY( m_Path, path, MAX_PATH );
		VT::KillEndingSlash( m_Path );
	}
	if( VTSTRLEN( m_Path ) )
		return VT::IsDirectory( m_Path );
	else
		return false;
}

bool CFilesystemDirectory::GetFileList( CStringArray *out )
{
	if( VT::FileExists(m_Path) ){
		CString mask( m_Path );
		mask << VTPATHSLASHT << VT_ALLFILESMASKT;
		VT::CDirectoryReader reader( mask );
		size_t nfiles = reader.GetFileCount();
		out->resize( nfiles, false );
		size_t outfiles=0;
		for( size_t i=0; i<nfiles; i++ ){
			LPCVTCHAR fn = reader.GetFile(i);
			if( fn[0] &&
				VTSTRCMP( fn, _T(".")) &&
				VTSTRCMP( fn, _T("..")) ){
				out->get(outfiles++).CopyFrom( fn );				
			}
		}
		out->resize( outfiles, true );
		return true;
	}else{
		out->resize( 0 );
		return false;
	}
}

bool CFilesystemDirectory::IsFileExists( LPCVTCHAR fileName )
{
	CString fullPath( m_Path );
	fullPath<<VTPATHSLASHT<<fileName;
	return VT::FileExists( fullPath );
}


bool CFilesystemDirectory::IsSubdirectory( LPCVTCHAR fileName )
{
	CString fullPath( m_Path );
	fullPath<<VTPATHSLASHT<<fileName;
	bool ret = VT::IsDirectory( fullPath );
	return ret;
}

CFile* CFilesystemDirectory::OpenFile( LPCVTCHAR fileName, LPCVTCHAR mode )
{
	CString fullPath( m_Path );
	fullPath<<VTPATHSLASHT<<fileName;
	CDiskFile *file = new CDiskFile( fullPath, mode );
	if( file!=NULL ){
		if( !file->IsOK() ){ // Did not open correctly
			delete file;
			return NULL;
		}			
	}
	return file;
}

CFile* CFilesystemDirectory::OpenFileForReading( LPCVTCHAR fileName )
{
	return OpenFile( fileName, _T("rb") );
}

CFile* CFilesystemDirectory::OpenFileForWriting( LPCVTCHAR fileName )
{
	return OpenFile( fileName, _T("wb") );
}


CDirectory* CFilesystemDirectory::OpenSubdirectory( LPCVTCHAR fileName )
{
	VTASSERT( fileName!=NULL );
	VTASSERT( VTSTRLEN(fileName)>0 );

	if( !IsSubdirectory( fileName ) )
		return NULL;
	
	CString fullPath( m_Path );
	fullPath<<VTPATHSLASHT<<fileName;
	CFilesystemDirectory *dir = new CFilesystemDirectory( fullPath );
	return dir;
}

VTFSIZE CFilesystemDirectory::GetFileSize( LPCVTCHAR fileName )
{
	CString fullPath( m_Path );
	fullPath<<VTPATHSLASHT<<fileName;
	return VT::GetFileSize( fullPath );
}




bool VT::PumpFileContents( CFile *src, CFile *dest, VTFSIZE maxBytes, size_t bufferSize )
{
	if( !src->IsOK() || !dest->IsOK() )
		return false;
	if( maxBytes==0 )
		return true;
	CArray<BYTE> buffer( bufferSize );
	VTFSIZE count = 0, maxBytesLeft = maxBytes;
	size_t toRead = bufferSize;
	while( !src->eof() ){
		if( maxBytes>0 )
			if( maxBytesLeft<VTFSIZE(toRead) ){
				toRead = size_t( maxBytesLeft );
				maxBytesLeft -= toRead;
			}
		size_t readBytes = src->read( buffer.getDataPtr(), 1, toRead );
		if( readBytes ){
			if( dest->write( buffer.getRODataPtr(), 1, readBytes ) != readBytes )
				return false; // Write error!
		}else
			break;
	}
	return true;
}

bool VT::PumpFileContents( CFile *src, LPCVTCHAR destName, VTFSIZE maxBytes, size_t bufferSize )
{
	CDiskFile file( destName, _T("wb") );
	if( !file.IsOK() )
		return false;
	return PumpFileContents( src, &file, maxBytes, bufferSize );
}

bool VT::PumpFileContents( LPCVTCHAR srcName, CFile *dest, VTFSIZE maxBytes, size_t bufferSize )
{
	CDiskFile file( srcName, _T("rb") );
	if( !file.IsOK() )
		return false;
	return PumpFileContents( &file, dest, maxBytes, bufferSize );
}


