//-----------------------------------------------------------------------------
//
// Filesystem abstraction layer.
//
//-----------------------------------------------------------------------------

/*
	VT Framework License (BSL 1.0)

	Copyright (C) 2006 Sergey A. Galin 
	E-mail: sergey.galin@gmail.com

	Permission is hereby granted, free of charge, to any person or organization
	obtaining a copy of the software and accompanying documentation covered by
	this license (the "Software") to use, reproduce, display, distribute,
	execute, and transmit the Software, and to prepare derivative works of the
	Software, and to permit third-parties to whom the Software is furnished to
	do so, all subject to the following:

	The copyright notices in the Software and this entire statement, including
	the above license grant, this restriction and the following disclaimer,
	must be included in all copies of the Software, in whole or in part, and
	all derivative works of the Software, unless such copies or derivative
	works are solely in the form of machine-executable object code generated by
	a source language processor.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
	SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
	FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
	ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
	DEALINGS IN THE SOFTWARE.
*/

#ifndef VT_FILE_H
#define VT_FILE_H

#include <stdio.h>
#include "vt_portable.hpp"
#include "vt_string.hpp"

#if defined(_MSC_VER) && !defined(_CRT_SECURE_NO_DEPRECATE)
	#define _CRT_SECURE_NO_DEPRECATE
#endif


namespace VT{

	class CString;
	class CStringArray;

	//
	// The basic interface of a File.
	// Function prototypes are mostly based on usual stdio stuff.
	// Note there are no functions for opening, closing, and etc. in this class yet.
	// Just the functions which must be available for data readers/writers.
	//
	class CFile
	{
	public:	
		CFile(){}
		virtual ~CFile(){}

		// Descendant classes should normally overload the following two versions.
		// Like standard C functions, they receive record size and number of
		// records to read and try to coupe with as many records as they can;
		// the returned result is the number of records successfully read/written.
		virtual size_t read( void* ptr, size_t size, size_t n ) = 0; // fread-like behaviour
		virtual size_t write( const void* ptr, size_t size, size_t n ) = 0; // fwrite-like behaviour

		virtual bool read( void* ptr, size_t size ){ return read(ptr, size, 1)==1; } // Returns false on error
		virtual bool write( const void* ptr, size_t size ){ return write(ptr, size, 1)==1; } // Returns false on error

		virtual int seek( VTFSIZE offset, int origin=SEEK_SET ) = 0; // returns 0 on success; origin = SEEK_SET/SEEK_CUR/SEEK_END
		virtual VTFSIZE tell() const = 0; // Tell file pointer position
		virtual VTFSIZE getSize() =0; // Get file size
		virtual bool IsOK() const = 0; // Is the file opened successfully?
		virtual bool eof() const = 0; // Is file pointer at end of the file?
	};


	//
	// File redirector - passes all operations to another file object.
	// When file redirector is deleted, the target file stays.
	//
	class CFileRedirector: public CFile
	{
	public:
		CFileRedirector( CFile *file=NULL ): m_pFile( file ){}
		void SetFile( CFile *file ){ m_pFile = file; }
		virtual size_t read( void* ptr, size_t size, size_t n ){ return m_pFile->read( ptr, size, n ); }
		virtual size_t write( const void* ptr, size_t size, size_t n ){ return m_pFile->write( ptr, size, n ); }
		virtual int seek( VTFSIZE offset, int origin=SEEK_SET ){ return m_pFile->seek( offset, origin ); }
		virtual VTFSIZE tell() const { return m_pFile->tell(); }
		virtual VTFSIZE getSize() { return m_pFile->getSize(); }
		virtual bool IsOK() const { if( m_pFile==NULL ) return false; else return m_pFile->IsOK(); }
		virtual bool eof() const { return m_pFile->eof(); }
	protected:
		CFile *m_pFile;
	};

	//
	// Re-seeking file redirector - passes all operations to another file object.
	// When file redirector is deleted, the target file stays.
	// Keeps internal file pointer and restores file position before each operation;
	// this makes it slow, but multiple processes can have access to file (as long
	// as they don't do that simultaneously; no support for built-in blocking file via
	// critical section yet!!!!!!!!!!!)
	//
	class CReseekingFileRedirector: public CFile
	{
	public:
		CReseekingFileRedirector( CFile *file=NULL ): m_pFile( file ), m_iPosition( 0 ){}
		void SetFile( CFile *file ){ m_pFile = file; }
		
		virtual size_t read( void* ptr, size_t size, size_t n );
		virtual size_t write( const void* ptr, size_t size, size_t n );
		virtual int seek( VTFSIZE offset, int origin=SEEK_SET );
		virtual bool eof() const;
		
		virtual VTFSIZE tell() const { return m_iPosition; }
		virtual VTFSIZE getSize() { return m_pFile->getSize(); }
		virtual bool IsOK() const { if( m_pFile==NULL ) return false; else return m_pFile->IsOK(); }
		
	protected:
		CFile *m_pFile;
		VTFSIZE m_iPosition;
	};



	//
	// FILE* I/O wrapper (turns FILE* into CFile).
	// The FILE pointer must be opened and closed somewhere outside of the class.
	//
	class CFILEFile: public CFile
	{
	public:
		CFILEFile(FILE *f): fp(f){}
		virtual ~CFILEFile(){}

		virtual size_t read( void* ptr, size_t size, size_t n ){ return fread( ptr, size, n, fp ); }
		virtual size_t write( const void* ptr, size_t size, size_t n ){ return fwrite( ptr, size, n, fp ); }
		virtual int seek( VTFSIZE offset, int origin=SEEK_SET ){ return VTFSEEK( fp, offset, origin ); }
		virtual VTFSIZE tell() const { return VTFTELL( fp ); } 
			
		virtual VTFSIZE getSize(){
			// Measure file by seeking to its end and then telling the position.
			// This method is ugly but it works anywhere. System-dependent methods
			// can be added later using #ifdefs.
			VTFSIZE pos = tell();
			seek( 0, SEEK_END );
			VTFSIZE ret = tell();
			seek( pos, SEEK_SET );
			return ret;
		}

		virtual bool IsOK() const{ return fp!=NULL; }
		virtual bool eof() const { return feof(fp)!=0; }

	protected:
		FILE *fp;
	};


	//
	// Standard filesystem file, opened via filename and 
	//
	class CDiskFile: public CFILEFile
	{
	public:
		CDiskFile( LPCVTCHAR name, LPCVTCHAR mode ):
		  CFILEFile( VTFOPEN(name, mode) )
		{}
	    
		virtual ~CDiskFile(){
			if( fp!=NULL ){
				fclose( fp );
				fp = NULL;
			}
		}
	};


	class CMemoryFile: public CFile, protected CArray<BYTE>
	{
	public:
		static const size_t kDefGrowSize = 1024;

		CMemoryFile(size_t sz = 0, size_t alloc = kDefGrowSize, size_t growSize = kDefGrowSize);	
		CMemoryFile( const CMemoryFile& srcFile ); // Copy constructor!!

		virtual size_t GetGrowSize() const { return m_uGrowSize; }
		virtual void SetGrowSize( size_t sz ){ m_uGrowSize = sz; }

		virtual size_t read( void* ptr, size_t size, size_t n );
		virtual size_t write( const void* ptr, size_t size, size_t n );
		virtual int seek( VTFSIZE offset, int origin=SEEK_SET );
		
		virtual VTFSIZE tell() const { return VTFSIZE(m_uPointer); } 		
		virtual VTFSIZE getSize(){ return VTFSIZE(CArray<BYTE>::getSize()); }

		// Since all memory files have size_t size and can return their
		// size without altering any fileds, this size_t and const function
		// is provided.
		virtual size_t getMFSize() const { return CArray<BYTE>::getSize(); } 
		
		virtual bool IsOK() const{ return true; }
		virtual bool eof() const { return m_uPointer>=CArray<BYTE>::getSize(); }
		
		virtual void Truncate( VTFSIZE newLength=0 ){ Resize(0); }
		virtual void Resize( VTFSIZE newLength );
		virtual void Fill( BYTE fill=0 );

		// Memory file can be loaded/saved from/to another file.
		virtual bool load( CFile *file, VTFSIZE size = -1 ); // If size<0, then whole rest of the file is loaded
		virtual bool save( CFile *file ) const;
		virtual bool load( LPCVTCHAR fn );
		virtual bool save( LPCVTCHAR fn ) const;

		const BYTE* getRODataPtr() const { return CArray<BYTE>::getRODataPtr(); }
		BYTE *getDataPtr(){ return CArray<BYTE>::getDataPtr(); }

		#if defined(VT_WINDOWS)
			virtual bool LoadResource(UINT iRes, LPCTSTR lpType=_T("MEMFILE"), HMODULE hModule=NULL);
			virtual bool LoadResource(HRSRC hRes, HMODULE hModule);	
		#endif

	protected:	
		size_t m_uGrowSize, m_uPointer;
	};




	class CDirectory
	{
	public:
		
		// Default implememtation returns true under Windows and false under *NIX
		virtual bool IsCaseSensitive();
		
		// Read file list to out (out is emptied before reading file list).
		virtual bool GetFileList( CStringArray *out )=0;

		// Verify that file exists
		virtual bool IsFileExists( LPCVTCHAR fileName )=0;

		// Is it a sub-directory or file?
		virtual bool IsSubdirectory( LPCVTCHAR fileName )=0;

		// Open a file and returns pointer to CFile descendant.
		// The returned object must be deleted by the caller.
		// May return NULL in case of error.
		virtual CFile* OpenFileForReading( LPCVTCHAR fileName )=0;
		virtual CFile* OpenFileForWriting( LPCVTCHAR fileName )=0;
		
		virtual CDirectory* OpenSubdirectory( LPCVTCHAR fileName )=0;

		virtual LPCVTCHAR GetPath() const = 0;	

		virtual VTFSIZE GetFileSize( LPCVTCHAR fileName )=0;
	};


	//
	// Normal filesystem directory. Heavily uses VT stuff.
	//
	class CFilesystemDirectory: public CDirectory
	{
	public:
		CFilesystemDirectory( LPCVTCHAR path=NULL );
		bool Open( LPCVTCHAR path );
		virtual bool GetFileList( CStringArray *out );
		virtual bool IsFileExists( LPCVTCHAR fileName );
		virtual bool IsSubdirectory( LPCVTCHAR fileName );
		virtual CFile* OpenFile( LPCVTCHAR fileName, LPCVTCHAR mode );
		virtual CFile* OpenFileForReading( LPCVTCHAR fileName );
		virtual CFile* OpenFileForWriting( LPCVTCHAR fileName );
		virtual CDirectory* OpenSubdirectory( LPCVTCHAR fileName );
		virtual LPCVTCHAR GetPath() const { return m_Path; }
		virtual VTFSIZE GetFileSize( LPCVTCHAR fileName );
	protected:
		VTCHAR m_Path[MAX_PATH+1]; // Without ending slash!
		
	};

	//
	// Global functions
	//
	// Read data from source file and write it to destination file
	bool PumpFileContents( CFile *src, CFile *dest, VTFSIZE maxBytes = -1, size_t bufferSize=65536 );
	bool PumpFileContents( CFile *src, LPCVTCHAR destName, VTFSIZE maxBytes = -1, size_t bufferSize=65536 );
	bool PumpFileContents( LPCVTCHAR srcName, CFile *dest, VTFSIZE maxBytes = -1, size_t bufferSize=65536 );

} // namespace

#endif



