//-----------------------------------------------------------------------------
//
// Cross-platform disk filesystem functions
//
//-----------------------------------------------------------------------------

/*
	VT Framework License (BSL 1.0)

	Copyright (C) 2006 Sergey A. Galin 
	E-mail: sergey.galin@gmail.com

	Permission is hereby granted, free of charge, to any person or organization
	obtaining a copy of the software and accompanying documentation covered by
	this license (the "Software") to use, reproduce, display, distribute,
	execute, and transmit the Software, and to prepare derivative works of the
	Software, and to permit third-parties to whom the Software is furnished to
	do so, all subject to the following:

	The copyright notices in the Software and this entire statement, including
	the above license grant, this restriction and the following disclaimer,
	must be included in all copies of the Software, in whole or in part, and
	all derivative works of the Software, unless such copies or derivative
	works are solely in the form of machine-executable object code generated by
	a source language processor.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
	SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
	FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
	ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
	DEALINGS IN THE SOFTWARE.
*/

#if !defined(_CRT_SECURE_NO_DEPRECATE)
	#define _CRT_SECURE_NO_DEPRECATE
#endif

#include "vt_fileutils.hpp"

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#if defined(VT_WINDOWS)
	#include <direct.h>
	#include <shlobj.h>
	#include <Shellapi.h>
	#include <io.h>
	#include <errno.h>
	#include <locale>
#else
	#include <sys/types.h>
	#include <sys/stat.h>
	#include <errno.h>
	#include <dirent.h>
	#include <fcntl.h>
	#include <unistd.h>
#endif

using namespace VT;

void VT::CutoffFileName(LPVTCHAR out)
{
	if(!out) return;
	register LPVTCHAR p;
	for(p=out; *p; p++);
	for(; (*p)!=VTPATHSLASHCHT; p--)
		if(p==out){
			*out=0;
			return;
		}
	*(++p) = 0;
}


void VT::CutoffFileExt(LPVTCHAR out)
{
	if(!out) return;
	register LPVTCHAR p;
	for(p=out; *p; p++); // Seek to the trailing '\0'
	for(; (*p)!=_T('.'); p--){
		if(p==out) 
			return;
		if(*p==VTPATHSLASHCHT)
			return;
	}
	if( p==out ) // save .xxxx filenames
		return;
	if( p>out )
		if( *(p-1)==VTPATHSLASHCHT )
			return;
	*p = 0; // Kill the dot
}

void VT::ExtractFileName(LPCVTCHAR in, LPVTCHAR out)
{
	if(!out) return;
	if(!in){ *out=0; return; }
	register LPCVTCHAR p;
	for(p=in; *p; p++);
	for(; (*p)!=VTPATHSLASHCHT; p--)
		if(p==in)
			break;
	if( (*p) == VTPATHSLASHCHT )
		p++;
	for(; *p; p++)
		*(out++) = *p;
	*out = 0;		
}

void VT::ExtractFileNameWOExt(LPCVTCHAR in, LPVTCHAR out)
{
	ExtractFileName(in, out);
	CutoffFileExt(out);
}

void VT::ExtractFileExt(LPCVTCHAR in, LPVTCHAR out)
{
	if(!out) return;
	if(!in){ *out=0; return; }
	register LPCVTCHAR p;
	for(p=in; *p; p++);
	for(; (*p)!=_T('.'); p--)
		if(p==in || *p==VTPATHSLASHCHT){
			*out = 0;
			return;
		}
	if( p==in ){
		*out = 0;
		return;
	}
	if( p>in )
		if( *(p-1)==VTPATHSLASHCHT ){
			*out = 0;
			return;
		}
	if( (*p) == _T('.') )
		p++;
	for(; *p; p++)
		*(out++) = *p;
	*out = 0;	
}

void VT::ExtractDirectoryName(LPCVTCHAR in, LPVTCHAR out)
{
	if(!out)
		return;
	if(!in){
		*out = 0;
		return;
	}
	VTSTRNCPY( out, in, MAX_PATH );
	CutoffFileName( out );
}

bool VT::PathStartsWithDirUp(LPCVTCHAR str)
{
	if( *str != _T('.') )
		return false;
	if( *(++str) != _T('.') )
		return false;
	if( *(++str) != VTPATHSLASHCHT )
		return false;
	return true;
}

// Convert absolute file path fileName into path relative to startDir.
bool VT::RelativePathFromAbs(LPCVTCHAR startDir, LPCVTCHAR fileName, LPVTCHAR out)
{
	if(!out) return false;
	if(!startDir || !fileName){ *out=0; return false; }
	if( !IsAbsolutePath(startDir) ||
	    !IsAbsolutePath(fileName) ||
	    !EndsWithSlash(startDir) ){
		// It can't handle that :-(
		VTSTRNCPY(out, fileName, MAX_PATH);
		return false;
	}
	LPCVTCHAR sd = startDir, fn = fileName;
	// Find the last character which is the same for both sd and fn
	for(; *sd && *fn && *fn==*sd; fn++, sd++);
	// Now, step both pointers back until the last VTPATHSLASHCHT
	for(; fn>fileName && *fn!=VTPATHSLASHCHT; fn--, sd--);
	// If the paths have nothing common
	// (But for paths starting with slash we still assume they have the same root dir)
	if( fn==fileName && *fn!=VTPATHSLASHCHT ){ 
		VTSTRNCPY(out, fileName, MAX_PATH);
		return true;
	}
	// If the file is under the directory (then sd points to the last slash of startDir)
	if( !*(sd+1) ){
		VTSTRNCPY(out, ++fn, MAX_PATH);
		return true;
	}
	// Some dirs in startDir and fileName are common.
	*out=0;
	for(++sd; *sd; sd++)
		if( *sd==VTPATHSLASHCHT ) 
			VTSTRNCAT(out, _T("..") VTPATHSLASHT, MAX_PATH);
	VTSTRNCAT(out, ++fn, MAX_PATH);
	
	return true;
}

// Convert relative file path fileName into absolute path, using startDir as the start.
// It doesn't just concatenate startDir and fileName, but also removes ..\ elements
// from path.
bool VT::AbsolutePath(LPCVTCHAR startDir, LPCVTCHAR fileName, LPVTCHAR out)
{
	if(!out) return false;
	if(!startDir || !fileName){ *out=0; return false; }
	// fileName is an absolute path by itself => ignore startDir
	if( IsAbsolutePath(fileName) ){
		VTSTRNCPY(out, fileName, MAX_PATH);
		return true;
	}

	// Combine the paths
	LPCVTCHAR fn = fileName;
	VTSTRNCPY(out, startDir, MAX_PATH);
	while( PathStartsWithDirUp(fn) ){
		if(!(*out)){
			VTSTRNCPY(out, fileName, MAX_PATH);
			return false;
		}
		// Remove one directory name from out and one ..\ from fileName.
		KillEndingSlash(out);
		CutoffFileName(out);
		fn+=3;
	}
	EndsWithSlash(out);
	VTSTRNCAT(out, fn, MAX_PATH);
	return true;
}

bool VT::RemoveDirectory(LPCVTCHAR dirName)
{
	if(!dirName) return true;
	if(!dirName[0]) return true;
	#if defined(VT_WINDOWS)

		VTCHAR searchMask[MAX_PATH+1], subDir[MAX_PATH+1];
		bool addSlash; 
		size_t dirNameLastChar;
		WIN32_FIND_DATA fileData;
	
		// Creating search mask
		VTSTRNCPY( searchMask, dirName, MAX_PATH );
		if( dirName[ dirNameLastChar = (VTSTRLEN(dirName)-1) ] != _T('\\') ){
			// dirName doesn't end with a slash, so we have to add it when 
			// creating sub-searches and etc.
			addSlash = true;
			VTSTRNCAT( searchMask, _T("\\"), MAX_PATH );
		}else
			addSlash = false;
		VTSTRNCAT( searchMask, _T("*"), MAX_PATH );
		
		HANDLE hFile = FindFirstFile(searchMask, &fileData);
		if (hFile != INVALID_HANDLE_VALUE){
			if (fileData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY){		
				if( fileData.cFileName[0] != _T('.') ){	
					VTSTRNCPY( subDir, dirName, MAX_PATH );
					if( addSlash )
						VTSTRNCAT( subDir, _T("\\"), MAX_PATH );
					VTSTRNCAT( subDir, fileData.cFileName, MAX_PATH );
					VTSTRNCAT( subDir, _T("\\"), MAX_PATH );
					RemoveDirectory( subDir );
				}
			}else{
				VTSTRNCPY( subDir, dirName, MAX_PATH );
				if( addSlash )
					VTSTRNCAT( subDir, _T("\\"), MAX_PATH );
				VTSTRNCAT( subDir, fileData.cFileName, MAX_PATH );
				VTUNLINK( subDir );
			}

			while (FindNextFile(hFile, &fileData) != NULL)
				if (fileData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY){		
					if( fileData.cFileName[0] != _T('.') ){	
						VTSTRNCPY( subDir, dirName, MAX_PATH );
						if( addSlash )
							VTSTRNCAT( subDir, _T("\\"), MAX_PATH );
						VTSTRNCAT( subDir, fileData.cFileName, MAX_PATH );
						VTSTRNCAT( subDir, _T("\\"), MAX_PATH );
						RemoveDirectory( subDir );
					}
				}else{
					VTSTRNCPY( subDir, dirName, MAX_PATH );
					if( addSlash )
						VTSTRNCAT( subDir, _T("\\"), MAX_PATH );
					VTSTRNCAT( subDir, fileData.cFileName, MAX_PATH );
					VTUNLINK( subDir );
				}	
			
			FindClose(hFile);
		}

		// Destroying the original directory.
		// Remove "\\" symbol, using searchMask as we don't need its value anymore.
		VTSTRNCPY( searchMask, dirName, MAX_PATH );
		if( addSlash )
			searchMask[ dirNameLastChar ] = 0;
		return VTRMDIR( searchMask )==0;
	#else
		perror("RemoveDirectory() is not implemented!\n");
	#endif
}


void VT::EndWithSlash(LPVTCHAR out)
{
	if(!out) return;
	size_t len=VTSTRLEN(out);
	if(len) 
		if( out[len-1] == VTPATHSLASHCHT )
			return;
	if(!len || len >= MAX_PATH-1) 
		return;
	out[len] = VTPATHSLASHCH;
	out[len+1] = 0;
}


bool VT::EndsWithSlash(LPCVTCHAR str)
{
	if(!str) return false;
	size_t len=VTSTRLEN(str);
	if(len)
		return str[len-1]==VTPATHSLASHCHT;
	else
		return false;
}
	
bool VT::KillEndingSlash(LPVTCHAR out)
{
	if(!out) return false;
	size_t len=VTSTRLEN(out);
	if(len)
		if(out[len-1]==VTPATHSLASHCHT){
			out[len-1]=0;
			return true;
		}
	return false;
}

bool VT::IsAbsolutePath(LPCVTCHAR path)
{
	if(!path) return false;
	if(!path[0]) return false;
	#if defined(VT_WINDOWS)
		// Windows: how the hell can we do that? Let's make
		// some guessing work. If it starts with <letters>:<anything>,
		// it must be absolute.		
		if( path[0]==_T('/') || path[0]==_T('\\') ) // See the comment in .h filr
			return true;
		// OK, it may be a device name
		int i;
		for(i=0; path[i] && isalpha(path[i]); i++);
		if( path[i]==_T(':') ) 
			return true;
		return false;
	#else
		// UNIX: here, it's very simple ;-))
		return path[0]==VTPATHSLASHCH;
	#endif
}

#if defined(VT_WINDOWS)
void VT::GetProgramFilesDir(LPVTCHAR out)
{
	if(!out) return;
	if(!SHGetSpecialFolderPath(NULL, out, CSIDL_PROGRAM_FILES, false))
		VTSTRNCPY(out, _T("C:\\Program Files\\"), MAX_PATH);
	else
		EndWithSlash(out);
}
#endif

void VT::GetTempDir(LPVTCHAR out)
{
	if(!out) return;
	#if defined(VT_WINDOWS)
		if(!GetTempPath(MAX_PATH, out))
			GetExeDir(out);
		else
			EndWithSlash(out);
	#else
		const char *var = getenv("TMP");
		if(!var)
			VTSTRNCPY(out, "/tmp/", MAX_PATH);
		else{
			for(; *var && *var!='='; var++);
			if( *var=='=' ){
				VTSTRNCPY(out, ++var, MAX_PATH);
				EndWithSlash(out);
			}else
				VTSTRNCPY(out, "/tmp/", MAX_PATH);
		}
	#endif
}

void VT::GetApplicationDataDir(LPVTCHAR out, bool systemWide)
{
	if(!out) return;
	#if defined(VT_WINDOWS)
		if(!SHGetSpecialFolderPath(NULL, out, 
			(systemWide)? CSIDL_COMMON_APPDATA : CSIDL_APPDATA, 
			false)){
			// Failed to get application data directory, falling back
			// to directory where the program executable file is located.
			GetExeDir(out);
		}else
			EndWithSlash(out);
	#else
		GetUserDocDir(out);
	#endif
}

void VT::GetUserDocDir(LPVTCHAR out)
{
	if(!out) return;
	#if defined(VT_WINDOWS)
		if(!SHGetSpecialFolderPath(NULL, out, CSIDL_PERSONAL, false))
			VTSTRNCPY(out, _T("C:\\My Documents\\"), MAX_PATH);
		else
			EndWithSlash(out);
	#else
		// Return $HOME, if it's not set - just "~/"
		const char *var = getenv("HOME");
		if(!var)
			VTSTRNCPY(out, "~/", MAX_PATH);
		else{
			for(; *var && *var!='='; var++);
			if( *var=='=' ){
				VTSTRNCPY(out, ++var, MAX_PATH);
				EndWithSlash(out);
			}else
				VTSTRNCPY(out, "~/", MAX_PATH);
		}
	#endif
}


void VT::GetExeName(LPVTCHAR out)
{
	if(!out) return;
	#if defined(VT_WINDOWS)
		/*
		int pNumArgs;
		#if defined(UNICODE)
			LPVTCHAR *argvs = CommandLineToArgvW( GetCommandLineW(), &pNumArgs );
			if( pNumArgs ){
				VTSTRNCPY(out, argvs[0], MAX_PATH);
			}else
				out[0]=0;
		#else
			LPWSTR *argvsw = CommandLineToArgvW( GetCommandLineW(), &pNumArgs );
			if( pNumArgs ){
				char buf[MAX_PATH+1];
				BOOL charused;
				// CP_OEMCP ??
				WideCharToMultiByte(CP_ACP, 0, argvsw[0], -1, buf, MAX_PATH, NULL, &charused);
			}else
				out[0]=0;			
		#endif
		*/
		if( GetModuleFileName(NULL, out, MAX_PATH)==0 )
			*out = 0;
	#else
		if(vt_argc && vt_argv)
			VTSTRNCPY(out, vt_argv[0], MAX_PATH);
		else{
			VTCheckGlobalArgs();
			out[0]=0;
		}
	#endif
}

void VT::GetExeDir(LPVTCHAR out)
{	
	if(!out) return;
	#if defined(VT_WINDOWS)
		GetExeName(out);
		if( IsAbsolutePath(out) ){
			// Delete executable name, leave directory name only
			for(int i=int(VTSTRLEN(out)); i>=0 && out[i]!=_T('\\'); i--)
				out[i]=0;

		}else{
			VTCHAR buf[MAX_PATH+1];
			VTSTRNCPY(buf, out, MAX_PATH);
			GetCWD(out);
			VTSTRNCAT(out, buf, MAX_PATH);
			
			// Check that our guessing is correct and the file
			// really present in the directory.
			// Set buf to the same with added ".exe" ending.
			VTSTRNCPY(buf, out, MAX_PATH);
			VTSTRNCAT(buf, _T(".exe"), MAX_PATH);
			if( FileExists( out ) || FileExists( buf ) ){
				// That's cool, we found it.
				// Delete executable name, leave directory name only
				for(int i=int(VTSTRLEN(out)); i>=0 && out[i]!=_T('\\'); i--)
					out[i]=0;
			}else{
				// Can't do anything. 
				// Report the current directory.
				GetCWD(out);
			}
		}
	#else
		perror("GetExePath() is not implemented!\n");
	#endif
}

bool VT::GetCWD(LPVTCHAR out)
{
	if(!out) return false;
	VTCHAR buf[MAX_PATH+1];
	out[0] = 0;
	#if defined(VT_WINDOWS)
		#if defined(UNICODE)
			if( _wgetcwd(buf, MAX_PATH)==NULL ) 
				return false;
		#else
			if( _getcwd(buf, MAX_PATH)==NULL ) 
				return false;
		#endif
	#else
		if( getcwd(buf, MAX_PATH)==NULL ) 
			return false;
	#endif
	VTSTRNCPY(out, buf, MAX_PATH);
	EndWithSlash(out);
	return true;
}

bool VT::FileExists(LPCVTCHAR fn)
{
	if(!fn) return false;
	#if defined(VT_WINDOWS)
		#if defined(UNICODE)
			_wfinddata64_t fd;
			intptr_t h = _wfindfirst64( (wchar_t*)fn, &fd );
			if( h != -1 ){
				_findclose( h );
				return true;
			}else
				return false;
		#else
			__finddata64_t fd;
			intptr_t h = _findfirst64( fn, &fd );
			if( h != -1 ){
				_findclose( h );
				return true;
			}else
				return false;
		#endif
	#else
		#if defined(VT_64BIT_FILEOIO)
			struct stat64 st;
			return (stat64(fn, &st) == 0);
		#else
			struct stat st;
			return (stat(fn, &st) == 0);
		#endif
	#endif
}

bool VT::IsDirectory( LPCVTCHAR fn )
{
	if(!fn) return false;
	#if defined(VT_WINDOWS)
		#if defined(UNICODE)
			_wfinddata64_t fd;
			intptr_t h = _wfindfirst64( (wchar_t*)fn, &fd );
			if( h != -1 ){
				_findclose( h );
				return (fd.attrib & _A_SUBDIR) == _A_SUBDIR;
			}else
				return false;
		#else
			__finddata64_t fd;
			intptr_t h = _findfirst64( fn, &fd );
			if( h != -1 ){
				_findclose( h );				
				bool ret = (fd.attrib & _A_SUBDIR) == _A_SUBDIR;
				return ret;
			}else
				return false;
		#endif
	#else
		#if defined(VT_64BIT_FILEOIO)
			struct stat64 st;
			if(stat64(fn, &st) != 0)
				return false;
		#else
			struct stat st;
			if(stat(fn, &st) != 0)
				return false;		
		#endif
		return (st.st_mode & S_IFDIR) == S_IFDIR;
	#endif
}

// This function retruns date when file was created or its contents modified.
time_t VT::FileCDate(LPCVTCHAR fn)
{
	if(!fn) return 0;
	#if defined(VT_WINDOWS)
		#if defined(UNICODE)
			_wfinddata64_t fd;
			intptr_t h = _wfindfirst64( (wchar_t*)fn, &fd );
			if( h != -1 ){
				_findclose( h );
				return fd.time_create;
			}else
				return false;
		#else
			__finddata64_t fd;
			intptr_t h = _findfirst64( fn, &fd );
			if( h != -1 ){
				_findclose( h );
				return fd.time_create;
			}else
				return false;
		#endif
	#else
		#if defined(VT_64BIT_FILEOIO)
			struct stat64 st;
			if(stat64(fn, &st) != 0)
				return 0;
		#else
			struct stat st;
			if(stat(fn, &st) != 0)
				return 0;		
		#endif
		return st.st_mtime;
	#endif
}

VTFSIZE VT::GetFileSize(LPCVTCHAR fn)
{
	if(!fn) return 0;
	#if defined(VT_WINDOWS)
		#if defined(UNICODE)
			_wfinddata64_t fd;
			intptr_t h = _wfindfirst64( (wchar_t*)fn, &fd );
			if( h != -1 ){
				_findclose( h );
				return VTFSIZE(fd.size);
			}else
				return 0;
		#else
			__finddata64_t fd;
			intptr_t h = _findfirst64( fn, &fd );
			if( h != -1 ){
				_findclose( h );
				return VTFSIZE(fd.size);
			}else
				return 0;
		#endif
	#else
		#if defined(VT_64BIT_FILEOIO)
			struct stat64 st;
			if(stat64(fn, &st) != 0)
				return 0;
		#else
			struct stat st;
			if(stat(fn, &st) != 0)
				return 0;		
		#endif
		return st.st_size;
	#endif
}

bool VT::MkDir(LPCVTCHAR dirName, int unixMode)
{
	if( !dirName ) 
		return false;
	if( FileExists( dirName ) )
		return IsDirectory( dirName );
	
	#if defined(VT_WINDOWS)
		#if defined(UNICODE)
			if( _wmkdir( dirName ) )
				return errno == EEXIST;
			else
				return true;		
		#else
			if( _mkdir( dirName ) )
				return errno == EEXIST;
			else
				return true;		
		#endif
	#else
		if( mkdir( dirName, unixMode ) )
			return errno == EEXIST;
		else
			return true;
	#endif
}

// Create temp file name. On input, prefx must contain directory
// path and filename prefix, e.g.: C:\Temp\ProgName
// Works like mktemp, i.e. prefx is used to store the name;
// but don't add 'XXXXXX' at end of the prefix like with mktemp.
// prefx must be at least MAX_PATH bytes long.
// Returns prefx on success, NULL on error.
LPVTCHAR VT::MkTemp( LPVTCHAR prefx )
{
	if(!prefx) 
		return NULL;
	VTSTRNCAT(prefx, _T("XXXXXX"), MAX_PATH-VTSTRLEN(prefx)-1);
	#if defined(_MSC_VER)
		#if defined(UNICODE)
			return _wmktemp( prefx );	
		#else
			return _mktemp( prefx );	
		#endif
	#else
		// As the most simple solution, we could use mktemp(),
		// but it's considered insecure.
		// return mktemp( prefx );
		VTCHAR tmp[MAX_PATH];
		if( UniqueName( prefx, tmp ) ){
			VTSTRNCPY( prefx, tmp, MAX_PATH );
			return prefx;
		}else
			return NULL;
	#endif
}

LPVTCHAR VT::GetTempDirecory( LPVTCHAR out )
{
	VTCHAR *temp = NULL;
	temp = VTGETENV(_T("TEMP"));
	if( temp==NULL )
		temp = VTGETENV(_T("TMP"));
	if( temp[0]==0 )
		temp = NULL;
	if( temp==NULL ){
		#if defined(VT_WINDOWS)
			temp = _T("C:\\");
		#else
			temp = _T("/tmp/");
		#endif
	}
	VTSTRNCPY(out, temp, MAX_PATH);
	EndWithSlash(out);
	return out;
}

LPVTCHAR VT::MkSysTemp( LPCVTCHAR namePrefx, LPVTCHAR out )
{
	GetTempDirecory(out);
	VTSTRNCAT(out, namePrefx, MAX_PATH-VTSTRLEN(out)-1);
	return MkTemp(out);
}

bool VT::UniqueName( LPCVTCHAR fileName, LPVTCHAR uniqueName )
{
	if( FileExists(fileName) ){
		VTCHAR dir[MAX_PATH+1], fn[MAX_PATH+1], ext[MAX_PATH+1], addon[32];
		ExtractDirectoryName(fileName, dir);
		ExtractFileNameWOExt(fileName, fn);
		ExtractFileExt(fileName, ext);
		size_t dirlen = VTSTRLEN(dir), fnlen = VTSTRLEN(fn);
		for(int attempts=1; attempts<10000; attempts++){
			size_t sz = MAX_PATH;
			VTSTRNCPY(uniqueName, dir, sz);
			VTSTRNCAT(uniqueName, fn, sz-=dirlen);
			VTSPRINTF(addon, _T("_%03d%04d"), attempts, rand()%10000);
			VTSTRNCAT(uniqueName, addon, sz-=sizeof(addon)-1);
			if( ext[0] ){
				VTSTRNCAT(uniqueName, _T("."), --sz);
				VTSTRNCAT(uniqueName, ext, sz);
			}
			if( !FileExists(uniqueName) )
				return true;
		}
		return false;
	}else{
		VTSTRNCPY(uniqueName, fileName, MAX_PATH);
		return true;
	}
}

#if defined(VT_WINDOWS)
bool VT::GetShortName( LPCVTCHAR fileName, char *out, bool bCreateFile )
{
	bool bFileCreated = false;
	if( bCreateFile && !FileExists( fileName ) ){
		FILE *f = VTFOPEN( fileName, _T("w") );
		if( f==NULL )
			return false;
		fclose(f);
		bFileCreated = true;
	}
	#if defined(UNICODE)
		TCHAR gsnshortdestMP3File[MAX_PATH];
	#else
		#define gsnshortdestMP3File out
	#endif
	if(!GetShortPathName(fileName, gsnshortdestMP3File, MAX_PATH)){
		if( bFileCreated )
			VTUNLINK( fileName );
		return false;
	}		
	#if defined(UNICODE)
		if(!WideCharToMultiByte(CP_ACP, 0, gsnshortdestMP3File, -1, out, MAX_PATH, NULL, NULL)){
			if( bFileCreated )
				VTUNLINK( fileName );
			return false;
		}
	#endif
	return true;
}
#endif


VT::CDirectoryReader::CDirectoryReader():
	stringBuffer( NULL ),
	stringPtrs( NULL ),
	stringBufferSize( 0 ),
	stringPtrsSize( 0 ), 
	stringCount( 0 ),
	nextChar( 0 ) 
{
}


// Filter is a comma-separated string with a list of allowed 
// extenstions. Note that the extenstions are not case sensitive.
bool VT::CDirectoryReader::Read(LPCVTCHAR mask, LPCVTCHAR filter)
{	
	bool bRet = true;

	// Allocating initial amount of memory
	if( stringBuffer!=NULL && stringPtrs!=NULL ){
		// Let's start with already allocated amount of memory;
		// just reset the counters.
		stringCount =0;
		nextChar = 0;
	}else{
		// Memory must be re-allocated.
		Free();
		stringBufferSize = kBufferIncrement;
		stringBuffer = LPVTCHAR( malloc( stringBufferSize*sizeof(VTCHAR) ) );
		if( !stringBuffer ){
			Free();
			throw _T("Out of memory.");
		}
		stringPtrsSize = kPointersIncrement;
		stringPtrs = (size_t*) malloc( stringPtrsSize*sizeof(size_t) );
		if( !stringPtrs ){
			Free();
			throw _T("Out of memory.");
		}
	}
	
	size_t filterStrLen, filterCount;
	LPCVTCHAR *masks; 
	LPVTCHAR filterParsed;
	if( filter==NULL ){
		filterStrLen = 0;
		filterCount = 0;
		masks = NULL;
		filterParsed = NULL;
	}else{
		filterStrLen = VTSTRLEN( filter );
		filterParsed = LPVTCHAR( malloc( (filterStrLen+1)*sizeof(VTCHAR) ) );
		masks = (LPCVTCHAR*) malloc( (filterStrLen+1)*sizeof(LPVTCHAR) );
		VTSTRNCPY( filterParsed, filter, filterStrLen+1 );
		filterCount = 0;
		for( LPVTCHAR ptr=filterParsed; *ptr; ptr++ ){
			masks[ filterCount++ ] = ptr;
			#if defined(VT_CRTDEBUG)
				printf("*** FILTER: %s\n", ptr);
			#endif
			for(; *ptr && *ptr!=_T(','); ptr++);
			if( !(*ptr) )
				break; // Otherwise ptr will be incremented over 0 by the for().
			(*ptr) = 0;
		}		
		masks[ filterCount ] = NULL;
	}


	#if defined(VT_WINDOWS)
		#if defined(UNICODE)
			_wfinddata64_t fd;
			intptr_t h = _wfindfirst64( (wchar_t*)mask, &fd );
			if( h != -1 ){		
				addFile( fd.name, masks );
				while( _wfindnext64( h, &fd )==0 )
					addFile( fd.name, masks );
				_findclose( h );
				bRet = true;
			}else
				bRet = false;
		#else
			__finddata64_t fd;
			intptr_t h = _findfirst64( mask, &fd );
			if( h != -1 ){
				addFile( fd.name, masks );
				while( _findnext64( h, &fd )==0 )
					addFile( fd.name, masks );
				_findclose( h );
				bRet = true;
			}else
				bRet = false;
		#endif		
	#else
		#warning "VT::CDirectoryReader::Read() is not implemented!"
	#endif

	if( masks!=NULL )
		free( masks );
	if( filterParsed!=NULL )
		free( filterParsed );

	#if defined(VT_CRTDEBUG)
		printf("*** Found %d files.\n", int( stringCount ) );
	#endif

	return bRet;
}

// masks is a NULL-terminated array of pointers to allowed extenstions
void VT::CDirectoryReader::addFile(LPCVTCHAR name, LPCVTCHAR *masks)
{
	// Check if we really have to add this
	if( masks!=NULL ){
		bool bMaskFound = false;
		VTCHAR ext[MAX_PATH];
		ExtractFileExt( name, ext );
		for( LPCVTCHAR *mask = masks; (*mask)!=NULL; mask++ )
			if( !VTSTRICMP(*mask,  ext ) ){
				bMaskFound = true;
				break;
			}
		if( !bMaskFound )
			return;
	}

	//printf("!!!! %s\n", name);

	// Check space in pointer buffer; increase it if necessary
	if( stringCount>=stringPtrsSize ){
		#if defined(VT_CRTDEBUG)
			printf("Increasing pointer block size from %d items.\n", int(stringPtrsSize));
		#endif
		stringPtrsSize += kPointersIncrement;
		void *newStringPtrs = realloc( stringPtrs, stringPtrsSize*sizeof(size_t) );
		if( newStringPtrs==NULL ){
			Free();
			throw _T("Out of memory.");
		}
		stringPtrs = (size_t*) newStringPtrs;
	}
	// Check space in text buffer; increase it if necessary
	size_t currentFreeBufferSize = stringBufferSize;
	if( stringCount )
		currentFreeBufferSize -= nextChar;
	size_t nLen = VTSTRLEN( name );
	// Increase buffer...
	bool bMustEatMemory = false;
	while( nLen>=currentFreeBufferSize-1 ){
		stringBufferSize += kBufferIncrement;
		currentFreeBufferSize += kBufferIncrement;
		bMustEatMemory = true;
	}
	if( bMustEatMemory ){
		#if defined(VT_CRTDEBUG)
			printf("Increasing text buffer size to %d chars.\n", int(stringBufferSize));
		#endif
		void *newStringBuffer = realloc( stringBuffer, stringBufferSize*sizeof(VTCHAR) );
		if( newStringBuffer==NULL ){
			Free();
			throw _T("Out of memory.");
		}
		stringBuffer = LPVTCHAR( newStringBuffer );
	}
	// Now we have enough space, add the string
	// Note the +1, otherwise trailing 0 not copied
	VTSTRNCPY( stringBuffer+nextChar, name, nLen+1 ); 
	stringPtrs[ stringCount ] = nextChar;
	stringCount++;
	nextChar += nLen+1;	
}

VT::CDirectoryReader::CDirectoryReader(LPCVTCHAR mask, LPCVTCHAR filter):
	stringBuffer( NULL ),
	stringPtrs( NULL ),
	stringBufferSize( 0 ),
	stringPtrsSize( 0 ), 
	stringCount( 0 ),
	nextChar( 0 )
{
	Read(mask, filter);
}


void VT::CDirectoryReader::Free()
{
	if( stringBuffer!=NULL ){
		free( stringBuffer );
		stringBuffer = NULL;
	}
	if( stringPtrs!=NULL ){
		free( stringPtrs );
		stringPtrs = NULL;
	}
	stringBufferSize = 0;
	stringPtrsSize = 0;
	stringCount = 0;
	nextChar = 0;
}














#if defined(USE_LSSTRING)

LString VT::EndWithSlash(const LString& str)
{
	if( !str.IsEmpty() ){
		LString ret;
		ret<<str;
		if(!EndsWithSlash(ret))
			ret.ConcatChar(VTPATHSLASHCHT);
		return ret;
	}else{
		return LString();
	}
}

bool VT::EndsWithSlash(const LString &str)
{
	return str.LastChar()==VTPATHSLASHCHT;
}

LString VT::KillEndingSlash(const LString &str)
{
	VTCHAR buf[MAX_PATH+1];
	VTSTRNCPY(buf, LPCTSTR(str)?LPCTSTR(str):_T(""), MAX_PATH);
	KillEndingSlash(buf);
	return LString(buf);
}

LString VT::CutoffFileName(const LString &in)
{
	VTCHAR buf[MAX_PATH+1];
	VTSTRNCPY(buf, LPCTSTR(in)?LPCTSTR(in):_T(""), MAX_PATH);
	CutoffFileName(buf);
	return LString(buf);
}

LString VT::CutoffFileExt(const LString &in)
{
	VTCHAR buf[MAX_PATH+1];
	VTSTRNCPY(buf, LPCTSTR(in)?LPCTSTR(in):_T(""), MAX_PATH);
	CutoffFileExt(buf);
	return LString(buf);
}

LString VT::ExtractFileName(LPCVTCHAR in)
{
	VTCHAR buf[MAX_PATH+1];
	ExtractFileName(in, buf);
	return LString(buf);
}

LString VT::ExtractFileNameWOExt(LPCVTCHAR in)
{
	VTCHAR buf[MAX_PATH+1];
	ExtractFileNameWOExt(in, buf);
	return LString(buf);
}

LString VT::ExtractFileExt(LPCVTCHAR in)
{
	VTCHAR buf[MAX_PATH+1];
	ExtractFileExt(in, buf);
	return LString(buf);
}

LString VT::ExtractDirectoryName(LPCVTCHAR in)
{
	VTCHAR buf[MAX_PATH+1];
	ExtractDirectoryName(in, buf);
	return LString(buf);
}

LString VT::RelativePathFromAbs(LPCVTCHAR startDir, LPCVTCHAR fileName)
{
	VTCHAR buf[MAX_PATH+1];
	RelativePathFromAbs(startDir, fileName, buf);
	return LString(buf);
}

LString VT::AbsolutePath(LPCVTCHAR startDir, LPCVTCHAR fileName)
{
	VTCHAR buf[MAX_PATH+1];
	AbsolutePath(startDir, fileName, buf);
	return LString(buf);
}


#if defined(VT_WINDOWS)
LString VT::GetProgramFilesDir()
{
	VTCHAR buf[MAX_PATH+1];
	GetProgramFilesDir(buf);
	return LString(buf);
}
#endif

LString VT::GetApplicationDataDir()
{
	VTCHAR buf[MAX_PATH+1];
	GetApplicationDataDir(buf);
	return LString(buf);
}

LString VT::GetUserDocDir()
{
	VTCHAR buf[MAX_PATH+1];
	GetUserDocDir(buf);
	return LString(buf);
}

LString VT::GetTempDir()
{
	VTCHAR buf[MAX_PATH+1];
	GetTempDir(buf);
	return LString(buf);
}

LString VT::GetExeName()
{
	VTCHAR buf[MAX_PATH+1];
	GetExeName(buf);
	return LString(buf);
}

LString VT::GetExeDir()
{
	VTCHAR buf[MAX_PATH+1];
	GetExeDir(buf);
	return LString(buf);
}

LString VT::GetCWD()
{
	VTCHAR buf[MAX_PATH+1];
	if( GetCWD(buf) )
		return LString(buf);
	else
		return LString();
}

bool VT::MkTemp(LPCVTCHAR prefx, LString &out_name)
{
	VTCHAR buf[MAX_PATH+1];
	VTSTRNCPY( buf, prefx, MAX_PATH );
	bool ret = (MkTemp(buf) == buf);
	if( ret ) 
		out_name = buf;
	return ret;
}

LString VT::GetTempDirecory()
{
	VTCHAR buf[MAX_PATH+1];
	GetTempDirecory( buf );
	return LString(buf);
}

bool VT::MkSysTemp(LPCVTCHAR namePrefx, LString &out_name)
{
	VTCHAR buf[MAX_PATH+1];
	bool ret = (MkSysTemp(namePrefx, buf) == buf);
	if(ret)
		out_name = buf;
	return ret;
}

bool VT::UniqueName( LPCVTCHAR fileName, LString &uniqueName )
{
	VTCHAR buf[MAX_PATH+1];
	if(!UniqueName(fileName, buf))
		return false;
	uniqueName = buf;
	return true;
}


#endif

