//-----------------------------------------------------------------------------
//
// Cross-platform disk filesystem functions
//
//-----------------------------------------------------------------------------

/*
	VT Framework License (BSL 1.0)

	Copyright (C) 2006 Sergey A. Galin 
	E-mail: sergey.galin@gmail.com

	Permission is hereby granted, free of charge, to any person or organization
	obtaining a copy of the software and accompanying documentation covered by
	this license (the "Software") to use, reproduce, display, distribute,
	execute, and transmit the Software, and to prepare derivative works of the
	Software, and to permit third-parties to whom the Software is furnished to
	do so, all subject to the following:

	The copyright notices in the Software and this entire statement, including
	the above license grant, this restriction and the following disclaimer,
	must be included in all copies of the Software, in whole or in part, and
	all derivative works of the Software, unless such copies or derivative
	works are solely in the form of machine-executable object code generated by
	a source language processor.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
	SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
	FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
	ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
	DEALINGS IN THE SOFTWARE.
*/


#ifndef VT_FILEUTILS_H
#define VT_FILEUTILS_H

#include "vt_portable.hpp"

#if defined(USE_LSGUI) && !defined(USE_LSSTRING)
	#define USE_LSSTRING
#endif

#if defined(USE_LSSTRING)
	#include "LSString.h"
#endif

// NOTE: all functions which return directory names return them with slash at the end,
// except for situation when a direcrory name is an empty string (then, empty string
// is returned).

namespace VT {
	// Recursively remove directory. Returns true on success and false on error.
	// Always attempts to delete as much as possible (i.e. doesn't stop on errors).
	bool RemoveDirectory(LPCVTCHAR dirName);

	#if defined(VT_WINDOWS)
		void GetProgramFilesDir(LPVTCHAR out); 
		#if defined(USE_LSSTRING)
			LString GetProgramFilesDir(); 
		#endif
	#endif

	// Adds a slash at end of out, if it doesn't end with it yet.
	// out is assumed to be at least MAX_PATH bytes.
	// Note that if 'out' is an empty string, slash will not be added.
	void EndWithSlash(LPVTCHAR out);

	// Check whether the string ends with a slash.
	// Note: returns 'false' for empty strings.
	bool EndsWithSlash(LPCVTCHAR str);
	
	// Kill ending slash in the string.
	// If 'out' is empty, do nothing.
	// Returns true if there was a slash.
	bool KillEndingSlash(LPVTCHAR out);

	// Returns true if str starts with "../" or "..\"
	bool PathStartsWithDirUp(LPCVTCHAR str);

	// Returns true for absolute path, false for relative path.
	// On Windows, paths starting with '\\' and '/' also considered absolute,
	// because these might be network paths and are incorrect as relative 
	// paths anyway. E.g.: \\SomeComputer\ShareName
	bool IsAbsolutePath(LPCVTCHAR path);

	// File name operations.
	// All input vars must be no longer than MAX_PATH, all output
	// vars must have space for at least MAX_PATH chars.
	void CutoffFileName(LPVTCHAR out); // Also kills the period: "xxxx.yyy" => "xxxx"
	void CutoffFileExt(LPVTCHAR out);
	void ExtractFileName(LPCVTCHAR in, LPVTCHAR out);
	void ExtractFileNameWOExt(LPCVTCHAR in, LPVTCHAR out);
	void ExtractFileExt(LPCVTCHAR in, LPVTCHAR out);
	void ExtractDirectoryName(LPCVTCHAR in, LPVTCHAR out);

	
	// Convert absolute file path fileName into path relative to startDir.
	// Returns false on error, e.g. if startDir or fileName are not
	// absolute paths. Also, in case of error fileName is copied to out.
	// If fileName cannot be relative to startDir (e.g. on Windows if they are
	// on different logical drives) fileName is copied to out and true is returned.
	// Warning: this function is case-sensitive, i.e. if 
	// startDir = "C:\\Something" and fileName = "c:\\something\\better",
	// it will not detect that fileName is under startDir and simply 
	// return fileName.
	// Note: startDir must end with a slash. Double slashes are not allowed.
	bool RelativePathFromAbs(LPCVTCHAR startDir, LPCVTCHAR fileName, LPVTCHAR out);

	// Convert relative file path fileName into absolute path, using startDir as the start.
	// It doesn't just concatenate startDir and fileName, but also removes ..\ elements
	// from path. Doesn't do anything with '.' directory entries yet.
	// Returns 'false' on errors. Also, on errors fileName is copied to out.
	bool AbsolutePath(LPCVTCHAR startDir, LPCVTCHAR fileName, LPVTCHAR out);

	// !! All directory name functions return string ended with a slash !!

	// Application data directory (if 'systemWide', then the directory 
	// which is common for all users; if not, the user's private app data 
	// directory). This is a place where program configuration can be stored
	// (instead of Windows System Registry).
	void GetApplicationDataDir(LPVTCHAR out, bool systemWide = false);
	
	// User documents directory
	void GetUserDocDir(LPVTCHAR out);

	// Temp directory
	void GetTempDir(LPVTCHAR out);

	// This function should not be used to determine full path
	// to the executable. Sometimes it may return just a command
	// used to run it.
	void GetExeName(LPVTCHAR out);

	// This function tries to detect full, absolute path to the
	// directory where the executable is. If it fails, it
	// returns current directory (GetCWD()).
	void GetExeDir(LPVTCHAR out);	

	// Get Current Working Directory.
	bool GetCWD(LPVTCHAR out);

	// Checks whether file or directory exists
	bool FileExists(LPCVTCHAR fn);

	// Get file creation or content modification date
	time_t FileCDate(LPCVTCHAR fn);

	// Get file creation or content modification date
	VTFSIZE GetFileSize(LPCVTCHAR fn);

	// Create a directory. Returns true on success or if it already exists.
	// (If a normal file exists with the spefied name, returns false.)
	// unixMode is fully ignored under Windows, at the moment.
	bool MkDir(LPCVTCHAR dirName, int unixMode = VTDEFMKDIRMODE);

	// Create temp file name. On input, prefx must contain directory
	// path and filename prefix, e.g.: C:\Temp\ProgName
	// Works like mktemp, i.e. prefx is used to store the name;
	// but don't add 'XXXXXX' at end of the prefix like with mktemp.
	// prefx must be at least MAX_PATH bytes long.
	// Returns prefx on success, NULL on error.
	LPVTCHAR MkTemp( LPVTCHAR prefx ); 

	// Get system temporary file directory. Always returns out.
	// The directory is padded with trailing slash if neccessary.
	// out must be at least MAX_PATH bytes in size, as usually.
	LPVTCHAR GetTempDirecory( LPVTCHAR out );

	// Cteate temp file name in system temp file directory. The name 
	// is written into out. Returns out on success, NULL on error.
	// Note that in case of error old contents of out is not 
	// preseved.
	LPVTCHAR MkSysTemp( LPCVTCHAR namePrefx, LPVTCHAR out ); 

	// Modifies fileName by adding digits into it, so it becomes
	// unique (not existing yet) within the directory. If the fileName 
	// file doesn't exist, fileName is returned (copied to uniqueName) 
	// without modification.
	bool UniqueName( LPCVTCHAR fileName, LPVTCHAR uniqueName );

	// Convert long file name into short, 8-bit file name.
	// This is a very useful function for MS Windows, to pass
	// file names to libraries which want char*. The operation may be performed 
	// only if the file already exists; you can also set bCreateFile to allow 
	// the function to create the file if it doesn't exist first.
	#if defined(VT_WINDOWS)
		bool GetShortName( LPCVTCHAR fileName, char *out, bool bCreateFile=false );
	#endif

	bool IsDirectory( LPCVTCHAR fn );

	#if defined(USE_LSSTRING)
		LString EndWithSlash(const LString& str);
		bool EndsWithSlash(const LString &str);
		LString KillEndingSlash(const LString &str);

		LString CutoffFileName(const LString &in);
		LString CutoffFileExt(const LString &in);
		LString ExtractFileName(LPCVTCHAR in);
		LString ExtractFileNameWOExt(LPCVTCHAR in);
		LString ExtractFileExt(LPCVTCHAR in);
		LString ExtractDirectoryName(LPCVTCHAR in);
		LString RelativePathFromAbs(LPCVTCHAR startDir, LPCVTCHAR fileName);
		LString AbsolutePath(LPCVTCHAR startDir, LPCVTCHAR fileName);

		LString GetApplicationDataDir();
		LString GetUserDocDir();
		LString GetTempDir(); 
		LString GetExeName();
		LString GetExeDir();	
		LString GetCWD();
		bool MkTemp(LPCVTCHAR prefx, LString &out_name);
		LString GetTempDirecory();
		bool MkSysTemp(LPCVTCHAR namePrefx, LString &out_name); 
		bool UniqueName( LPCVTCHAR fileName, LString &uniqueName );
	#endif

	//
	// Directory reader class. Example of usage:
	// CDirectoryReader dir(someDir, "exe,com,bat");
	// for(int i=0; i<dir.GetFileCount(); i++)
	//     printf("Found an executable: %s\n", dir.GetFile(i));
	// Read() can be called multiple times; old list
	// of files is deleted from memory before each call.
	//
	// Also look at vt_file -> CFilesystemDirectory class 
	// (which BTW uses VT and CDirectoryReader) which offers 
	// some additional functionality.
	//
	class CDirectoryReader
	{
	public:
		CDirectoryReader();
		// Filter is a comma-separated string with a list of allowed 
		// extenstions. Note that the extenstions are not case sensitive.
		// This function can be called multiple times; the list of files 
		// is automatically emptied before next reading.
		bool Read(LPCVTCHAR mask, LPCVTCHAR filter=NULL); 
		// Create and Read() directory listing.
		CDirectoryReader(LPCVTCHAR mask, LPCVTCHAR filter=NULL);		
		VTINLINE size_t GetFileCount() const { return stringCount; }
		VTINLINE LPCVTCHAR GetFile(size_t idx) const { return stringBuffer+stringPtrs[idx]; }
		void Free(); // Deallocate memory buffers
		size_t GetDataStorageSize() const { return nextChar; } // Size of memory occupied by all file names
		~CDirectoryReader(){ Free(); }
	private:
		LPVTCHAR stringBuffer; // All strings are dumped into one buffer, 0-separated
		size_t *stringPtrs; // Indexes of first characters of strings in the buffer
		size_t stringBufferSize, stringPtrsSize, stringCount, nextChar;	
		static const size_t kBufferIncrement = 1024*16, kPointersIncrement = 1024;		
	protected:
		void addFile(LPCVTCHAR name, LPCVTCHAR *masks);
	};
	

} // namespace



#endif
