//-----------------------------------------------------------------------------
//
// Polyphase Sinc Filter array
//
//-----------------------------------------------------------------------------

/*
	VT Framework License (BSL 1.0)

	Copyright (C) 2008 Sergey A. Galin 
	E-mail: sergey.galin@gmail.com

	Permission is hereby granted, free of charge, to any person or organization
	obtaining a copy of the software and accompanying documentation covered by
	this license (the "Software") to use, reproduce, display, distribute,
	execute, and transmit the Software, and to prepare derivative works of the
	Software, and to permit third-parties to whom the Software is furnished to
	do so, all subject to the following:

	The copyright notices in the Software and this entire statement, including
	the above license grant, this restriction and the following disclaimer,
	must be included in all copies of the Software, in whole or in part, and
	all derivative works of the Software, unless such copies or derivative
	works are solely in the form of machine-executable object code generated by
	a source language processor.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
	SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
	FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
	ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
	DEALINGS IN THE SOFTWARE.
*/

#ifndef VT_FIR_POLYPHASE_H
#define VT_FIR_POLYPHASE_H

#include "vt_firs.hpp"
#include "vt_vectors.hpp"
#include "vt_assert.hpp"

namespace VT {
	namespace FIR {

		//
		// Base class for all polyphase filters.
		//
		// All filter vectors are stacked into one vector, one by one 
		// from low phase to high, for more efficient memory management.
		//
		// Note that there's no public constructor.
		//
		template<class T> class CPolyphaseFilterSet: protected CVector<T>
		{
		public:	
			size_t GetFilterSize() const { return m_uFilterSize; } // Get length of one filter (single-phase).
			size_t GetNPhases() const { return m_uPhases; } // Get number of phases
			
			// Get address of one filter by its index
			VTINLINE const T* GetFilterIdx( size_t iphase ) const {
				VTASSERT2T( iphase>=0 && iphase<m_uPhases, "Integer FIR filter index out of range." );
				return this->getRODataPtr() + m_uFilterSize * iphase;
			}

			// Get address of one filter by phase, where phase is [0..1)
			VTINLINE const T* GetFilter( T phase01 ) const {
				VTASSERT2( phase01>=0 && phase01<=1.0, CString(phase01)<<_T(" <- the phase") );
				return this->GetFilterIdx( URound( phase01 * T(this->m_uPhases-1) ) );
			}

		protected:
			size_t m_uPhases, m_uFilterSize;

			// Constructor
			CPolyphaseFilterSet<T>(size_t filtersize, size_t phases):
				CVector<T>( filtersize*phases ),
				m_uPhases( phases ),
				m_uFilterSize( filtersize )
			{
				VTASSERT( phases>0 );
				VTASSERT( filtersize>3 );
			}
			
		};


		//
		// Implementation of standard low-pass Sinc filter set.
		// Such kernels are widely used for resampling and speed changing.
		//
		template<class T> class CPolyphaseSincFilterSet: public CPolyphaseFilterSet<T>
		{
		public:
			CPolyphaseSincFilterSet<T>( size_t filtersize, size_t phases, T fc ):
				CPolyphaseFilterSet<T>( filtersize, phases ),
				m_Fc( fc )
			{
				CVector<T> windowVector( filtersize );
				windowVector.fill(T(1.0));
				KaiserWindow<T, double>( windowVector );
				T *firDataPtr = this->getDataPtr();
				for( size_t phasei=0; phasei<int(phases); phasei++ ){
					T phaseShift = T(phasei) / T(phases);
					FIR::LowPass<T>( firDataPtr, filtersize, fc, phaseShift, windowVector.getDataPtr() );
					firDataPtr += filtersize;
				}
			}
			
			T GetFc() const { return m_Fc; } // Get cutoff frequency
		protected:
			T m_Fc;
		};

		//
		// Definition of standard polyphase sinc filter set types.
		//
		typedef CPolyphaseSincFilterSet<float> CPolyphaseSincFilterSetSP;
		typedef CPolyphaseSincFilterSet<double> CPolyphaseSincFilterSetDP;
		typedef CPolyphaseSincFilterSet<long double> CPolyphaseSincFilterSetLDP;

	} // namespace FIR
} // namespace VT

#endif

