//-----------------------------------------------------------------------------
//
// FIR filter caclulations
//
//-----------------------------------------------------------------------------

/*
	VT Framework License (BSL 1.0)

	Copyright (C) 2006 Sergey A. Galin 
	E-mail: sergey.galin@gmail.com

	Permission is hereby granted, free of charge, to any person or organization
	obtaining a copy of the software and accompanying documentation covered by
	this license (the "Software") to use, reproduce, display, distribute,
	execute, and transmit the Software, and to prepare derivative works of the
	Software, and to permit third-parties to whom the Software is furnished to
	do so, all subject to the following:

	The copyright notices in the Software and this entire statement, including
	the above license grant, this restriction and the following disclaimer,
	must be included in all copies of the Software, in whole or in part, and
	all derivative works of the Software, unless such copies or derivative
	works are solely in the form of machine-executable object code generated by
	a source language processor.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
	SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
	FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
	ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
	DEALINGS IN THE SOFTWARE.
*/

#ifndef VT_FIRS_HPP
#define VT_FIRS_HPP

#include <math.h>
#include "vt_assert.hpp"
#include "vt_audiomath.hpp"
#include "vt_buffers.hpp"
#include "vt_localization.hpp"

namespace VT {
	namespace FIR {

		// Calculate half of length of windowed-sinc FIR filter from its bandwidth
		VTINLINE long Lowpass_BW2Size2(double BW){ return URound(2.0 / BW); }

		// Calclulate bandwidth of a windowed-sinc FIR filter from its length
		VTINLINE double Lowpass_SizeToBW(long sz){ return 4.0/(sz-1.0); }

		// Calclulate bandwidth, in Hz, of a windowed-sinc FIR  filter from its length
		VTINLINE double Lowpass_SizeToBWHz(long sz, double srate){ return Lowpass_SizeToBW(sz)*srate; }

		// Construct simple Sinc low-pass FIR filter, which is determined by:
		// fc - cutoff frequency, as a fraction of sample rate (0..0.5)
		// size of H (H.getSize()) defines size of the filter. Must be 
		// an odd number.
		// Don't forget to apply windowing to the created filter.
		template<class T> inline void LowPass(CVector<T> &H, T fc)
		{	
			if( H.getSize() % 1 != 0 )
				throw LCT("FIR size must an odd number.");
			// Hi  =  sin( 2 pi fc i) / i pi  =  2 fc sinc( 2 i fc )
			long M2 = long(H.getSize()) / 2; // M/2
			long idx = 0;
			for(long i = -M2; i<0; i++) // Left coefficients
				H.set(idx++, T( sin( 2.0 * kPi * fc * i ) / (i*kPi)));	
			H.set(idx, T(2.0) * fc); // Central coefficient. Note it's always positive.
			H.symmet(); // The rest is symmetrical around the center
		}


		// Construct simple Sinc low-pass FIR filter, which is determined by:
		// fc - cutoff frequency, as a fraction of sample rate (0..0.5)
		// BW - transition band width, as fraction of sample rate (0..0.5)
		// Don't forget to apply windowing to the created filter.
		template<class T> inline void LowPass(CVector<T> &H, T fc, T BW)
		{	
			// Hi  =  sin( 2 pi fc i) / i pi  =  2 fc sinc( 2 i fc )
			long M2 = Lowpass_BW2Size2(BW); // M/2
			long idx = 0;
			H.resize( M2+M2+1 );
			for(long i = -M2; i<0; i++) // Left coefficients
				H.set(idx++, T( sin( 2.0 * kPi * fc * i ) / (i*kPi)));	
			H.set(idx, T(2.0) * fc); // Central coefficient. Note it's always positive.
			H.symmet(); // The rest is symmetrical around the center
		}


		// Construct simple Sinc low-pass FIR filter, which is determined by:
		// fc - cutoff frequency, as a fraction of sample rate (0..0.5).
		// size - length of the filter (an odd value).
		// phaseShift - a shift of phase, supposedly in (-1..1).
		// window - a pointer to array which contains windowing coefficients.
		template<class T> VTINLINE void LowPass(T* data, size_t size, T fc,
			T phaseShift, const T* window)
		{
			VTASSERT( size % 1 == 0 );
			VTASSERT( phaseShift>=-1 && phaseShift<=1 );
			T twofc = T(2.0 * fc), x = T(T(-long(size/2)) - phaseShift);
			for(size_t i = 0; i<size; i++, x+=T(1.0))
				data[i] = T( twofc * NormSinc( twofc * x ) ) * window[i];	
		}


		// The same with parameters in Hz
		template<class T> VTINLINE void LowPass(CVector<T> &H, T fchz, T bwhz, T srate)
		{
			FIR::LowPass(H, T(fchz/srate), T(bwhz/srate));
		}

		template<class T> void HammingWindow(CVector<T> &H)
		{
			// Hamming window: 
			// w(n) = 0.53836 - 0.46164 cos(( 2 pi n ) / N - 1 )
			T N1 = H.getSize()-1, tmp;
			size_t end = (H.getSize()+1)/2, j=H.getSize()-1;
			for(size_t n=0; n<end; n++){
				tmp = T(0.53836 - 0.46164 * cos(2.0 * kPi * T(n) / N1 ));
				H.set(n, H.get(n) * tmp); // Left wing
				if(n!=j){ // Right wing?
					H.set(j, H.get(j) * tmp);
					j--;
				}
			}
		}

		template<class T> void BlackmanWindow(CVector<T> &H)
		{
			// Blackman window:
			// w(n) = a0 - a1 cos( (2pi n) / (N-1)) + a2 cos( (4pi n) / (N-1))
			// a0 = 0.42 a1 = 0.5 a2 = 0.08	
			T N1 = H.getSize()-1, tmp;
			size_t end = (H.getSize()+1)/2, j=H.getSize()-1;
			for(size_t n=0; n<end; n++){
				tmp = T( 0.42 - 0.5 * cos( 2.0*kPi*T(n) / N1 ) + 0.08 * cos( 4.0*kPi*T(n) / N1 ) );
				H.set(n, H.get(n) * tmp); // Left wing
				if(n!=j){ // Right wing?
					H.set(j, H.get(j) * tmp); 
					j--;
				}	
			}
		}


		// I0 needed to calculate Kaiser window
		// Jalpha(x) = SUM[m:=0..oo]( (-1)^m * (x/2)^(2m+alpha) / (m! * Gamma(m+alpha+1)
		// Where Gamma(x) == (x-1)! (for integers), i.e.:
		// Gamma(m+alpha+1) = (m+alpha)!
		template<class T> VTINLINE T KaiserI0(T x, T epsilon /*, int alpha = 0*/ )
		{
		   T sum = T(1.0), u = T(1.0), n = T(1.0), x2 = T(x/2.0);
		   do{
			  T t = x2/n;      
			  u *= t*t;
			  sum += u;
			  n += T(1.0);
		   }while( u >= epsilon*sum );
		   return sum;
		}

		// Kaiser window.
		// I0T may have bigger precision than T, in case you want to calculate I0 more
		// precisely. E.g. long double.
		// beta defines window width (0 - rectangular window, large values change its 
		// shape toward Gaussian curve).
		// epsilon defines precision of I0().
		#define VT_KAISER_DEFAULT_BETA		5.6533
		#define VT_KAISER_DEFAULT_EPSILON	1e-21

		template<class T, class I0T> inline 
			void KaiserWindow(CVector<T> &H, I0T beta=I0T(VT_KAISER_DEFAULT_BETA), 
			I0T epsilon=I0T(VT_KAISER_DEFAULT_EPSILON))
		{
			T Ibeta = T(T(1.0)/KaiserI0(beta, epsilon));
			int N = int(H.getSize());
			int N2 = int(H.getSize())/2;
			I0T inm1 = I0T(1.0)/I0T(N2);
			int idx1 = N2+1, idx2 = N2-1;
			for(int i=1; i<=N2; i++){
			  I0T temp = I0T(i) * inm1;
			  temp = I0T(1.0) - temp*temp;
			  if( temp<I0T(0) )
				  temp = I0T(0);
			  temp = KaiserI0(I0T(beta*sqrt(temp)), epsilon) * Ibeta;
			  H.set( idx1, T(H.get(idx1) * temp));
			  H.set( idx2, T(H.get(idx2) * temp));
			  idx1++;
			  idx2--;
		   }
		}

		// Invert spectrum of a FIR filter, e.g. to turn low pass into high pass and vice versa.
		template<class T> inline bool SpectralInversion(CVector<T> &H)
		{
			if( !(H.getSize()%2) )
				return false;
			size_t center = H.getSize() / 2;
			for(size_t n=0; n<H.getSize(); n++)
				H.set(n, -H.get(n));
			H.set(center, 1.0+H.get(center));
			return true;
		}

		/*
		Something's wrong here and it doesn't work. Use only SpectralInversion() for now.

		// Reverse spectrum of a FIR filter, e.g. turns low pass into high pass and vice versa.
		template<class T> bool SpectralReversal(CVector<T> &H)
		{
			if( H.getSize() < 3 || !(H.getSize()%2) )
				return false;	
			size_t 
				center = H.getSize() / 2,
				i=center-1, j=center+1;
			for(;;){
				H[i] = -H[i];
				H[j] = -H[j];
				if(i<2) 
					return true;
				i--; i--; 
				j++; j++;
			}
		}*/

		// Resize frequency response array for another sample rate.
		// 'srate' is the current sample rate, for which we might want to build a FIR filter
		// 'FRsrate' is the sample rate for which the 'fr' was generated
		template<class T> VTINLINE void ResizeFRArray( CVector<T> &fr, double srate, double FRsrate )
		{	
			// step == FRsrate/(fr.getSize()-1) (step between FR samples)
			// AND:
			// step == srate / (newsize-1)
			//
			// FRsrate/(fr.getSize()-1) == srate / (newsize-1)
			// newsize-1 = (fr.getSize()-1) * srate / FRsrate
			size_t 
				oldsize = fr.getSize(),	
				//newsize = size_t(URound(srate / FRsrate * double(fr.getSize())));
				newsize = size_t(URound(  double(fr.getSize()-1) * srate / FRsrate ) +1);
			fr.resize( newsize, true );		
			// Pad the array at right, if neccessary
			if( oldsize<newsize){ 
				T rightdata = fr.get( fr.getSize()-1 );
				for(size_t i=oldsize; i<newsize; i++)
					fr.set( i, rightdata );
			}
		}

		// Create a zero-phase FIR filter from the given frequency response table contained in fr.
		// The fr contains points which determine amplitude for frequencies spread from 
		// 0 to 0.5 of the sampling frequency and it's longer than 3. 
		// Length of the resulting filter is out_fir.getSize(); make sure it's an odd number!
		// This procedure is based on stuff from DSP Guide, chapters 8 and 17.
		// The filter is windowed using Hamming window.
		template<class T> void FR2FIR( const CVector<T> &fr, CVector<T> &out_fir )
		{
			// FIR == impulse response == IDFT(of frequency response).
			// Polar to rectanglar conversion is:
			// ReX[k] = MagX[k] cos(Phase[k]) = MagX[k] = fr[]
			// ImX[k] = MagX[k] sin(Phase[k]) == 0 (if Phase[k]==0)
			// So, we are simply feeding fr into real IDFT with the sinus part thrown away.
			
			// Now with indexes: fr.getSize()-1 =: N/2 => N=2*(fr.getSize()-1)
			// 'i' runs through 0..N-1 and 'k' runs through 0..N/2	
			// M := out_fir.getSize()-1 (even)	
			// fr.getSize() == M+1 = M2+M2+1
			
			// We don't follow the "caclulate N sample IR then shift it 
			// then truncate it to M+1" stuff, instead, we just calculate M+1
			// element of the output FIR vector and write them right to their
			// places, which may save a lot of computations.
			
			size_t 
				N2 = fr.getSize()-1, 
				N = N2+N2, 
				M =out_fir.getSize()-1,
				M2 = M/2;
			
			T w1 = T(2.0 * kPi / T(N)); 
				


		#if 0
			register size_t i=0, k, out_i=M2; // out_fir output index
				
			// We use M2 samples from the end of the row.
			// E.g.: M2 == 1 => use N-1..N-1 samples. M2 == 2 => N-2..N-1. And etc.
			// So the starting index is:	
			i = N-M2;
			for(out_i=0; out_i<out_fir.getSize(); out_i++){
			
				register T sum=0;
				T w=T(i)*w1;
				// Yes, 'k' runs through 0..N/2... and N/2 == fr.getSize()-1
				// But, remember that 0th and N/2th steps use another coefficients,
				// so let's calculate them out of the loop.
				for(k=1; k<N2; k++)
					sum += fr.get( k ) * cos( w*k );
				sum *= 2.0;
				// Now, add the edge components of the sum, which are not
				// multiplied by 2.
				// cos(0) == 1, throw it away:
				sum += fr.get(0);
				// cos(i*pi) == (-1)^i == { -1, i is odd; 1, i is even }.
				if( i % 2 ) // i is odd
					sum -= fr.get(N2);
				else // i is even
					sum += fr.get(N2);

				out_fir.set(out_i, sum/T(N));
				
				if( ++i == N )
					i=0;
			}
			// Didn't look like it was big deal, but it was... ;-)
			HammingWindow( out_fir ); // And the final touch is...*/




		#else

			register size_t i=0, out_i, j=M2; 
			T wWindow = T(kPi / T(M2)), tN = T(N);
			//j=M2;
			out_i=M2;
			for(i=0; i<=M2; i++){
				register T sum=0;
				T w=T(i)*w1;
				// Yes, 'k' runs through 0..N/2... and N/2 == fr.getSize()-1
				// But, remember that 0th and N/2th steps use another coefficients,
				// so let's calculate them out of the loop.
				for(register size_t k=1; k<N2; k++)
					sum += fr.get( k ) * cos( w*k );
				sum *= 2.0;
				// Now, add the edge components of the sum, which are not
				// multiplied by 2. 
				// cos(0) == 1, throw it away:
				sum += fr.get(0);
				// cos(i*pi) == (-1)^i == { -1, i is odd; 1, i is even }.
				if( i % 2 ) // i is odd
					sum -= fr.get(N2);
				else // i is even
					sum += fr.get(N2);
				
				// Multply by Hamming window and divide by N. Hamming window is: 
				// w(n) = 0.53836 - 0.46164 cos(( 2 pi n ) / N - 1 ), or, in this case:
				// w(n) = 0.53836 - 0.46164 cos(( pi out_i ) / M2 ) 
				sum *= T( ( 0.53836 - 0.46164 * cos(wWindow * T(out_i)) ) / tN );

				// Despicably use the fact that any zero-phase filter 
				// is symmetrical around its center to halve calculations.
				// Central element is set twice... no problem.
				out_fir.set(out_i++, sum);
				out_fir.set(j--, sum);
			}
		#endif
		}



		// Create a zero-phase FIR filter from the given frequency response table contained in fr.
		// The fr contains points which determine amplitude for frequencies spread from 
		// 0 to 0.5 of the sampling frequency and it's longer than 3. 
		// Length of the resulting filter is out_fir.getSize(); make sure it's an odd number!
		// This procedure is based on stuff from DSP Guide, chapters 8 and 17.
		// This function used 'window' to window the filter.
		// See the previous implementation for more comments on the math.
		template<class T> void FR2FIR( const CVector<T> &fr, CVector<T> &out_fir, const T* window )
		{
			size_t N2 = fr.getSize()-1, N = N2+N2, M =out_fir.getSize()-1, M2 = M/2;	
			T tN = T(N), w1 = 2.0 * kPi / tN; 
			register size_t out_i=M2, j=M2; 
			for(register size_t i=0; i<=M2; i++){	
				register T sum=0;
				T w=T(i)*w1;
				for(register size_t k=1; k<N2; k++)
					sum += fr.get( k ) * cos( w*k );
				sum *= 2.0;
				sum += fr.get(0);
				if( i % 2 ) // i is odd
					sum -= fr.get(N2);
				else // i is even
					sum += fr.get(N2);
				sum *= window[out_i] / tN;
				// Central element is set twice... no problem.
				out_fir.set(out_i++, sum);
				out_fir.set(j--, sum);	
			}
		}

	} // namespace FIR
} // namespace VT


#endif

