//-----------------------------------------------------------------------------
//
// Portable "Message boxing" system provides system-independent UI messages.
//
//-----------------------------------------------------------------------------

/*
	VT Framework License (BSL 1.0)

	Copyright (C) 2006, 2007, 2008, 2009 Sergey A. Galin 
	E-mail: sergey.galin@gmail.com

	Permission is hereby granted, free of charge, to any person or organization
	obtaining a copy of the software and accompanying documentation covered by
	this license (the "Software") to use, reproduce, display, distribute,
	execute, and transmit the Software, and to prepare derivative works of the
	Software, and to permit third-parties to whom the Software is furnished to
	do so, all subject to the following:

	The copyright notices in the Software and this entire statement, including
	the above license grant, this restriction and the following disclaimer,
	must be included in all copies of the Software, in whole or in part, and
	all derivative works of the Software, unless such copies or derivative
	works are solely in the form of machine-executable object code generated by
	a source language processor.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
	SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
	FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
	ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
	DEALINGS IN THE SOFTWARE.
*/

#include "vt_messageboxes.hpp"
#include "vt_threading.hpp"

#if defined(_MSC_VER) && !defined(_CRT_SECURE_NO_DEPRECATE)
	#define _CRT_SECURE_NO_DEPRECATE
#endif

using namespace VT;

// Global VTMB objects
VTMB *volatile VT::VTMB::s_pMB = NULL;
CCriticalSection VT::VTMB::m_CritS;
VTMB VT::g_VTMB;

VTMB::VTMB(bool bMakeMeDefault)
{ 
	if(bMakeMeDefault) 
		s_pMB = this; 
}

VTMB::~VTMB()
{ 
	if(s_pMB == this) 
		s_pMB = NULL; 
}

void VTMB::SetAsDefault()
{
	s_pMB = this;
}

void VTMB::ClearDefault()
{
	s_pMB = NULL;
}

VTMB::TDialogResult VTMB::VMsgBox(LPCVTCHAR message, LPCVTCHAR caption, 
								  VTMB::TDialogType type, VTMB::TDialogButtons btns, bool bThreadSafe)
{		
    #if defined(VT_WINDOWS)
		//
		// Windows GUI implementation
		//
		const LPCVTCHAR kNullPointer = _T("[NULL POINTER]");
		if( message==NULL )
			message = kNullPointer;
		if( caption==NULL )
			caption = kNullPointer;

		try{
			UINT mbtype = MB_APPLMODAL;
			switch( type ){
				case VTMB::toadError: 				mbtype|=MB_ICONSTOP;			break;
				case VTMB::toadWarning:				mbtype|=MB_ICONEXCLAMATION;		break;
				case VTMB::toadInformation:			mbtype|=MB_ICONINFORMATION;		break;
				case VTMB::toadQuestion:			mbtype|=MB_ICONQUESTION;		break;
			}
			switch( btns ){
				case VTMB::toabOK:					mbtype|=MB_OK;					break;
				case VTMB::toabOKCancel:			mbtype|=MB_OKCANCEL;			break;
				case VTMB::toabYesNo:				mbtype|=MB_YESNO;				break;
				case VTMB::toabYesNoCancel:			mbtype|=MB_YESNOCANCEL;			break;
				case VTMB::toabAbortRetryIgnore:	mbtype|=MB_ABORTRETRYIGNORE;	break;
				case VTMB::toabRetryCancel:			mbtype|=MB_RETRYCANCEL;			break;
			}

			HWND hWnd;
			if( s_pMB )
				hWnd = s_pMB->GetMainWindowHandle( bThreadSafe );
			else
				hWnd = NULL;

			int res = MessageBox(hWnd, message, caption, mbtype);

			switch( res ){
				case IDABORT:		return toarAbort;
				case IDCANCEL:		return toarCancel;
				#ifdef _MSC_VER
					case IDCONTINUE:	return toarContinue;
				#endif
				case IDIGNORE:		return toarIgnore;
				case IDNO:			return toarNo;
				case IDOK:			return toarOK;
				case IDRETRY:		return toarRetry;
				#ifdef _MSC_VER
					case IDTRYAGAIN:	return toarRetry; // !!
				#endif
				case IDYES:			return toarYes;
				default:			return toarStub;
			}	
		}catch(...){ 
			MessageBox(0, _T("[Error in MsgBox function #1.]"), _T("Error"), MB_ICONSTOP);
		};
		return toarStub;
    
	#else
		//	
		// Portable console implementation
		//

		// !!!!!!!!!!!!!!!!! INCOMPLETE !!!!!!!!!!!!!!!!!!
		printf("MESSAGE[%s]: %s\n", caption, message);
		return toarStub;
    
    #endif
}

VTMB::TDialogResult VTMB::MsgBox(LPCVTCHAR message, LPCVTCHAR caption, TDialogType type, 
								 TDialogButtons btns, bool bThreadSafe)
{	
	CCSLock lock(&m_CritS);
	try{
		if( s_pMB!=NULL )
			return s_pMB->VMsgBox(message, caption, type, btns, bThreadSafe);
		else
			return g_VTMB.VMsgBox(message, caption, type, btns, bThreadSafe);
	}catch(...){ 
	    #if defined(VT_WINDOWS)
		MessageBox(0, _T("[Error in MsgBox function #1.]"), _T("Error"), MB_ICONSTOP);
	    #else
		printf("[Error in MsgBox function #1.]\nMESSAGE[%s]: %s\n", caption, message);
	    #endif
	};	
	return toarStub;
}


void VTMB::Error(LPCVTCHAR message, bool bThreadSafe)
{
	MsgBox(message, LCT("Error"), toadError, toabOK, bThreadSafe); 
}

void VTMB::UserError(LPCVTCHAR message, bool bThreadSafe)
{
	MsgBox(message, LCT("Error"), toadError, toabOK, bThreadSafe); 
}

void VTMB::Warning(LPCVTCHAR message, bool bThreadSafe)
{
	MsgBox(message, LCT("Warning"), toadWarning, toabOK, bThreadSafe); 
}

void VTMB::Info(LPCVTCHAR message, bool bThreadSafe)
{
	MsgBox(message, LCT("Information"), toadInformation, toabOK, bThreadSafe); 
}

bool VTMB::Confirm(LPCVTCHAR message, bool bThreadSafe)
{
	return MsgBox(message, LCT("Confirm"), toadQuestion, toabYesNo, bThreadSafe)==toarYes; 
}

VTMB::TDialogResult VTMB::YesNo(LPCVTCHAR message, bool bThreadSafe)
{
	return MsgBox(message, LCT("Confirm"), toadQuestion, toabYesNo, bThreadSafe); 
}

VTMB::TDialogResult VTMB::YesNo(LPCVTCHAR message, LPCVTCHAR caption, bool bThreadSafe)
{
	return MsgBox(message, caption, toadQuestion, toabYesNo, bThreadSafe); 
}

void VTMB::Crash(LPCVTCHAR message)
{
	Error(message, true); 
	#if defined(VT_WINDOWS) && defined(_MSC_VER)
		terminate(); 
	#else
		exit(255);
	#endif
}

void VTMB::DebugMsgBox(LPCVTCHAR message, LPCVTCHAR file, int line, LPCVTCHAR function, 
					   LPCVTCHAR header) // header can be NULL!
{
	static VTCHAR buffer[1024], buffer2[256];
	VTSTRNCPY( buffer,
		_T("Debug Message:") VTENDLT VTENDLT
		_T("------------------------------------------------------------------------") VTENDLT, 256 );
	VTSTRNCAT(buffer, message, 500);
	
	if( header!=NULL ){
		VTSTRNCAT(buffer, 
			VTENDLT 
			_T("------------------------------------------------------------------------") VTENDLT, 256 );
		VTSTRNCAT(buffer, header, 500);	
	}
	
	VTSTRNCAT(buffer, 
		VTENDLT
		_T("------------------------------------------------------------------------") VTENDLT VTENDLT, 256);
	VTSPRINTF(buffer2, _T("File: %s") VTENDLT _T("Line: %d\r\nFunction: %s") VTENDLT, file, line, function);
	VTSTRNCAT(buffer, buffer2, 200);
	VTSTRNCAT(buffer, _T("Select OK to continue, Cancel to abort application.") VTENDLT, 200);
	
	if( VTMB::MsgBox(buffer, (header==NULL)?_T("Debug Information"):header, 
		toadInformation, toabOKCancel, true) == toarCancel ){
		#if defined(VT_WINDOWS) && defined(_MSC_VER)
		    terminate();
		#else
		    exit(255);
		#endif
	}
}


