//-----------------------------------------------------------------------------
//
// Portable "Message boxing" system provides system-independent UI messages.
//
//-----------------------------------------------------------------------------

/*
	VT Framework License (BSL 1.0)

	Copyright (C) 2006, 2007, 2008, 2009 Sergey A. Galin 
	E-mail: sergey.galin@gmail.com

	Permission is hereby granted, free of charge, to any person or organization
	obtaining a copy of the software and accompanying documentation covered by
	this license (the "Software") to use, reproduce, display, distribute,
	execute, and transmit the Software, and to prepare derivative works of the
	Software, and to permit third-parties to whom the Software is furnished to
	do so, all subject to the following:

	The copyright notices in the Software and this entire statement, including
	the above license grant, this restriction and the following disclaimer,
	must be included in all copies of the Software, in whole or in part, and
	all derivative works of the Software, unless such copies or derivative
	works are solely in the form of machine-executable object code generated by
	a source language processor.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
	SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
	FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
	ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
	DEALINGS IN THE SOFTWARE.
*/

#ifndef VT_MESSAGEBOXES_H
#define VT_MESSAGEBOXES_H

#define VT_INSIDE_MESSAGEBOXES_H

#include "vt_portable.hpp"
#include "vt_localization.hpp"
#include "vt_threading.hpp"

// Define this for detailed information about VTASSERT macros in compiler's log.
// #define VERBOSE 

#if defined(VT_WINDOWS)
	#if defined(_MSC_VER)
		#include <eh.h>
	#endif
	#include <process.h>
#endif

#if defined(USE_CASSERT)
	#pragma message("Flagging to use C assert() function in VTASSERT because USE_CASSERT is defined externally.")
#endif

#if defined(CONSOLE)
	#if !defined(USE_CASSERT)
		#define USE_CASSERT
	#endif
	#pragma message("Flagging to use C assert() function in VTASSERT because CONSOLE is defined.")
#endif

#if defined(VT_LINUX) 
	#if !defined(USE_CASSERT)
		#define USE_CASSERT
	#endif
	#pragma message("Flagging to use C assert() function in VTASSERT because VT_LINUX is defined.")	
#endif
	
#if defined(USE_CASSERT)
	#include <assert.h>
#endif


namespace VT {

	class VTMB;

	//        ;
	//      .
	//    ,         
	//  ,         
	//   .
	//     ,    
	//  -  ,    "" VTMB::Error()  throw.
	#if defined(_DEBUG)
		#define VTDEBUGMSG(text) VT::VTMB::DebugMsgBox((text), _T(__FILE__), \
				__LINE__, _T(__FUNCTION__))
		#define VTDEBUGMSG2(text, header) VT::VTMB::DebugMsgBox((text), _T(__FILE__), \
				__LINE__, _T(__FUNCTION__), header)
		#define VTDEBUGMSGT(text)	VTDEBUGMSG(_T(text))
		#define VTDEBUGTHROW(text)	throw((const wchar_t*)text)
		#define VTDEBUGTHROWT(text) VTDEBUGTHROW(_T(text))
		#define VTDEBUGMSGBOX		VTDEBUGMSGT("(no special message)")

		#if defined(USE_CASSERT)
			#define VTASSERT(cond)			assert(cond)
			#define VTASSERT2(cond, msg)	assert(cond)
			#define VTASSERT2T(cond, msg)	assert(cond)
			#ifdef VERBOSE
				#ifdef _MSC_VER
					#pragma message("messageboxes.h: VTASSERT set to use C assert().")
				#else
					#warning VTASSERT set to use C assert().
				#endif#
			#endif
		#else
			#define VTASSERT(cond) {if(!(cond)){ VTDEBUGMSG(_T("Assertion failed: ") _T(#cond)); }}	
			#define VTASSERT2(cond, msg) {if(!(cond)){ VTDEBUGMSG2(_T("Assertion failed: ") _T(#cond), msg); }}	
			#define VTASSERT2T(cond, msg) VTASSERT2( cond, _T(msg))
			#ifdef VERBOSE
				#ifdef _MSC_VER
					#pragma message("VTASSERT macros are enabled.")
				#else
					#warning messageboxes.h: VTASSERT macros are enabled.
				#endif
			#endif
		#endif
	#else	
		#define VTDEBUGMSG(text)		{}
		#define VTDEBUGMSG2(text, msg)	{}
		#define VTDEBUGMSGT(text)		{}
		#define VTDEBUGTHROW(text)		{}
		#define VTDEBUGTHROWT(text)		{}
		#define VTDEBUGMSGBOX			{}
		#define VTASSERT(cond)			{}
		#define VTASSERT2(cond, msg)	{}
		#define VTASSERT2T(cond, msg)	{}
		#ifdef VERBOSE
			#ifdef _MSC_VER
				#pragma message("messageboxes.h: VTASSERT set empty.")
			#else
				#warning VTASSERT set empty.
			#endif
		#endif
	#endif

	#if !defined(VTASSERT) || !defined(VTASSERT2) || !defined(VTASSERT2T)
		#error "Something went wrong with VTASSERT right IN messageboxes.h!!!"
	#endif

	//    ,      .
	//    ,    .    
	//  .
	#define VTTHROW(text) {VT::VTMB::Error(text, true); throw (text);}
	#define VTTHROWT(text) VTTHROW(_T(text))



		
	// Please note the behaviour of this class. It contains a global pointer to
	// current default VTMB (or descendant) class object - M/B handler. 
	// If the pointer is set, any message box calls are forwarded to the handler; 
	// if the pointer is NULL, then the calls are forwarded to global VTMB object. 
	// The handler pointer can be set via the constructor or 
	// SetAsDefault() / ClearDefault().
	// All M/B functions are static. So there's no difference if you call them
	// as e.g. VTMB::Error() or SomeDescendantOfVTMB::Error().
	// The only non-static M/B function is VMsgBox().
	class VTMB
	{
	public:
		enum TDialogType
		{ 
			toadError, 
			toadInformation, 
			toadWarning, 
			toadQuestion, 
			toadNone 
		};

		enum TDialogButtons
		{
			toabOK, 
			toabOKCancel, 
			toabYesNo, 
			toabYesNoCancel,
			toabAbortRetryIgnore, 
			toabRetryCancel 
		};
		
		enum TDialogResult{ 
			toarStub=0, 
			toarOK, 
			toarCancel, 
			toarYes, 
			toarNo, 
			toarAbort, 
			toarRetry, 
			toarIgnore, 
			toarContinue 
		};

		VTMB(bool bMakeMeDefault=FALSE);
		~VTMB();
		void SetAsDefault();
		void ClearDefault();

		// This is the only virtual function, which is the creator of
		// all system message boxes. All other message functions are static.
		// If you need to change way the messages are displayed or add some
		// special handling (e.g. logging), override this function.
		// And don't forget to do SetAsDefault()!
		// bThreadSafe indicates that message box is fired from a non-GUI thread.
		virtual TDialogResult VMsgBox(LPCVTCHAR message, LPCVTCHAR caption, 
			TDialogType type=toadInformation, TDialogButtons btns=toabOK, bool bThreadSafe=false);

		// Shows a message box using currently set message box handler.
		// This function dispatches message either to s_pMB's VMsgBox(),
		// or to this's implementation of VMsgBox() (if s_pMB is NULL).
		static TDialogResult MsgBox(LPCVTCHAR message, LPCVTCHAR caption, 
			TDialogType type=toadInformation, TDialogButtons btns=toabOK, bool bThreadSafe=false);

		// Some handy functions which allow you to avoid typing
		// captions, icons and button sets for all typical dialogs;
		// they simply call MsgBox() with all required parameters.
		static void Error(LPCVTCHAR message, bool bThreadSafe=false);
		static void UserError(LPCVTCHAR message, bool bThreadSafe=false);
		static void Warning(LPCVTCHAR message, bool bThreadSafe=false);
		static void Info(LPCVTCHAR message, bool bThreadSafe=false);
		static bool Confirm(LPCVTCHAR message, bool bThreadSafe=false);
		static TDialogResult YesNo(LPCVTCHAR message, bool bThreadSafe=false);
		static TDialogResult YesNo(LPCVTCHAR message, LPCVTCHAR caption, bool bThreadSafe=false);

		// Show error message and terminate application.
		static void Crash(LPCVTCHAR message);

		// Formats debug message and shows a message box where user
		// is offered to choose: continue or abort execution of the program.
		// This function is used by VTASSERT macro.
		static void DebugMsgBox(
			LPCVTCHAR message,			// Message
			LPCVTCHAR file,				// Source code file name
			int line,					// Source code line number
			LPCVTCHAR function,			// Function name
			LPCVTCHAR header = NULL);   // Window header

	protected:
		static VTMB *volatile s_pMB;
		static CCriticalSection m_CritS;
		#if defined(VT_WINDOWS)
			virtual HWND GetMainWindowHandle(bool bThreadSafe=false){ return NULL; }	    
		#endif
	};


	extern VTMB g_VTMB;

} // namespace

#undef VT_INSIDE_MESSAGEBOXES_H

#endif


