//-----------------------------------------------------------------------------
//
// A collection of volume level metering classes (amplitude value, 
// RMS, precise metering, fast metering...)
//
//-----------------------------------------------------------------------------

/*
	VT Framework License (BSL 1.0)

	Copyright (C) 2006 Sergey A. Galin 
	E-mail: sergey.galin@gmail.com

	Permission is hereby granted, free of charge, to any person or organization
	obtaining a copy of the software and accompanying documentation covered by
	this license (the "Software") to use, reproduce, display, distribute,
	execute, and transmit the Software, and to prepare derivative works of the
	Software, and to permit third-parties to whom the Software is furnished to
	do so, all subject to the following:

	The copyright notices in the Software and this entire statement, including
	the above license grant, this restriction and the following disclaimer,
	must be included in all copies of the Software, in whole or in part, and
	all derivative works of the Software, unless such copies or derivative
	works are solely in the form of machine-executable object code generated by
	a source language processor.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
	SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
	FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
	ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
	DEALINGS IN THE SOFTWARE.
*/


#ifndef VT_METER_HPP
#define VT_METER_HPP

#include <math.h>
#include "vt_buffers.hpp"
#include "vt_audiomath.hpp"
#include "vt_debug.hpp"


namespace VT{

	template<class T> class CAudioLevelMeter
	{
	public:

		VTINLINE virtual T getRawValue() const { return 0; }

		// These functions can be implemented faster than using getRawValue(), 
		// so try to use them when possible.
		VTINLINE virtual bool levelOver(T lev){ return getRawValue()>=lev; }
		VTINLINE virtual bool levelBelow(T lev){ return !levelOver(lev); }
		
		// Returns replaced sample, or 0 if the meter doesn't have any buffer replacements etc.
		virtual T pushSample(T newsample)=0;

		// Push audio stream sample, applying a conversion function to it.
		virtual T pushAudioSample(T newsample){ return pushSample(rawToBuf(newsample)); }

		// Playing is restarted
		virtual void restart()=0;

		template<class U> void reloadFrom(CCircularBuffer<U> *bu, int offsetfromnewest);

		// How many latest samples affect current meter value.
		// Return -1 if not applicable (e.g. all samples input since
		// restart() may affect the value).
		virtual SDWORD measuringSamples() const { return -1; }

		// This function should recalculate all re-calculatable variables using
		// buffers and etc.
		virtual void revalidate(){}
	protected:
		virtual T bufToRaw(T x){ return x; }
		virtual T rawToBuf(T x){ return x; }
	};

	// Simple meter which just keeps the maximum amplitude of the input signal
	template<class T> class CAudioMaxLevelMeter: public CAudioLevelMeter<T>
	{
	public:
		CAudioMaxLevelMeter<T>(): maxValue(0){}
		VTINLINE virtual T getRawValue() const { return maxValue; }
		virtual void restart(){ maxValue=0; }
		virtual T pushSample(T newsample){ if(newsample>maxValue) maxValue=newsample; }
	protected:
		T maxValue;
		virtual T rawToBuf(T x){ return Abs(x); }
	};

	template<class T> class CBufferedAudioLevelMeter: public CAudioLevelMeter<T>
	{
	public:
		static const int cDefaultBufferSize = 10;

		// Giving memory allocation size in constructor creates a buffer
		// which will never reallocate memory (fixed-size).
		CBufferedAudioLevelMeter(SDWORD nsamples=cDefaultBufferSize, SDWORD allocate=0):
			buffer(0), bufferSize(0), bufferFill(0), bufferPos(0), bufferAllocSize(allocate)
		{ resize(nsamples); }

		SDWORD getBufferSize() const { return bufferSize; }
		SDWORD getBufferFill() const { return bufferFill; }
		bool isFull() const { return bufferFill>=bufferSize; }
		virtual SDWORD measuringSamples() const { return bufferSize; }

		// Resize buffer.
		// This function calls revalidate() after the resize.
		// Also, remember that resize() should not be done during processReplacing() unless
		// the buffer is fixed-size.
		// For buffers allowing reallocation, 'onthefly' allows resizing of the buffer
		// without losing its contents (or the most recent contents, when it's downsized).
		// Returns actual buffer size (in samples) (can be less than nsamples if bufferAllocSize
		// limits that).
		virtual SDWORD resize(SDWORD nsamples, bool onthefly=false);

		virtual SDWORD resize(double msec, double srate, bool onthefly=false){ 
			return resize( SDWORD(ms2samples(msec, srate)), onthefly ); }

		// Returns replaced sample
		virtual T pushSample(T newsample);

		virtual void restart(){ bufferFill = 0; bufferPos = 0; }

		T bufferMax();
		T bufferSum();
		long double lBufferSum() const;

		template<class U> void reloadFrom(CCircularBuffer<U> *bu, int offsetfromnewest);
		
		// Load from another buffer of the same size
		virtual bool loadFromSameSize(CBufferedAudioLevelMeter *mtr);
			
		~CBufferedAudioLevelMeter(){ if(buffer){ delete buffer; buffer=0; } }
	protected:
		T *buffer; 
		SDWORD bufferFill, bufferSize, bufferPos, bufferAllocSize;	

		SDWORD getLastInsertPos() const { return bufferPos-1; }

		VTINLINE SDWORD getPos(SDWORD fromoldest) const { 
			SDWORD pos = bufferPos+fromoldest;
			return ( pos >= bufferFill )?			
				(( bufferFill >= bufferSize )? pos-bufferSize: -1) :
				pos;
		}

		// Returns position of the oldest max sample; the sample itself is written into *mx
		SDWORD bufferMax(T *mx);
	};


	// A very useful meter but it may cause very high CPU load if fed with a shallow slope
	// instead of normal sound, because it causes rescanning the buffer on each tick.
	template<class T> class CAudioPeakMeter: public CBufferedAudioLevelMeter<T>
	{
	public:
		CAudioPeakMeter(SDWORD nsamples=16, SDWORD allocate=0):
		  CBufferedAudioLevelMeter<T>(nsamples, allocate), maxVal(0), maxPos(-1){}	
		virtual T pushSample(T newsample);
		VTINLINE virtual T getRawValue() const { return maxVal; }
		virtual void restart(){ CBufferedAudioLevelMeter<T>::restart(); maxVal = 0; maxPos = 0; }
		virtual void revalidate(){ maxPos = bufferMax( &maxVal ); }

		template<class U> void reloadFromStereo(CCircularBuffer<U> *bu1, CCircularBuffer<U> *bu2, int offsetfromnewest);

		virtual SDWORD getMaxSampleAge(){ 
			return (maxPos<this->bufferPos)? 
				(this->bufferPos-maxPos-1): (this->bufferFill-maxPos+this->bufferPos-1);
		}

	protected:
		T maxVal;
		SDWORD maxPos;
		virtual T bufToRaw(T x){ return x; }
		virtual T rawToBuf(T x){ return (x>=0)? x: -x; }
	};

	typedef CAudioPeakMeter<double> CDefaultAudioPeakMeter;


	// More complex device than a normal peak meter with an advancement to solve
	// the CAudioPeakMeter problem with slopes.
	template<class T> class CAudioBlockPeakMeter: public CAudioPeakMeter<T>
	{
	public:
		CAudioBlockPeakMeter(SDWORD nsamples=16, SDWORD allocate=0, SDWORD blocksize=256);
		~CAudioBlockPeakMeter(){ if(blockMax) delete blockMax; if(blockMaxPos) delete blockMaxPos; }
		virtual SDWORD resize(SDWORD nsamples, bool onthefly=false);
		//virtual void restart();
		virtual void revalidate();
		virtual T pushSample(T newsample);
	protected:
		T *blockMax, ovwrMax;
		SDWORD *blockMaxPos, blockSize, ovwrMaxPos;


		// Number of blocks to allocate
		SDWORD numBlocks(SDWORD nsamples){ return (nsamples+blockSize-1)/blockSize; }
		// Index of a block which contains the sample
		SDWORD blockNumber(SDWORD pos){ return pos/blockSize; }
		// Index inside the block
		SDWORD inBlockNumber(SDWORD pos){ return pos % blockSize; }
		
	};

	typedef CAudioBlockPeakMeter<double> CDefaultBlockPeakMeter;




	template<class T, class U> class CAudioRMSLevelMeter: public CBufferedAudioLevelMeter<T>
	{
	public:
		CAudioRMSLevelMeter(SDWORD nsamples=16, SDWORD allocate=0):
			CBufferedAudioLevelMeter<T>(nsamples, allocate), 
			dBufSum(0.0) 
			//count(0)
		{}

		virtual T pushSample(T newsample){ 
			T sq = newsample*newsample;
			T repl = CBufferedAudioLevelMeter<T>::pushSample(sq); 

			// Split to 2 steps for higher precision (dBufSum is quite big)
			dBufSum-=repl;
			dBufSum+=sq;
				
			/*if( count++ > 200000 ){
				count = 0;
				revalidate();
			}*/
			
			if( dBufSum < 0.0 ) 
				dBufSum = 0.0;

			return repl;
		}

		T pushAudioSample(T newsample){ return pushSample(newsample); }

		VTINLINE T GetMeanSquare() const { return (this->bufferFill)? T(dBufSum)/T(this->bufferFill): 0; }
		virtual T getRawValue() const { return sqrt(GetMeanSquare()); }
		VTINLINE virtual bool levelOver(T lev){ return GetMeanSquare()>=(lev*lev); }
		virtual void restart(){ dBufSum = 0.0; /*count=0;*/ CBufferedAudioLevelMeter<T>::restart(); }
		virtual void revalidate(){ dBufSum = U(this->lBufferSum()); }
	protected:
		U dBufSum;
		//long count;
		virtual T bufToRaw(T x){ return sqrt((x>=0)?x:-x); }
		virtual T rawToBuf(T x){ return x*x; }
	};

	typedef CAudioRMSLevelMeter<double, long double> CDefaultRMSLevelMeter;


	// Simple "capacitor" meter which charges from input samples and discharges each tick.
	// The "capacitor" discharges linearly (not like a real capacitor, which is exponential).
	// This is the best solution for VU meters.
	template<class T> class CAudioLevelCapacitorMeter: public CAudioLevelMeter<T>
	{
	public:
		CAudioLevelCapacitorMeter(T flloff=T(0.00005)):
		  falloff(flloff),
		  cap(0),
		  measSamples(0)
		{}
	    
		virtual void dischargeSmp(T nsamples){ falloff = T(1.0)/T(nsamples); }
		virtual void dischargeMs(double msec, double srate){ 
			falloff = T(1.0)/T(ms2samples(msec, srate)); }
		
		T getLevelFalloff(T level, SDWORD nsamples){ return level-T(nsamples)*falloff; }
		T getLevelFalloff(T level, T msec, T srate){ return level-T(ms2samples(msec, srate))*falloff; }

		VTINLINE virtual T getRawValue() const { return (cap>T(1.0))?T(1.0):cap; }
		VTINLINE void capDischarge(){ cap-=falloff; if(cap<T(0.0)) cap=T(0.0); }

		virtual SDWORD measuringSamples() const { return measSamples; }

		virtual T pushSample(T newsample){ 
			capDischarge();
			cap=Limit01(Max(cap, Abs(newsample))); 
			return 0; 
		}	
		
		T pushAudioSample(T newsample){ return pushSample(newsample); }
		virtual void restart(){ cap=0.0; };

		T getCharge() const { return cap; }
		T getFalloff() const { return falloff; }

	protected:
		T cap, falloff;
		virtual T rawToBuf(T x){ return Abs(x); }
		SDWORD measSamples;
	};

	typedef CAudioLevelCapacitorMeter<double> CDefaultCapacitorMeter;


	// "Capacitor" level meter designed specially to measure if level was 
	// over a specified threshold or not, during 'ttl' ms time.
	template<class T> class CThresholdAudioLevelCapacitorMeter: public CAudioLevelCapacitorMeter<T>
	{
	public:
		CThresholdAudioLevelCapacitorMeter():
			CAudioLevelCapacitorMeter<T>(),
			threshold(1.0),
			capacity(2.0)
		{}  

		// Setup the meter, so when 'thresh' or higher level is entered, the cap
		// will stay over 'thresh' for 'ttl' milliseconds.
		virtual void setup(double thresh, double ttl, double srate){ 		
			threshold = T(thresh);
			capacity = T(threshold * 2.0); 
			this->measSamples = SDWORD(ms2samples(ttl, srate));
			this->falloff = T( thresh / double(this->measSamples)); 
			if( this->cap > capacity )
				this->cap = capacity;
		}

		virtual T pushSample(T newsample){ 
			this->capDischarge();
			T absval=(newsample>=0)?newsample:-newsample;
			if( absval >= threshold ){
				//T set=absval * multiplier;
				//if( set>capacity )
					this->cap = capacity;
				//else if( set>cap)
				//	cap = set;
			}
			return 0; 
		}

		T getCapacity() const { return capacity; }

	protected:
		T threshold, capacity;		
	};


	typedef CThresholdAudioLevelCapacitorMeter<double> CDefaultThresholdAudioLevelCapacitorMeter;






	//////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////
	//
	// Implementation of the template class functions
	//







	template<class T> template<class U> void CAudioLevelMeter<T>::reloadFrom(CCircularBuffer<U> *bu, int offsetfromnewest)
	{
		restart();
		for(int i=measuringSamples()-1; i>=0; i--)
			pushAudioSample( T(bu->sampleFN( offsetfromnewest+i )) );
		this->revalidate();
	}





	template<class T> SDWORD CBufferedAudioLevelMeter<T>::resize(SDWORD nsamples, bool onthefly)
	{	
		if( !buffer || nsamples!=bufferSize || !onthefly || bufferAllocSize){
			
			if( bufferAllocSize>0 ){

				// "Static" buffer.
				if( nsamples > bufferAllocSize )
					nsamples = bufferAllocSize;
				if( !buffer )
					buffer = new T[ bufferAllocSize ];
				bufferFill = 0; 	
				bufferPos = 0; 
				bufferSize = nsamples;		
				this->revalidate();
				return nsamples;
			
			}else{

				// Moveable buffer
				if( buffer )
					if( bufferFill>0 && onthefly){
						T *oldbuf = buffer;
						buffer = new T[ nsamples ];
						int copysamples=(bufferFill>nsamples)?nsamples:bufferFill;
						int copysrcptr=bufferPos;				
						for(int i=0; i<copysamples; i++){
							if( copysrcptr < 0 )
								copysrcptr = bufferSize-1;
							buffer[i]=oldbuf[copysrcptr--];
						}
						delete oldbuf;
						bufferFill = copysamples;
						bufferPos = copysamples;
						bufferSize = nsamples;
						this->revalidate();
						return nsamples; // all done
					}else
						delete buffer;

				buffer = new T[ nsamples ];
				bufferFill = 0; 	
				bufferPos = 0; 
				bufferSize = nsamples;		
				this->revalidate();
				return bufferSize;
			}
		}
		this->revalidate();
		return nsamples;
	}

	template<class T> T CBufferedAudioLevelMeter<T>::pushSample(T newsample)
	{
		T replacedSample;
		if( bufferFill == bufferSize ){
			// Buffer is full, so cycling through it using bufferPos
			if( bufferPos == bufferFill ){
				bufferPos = 0;
				replacedSample = buffer[0];
				buffer[0] = newsample;
			}else{
				replacedSample = buffer[ bufferPos ];
				buffer[ bufferPos++ ] = newsample;
			}
		}else{
			// Buffer is still being filled
			bufferFill++;
			buffer[ bufferPos++ ] = newsample;
			replacedSample = 0.0;
		}	
		return replacedSample;
	}

	template<class T> T CBufferedAudioLevelMeter<T>::bufferMax()
	{
		T max=0;
		for(int i=0; i<bufferFill; i++)
			if( buffer[i]>max ) 
				max = buffer[i];
		return max;	
	}

	template<class T> SDWORD CBufferedAudioLevelMeter<T>::bufferMax(T *mx)
	{
		T max=0;
		SDWORD pos=-1;
		int i;
		for(i=bufferPos; i<bufferFill; i++)
			if( buffer[i]>max ){
				max = buffer[i];
				pos = i;
			}
		for(i=0; i<bufferPos; i++)
			if( buffer[i]>max ){
				max = buffer[i];
				pos = i;
			}
		*mx = max;
		return pos;	
	}

	template<class T> T CBufferedAudioLevelMeter<T>::bufferSum()
	{
		T sum=0;
		for(int i=0; i<bufferFill; i++)
			sum+=buffer[i];
		return sum;	
	}

	template<class T> long double CBufferedAudioLevelMeter<T>::lBufferSum() const
	{
		long double sum=0;
		for(int i=0; i<bufferFill; i++)
			sum+=(long double)buffer[i];
		return sum;	
	}


	template<class T> template<class U> void CBufferedAudioLevelMeter<T>::reloadFrom(CCircularBuffer<U> *bu, int offsetfromnewest)
	{
		for(int i=0; i<bufferSize; i++)
			buffer[bufferSize-1-i] = rawToBuf(T(bu->sampleFN( offsetfromnewest+i )));
		bufferPos = bufferSize;
		bufferFill = bufferSize;
		this->revalidate();
	}


	template<class T> bool CBufferedAudioLevelMeter<T>::loadFromSameSize(CBufferedAudioLevelMeter *mtr)
	{
		if( mtr->bufferSize != bufferSize )
			return false;
		bufferFill = mtr->bufferFill;
		bufferPos = mtr->bufferPos;
		for(int i=0; i<bufferFill; i++)
			buffer[ i ] = rawToBuf( mtr->bufToRaw( mtr->buffer[i] ) );
		this->revalidate();
		return true;
	}


	template<class T> T CAudioPeakMeter<T>::pushSample(T newsample)
	{
		T absval = (newsample>0)?newsample:-newsample;
		T replacedSample = CBufferedAudioLevelMeter<T>::pushSample( absval ); 	
		if( absval >= maxVal ){ // Level is raising or the same
			maxVal = absval;
			maxPos = this->getLastInsertPos();
		
		}else{
			//if( replacedSample >= maxVal && replacedSample > 0 ) 
			if( this->getLastInsertPos()==maxPos )
				maxPos = bufferMax( &maxVal );	
			// else - do nothing, the maximum is still somewhere inside the buffer
		}
		return replacedSample;
	}


	template<class T> template<class U> 
		void CAudioPeakMeter<T>::reloadFromStereo(CCircularBuffer<U> *bu1, CCircularBuffer<U> *bu2, int offsetfromnewest)
	{
		for(int i=0; i<this->bufferSize; i++)
			this->buffer[this->bufferSize-1-i] = 
				T(Max( Abs(bu1->sampleFN( offsetfromnewest+i )), Abs(bu2->sampleFN( offsetfromnewest+i )) ));
		this->bufferPos = this->bufferSize;
		this->bufferFill = this->bufferSize;
		revalidate();
	}






	template<class T> CAudioBlockPeakMeter<T>::CAudioBlockPeakMeter(SDWORD nsamples, SDWORD allocate, SDWORD blocksize):
		CAudioPeakMeter<T>(nsamples, allocate),
		blockSize(blocksize),
		ovwrMax(0),
		ovwrMaxPos(-1)
	{
		SDWORD nb = numBlocks( ( allocate )? allocate: nsamples );
		blockMax = new T[ nb ];	
		blockMaxPos = new SDWORD [ nb ];
	}


	template<class T> SDWORD CAudioBlockPeakMeter<T>::resize(SDWORD nsamples, bool onthefly)
	{
		if( this->bufferAllocSize<=0 ){ // Have to resize blockMax
			if(blockMax) delete blockMax;
			if(blockMaxPos) delete blockMaxPos;
			SDWORD nb = numBlocks( nsamples );
			blockMax = new T[ nb ];	
			blockMaxPos = new SDWORD [ nb ];		
		}
		return CAudioPeakMeter<T>::resize(nsamples, onthefly);	
	}


	template<class T> void CAudioBlockPeakMeter<T>::revalidate()
	{
		SDWORD blocknum = 0, in_block_pos = 0;
		T max = 0;
		this->maxVal = 0;
		ovwrMax = 0;
		ovwrMaxPos = -1;

		for(SDWORD i=0; i<this->bufferFill; i++){
			if( this->buffer[i] > max ){
				max = this->buffer[i];
				blockMaxPos[ blocknum ] = i;
				if( max >= this->maxVal ){
					this->maxVal = max;
					this->maxPos = i;
				}
			}		
			if(++in_block_pos == blockSize){
				blockMax[ blocknum ] = max;
				in_block_pos = 0;
				blocknum++;
				max=0;
			}
		}
	}

	template<class T> T CAudioBlockPeakMeter<T>::pushSample(T newsample)
	{
		T absval = (newsample>0)?newsample:-newsample;
		T replacedSample = CBufferedAudioLevelMeter<T>::pushSample( absval ); 	

		SDWORD lastinsert = this->getLastInsertPos();
		SDWORD blockidx = blockNumber( lastinsert );
		SDWORD inbn = inBlockNumber( lastinsert );
			
		
		if( !inbn || absval > blockMax[ blockidx ] ){	

			// We just got into a new block

			// If we didn't just overwrite block maxinum, we can use it for ovwr...
			if( !inbn && this->isFull() && blockMaxPos[ blockidx ]!=lastinsert ){			
				ovwrMax = blockMax[ blockidx ];
				ovwrMaxPos = blockMaxPos[ blockidx ];
			}else
				ovwrMaxPos = -1;

			// Restart block's max 

			blockMax[ blockidx ] = absval;
			blockMaxPos[ blockidx ] = lastinsert;		
		
		}else{
			// We are moving inside a block
			if( lastinsert==ovwrMaxPos)
				ovwrMaxPos = -1;
		}

		if( absval >= this->maxVal ){ 
			this->maxVal = absval;
			this->maxPos = lastinsert;
		
		}else{
			if( lastinsert==this->maxPos ){
				// Look for max position through blocks which are not
				// currently being overwritten
				SDWORD nb = numBlocks( this->bufferSize );

				this->maxVal = 0;
				//maxPos = 0;
				for(SDWORD b=0; b<nb; b++)
					if( blockMax[b]>=this->maxVal ){
						this->maxVal = blockMax[b];
						this->maxPos = blockMaxPos[b];
					}

				// Look for max postion through the block which is being overwritten
				if( this->isFull() ){
					if( ovwrMaxPos>=0 ){
						if( ovwrMax>this->maxVal ){
							this->maxVal = ovwrMax;
							this->maxPos = ovwrMaxPos;
						}
					}else{

						SDWORD blockend = (blockidx+1) * blockSize - 1;
						ovwrMax = 0;
						for(SDWORD i=lastinsert+1; i<=blockend; i++){
							if( this->buffer[i] >= this->maxVal ){
								this->maxVal = this->buffer[i];
								this->maxPos = i;
							}
							if( this->buffer[i] >= ovwrMax ){
								ovwrMax = this->buffer[i];
								ovwrMaxPos = i;
							}
						}
					}
				}
			}
		}

		return replacedSample;

	}

} // namespace VT

#endif

