//-----------------------------------------------------------------------------
//
// This file contains misc stuff which is useful for any MIDI processing.
//
//-----------------------------------------------------------------------------

/*
	VT Framework License (BSL 1.0)

	Copyright (C) 2007 Sergey A. Galin 
	E-mail: sergey.galin@gmail.com

	Permission is hereby granted, free of charge, to any person or organization
	obtaining a copy of the software and accompanying documentation covered by
	this license (the "Software") to use, reproduce, display, distribute,
	execute, and transmit the Software, and to prepare derivative works of the
	Software, and to permit third-parties to whom the Software is furnished to
	do so, all subject to the following:

	The copyright notices in the Software and this entire statement, including
	the above license grant, this restriction and the following disclaimer,
	must be included in all copies of the Software, in whole or in part, and
	all derivative works of the Software, unless such copies or derivative
	works are solely in the form of machine-executable object code generated by
	a source language processor.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
	SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
	FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
	ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
	DEALINGS IN THE SOFTWARE.
*/

#include <math.h>
#include "vt_midi.hpp"
#include "vt_debug.hpp"
#include "vt_localization.hpp"

#if defined(VT_MIDI_OCTAVE_0_11)
	#ifdef _MSC_VER
		#pragma message("VT_MIDI: enabled legacy octave numbering (0..11)")
	#else
		#warning VT_MIDI: enabled legacy octave numbering (0..11)
	#endif
#endif

using namespace VT;
using namespace VT::MIDI;

const char Notes::kNoteNames[12][3]={ 
	"C", "C#", "D", "D#", "E", "F", "F#", "G", "G#", "A", "A#","B" 
};

/*
	The MIDI system for electronic musical instruments and computers uses a 
	straight count starting with note 0 for C-1 at 8.1758 Hz up to note 
	127 for G9 at 12,544 Hz.
*/

// Wasting 512 bytes in executable module, yeah!

#ifdef VT_MIDI_OCTAVE_COMPLIANCE

// MIDI Table generator, Oct 10 2008
const float Notes::kNoteFrequencies[128] = {
	8.17579936981f,		// C-1		000		00
	8.66195678711f,		// C#-1		001		01
	9.17702388763f,		// D-1		002		02
	9.72271823883f,		// D#-1		003		03
	10.30086135864f,	// E-1		004		04
	10.91338253021f,	// F-1		005		05
	11.56232547760f,	// F#-1		006		06
	12.24985694885f,	// G-1		007		07
	12.97827148438f,	// G#-1		008		08
	13.75000000000f,	// A-1		009		09
	14.56761741638f,	// A#-1		010		0a
	15.43385314941f,	// B-1		011		0b
	16.35159873962f,	// C0		012		0c
	17.32391357422f,	// C#0		013		0d
	18.35404777527f,	// D0		014		0e
	19.44543647766f,	// D#0		015		0f
	20.60172271729f,	// E0		016		10
	21.82676506042f,	// F0		017		11
	23.12465095520f,	// F#0		018		12
	24.49971389771f,	// G0		019		13
	25.95654296875f,	// G#0		020		14
	27.50000000000f,	// A0		021		15
	29.13523483276f,	// A#0		022		16
	30.86770629883f,	// B0		023		17
	32.70319747925f,	// C1		024		18
	34.64782714844f,	// C#1		025		19
	36.70809555054f,	// D1		026		1a
	38.89087295532f,	// D#1		027		1b
	41.20344543457f,	// E1		028		1c
	43.65353012085f,	// F1		029		1d
	46.24930191040f,	// F#1		030		1e
	48.99942779541f,	// G1		031		1f
	51.91308593750f,	// G#1		032		20
	55.00000000000f,	// A1		033		21
	58.27046966553f,	// A#1		034		22
	61.73541259766f,	// B1		035		23
	65.40639495850f,	// C2		036		24
	69.29565429688f,	// C#2		037		25
	73.41619110107f,	// D2		038		26
	77.78174591064f,	// D#2		039		27
	82.40689086914f,	// E2		040		28
	87.30706024170f,	// F2		041		29
	92.49860382080f,	// F#2		042		2a
	97.99885559082f,	// G2		043		2b
	103.82617187500f,	// G#2		044		2c
	110.00000000000f,	// A2		045		2d
	116.54093933105f,	// A#2		046		2e
	123.47082519531f,	// B2		047		2f
	130.81278991699f,	// C3		048		30
	138.59130859375f,	// C#3		049		31
	146.83238220215f,	// D3		050		32
	155.56349182129f,	// D#3		051		33
	164.81378173828f,	// E3		052		34
	174.61412048340f,	// F3		053		35
	184.99720764160f,	// F#3		054		36
	195.99771118164f,	// G3		055		37
	207.65234375000f,	// G#3		056		38
	220.00000000000f,	// A3		057		39
	233.08187866211f,	// A#3		058		3a
	246.94165039063f,	// B3		059		3b
	261.62557983398f,	// C4		060		3c
	277.18261718750f,	// C#4		061		3d
	293.66476440430f,	// D4		062		3e
	311.12698364258f,	// D#4		063		3f
	329.62756347656f,	// E4		064		40
	349.22824096680f,	// F4		065		41
	369.99441528320f,	// F#4		066		42
	391.99542236328f,	// G4		067		43
	415.30468750000f,	// G#4		068		44
	440.00000000000f,	// A4		069		45 (!!)
	466.16375732422f,	// A#4		070		46
	493.88330078125f,	// B4		071		47
	523.25115966797f,	// C5		072		48
	554.36523437500f,	// C#5		073		49
	587.32952880859f,	// D5		074		4a
	622.25396728516f,	// D#5		075		4b
	659.25512695313f,	// E5		076		4c
	698.45648193359f,	// F5		077		4d
	739.98883056641f,	// F#5		078		4e
	783.99084472656f,	// G5		079		4f
	830.60937500000f,	// G#5		080		50
	880.00000000000f,	// A5		081		51
	932.32751464844f,	// A#5		082		52
	987.76660156250f,	// B5		083		53
	1046.50231933594f,	// C6		084		54
	1108.73046875000f,	// C#6		085		55
	1174.65905761719f,	// D6		086		56
	1244.50793457031f,	// D#6		087		57
	1318.51025390625f,	// E6		088		58
	1396.91296386719f,	// F6		089		59
	1479.97766113281f,	// F#6		090		5a
	1567.98168945313f,	// G6		091		5b
	1661.21875000000f,	// G#6		092		5c
	1760.00000000000f,	// A6		093		5d
	1864.65502929688f,	// A#6		094		5e
	1975.53320312500f,	// B6		095		5f
	2093.00463867188f,	// C7		096		60
	2217.46093750000f,	// C#7		097		61
	2349.31811523438f,	// D7		098		62
	2489.01586914063f,	// D#7		099		63
	2637.02050781250f,	// E7		100		64
	2793.82592773438f,	// F7		101		65
	2959.95532226563f,	// F#7		102		66
	3135.96337890625f,	// G7		103		67
	3322.43750000000f,	// G#7		104		68
	3520.00000000000f,	// A7		105		69
	3729.31005859375f,	// A#7		106		6a
	3951.06640625000f,	// B7		107		6b
	4186.00927734375f,	// C8		108		6c
	4434.92187500000f,	// C#8		109		6d
	4698.63623046875f,	// D8		110		6e
	4978.03173828125f,	// D#8		111		6f
	5274.04101562500f,	// E8		112		70
	5587.65185546875f,	// F8		113		71
	5919.91064453125f,	// F#8		114		72
	6271.92675781250f,	// G8		115		73
	6644.87500000000f,	// G#8		116		74
	7040.00000000000f,	// A8		117		75
	7458.62011718750f,	// A#8		118		76
	7902.13281250000f,	// B8		119		77
	8372.01855468750f,	// C9		120		78
	8869.84375000000f,	// C#9		121		79
	9397.27246093750f,	// D9		122		7a
	9956.06347656250f,	// D#9		123		7b
	10548.08203125000f,	// E9		124		7c
	11175.30371093750f,	// F9		125		7d
	11839.82128906250f,	// F#9		126		7e
	12543.85351562500f,	// G9		127		7f
};


#else


const float Notes::kNoteFrequencies[128]={ 
	16.35159783129f,	// C0		000		00
	17.32391443605f,	// C#0		001		01
	18.35404799484f,	// D0		002		02
	19.44543648263f,	// D#0		003		03
	20.60172230705f,	// E0		004		04
	21.82676446456f,	// F0		005		05
	23.12465141948f,	// F#0		006		06
	24.49971474886f,	// G0		007		07
	25.95654359875f,	// G#0		008		08
	27.50000000000f,	// A0		009		09
	29.13523509488f,	// A#0		010		0a
	30.86770632851f,	// B0		011		0b
	32.70319566257f,	// C1		012		0c
	34.64782887211f,	// C#1		013		0d
	36.70809598968f,	// D1		014		0e
	38.89087296526f,	// D#1		015		0f
	41.20344461411f,	// E1		016		10
	43.65352892913f,	// F1		017		11
	46.24930283895f,	// F#1		018		12
	48.99942949772f,	// G1		019		13
	51.91308719749f,	// G#1		020		14
	55.00000000000f,	// A1		021		15
	58.27047018976f,	// A#1		022		16
	61.73541265702f,	// B1		023		17
	65.40639132515f,	// C2		024		18
	69.29565774422f,	// C#2		025		19
	73.41619197935f,	// D2		026		1a
	77.78174593052f,	// D#2		027		1b
	82.40688922822f,	// E2		028		1c
	87.30705785825f,	// F2		029		1d
	92.49860567791f,	// F#2		030		1e
	97.99885899544f,	// G2		031		1f
	103.82617439499f,	// G#2		032		20
	110.00000000000f,	// A2		033		21
	116.54094037952f,	// A#2		034		22
	123.47082531403f,	// B2		035		23
	130.81278265030f,	// C3		036		24
	138.59131548844f,	// C#3		037		25
	146.83238395870f,	// D3		038		26
	155.56349186104f,	// D#3		039		27
	164.81377845643f,	// E3		040		28
	174.61411571650f,	// F3		041		29
	184.99721135582f,	// F#3		042		2a
	195.99771799087f,	// G3		043		2b
	207.65234878997f,	// G#3		044		2c
	220.00000000000f,	// A3		045		2d
	233.08188075905f,	// A#3		046		2e
	246.94165062806f,	// B3		047		2f
	261.62556530060f,	// C4		048		30
	277.18263097687f,	// C#4		049		31
	293.66476791741f,	// D4		050		32
	311.12698372208f,	// D#4		051		33
	329.62755691287f,	// E4		052		34
	349.22823143300f,	// F4		053		35
	369.99442271163f,	// F#4		054		36
	391.99543598175f,	// G4		055		37
	415.30469757995f,	// G#4		056		38
	440.00000000000f,	// A4		057		39
	466.16376151809f,	// A#4		058		3a
	493.88330125612f,	// B4		059		3b
	523.25113060120f,	// C5		060		3c
	554.36526195374f,	// C#5		061		3d
	587.32953583482f,	// D5		062		3e
	622.25396744416f,	// D#5		063		3f
	659.25511382574f,	// E5		064		40
	698.45646286601f,	// F5		065		41
	739.98884542327f,	// F#5		066		42
	783.99087196350f,	// G5		067		43
	830.60939515989f,	// G#5		068		44
	880.00000000000f,	// A5		060		45
	932.32752303618f,	// A#5		070		46
	987.76660251225f,	// B5		071		47
	1046.50226120239f,	// C6		072		48
	1108.73052390749f,	// C#6		073		49
	1174.65907166963f,	// D6		074		4a
	1244.50793488832f,	// D#6		075		4b
	1318.51022765148f,	// E6		076		4c
	1396.91292573202f,	// F6		077		4d
	1479.97769084654f,	// F#6		078		4e
	1567.98174392700f,	// G6		079		4f
	1661.21879031978f,	// G#6		080		50
	1760.00000000000f,	// A6		081		51
	1864.65504607236f,	// A#6		082		52
	1975.53320502450f,	// B6		083		53
	2093.00452240479f,	// C7		084		54
	2217.46104781498f,	// C#7		085		55
	2349.31814333926f,	// D7		086		56
	2489.01586977665f,	// D#7		087		57
	2637.02045530296f,	// E7		088		58
	2793.82585146403f,	// F7		089		59
	2959.95538169308f,	// F#7		090		5a
	3135.96348785400f,	// G7		091		5b
	3322.43758063956f,	// G#7		092		5c
	3520.00000000000f,	// A7		093		5d	
	3729.31009214472f,	// A#7		094		5e
	3951.06641004899f,	// B7		095		5f
	4186.00904480958f,	// C8		096		60
	4434.92209562995f,	// C#8		097		61
	4698.63628667852f,	// D8		098		62
	4978.03173955329f,	// D#8		099		63
	5274.04091060592f,	// E8		100		64
	5587.65170292806f,	// F8		101		65
	5919.91076338615f,	// F#8		102		66
	6271.92697570799f,	// G8		103		67
	6644.87516127912f,	// G#8		104		68
	7040.00000000000f,	// A8		105		69
	7458.62018428944f,	// A#8		106		6a
	7902.13282009799f,	// B8		107		6b
	8372.01808961916f,	// C9		108		6c
	8869.84419125991f,	// C#9		109		6d
	9397.27257335704f,	// D9		110		6e
	9956.06347910659f,	// D#9		111		6f
	10548.08182121184f,	// E9		112		70
	11175.30340585613f,	// F9		113		71
	11839.82152677230f,	// F#9		114		72
	12543.85395141598f,	// G9		115		73
	13289.75032255825f,	// G#9		116		74
	14080.00000000000f,	// A9		117		75
	14917.24036857888f,	// A#9		118		76
	15804.26564019598f,	// B9		119		77
	16744.03617923831f,	// C10		120		78
	17739.68838251981f,	// C#10		121		79
	18794.54514671409f,	// D10		122		7a
	19912.12695821318f,	// D#10		123		7b
	21096.16364242368f,	// E10		124		7c
	22350.60681171225f,	// F10		125		7d
	23679.64305354461f,	// F#10		126		8e
	25087.70790283196f	// G10		127		7f
};

#endif

VT::MIDI::Notes::TBlackKeyShift VT::MIDI::Notes::kBlackKeyShifts[ VT::MIDI::Notes::kNumberOfNotesInOctave ] = 
{
	VT::MIDI::Notes::kBKSCen, // C
	VT::MIDI::Notes::kBKSLft, // C#
	VT::MIDI::Notes::kBKSCen, // D
	VT::MIDI::Notes::kBKSRgt, // D#
	VT::MIDI::Notes::kBKSCen, // E
	VT::MIDI::Notes::kBKSCen, // F
	VT::MIDI::Notes::kBKSLft, // F#
	VT::MIDI::Notes::kBKSCen, // G
	VT::MIDI::Notes::kBKSCen, // G#
	VT::MIDI::Notes::kBKSCen, // A
	VT::MIDI::Notes::kBKSRgt, // A#
	VT::MIDI::Notes::kBKSCen  // B
};







BYTE VT::MIDI::Notes::CountWhiteKeys( BYTE startingNote, BYTE count )
{
	if( count==0 )
		return 0;
	VTASSERT( startingNote<128 );
	VTASSERT( count<128 );
	VTASSERT( int(startingNote)+int(count)-1 < 128 );
	BYTE endNote = startingNote+count-1, ret = 0;
	for( BYTE i=startingNote; i<=endNote; i++ )
		if( VT::MIDI::Notes::IsWhiteKey( i ) )
			ret++;
	return ret;
}

void VT::MIDI::Notes::GetFullNoteName(BYTE iNote, LPVTCHAR out)
{
	VTSPRINTF( out, _T("%s%d"), GetNoteName( iNote ), 
		GetOctaveNumber( iNote ) );
}

void VT::MIDI::TMIDIMessage::ToDebugString( LPVTCHAR out, size_t out_size ) const
{
	if( !out_size || out==NULL ){
		VTDEBUGMSGT("No output buffer in MIDI message to string converter.");
		throw _T("No output buffer in MIDI message to string converter.");
	}
	VTCHAR buf[256], notebuf[VT::MIDI::Notes::kMaxFullNoteNameSize], note[16];
	note[0]=0;

	LPCVTCHAR msgtype=_T("??");

	if( IsNote() ){
		if( IsNoteOn() ){
			msgtype = _T("NOTE ON");
		}else if( IsNoteOff() ){
			msgtype = _T("NOTE OFF");
		}else
			msgtype = _T("NOTE ??");
		VT::MIDI::Notes::GetFullNoteName( GetNote(), notebuf );
		VTSPRINTF(note, _T(" %s"), notebuf );
	} 
	// !!!!!!!!!!!!!! support more types

	VTSPRINTF( buf, _T("%s%s; %02x:%02x:%02x:%02x"), 
		msgtype, note, command, channel, data1, data2 );

	VTSTRNCPY( out, buf, out_size-1 );
}

