//-----------------------------------------------------------------------------
//
// This file contains misc stuff which is useful for any MIDI processing.
//
//-----------------------------------------------------------------------------

/*
	VT Framework License (BSL 1.0)

	Copyright (C) 2007 Sergey A. Galin 
	E-mail: sergey.galin@gmail.com

	Permission is hereby granted, free of charge, to any person or organization
	obtaining a copy of the software and accompanying documentation covered by
	this license (the "Software") to use, reproduce, display, distribute,
	execute, and transmit the Software, and to prepare derivative works of the
	Software, and to permit third-parties to whom the Software is furnished to
	do so, all subject to the following:

	The copyright notices in the Software and this entire statement, including
	the above license grant, this restriction and the following disclaimer,
	must be included in all copies of the Software, in whole or in part, and
	all derivative works of the Software, unless such copies or derivative
	works are solely in the form of machine-executable object code generated by
	a source language processor.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
	SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
	FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
	ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
	DEALINGS IN THE SOFTWARE.
*/

#ifndef VT_MIDI_H
#define VT_MIDI_H

#include <math.h>
#include "vt_audiomath.hpp"

#if defined(VT_MIDI_OCTAVE_COMPLIANCE) && defined(VT_MIDI_OCTAVE_0_11)
	#error Both VT_MIDI_OCTAVE_COMPLIANCE and VT_MIDI_OCTAVE_0_11 cannot be defined!
#endif

#if !defined(VT_MIDI_OCTAVE_COMPLIANCE) && !defined(VT_MIDI_OCTAVE_0_11)
	#define VT_MIDI_OCTAVE_COMPLIANCE
#endif

namespace VT {
	namespace MIDI {
		const BYTE 
			// Commands
			MIDI_NOTEOFF				= 0x80, // Note off event
			MIDI_NOTEON					= 0x90, // Note on event
			MIDI_POLYPRESSURE			= 0xA0, // Polyphonic Key Pressure (Aftertouch)
			MIDI_CONTROLCHANGE			= 0xB0, // Control Change
			MIDI_PROGRAMCHANGE			= 0xC0, // Program change
			MIDI_CHANPRESSURE			= 0xD0, // Channel Pressure (After-touch).
			MIDI_PITCHBEND				= 0xE0, // Pitch Wheel Change
			
			// System Common messages 
			MIDI_SYSEX					= 0xF0, // Start of exclusive
			MIDI_SYSEXEND				= 0xF7, // End of exclusive
			MIDI_SONGPOINTER			= 0xF2, // Song position pointer
			MIDI_SONGSELECT				= 0xF3, // Song select
			MIDI_TUNEREQUEST			= 0xF6, // Tune request
			
			// System Real-Time Messages 
			MIDI_TIMINGCLOCK			= 0xF8, // Timing clock
			MIDI_RESET					= 0xFF, // Reset
			MIDI_ACTIVESENSING			= 0xFE, // Active sensing
			MIDI_START					= 0xFA, // Start the sequence
			MIDI_STOP					= 0xFC, // Stop the sequence


			// Control Indexes (when command == MIDI_CONTROLCHANGE)
			kControlModulation			= 0x01, // 001
			kControlVolume				= 0x07, // 007
			kControlPan					= 0x0A, // 010
			kControlExpression			= 0x0B, // 011
			kControlPedalSustain		= 0x40, // 064
			kControlSostunetoThumby		= 0x42, // 066
			kControlPedalSoft			= 0x43, // 067
			kControlHarmonicContent		= 0x47, // 071
			kControlReleaseTime			= 0x48, // 072
			kControlAttackTime			= 0x49, // 073
			kControlBrightness			= 0x4A, // 074
			kControlReverbDepth			= 0x5B, // 091
			kControlChorusDepth			= 0x5D, // 093
			kControlDataIncrement		= 0x60, // 096
			kControlDataDecrement		= 0x61, // 097		

			kDefaultVelocity			= 100,	// Default velocity
			kErrorNote					= 255;  // Note number to use as invalid note indicator

		const int
			kMinPitch					= -8192, // Pitch bend == 00 00 (see comments to Pitch Bend functions)
			kMaxPitch					= 8191;	 // Pitch bend == 7F 7F (see comments to Pitch Bend functions)

		namespace GM2{
			const BYTE
				kAllSoundOff			= 120, // 0x78
				kResetAllControllers	= 121, // 0x79
				kAllNotesOff			= 123; // 0x7B	
		};	

		struct TMIDIMessage
		{
			// !!!!PORTABILITY!!!! The bytes should have proper order in DWORD.
			BYTE 
				command, // MIDI command, when decoded from MIDI data it's contains no channel info
				channel, // Channel, 1-16 (0 = not applicable)
				data1,   // E.g.: Note number - meaning varies depending on command
				data2;   // E.g.: Velocity - meaning varies depending on command
			
			VTINLINE void ZeroMessage(){ *((DWORD*)this)=0; } 
			TMIDIMessage(){ ZeroMessage(); } // An empty MIDI message.
			TMIDIMessage(const BYTE* data){ Assign(data); }
			TMIDIMessage(BYTE status, BYTE databyt1, BYTE databyt2){ 
				Assign(status, databyt1, databyt2); }
			
			VTINLINE DWORD AsDWORD() const { return *((DWORD*)this); } 
			VTINLINE void FromDWORD(DWORD value) const { *((DWORD*)this)=value; }

			static VTINLINE bool StatusContainsChannel(BYTE status){
				return status>=0x80 && status<=0xEF;
			}
			
			// Decode TMIDIMessage from standard MIDI byte sequence:
			// [status][data1][data2], 'data' points to the sequence.
			void Assign(const BYTE* data){
				if( StatusContainsChannel(data[0]) ){
					command = data[0] & 0xf0; 
					channel = (data[0] & 0x0f) +1;
				}else{
					command = data[0];
					channel = 0; 
				}
				data1 = data[1];
				data2 = data[2];
			}

			// Decode TMIDIMessage from standard MIDI byte sequence, the bytes
			// of the sequence are passed as parameters.
			void Assign(BYTE status, BYTE databyt1, BYTE databyt2){
				if( StatusContainsChannel(status) ){
					command = status & 0xf0;
					channel = (status & 0x0f) +1;
				}else{
					command = status;
					channel = 0;
				}
				data1 = databyt1;
				data2 = databyt2;
			}

			// Check if the message changes currently played note.
			// MIDI_POLYPRESSURE is a different story!
			bool IsNote() const { 				
				return 
					command==MIDI_NOTEOFF || 
					command==MIDI_NOTEON; }

			// Some old-school devices use NOTE ON with zero velocity instead of NOTE OFF's.
			// Always use this function instead of checking for "command==MIDI_NOTEON".
			bool IsNoteOn() const {
				return command==MIDI_NOTEON && GetNoteVelocity()>0;
			}

			// Some old-school devices use NOTE ON with zero velocity instead of NOTE OFF's.
			// Always use this function instead of checking for "command==MIDI_NOTEOFF".
			bool IsNoteOff() const {
				return command==MIDI_NOTEOFF || 
					(command==MIDI_NOTEON && GetNoteVelocity()==0);
			}

			BYTE GetNote() const { return data1; }
			void SetNote( BYTE note ){ data1 = note; }
			BYTE GetNoteVelocity() const { return data2; }
			bool IsControllerChange() const { return command==MIDI_CONTROLCHANGE; }
			BYTE GetControllerNumber() const { return data1; }
			BYTE GetControllerValue() const { return data2; }

			bool IsPitchBend() const { return command==MIDI_PITCHBEND; }
			
			// Decode Pitch Bend
			// Pitch is encoded in data1 and data2, each varies from 0 to 0x7F (127).
			// Max value of pitch is encoded as 7F 7F: 127 + 127*128 == 16383
			// Default is 00 40: 0 + 64*128 == 8192
			// Returns 0..16383		
			unsigned GetPitchBendU() const { VTASSERT(data1<=0x7F && data1<=0x7F ); return (unsigned(data2)<<7) + unsigned(data1); }
			int GetPitchBend() const { return GetPitchBendU()-8192; } // Get pitch as -8192..8191		
			float GetPitchBendF() const { return float(GetPitchBend())/8192.0f; } // Get pitch as [-1..1)
			
			// Encode Pitch Bend
			void SetPitchBendU( unsigned pitch ){ command=MIDI_PITCHBEND; data2 = BYTE((pitch>>7)&0x7F); data1 = BYTE(pitch&0x7F); }
			void SetPitchBend( int pitch ){ SetPitchBendU( unsigned(Clip<int>(pitch+8192, 0, 16383)) ); }
			void SetPitchBendF( float pitch ){ SetPitchBend( Clip<int>( SRound( pitch*8192.0f ), -8192, 8191 ) ); }

			bool IsModulation() const { return command==MIDI_CONTROLCHANGE && data1==kControlModulation; }
			BYTE GetModulation() const { return data2; }
			float GetModulationF() const { return float(data2)/127.0f; }

			bool IsVolume() const { return command==MIDI_CONTROLCHANGE && data1==kControlVolume; }
			BYTE GetVolume() const { return data2; }
			float GetVolumeF() const { return float(data2)/127.0f; }

			void ToDebugString( LPVTCHAR out, size_t out_size ) const;

		};


		namespace Notes{
			
			// Some useful Note Frequency constants
			const double
				kNoteStep = 1.0594630943592952645618252949461, // Step between notes - root of 12 from 2			
				kA4freq = 440.0,                               // Standard frequency of A of 4th octave
				kC4freq = 261.62556530059863467784999352406,   // Standard frequency of  of 4th octave
				kC0freq = 16.351597831287414667365624595254,   // Standard frequency of  of 0th octave
				kNoteStepLog  = 0.057762265046662109118102676787859, // log(kNoteStep)
				kSonicSpeedMPS = 340.29;                       // Standard speed of sound, in meters per second

			extern const float kNoteFrequencies[128];	// Frequencies of all notes, in Hz
			extern const char kNoteNames[12][3];		// Table of note names matching TNoteNumber enumeration

			#if defined(VT_MIDI_OCTAVE_COMPLIANCE)
				// Standard MIDI octave range.
				const int kMinOctave = -1, kMaxOctave = 9;
			#else // defined VT_MIDI_OCTAVE_0_11
				// Octave range in some our old apps.
				const int kMinOctave = 0, kMaxOctave = 11;
			#endif
			
			// Note names enumerator
			enum TNoteNumber{
				C  = 0,		// C	do
				Cs = 1,		// C#	do-diez
				D  = 2,		// D	re
				Ds = 3,		// D#	re-diez
				E  = 4,		// E	mi
				F  = 5,		// F	fa
				Fs = 6,		// F#	fa-diez
				G  = 7,		// G	sol
				Gs = 8,		// G#	sol-diez
				A  = 9,		// A	la
				As = 10,	// A#	la-diez
				B  = 11,	// B	si
				kNumberOfNotesInOctave = 12
			};
		
			
			enum TBlackKeyShift{
				kBKSLft = -1,
				kBKSCen = 0,			
				kBKSRgt = 1
			};
			
			// How black keys are shifted in keyboard, relatively to space
			// between adjacement white keys. For white keys, all array
			// entries are kBKSCen.
			extern TBlackKeyShift kBlackKeyShifts[ kNumberOfNotesInOctave ];

			





			// Find octave of MIDI note 
			inline int GetOctaveNumber(BYTE iNote){ 
				#ifdef VT_MIDI_OCTAVE_COMPLIANCE
					return iNote / 12 - 1;
				#else
					return iNote / 12; 
				#endif
			}

			// Convert MIDI note number to in-octave note number
			inline int GetInOctaveNumber(BYTE iNote){ 
				return iNote % 12; 
			}

			inline TBlackKeyShift GetGeometryKeyShift(BYTE iNote){ return kBlackKeyShifts[GetInOctaveNumber(iNote)]; }
		
			// Get name of a note. Always returns correct name ignoring fact that
			// there are only 127 possible notes in MIDI.
			inline const char* GetNoteName(BYTE iNote){
				return kNoteNames[GetInOctaveNumber(iNote)];
			}

			BYTE CountWhiteKeys( BYTE startingNote, BYTE count );
			inline BYTE CountBlackKeys( BYTE startingNote, BYTE count ){ 
				return count-CountWhiteKeys( startingNote, count ); }

			const size_t kMaxFullNoteNameSize = 4;
			void GetFullNoteName(BYTE iNote, LPVTCHAR out);

			// This function isn't super fast, so better use GetNoteFrequency().
			// But you can use this function for calculating non-standard frequency grid
			// by passing another base frequency (C of 0th octave).
			inline double CalcNoteFrequencyHz(BYTE iNote, double kC0 = kC0freq){
				double zeroOctFreq = kC0 * pow(kNoteStep, double(GetInOctaveNumber(iNote)) );
				// return zeroOctFreq * double(1<<GetOctaveNumber(iNote)); // pow(2.0, GetOctaveNumber(iNote));
				return zeroOctFreq * pow(2.0, GetOctaveNumber(iNote));
			}

			// Safely find note frequency in Hertz (uses lookup table). 
			inline float GetNoteFrequency( BYTE iNote ){
				return (iNote<=127)? kNoteFrequencies[iNote]: float(CalcNoteFrequencyHz(iNote));
			}

			// Find note base frequency period length in samples (round if neccessary).
			inline float GetNoteWavelength( BYTE iNote, float sampleRate ){
				return sampleRate/GetNoteFrequency( iNote );
			}
			
			// If a key is not white, it is black ;-)
			inline bool IsBlackKey(BYTE iNote){ 
				// return !IsWhiteKey(iNote); 
				iNote = GetInOctaveNumber(iNote);
				return (iNote==Cs)||(iNote==Ds)||(iNote==Fs)||(iNote==Gs)||(iNote==As);
			}

			// Is it a "white" or "black" note?
			inline bool IsWhiteKey(BYTE iNote){
				// iNote = GetInOctaveNumber(iNote);
				// return (iNote==C)||(iNote==D)||(iNote==E)||(iNote==F)||(iNote==G)||(iNote==A)||(iNote==B);
				return !IsBlackKey(iNote); 
			}
		
			// Shift key / Transpose
			inline BYTE ChangeKey(BYTE iNote, int adjustment, int min=0, int max=127){
				VTASSERT( min<=127 && max<=127 );
				int newNote = int(iNote)+adjustment;
				while( newNote<min )
					newNote+=12;
				while( newNote>max )
					newNote-=12;
				return BYTE(newNote);
			}

			// Synonim for ChangeKey() ;)
			inline BYTE Transpose(BYTE iNote, int adjustment, int min=0, int max=127){
				return ChangeKey( iNote, adjustment, min, max ); }

			// Convert change in semitones to pitch change.
			// semitones: (-INF..+INF), pitch: (0..INF); semitones := 0 => pitch == 1.0
			// Unit test: notes.cpp.
			inline double SemitonesToPitch( double semitones ){	return pow( kNoteStep, semitones );	}		

			// Convert note frequency multiplier to tone change
			// pitch: (0..INF), semitones: (-INF..INF); pitch :=1 => semitones == 0
			// Unit test: notes.cpp.
			inline double PitchToSemitones( double pitch ){	return log( pitch )/kNoteStepLog; }

			inline double OctavesToPitch( double octaves ){	return pow( 2, octaves );	}
			inline double PitchToOctaves( double pitch ){ return log( pitch )/kLog_2; }


			// Calculate pitch difference between 2 notes. 
			// This function is quite fast because instead of using pow() it
			// simply divides note frequencies in the static note frequency table.
			inline float PitchDiffFst( BYTE srcNote, BYTE destNote ){
				VTASSERT( srcNote<128 && destNote<128 );
				return kNoteFrequencies[destNote]/kNoteFrequencies[srcNote];
			}

			// Calculate pitch difference between 2 notes (high precision).
			inline double PitchDiff( BYTE srcNote, BYTE destNote ){
				return pow( double(kNoteStep), double(int(destNote)-int(srcNote)) );
			}


		};


	} // namespace MIDI
} // namespace VT


#endif
