//-----------------------------------------------------------------------------
//
// MP3 encoder wrapper. Currently uses LAME DLL.
//
//-----------------------------------------------------------------------------

/*
	VT Framework License (BSL 1.0)

	Copyright (C) 2007 Sergey A. Galin 
	E-mail: sergey.galin@gmail.com

	Permission is hereby granted, free of charge, to any person or organization
	obtaining a copy of the software and accompanying documentation covered by
	this license (the "Software") to use, reproduce, display, distribute,
	execute, and transmit the Software, and to prepare derivative works of the
	Software, and to permit third-parties to whom the Software is furnished to
	do so, all subject to the following:

	The copyright notices in the Software and this entire statement, including
	the above license grant, this restriction and the following disclaimer,
	must be included in all copies of the Software, in whole or in part, and
	all derivative works of the Software, unless such copies or derivative
	works are solely in the form of machine-executable object code generated by
	a source language processor.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
	SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
	FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
	ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
	DEALINGS IN THE SOFTWARE.
*/

#include <stdlib.h>
#include "vt_messageboxes.hpp"
#include "vt_localization.hpp"
#include "vt_mp3encode.hpp"
#include "vt_fileutils.hpp"
#include "vt_riff.hpp"
#include "taglib.h"

using namespace VT;

const LPCVTCHAR kXMLHeader = _T("<?xml version=\"1.0\" encoding=\"utf-8\"?>\r\n");

const int MP3SimpleEncSettings::kSampleRates[] = {
	8000, 
	11025,
	12000,
	16000, 
	22050, 
	24000,
	32000,
	44100,
	48000,
	0
};

const int MP3SimpleEncSettings::kBitrates[] = { 
	8,  // 0, MPEG-2 only
	16, // 1, MPEG-2 only
	24, // 2, MPEG-2 only
	32, // 3
	40, // 4
	48, // 5
	56, // 6
	64, // 7
	80, // 8
	96, // 9
	112, // 10
	128, // 11 - used in FindBitrate()
	144, // 12, MPEG-2 only
	160, // 13
	192, // 14
	224, // 15
	256, // 16
	320, // 17
	0 
};

CCriticalSection CPCM2MP3FileEncoder::s_CS;

int MP3SimpleEncSettings::FindBitrate(int kbps, int def)
{
	for(int i=0; kBitrates[i]; i++)
		if( kBitrates[i]==kbps )
			return i;
	return def;
}

int MP3SimpleEncSettings::Idx2Bitrate(int idx)
{
	if( idx>=0 && idx<=17 )
		return kBitrates[idx];
	else
		return 128;
}

LPCVTCHAR MP3SimpleEncSettings::kBitratesStr[] = {
	_T("8 kbps (12 KHz or lower only)"),
	_T("16 kbps (12 KHz or lower only)"),
	_T("24 kbps (12 KHz or lower only)"),
	_T("32 kbps (AM Radio)"),
	_T("40 kbps"),
	_T("48 kbps"),
	_T("56 kbps"),
	_T("64 kbps"),
	_T("80 kbps"),	
	_T("96 kbps (FM Radio)"),
	_T("112 kbps"),
	_T("128 kbps (near-CD)"),	
	_T("144 kbps (12 KHz or lower only)"),	
	_T("160 kbps"),
	_T("192 kbps"),
	_T("224 kbps"),
	_T("256 kbps (studio)"),
	_T("320 kbps (extreme)"),
	NULL
};

void MP3SimpleEncSettings::ResetToDefaults()
{
	VTASSERT(sizeof(kBitratesStr) == sizeof(kBitrates));
	m_bVBR = true;
	m_bVBRHighQuality = false;
	m_bJointStereo = false;
	m_bHighQuality = true;
	m_bCopyright = false;
	m_bCRC = false;
	m_bOriginal = false;
	m_bPrivate = false;
	m_iMinBR = 128;
	m_iMaxBR = 256;
	m_iSrcNChannels = 2;
	m_iSrcSamplerate = 44100;
}

CPCM2MP3FileEncoder::CPCM2MP3FileEncoder(): 
	m_bDetectDataLength(false)
{ 
	m_szError[0]=0; 
}	

CPCM2MP3FileEncoder::CPCM2MP3FileEncoder(const MP3SimpleEncSettings& src): 
	m_bDetectDataLength(false)
{ 
	m_szError[0]=0; 
	GetSettings(src); 
}

void MP3SimpleEncSettings::CopyFrom(const MP3SimpleEncSettings& src)
{
	m_bVBR = src.m_bVBR;
	m_bVBRHighQuality = src.m_bVBRHighQuality;
	m_bJointStereo = src.m_bJointStereo;
	m_bHighQuality = src.m_bHighQuality;
	m_bCopyright = src.m_bCopyright;
	m_bCRC = src.m_bCRC;
	m_bOriginal = src.m_bOriginal;
	m_bPrivate = src.m_bPrivate;
	m_iMinBR = src.m_iMinBR;
	m_iMaxBR = src.m_iMaxBR;
	m_iSrcNChannels = src.m_iSrcNChannels;
	m_iSrcSamplerate = src.m_iSrcSamplerate;
}
	
void MP3SimpleEncSettings::CreateLAMESettings(BE_CONFIG* be)
{
	memset(be, 0, sizeof(BE_CONFIG));
	be->dwConfig = BE_CONFIG_LAME;
	be->format.LHV1.dwStructVersion = 1;
	be->format.LHV1.dwStructSize = 331;
	
	// Detect mp3 sample rate. 
	be->format.LHV1.dwSampleRate = 48000;
	for(int i=0; kSampleRates[i]>0; i++)
		if( kSampleRates[i]>=m_iSrcSamplerate ){
			be->format.LHV1.dwSampleRate = kSampleRates[i];
			break;
		}
	
	be->format.LHV1.dwReSampleRate = 0; // Decoder should choose how to resample input stream
	
	// mp3 channels. We don't support dual channel as nobody seems to be using it.
	if( m_iSrcNChannels == 1 )
		be->format.LHV1.nMode = BE_MP3_MODE_MONO;
	else{
		if( m_bJointStereo )
			be->format.LHV1.nMode = BE_MP3_MODE_JSTEREO;
		else
			be->format.LHV1.nMode = BE_MP3_MODE_STEREO;
	};
	
	if( FindBitrate(m_iMinBR, -1)==-1 ){
		#if defined(VT_CRTDEBUG)
			printf("*** Couldn't find min bitrate: %d kbps, resetting to 128.\n", m_iMinBR);
		#endif
		m_iMinBR = 128;
		Warning(LCT("Adjusting minimal mp3 encoding bitrate."));
	}
	if( FindBitrate(m_iMaxBR, -1)==-1 ){
		#if defined(VT_CRTDEBUG)
			printf("*** Couldn't find max bitrate: %d kbps, resetting to 256.\n", m_iMaxBR);
		#endif
		m_iMaxBR = 256;
		Warning(LCT("Adjusting maximal mp3 encoding bitrate."));	
	}

	be->format.LHV1.dwBitrate = m_iMinBR; // Min bitrate; also used for CBR
	be->format.LHV1.dwMaxBitrate = m_iMaxBR; // Max bitrate	

	bool bBitrateCorrected = false;

	// Decrease bitrates for smaller sample rates.
	// !!!!!!!!!!!!!!!!!!!!!!!!!!!
	// This is probably a wrong formula.
	int maxPossSampleRate = int(be->format.LHV1.dwSampleRate / 136); // 44100 / 320?
	if( int(be->format.LHV1.dwBitrate) > maxPossSampleRate ){
		for(int i = FindBitrate( be->format.LHV1.dwBitrate ); i>=0; i--)
			if( kBitrates[i]<=maxPossSampleRate ){
				be->format.LHV1.dwBitrate = kBitrates[i];
				break;
			}
		bBitrateCorrected = true;
	}
	if( int(be->format.LHV1.dwMaxBitrate) > maxPossSampleRate ){
		for(int i = FindBitrate( be->format.LHV1.dwMaxBitrate ); i>=0; i--)
			if( kBitrates[i]<=maxPossSampleRate ){
				be->format.LHV1.dwMaxBitrate = kBitrates[i];
				break;
			}
		bBitrateCorrected = true;
	}

	// Allow low bitrates only for low samplerates (MPEG-II.5)
	if( be->format.LHV1.dwSampleRate > kMPEG25MaxSampleRate ){ 
		// *** MPEG-1 or MPEG-2 mode ***
		// 144 kbps is not available, increase it to 160 kbps
		if( be->format.LHV1.dwBitrate == 144 ){
			#if defined(VT_CRTDEBUG)
				printf("*** Min bitrate == 144 kbps for MPEG-1 mode, correcting.\n");
			#endif
			be->format.LHV1.dwBitrate = 160;
			bBitrateCorrected = true;
		}
		if( be->format.LHV1.dwMaxBitrate == 144 ){
			#if defined(VT_CRTDEBUG)
				printf("*** Max bitrate == 144 kbps for MPEG-1 mode, correcting.\n");
			#endif
			be->format.LHV1.dwMaxBitrate = 160;
			bBitrateCorrected = true;
		}
		// Lowest bitrates are not available
		if( be->format.LHV1.dwBitrate < kMPEG1MinBitrate ){
			#if defined(VT_CRTDEBUG)
				printf("*** Min bitrate is too low for MPEG-1 mode, correcting.\n");
			#endif
			be->format.LHV1.dwBitrate = kMPEG1MinBitrate;
			bBitrateCorrected = true;
		}
		if( be->format.LHV1.dwMaxBitrate < kMPEG1MinBitrate ){
			#if defined(VT_CRTDEBUG)
				printf("*** Max bitrate is too low for MPEG-1 mode, correcting.\n");
			#endif
			be->format.LHV1.dwMaxBitrate = kMPEG1MinBitrate;
			bBitrateCorrected = true;
		}
	}else{ 
		// We're in MPEG-II.5 mode, highest bitrates are not available
		if( be->format.LHV1.dwBitrate > kMPEG25MaxBitrate ){
			#if defined(VT_CRTDEBUG)
				printf("*** Min bitrate is too hight for MPEG-2.5 mode, correcting.\n");
			#endif
			be->format.LHV1.dwBitrate = kMPEG25MaxBitrate;
			bBitrateCorrected = true;
		}
		if( be->format.LHV1.dwMaxBitrate > kMPEG25MaxBitrate ){
			#if defined(VT_CRTDEBUG)
				printf("*** Max bitrate is too hight for MPEG-2.5 mode, correcting.\n");
			#endif
			be->format.LHV1.dwMaxBitrate = kMPEG25MaxBitrate;
			bBitrateCorrected = true;
		}
	}
	if( bBitrateCorrected )
		Warning(LCT("Bitrate settings have been changed to match the sample rate."));
	// Ensure that min bitrate is less than max bitrate, otherwise set both to the maximum
	if( be->format.LHV1.dwBitrate > be->format.LHV1.dwMaxBitrate )
		be->format.LHV1.dwMaxBitrate = be->format.LHV1.dwBitrate;

	be->format.LHV1.nPreset = LQP_NOPRESET; // Don't use internal LAME quality presets
	be->format.LHV1.bCopyright = (m_bCopyright)?TRUE:FALSE;
	be->format.LHV1.bCRC = (m_bCRC)?TRUE:FALSE;
	be->format.LHV1.bOriginal = (m_bOriginal)?TRUE:FALSE;
	be->format.LHV1.bPrivate = (m_bPrivate)?TRUE:FALSE;
	
	// VBR settings
	if( m_bVBR){
		be->format.LHV1.nVbrMethod = VBR_METHOD_DEFAULT;
		be->format.LHV1.bWriteVBRHeader = TRUE;
		be->format.LHV1.bEnableVBR = TRUE;
		if( m_bVBRHighQuality )
			be->format.LHV1.nVBRQuality = 0; // Best quality
		else
			be->format.LHV1.nVBRQuality = 5; // Average quality
	}else{
		be->format.LHV1.nVbrMethod = VBR_METHOD_NONE;
		be->format.LHV1.bWriteVBRHeader = TRUE; // We'll write it anyway
		be->format.LHV1.bEnableVBR = FALSE;
		be->format.LHV1.nVBRQuality = 0;
	}

	be->format.LHV1.dwVbrAbr_bps = 0; // We don't support ABR mode yet
	be->format.LHV1.bNoRes = FALSE; // Let it be
	
	// Quality Setting, HIGH BYTE should be NOT LOW byte.
	if( m_bHighQuality )
		be->format.LHV1.nQuality = 0xFF00; // 0 means highest quality
	else
		be->format.LHV1.nQuality = 0xFA05; // 5 - average quality
}

void MP3SimpleEncSettings::Warning(LPCVTCHAR message)
{
	VTMB::Warning(message);
}

#if defined(USE_LSXML)
bool MP3SimpleEncSettings::Save(LXML *xml) const 
{
	try{
		xml->AddChildElem(XMLTag());
		if( xml->IntoElem() ){
			xml->AddChildElem(_T("vbr"), int(m_bVBR));
			xml->AddChildElem(_T("vbrhq"), int(m_bVBRHighQuality));
			xml->AddChildElem(_T("jointstereo"), int(m_bJointStereo));
			xml->AddChildElem(_T("hq"), int(m_bHighQuality));
			xml->AddChildElem(_T("copyrighted"), int(m_bCopyright));
			xml->AddChildElem(_T("crc"), int(m_bCRC));
			xml->AddChildElem(_T("original"), int(m_bOriginal));
			xml->AddChildElem(_T("private"), int(m_bPrivate));
			xml->AddChildElem(_T("minbr"), m_iMinBR);
			xml->AddChildElem(_T("maxbr"), m_iMaxBR);
			xml->OutOfElem();
		}else
			return false;
		return true;
	}catch(...){ VTDEBUGMSGBOX; }
	return false;
}

bool MP3SimpleEncSettings::Load(LXML *xml)
{
	try{
		ResetToDefaults();
		if( xml->FindChildElem(XMLTag()) ){
			if( xml->IntoElem() ){
				m_bVBR = xml->GetChildAsBool(_T("vbr"), TRUE)?true:false;
				m_bVBRHighQuality = xml->GetChildAsBool(_T("vbrhq"), FALSE)?true:false;
				m_bJointStereo = xml->GetChildAsBool(_T("jointstereo"), FALSE)?true:false;
				m_bHighQuality = xml->GetChildAsBool(_T("hq"), TRUE)?true:false;
				m_bCopyright = xml->GetChildAsBool(_T("copyrighted"), FALSE)?true:false;
				m_bCRC = xml->GetChildAsBool(_T("crc"), FALSE)?true:false;
				m_bOriginal = xml->GetChildAsBool(_T("original"), FALSE)?true:false;
				m_bPrivate = xml->GetChildAsBool(_T("private"), FALSE)?true:false;
				m_iMinBR = xml->GetChildAsInt(_T("minbr"), 128);
				m_iMaxBR = xml->GetChildAsInt(_T("maxbr"), 256);
				xml->OutOfElem();
			}else
				return false;
		}else
			return false;
		return true;
	}catch(...){ ResetToDefaults(); VTDEBUGMSGBOX; }
	return false;
}

bool MP3SimpleEncSettings::Save(LPCVTCHAR file) const 
{
	try{
		LXML xml(kXMLHeader);
		xml.AddElem( XMLFileTag() );
		bool bResult = Save(&xml);
		if( bResult )
			bResult = (xml.WriteToFile(file))?true:false;
		return bResult;		
	}catch(...){ VTDEBUGMSGBOX; }
	return false;
}

bool MP3SimpleEncSettings::Load(LPCVTCHAR file)
{
	try{
		ResetToDefaults();
		LXML xml;
		if( xml.ReadFromFile(file) ){
			if( xml.FindElem( XMLFileTag() ) )
				return (Load(&xml))?true:false;
			else
				return false;
		}else
			return false;
	}catch(...){ ResetToDefaults(); VTDEBUGMSGBOX; }
	return false;
}

#endif

LPCVTCHAR MP3SimpleEncSettings::GetEngineInfo() const 
{
	static VTCHAR info[256]; // 128 for homepage and 128 for anything else
	BE_VERSION v;	
	VTCHAR url[BE_MAX_HOMEPAGE+1];
	memset(&v, 0, sizeof(v));
	beVersion(&v); // Get version information from the DLL
	VTASCII2VTCHAR( v.zHomepage, url, BE_MAX_HOMEPAGE );
	VTSPRINTF(info,
		_T("LAME MP3 Encoder") VTENDLT
		_T("Copyright(C) by Mark Taylor and others.") VTENDLT
		_T("See About dialog for more information.") VTENDLT
		_T("Licensed under GNU LGPL version 2.1.") VTENDLT
		_T("%s: %d.%d") VTENDLT
		_T("%s: %d.%d") VTENDLT
		_T("%s: %02d-%02d-%02d") VTENDLT
		_T("%s: %s") //VTENDLT
		//_T("MMX: %s")
		,

		LCT("DLL Version"),
		int(v.byDLLMajorVersion),
		int(v.byDLLMinorVersion),
		LCT("Engine Version"),
		int(v.byMajorVersion),
		int(v.byMinorVersion),
		LCT("Release Date"),
		int(v.wYear),
		int(v.byMonth),
		int(v.byDay),		
		LCT("Homepage"),
		url//,
		//(v.byMMXEnabled)?LCT("Enabled"):LCT("Disabled")
	);
	return info;
}


bool CPCM2MP3FileEncoder::Encode(LPCVTCHAR srcPCMFile, LPCVTCHAR destMP3File)
{
	BE_CONFIG LAMEConfig;
	HBE_STREAM hbeStream;
	bool 
		bLameStreamOpen = false, // Flag used in clean-up on error: LAME stream needs closing
		bICreatedMP3 = false; // Flag used in clean-up on error: MP3 file must be deleted
	int lastPc = 0; // Last percentage sent to OnProgress()
	TRIFFWavHeader riff;	
	FILE *fpcm=NULL, *fmp3=NULL;
	DWORD dwSamples, dwBufferSize, dwDataSizeBytes, dwDataProcessed, 
		dwDataLeft, dwChunkSize, dwOutput, dwFileSize;
	char *inBuffer=NULL, *outBuffer=NULL;
	m_szError[0]=0;	
	riff.Init();
	CCSLock lock(&s_CS);
	try{
		m_bAborted = false;
		OnStart();
		OnStatus(LCT("Initializing MP3 encoding..."));
		OnProgress(0);

		// 1. Open source file, read PCM header, update settings and create LAME settings;
		// create LAME output stream.
		fpcm = VTFOPEN(srcPCMFile, _T("rb"));
		if( !fpcm )
			throw LCT("Failed to open input file.");

		fseek( fpcm, 0, SEEK_END );
		dwFileSize = ftell( fpcm );
		fseek( fpcm, 0, SEEK_SET );

		if( dwFileSize<=sizeof(riff) )
			throw LCT("Source file is too small to be correct PCM WAV file.");

		// Read RIFF WAV header
		if(fread(&riff, sizeof(riff), 1, fpcm)!=1)
			throw LCT("Failed to read RIFF file header.");
		if(!riff.IsValidPCMWav())
			throw LCT("Source file has no correct PCM WAV header.");
		if(riff.BitsPerSample!=16)
			throw LCT("Source file has wrong bit resolution (must be 16 bit signed PCM).");

		m_iSrcNChannels = riff.Channels; // Validated by riff.IsValidPCMWav()
		m_iSrcSamplerate = riff.SamplesPerSec;
		dwDataSizeBytes = riff.datalen;

		if( m_bDetectDataLength && 
			((dwDataSizeBytes==0) || (dwFileSize<dwDataSizeBytes+sizeof(riff))) ){			
			
			dwDataSizeBytes = dwFileSize - sizeof(riff);
		}else{
			if( dwDataSizeBytes<=0 )
				throw LCT("Source file contains no audio data.");
			if( dwFileSize<dwDataSizeBytes+sizeof(riff) )
				throw LCT("Source file is truncated.");
		}

		#if defined(VT_CRTDEBUG)
			long ftellpos = ftell(fpcm);
			printf(
				"*** Source channels: %d, Sample rate: %d, Data size: %d bytes\n"
				"*** Pos after reading header: %ld\n",
				int(m_iSrcNChannels), int(m_iSrcSamplerate), int(dwDataSizeBytes), ftellpos);
		#endif
		CreateLAMESettings(&LAMEConfig);

		// BE_ERR beInitStream( PBE_CONFIG pbeConfig, PDWORD dwSamples, PDWORD dwBufferSize, PHBE_STREAM phbeStream  )
		// pbeConfig Pointer at the struct containing encoder settings.
		// dwSamples Pointer at double word where number of samples to send to each beEncodeChunk() is returned.
		// dwBufferSize Pointer at double word where minimum size in bytes of output buffer is returned.
		// phbeStream Pointer at integer where Stream handle is returned.

		if( beInitStream(&LAMEConfig, &dwSamples, &dwBufferSize, &hbeStream)!=BE_ERR_SUCCESSFUL )
			throw LCT("Error initializing MP3 compression stream.");
		bLameStreamOpen = true;

		// Each sample is 2 bytes (16 bit) * number of channels
		dwChunkSize = dwSamples * 2; //m_iSrcNChannels

		// Allocate memory for buffers
		inBuffer = new char[dwChunkSize]; // always 2 bytes per sample
		outBuffer = new char[dwBufferSize];
		
		if( inBuffer == NULL || outBuffer == NULL )
			throw LCT("Memory allocation error.");

		// 2. Create output file
		fmp3 = VTFOPEN(destMP3File, _T("wb"));
		if( !fmp3 )
			throw LCT("Failed to open output file.");
		bICreatedMP3 = true;
		
		//printf("Chunk size: %d\n", int(dwChunkSize));

		// 3. Process source file by chunks
		OnStatus(LCT("Compressing..."));
		dwDataProcessed = 0;
		dwDataLeft = dwDataSizeBytes;
		while(dwDataLeft>0){
			DWORD toRead;
			if( dwDataLeft<dwChunkSize )
				toRead = dwDataLeft;
			else
				toRead = dwChunkSize;
			// Read chunk
			if( fread(inBuffer, size_t(toRead), 1, fpcm)!=1 ){
				//printf("To read: %d Pos: %d t=%d\n", int(toRead), int(ftell(fpcm)), t);
				throw LCT("Error reading input file.");
			}
			// Encode chunk
			// Encodes a chunk of samples. Please note that if you have set the output 
			// to generate mono MP3 files you must feed beEncodeChunk() with mono samples!
			if( beEncodeChunk(hbeStream, toRead/2, 
				(PSHORT)inBuffer, (PBYTE)outBuffer, &dwOutput)!=
				BE_ERR_SUCCESSFUL )
				throw LCT("Error encoding a chunk.");
			// Write encoded data
			if( dwOutput>0 )
				if( fwrite(outBuffer, size_t(dwOutput), 1, fmp3)!=1 )
					throw LCT("Error writing to output file.");
			dwDataLeft-=toRead;
			dwDataProcessed+=toRead;

			// Send event to display encoding progress
			int pc = int( float(dwDataProcessed)*100.0f / float(dwDataSizeBytes) );
			if( pc != lastPc ){
				lastPc = pc;
				OnProgress( pc );
			}
			
			// Check for user abort
			if( m_bAborted )
				throw LCT("Aborted.");
		}
		// Get the last piece of encoded data
		if( beDeinitStream(hbeStream, (PBYTE)outBuffer, &dwOutput)!=BE_ERR_SUCCESSFUL )
			throw LCT("Error de-intializing comression stream.");
		if( dwOutput>0 )
			if( fwrite(outBuffer, size_t(dwOutput), 1, fmp3)!=1 )
				throw LCT("Error writing to output file.");
		OnProgress(100); 

		// 4. Close in & out
		OnStatus(LCT("Closing..."));
		beCloseStream(hbeStream);
		bLameStreamOpen = false;

		fflush(fmp3);		
		fclose(fmp3);
		fmp3 = NULL;
		fclose(fpcm);
		fpcm = NULL;
		delete inBuffer;
		inBuffer = NULL;
		delete outBuffer;
		outBuffer = NULL;

		// 5. Write Xing VBR header
		// Create non-unicode file name for output file, if neccessary
		#if defined(VT_UNICODE) && defined(VT_WINDOWS)
			/*wchar_t shortdestMP3File[MAX_PATH];
			char chdestMP3File[MAX_PATH];
			if(!GetShortPathName(destMP3File, shortdestMP3File, MAX_PATH))
				throw LCT("Failed to determine short file name for output file.");
			if(!WideCharToMultiByte(CP_ACP, 0, shortdestMP3File, -1, chdestMP3File, MAX_PATH, NULL, NULL))
				throw LCT("Failed to convert short file of output file to ANSI charset.");
			*/
			char chdestMP3File[MAX_PATH];
			if( !VT::GetShortName( destMP3File, chdestMP3File, false ) )
				throw LCT("Failed to convert file name to ANSI charset.");
		#else
			#define chdestMP3File destMP3File
		#endif	
		beWriteVBRHeader(chdestMP3File);		
		
		OnStatus(LCT("Done."));
		OnEnd();
		if(!m_szError[0])
			VTSTRNCPY(m_szError, LCT("No error."), kHalfError);
		return true;
	}catch(LPCVTCHAR msg){			
		VTSTRNCPY(m_szError, msg, kHalfError);
		// Then proceed to the fire exit
	}catch(...){ 
		VTSTRNCPY(m_szError, LCT("Unknown exception occured in MP3 file encoder."), kHalfError);
		// Then proceed to the fire exit
	};

	// Fire exit
	try{
		if( fpcm != NULL ){
			fclose(fpcm);
			fpcm = NULL;
		}
		if( fmp3 != NULL ){
			fclose(fmp3);
			fmp3 = NULL;
		}
		if( bLameStreamOpen ){
			beCloseStream(hbeStream);
			bLameStreamOpen = false;
		}
		if( inBuffer != NULL ){
			delete inBuffer;
			inBuffer = NULL;
		}
		if( outBuffer != NULL ){
			delete outBuffer;
			outBuffer = NULL;
		}
		if( bICreatedMP3 ){
			#if !defined(VT_CRTDEBUG)
				VTUNLINK(destMP3File);
			#else
				printf("*** Keeping output MP3 file because of VT_CRTDEBUG.\n");
			#endif
		}
		OnEnd();
	}catch(...){
		VTSTRNCAT(m_szError, 
			LCT("\r\nAdditionaly, an error occured while de-initializing the encoder."), kHalfError);
	}
	
	return false;
}




