//-----------------------------------------------------------------------------
//
// MP3 encoder wrapper. Currently uses LAME DLL.
//
//-----------------------------------------------------------------------------

/*
	VT Framework License (BSL 1.0)

	Copyright (C) 2007 Sergey A. Galin 
	E-mail: sergey.galin@gmail.com

	Permission is hereby granted, free of charge, to any person or organization
	obtaining a copy of the software and accompanying documentation covered by
	this license (the "Software") to use, reproduce, display, distribute,
	execute, and transmit the Software, and to prepare derivative works of the
	Software, and to permit third-parties to whom the Software is furnished to
	do so, all subject to the following:

	The copyright notices in the Software and this entire statement, including
	the above license grant, this restriction and the following disclaimer,
	must be included in all copies of the Software, in whole or in part, and
	all derivative works of the Software, unless such copies or derivative
	works are solely in the form of machine-executable object code generated by
	a source language processor.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
	SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
	FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
	ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
	DEALINGS IN THE SOFTWARE.
*/

// Don't forget to link to the library, e.g.:
// #pragma comment(lib, "../Lib/lame_enc.lib")


#if !defined(VT_MP3ENCODE_H)
#define VT_MP3ENCODE_H

#if !defined(USE_LSGUI)
	#pragma message("Set USE_LSGUI globally to enable compilation of LString & LXML functions.")
#else
	// LS_GUI is defined
	#if !defined(USE_LSSTRING)
		#define USE_LSSTRING
	#endif
	#if !defined(USE_LSXML)
		#define USE_LSXML
	#endif
	#include "LSString.h"
	#include "LSXML.h"
#endif

#include "vt_portable.hpp"
#include "vt_threading.hpp"

#if defined(USE_LSGUI) 
	#include "vtls/vtls_musictags.hpp"
#endif

// Surprise! This is how LAME DLL header file is named!
// Note the _BLADEDLL define and that it must be included after system includes.
#define _BLADEDLL
#include "BladeMP3EncDLL.h"

namespace VT {

	// This class allows to set only flags a sane user would chose to set.
	// All stuff which is really not used by most people, like very fine tuning of quality
	// or choosing some ancient VBR preset is out. 
	class MP3SimpleEncSettings
	{
	public:
		// Do not use kBitrates directly - use the two functions to work with it.
		static const int kBitrates[]; // Allowed bitrates; 0-terminated array
		static int FindBitrate(int kbps, int def=11); // Returns BR index from kbps value, def if not found
		static int Idx2Bitrate(int idx); // Returns kbps value from index, or 128 kbps if not found
		// Bitrates with description. This array has the same size as kBitrates.
		static LPCVTCHAR kBitratesStr[]; // NULL-terminated
		// 32000, 44100 or 48000 for MPEG-I
		static const int kSampleRates[]; // 0-terminated	

		// MP3 capabilities
		static const int 
			kMinSampleRate = 8000,
			kMaxSampleRate = 48000,
			kMinBitrateKbps = 32,
			kMaxBitrateKbps = 320,		

			kMPEG25MaxSampleRate = 12000, 
			kMPEG1MinBitrate= 32,
			kMPEG25MaxBitrate = 160,
			kMPEG25OnlyMaxBitrate = 24; // Allow "<=" bitrates only for samplerates<=kMPEG25MaxSampleRate 

		// Settings which user should (or can) choose 
		bool
			m_bVBR,            // true = VBR, false = CBR
			m_bVBRHighQuality, // VBR bitrate bias? Enable when m_bVBR is set
			m_bJointStereo,    // Joint-Stereo
			m_bHighQuality,	   // High-quality encoding (slow)
			m_bCopyright,      // Mark as copyrighted
			m_bCRC,            // Add CRC
			m_bOriginal,       // Mark as original
			m_bPrivate;        // Mark as private

		int 
			m_iMinBR,          // Min bitrate, enable only when m_bVBR is set
			m_iMaxBR;          // Max bitrate, enable only when m_bVBR is set		

		// Settings which come from source WAV file. These are also neccessary 
		// for CreateLAMESettings().
		// Note that these variables are not used by Save()/Load() routines.
		int 
			m_iSrcNChannels,   // Number of channels to encode (1..2) (we always encode to the same n.of ch.)
			m_iSrcSamplerate;  // Sample rate (we encode to the same sample rate or better, or 48000 if src is higher)
		
		MP3SimpleEncSettings(){ ResetToDefaults(); } // Constructor fills vars with default values
		virtual void ResetToDefaults();
		virtual void CopyFrom(const MP3SimpleEncSettings& src);
		virtual void CopyTo(MP3SimpleEncSettings *dest) const { (*dest).CopyFrom(*this); }
		virtual void CreateLAMESettings(BE_CONFIG* be); // Fills *be with all settings 
		
		// By default, this function shows a warning message box via VTMB.
		// Override it in descendant class if this behaviour should be enhanced or changed.
		virtual void Warning(LPCVTCHAR message);

		// If compiled with LS_GUI, settings can be stored in XML file
		#if defined(USE_LSXML)
			static LPCVTCHAR XMLTag(){ return _T("MP3Compressor"); }
			static LPCVTCHAR XMLFileTag(){ return _T("MP3Encoding"); }
			virtual bool Save(LXML *xml) const;
			virtual bool Load(LXML *xml);
			virtual bool Save(LPCVTCHAR file) const;
			virtual bool Load(LPCVTCHAR file);		
		#endif	

		LPCVTCHAR GetEngineInfo() const;

	};

	// This class encodes PCM (16-bit signed) file using configuration loaded via
	// MP3SimpleEncSettings (see GetSettings()) into specified mp3 file.
	// It can be used as simply as:
	//
	// VTMP3FileEncoder encoder(m_ProjectMP3Settings);
	// if( !encoder.Encode(srcFile, destFile) ){
	//   VTMB::Error(encoder.GetErrorMessage());
	//
	class CPCM2MP3FileEncoder: protected MP3SimpleEncSettings
	{
	public:
		// Set to true before calling Encode() to enable auto-detection 
		// of audio data length (as (file size)-(header size)).
		bool m_bDetectDataLength;

		CPCM2MP3FileEncoder();
		// Load settings from another settings object.
		void GetSettings(const MP3SimpleEncSettings& src){ MP3SimpleEncSettings::CopyFrom(src); }
		// Initialize and load encoding settings.
		CPCM2MP3FileEncoder(const MP3SimpleEncSettings& src);

		// Basically, this is the only function which should be called.
		bool Encode(LPCVTCHAR srcPCMFile, LPCVTCHAR destMP3File);
		
		// Callback functions, override them to implement some logging or visual feedback to user.	
		virtual void OnStart(){} // Called before any OnStatus(), OnProgress() calls
		virtual void OnEnd(){} // Called after any OnStatus(), OnProgress() calls
		virtual void OnStatus(LPCVTCHAR message){}
		virtual void OnProgress(int percentDone){}

		LPCVTCHAR GetErrorMessage() const { return m_szError; }
		LPCVTCHAR GetEngineInfo() const { return MP3SimpleEncSettings::GetEngineInfo(); }
		void Abort(){ m_bAborted = true; }
	protected:
		volatile bool m_bAborted;
		static const int kMaxError = 255, kHalfError = 127;
		VTCHAR m_szError[kMaxError+1];
		static CCriticalSection s_CS; // Only one instance of encoder may be running at a time.
	};

} // namespace

#endif


