//-----------------------------------------------------------------------------
//
// Parameter classes, for interfacing program variables and GUI controls.
// The parameters were originally introduced to ease writing of VST plugins,
// hence the limitations they have to maintain (e.g. "float" values, 
// short names...)
//
//-----------------------------------------------------------------------------

/*
	VT Framework License (BSL 1.0)

	Copyright (C) 2006 Sergey A. Galin 
	E-mail: sergey.galin@gmail.com

	Permission is hereby granted, free of charge, to any person or organization
	obtaining a copy of the software and accompanying documentation covered by
	this license (the "Software") to use, reproduce, display, distribute,
	execute, and transmit the Software, and to prepare derivative works of the
	Software, and to permit third-parties to whom the Software is furnished to
	do so, all subject to the following:

	The copyright notices in the Software and this entire statement, including
	the above license grant, this restriction and the following disclaimer,
	must be included in all copies of the Software, in whole or in part, and
	all derivative works of the Software, unless such copies or derivative
	works are solely in the form of machine-executable object code generated by
	a source language processor.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
	SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
	FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
	ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
	DEALINGS IN THE SOFTWARE.
*/

#include "vt_parameters.hpp"

using namespace VT;
using namespace Params;

CParameter::CParameter(CParameterValidator *ownr, const VTCHAR *nme, const VTCHAR *lbl, 
							   float min, float max, float def, bool logH, bool gridH):
	VT::CAssociatedValue<float>( def ),
	owner(ownr), 
	fMin(min), 
	fMax(max), 
	fDefault(def), 
	//fValue(def), 	
	paramPos(-1), 
	m_Tag(0),
	logHandler(logH), 
	logMod(kDefaultLogMod),
	gridHandler(gridH),
	gridSteps(100.0f),
	iGridSteps(100),
	handleLookupTable(NULL),
	showPlusSign(false),
	m_uRangeChange( 0 )
	#if defined(VTEP_FAKE01)
		, bReturnFakeValue(false)
	#endif
{
	VTASSERT( nme!=NULL );
	VTASSERT( lbl!=NULL );
	VTASSERT( min<=max );
	VTASSERT( def>=min );
	VTASSERT( def<=max ); 

	VTSTRNCPY(name, nme, sizeof(name)-1);
	VTSTRNCPY(label, lbl, sizeof(label)-1);
	
	VTSTRNCPY(longName, nme, sizeof(longName)-1);
	VTSTRNCPY(codeName, nme, sizeof(codeName)-1);
	
	#if defined(VTEP_FAKE01)
		fFake01Value = as01(def); // Not just as01(), because it might return fFake01Value
	#endif
}

CParameter::~CParameter()
{
	deleteHandleLookupTable();
}

void CParameter::setLongName( LPCVTCHAR newName )
{
	VTASSERT( newName!=NULL );
	VTASSERT( VTSTRLEN( newName )<sizeof(longName) );
	VTSTRNCPY(longName, newName, sizeof(longName)-1);
}

void CParameter::setCodeName( LPCVTCHAR newName )
{
	VTASSERT( newName!=NULL );
	VTASSERT( VTSTRLEN( newName )<sizeof(codeName) );
	VTSTRNCPY(codeName, newName, sizeof(codeName)-1);
}

float CParameter::from01(float x) const 
{
	if(handleLookupTable!=NULL){
		
		// Use lookup table
		#if defined(VTEP_USECRITICALSECTION)
			if(owner!=NULL) 
				owner->effectCriticalSection.enter();
		#endif
		float ret = (handleLookupTable!=NULL)?
			handleLookupTable[ Clip(URound(gridSteps*x), SDWORD(0), iGridSteps) ]:
			from01_NLTG(x); // Let's just ignore the grid stuff in this case
		#if defined(VTEP_USECRITICALSECTION)
			if(owner!=NULL) 
				owner->effectCriticalSection.leave();
		#endif
		return ret;
	}else{

		// Just calculate 
		if(gridHandler) 
			x=GridRound01(x, gridSteps);
		return from01_NLTG(x);
	}
}

void CParameter::setValue(float x)
{ 	
	setValueNE(x);
	if( owner!=NULL )
		owner->VTValidateParams(this); 
}

#if defined(VTEP_FAKE01)
void CParameter::setValueNF01(float x)
{ 	
	setValueNENF01(x);
	if( owner!=NULL )
		owner->VTValidateParams(this); 
}
#endif

void CParameter::adjustLogHandler(float x, float y, double prec, double smin, double smax)
{
	float as_01 = as01_raw(x);
	logMod = CalculateLogMod(as_01, y, prec, smin, smax);
	
	#if defined(_DEBUG)
		if( logHandler ){
			float new_as_01 = as01( x );
			VTASSERT( Abs( new_as_01 - y )<prec );
		}
	#endif
}


void CParameter::createHandleLookupTable()
{
	deleteHandleLookupTable();
	handleLookupTable = new float[ iGridSteps+1 ];
	// Avoiding rounding errors at ends of the range - 
	// setting their values directly from fMin and fMax.
	handleLookupTable[0] = fMin;
	for(SDWORD step=1; step<iGridSteps; step++)
		handleLookupTable[ step ] = from01_NLTG( float(step)/gridSteps );
	handleLookupTable[iGridSteps] = fMax;
}

void CParameter::copyHandleSettings(const CParameter *srcParam)
{
	deleteHandleLookupTable();
	logHandler = srcParam->logHandler;
	gridHandler = srcParam->gridHandler;
	logMod = srcParam->logMod;
	gridSteps = srcParam->gridSteps;
	iGridSteps = srcParam->iGridSteps;
	#if defined(VTEP_FAKE01)
		bReturnFakeValue = srcParam->bReturnFakeValue;
	#endif
	if( srcParam->handleLookupTable!=NULL ){
		handleLookupTable = new float[ iGridSteps+1 ];
		memcpy( handleLookupTable, srcParam->handleLookupTable, 
			(iGridSteps+1)*sizeof(float) );
	}
}

void CParameter::deleteHandleLookupTable()
{
	if( handleLookupTable!=NULL ){
		#if defined(VTEP_USECRITICALSECTION)
			if(owner!=NULL) 
				owner->effectCriticalSection.enter();
		#endif
		delete handleLookupTable;
		handleLookupTable = NULL;
		#if defined(VTEP_USECRITICALSECTION)
			if(owner!=NULL) 
				owner->effectCriticalSection.leave();
		#endif
	}
}

void CParameter::getAsTextForVST(VTCHAR *out) const 
{
	#if defined(VTEP_FAKE01)
		if(bReturnFakeValue)
			getAsText(from01_NLTG(fFake01Value), out);
		else
			getAsText(out);
	#else
		getAsText(out);
	#endif
}

size_t CParameter::GetNValues()
{
	throw _T("Attempt to get number of values from a parameter which doesn't support that.");
	return 0;
}

float CParameter::GetValueByIdx(size_t idx)
{
	throw _T("Attempt to value-by-index from a parameter which doesn't support that.");
	return 0;
}

size_t CParameter::GetIdxOfValue(float value)
{
	throw _T("Attempt to index-of-value from a parameter which doesn't support that.");
	return 0;
}

void CParameter::setMin(float m)
{ 
	fMin=m; 
	if( getValue()<fMin ) 
		setValueNE(fMin); 
	RangeChanged();
}

void CParameter::setMax(float m)
{ 
	fMax=m; 
	if( getValue()>fMax ) 
		setValueNE(fMax); 
	RangeChanged();
}

void CParameter::setFromSaveValue( float value, bool notifyOwner )
{
	if( notifyOwner )
		setValue( value );
	else
		setValueNE( value );
}












CParameter_combo::CParameter_combo(CParameterValidator *ownr, const VTCHAR *nme, 
	const VTCHAR *const *values, int def):
	CParameter_int(ownr, nme, _T(""), 0, 
		def, // temp max value
		def ),
	labels( NULL )
{
	SetNewValues( values, false );
}

void CParameter_combo::SetNewValues( const VTCHAR *const *values, bool notifyIfValueChanged )
{
	float value = getValue();
	// Delete old labels (if necessary)
	DeleteLabels();
	if( values!=NULL ){
		// Count new labels
		int count=0;
		const VTCHAR *const *ptr=values;
		while(*ptr){
			count++;
			ptr++;
		}
		// Allocate memory for pointer array
		labels = new LPVTCHAR[ count+1 ];
		if( labels==NULL )
			throw _T("Memory allocation error.");
		// Create copies of values
		ptr=values;
		size_t len;
		for(int i=0; i<count; i++){
			len = strlen( *ptr );
			labels[i] = new VTCHAR[ len+1 ];
			if( labels[i]==NULL )
				throw _T("Memory allocation error.");
			strcpy( labels[i], *ptr );
			ptr++;
		}
		// Last label pointer is a NULL-terminator
		labels[ count ] = 0;
		// Updating range
		setMin( 0.0f );
		setMax( (count>0.0f)? float(count-1): 0.0f );
		
	}	
	if( SRound(value) > SRound( getMax() ) ){
		if( notifyIfValueChanged )
			setValue( value ); // This
		else
			setValueNE( value );
	}
	// Flagging the update
	RangeChanged(); // Don't trust setMax() on that ;-)
}


CParameter_combo::~CParameter_combo()
{
	DeleteLabels();
}

void CParameter_combo::DeleteLabels()
{
	if( labels!=NULL ){
		for(VTCHAR **ptr=labels; *ptr; ptr++)
			delete *ptr;
		delete labels;
		labels = NULL;
	}
}

void CParameter_combo::visibleText2Visible(VTCHAR *string, float &output) const
{
	for(int i=0; labels[i]; i++)
		if( !strcmp( labels[i], string ) ){
			output = float(i);
			return;
		}
}

void CParameter_combo::getAsText(float x, VTCHAR *out, size_t maxLen) const 
{
	LPCVTCHAR src = NULL;
	if( labels!=NULL ){
		size_t idx = URound(validate(x));
		src = labels[idx];
	}
	if( src==NULL )
		src = LCT("[Empty]");
	VTSTRNCPY(out, src, maxLen); 
}









CParameter_strval_combo::CParameter_strval_combo(
	CParameterValidator *ownr, 
	const VTCHAR *nme, const VTCHAR *lbl,
	const VTCHAR *const *values, 
	const float *realValues,
	float def ):
	
	CParameter( ownr, nme, lbl, 
		Min(0.0f, def), // temp min value
		def, // temp max value
		def, // default value
		false, // logH
		false  // gridH
	)
{
	m_uEntries = 0;
	const VTCHAR *const *ptr=values;
	// Counting values
	while(*ptr){
		m_uEntries++;
		ptr++;
	}
	if( m_uEntries==0 ){
		VTDEBUGMSGT("No entries!");
		throw _T("No entries.");
	}

	// Allocating memory
	m_pEntries = new TEntry[ m_uEntries ];
	if( m_pEntries==NULL )
		throw _T("Memory allocation error.");
	
	// Copying labels & parsing values
	for( size_t i=0; i<m_uEntries; i++ ){
		VTSTRNCPY( m_pEntries[i].m_Label, values[i], kMaxLabel );
		if( realValues==NULL ){
			// Parsing from string
			m_pEntries[i].m_fValue = 0.0f;
			int scanfresult = VTSSCANF( values[i], _T("%f"), &m_pEntries[i].m_fValue );
			// VTASSERT2T( scanfresult==1, values[i] );
		}else
			m_pEntries[i].m_fValue = realValues[i];
	}

	// Setting min and max values
	float min = m_pEntries[0].m_fValue, max = m_pEntries[0].m_fValue;
	size_t minidx = 0, maxidx = 0;
	for( size_t i=1; i<m_uEntries; i++ ){
		if( m_pEntries[i].m_fValue<min ){
			min = m_pEntries[i].m_fValue;
			minidx = i;
		}
		if( m_pEntries[i].m_fValue>max ){
			max = m_pEntries[i].m_fValue;
			maxidx = i;
		}
	}
	setMin( min );
	setMax( max );
}


CParameter_strval_combo::~CParameter_strval_combo()
{
	delete[] m_pEntries;
}

void CParameter_strval_combo::getAsText(float x, VTCHAR *out, size_t maxLen) const 
{ 
	VTSTRNCPY(out, m_pEntries[ FindClosestEntry(x) ].m_Label, maxLen); 
}

void CParameter_strval_combo::textFromVisible(float x, VTCHAR *out, size_t maxLen) const 
{ 
	VTSTRNCPY( out, m_pEntries[ FindClosestEntry(x) ].m_Label, maxLen );
}

float CParameter_strval_combo::getVisibleValue(float x) const
{
	return m_pEntries[ FindClosestEntry( x ) ].m_fValue;
}

float CParameter_strval_combo::fromVisibleValue(float x) const
{
	return m_pEntries[ FindClosestEntry(x) ].m_fValue;
}

// Find an entry which has a value closest to 'search'.
size_t CParameter_strval_combo::FindClosestEntry( float search ) const
{	
	if( m_uEntries==0 || m_pEntries==NULL ){
		VTDEBUGMSGT("Combobox: no entries!");
		throw _T("Combobox: no entries!");
	}	
	size_t closestMatch = 0;	
	float closestDiff = Abs( m_pEntries[0].m_fValue - search );
	for( size_t idx=1; idx<m_uEntries; idx++ ){
		float diff = Abs( m_pEntries[idx].m_fValue - search );
		if( diff<closestDiff ){
			closestDiff = diff;
			closestMatch = idx;
		}
	}
	return closestMatch;
}




