//-----------------------------------------------------------------------------
//
// Parameter classes, for interfacing program variables and GUI controls.
// The parameters were originally introduced to ease writing of VST plugins,
// hence the limitations they have to maintain (e.g. "float" values, 
// short names...)
//
//-----------------------------------------------------------------------------

/*
	VT Framework License (BSL 1.0)

	Copyright (C) 2006 Sergey A. Galin 
	E-mail: sergey.galin@gmail.com

	Permission is hereby granted, free of charge, to any person or organization
	obtaining a copy of the software and accompanying documentation covered by
	this license (the "Software") to use, reproduce, display, distribute,
	execute, and transmit the Software, and to prepare derivative works of the
	Software, and to permit third-parties to whom the Software is furnished to
	do so, all subject to the following:

	The copyright notices in the Software and this entire statement, including
	the above license grant, this restriction and the following disclaimer,
	must be included in all copies of the Software, in whole or in part, and
	all derivative works of the Software, unless such copies or derivative
	works are solely in the form of machine-executable object code generated by
	a source language processor.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
	SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
	FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
	ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
	DEALINGS IN THE SOFTWARE.
*/

#ifndef VT_PARAMETERS_H
#define VT_PARAMETERS_H

#include "vt_audiomath.hpp"
#include "vt_threading.hpp"
#include "vt_localization.hpp"
#include "vt_paramstorage.hpp"
#include "vt_assert.hpp"

// Enables code which fools parameter owner in making it believe that 
// some params (marked using setReturnFake01()) are continous, 
// while in fact they are discrete. This makes it using "smooth" 
// parameter automation controls, while otherwise it would use 
// "stair" controls.
// #define VTEP_FAKE01

// Enable to have log table operations to use critical sections.
// #define VTEP_USECRITICALSECTION


// #define VTEP_GETVALUE_CLIPPING

namespace VT {
	namespace Params {
		const size_t kMaxParamStrLen = 8; // VST compatibility

		class CParameter;

		//
		// A class which can receive events from CParameter.
		//
		class CParameterValidator
		{
		public:

			//
			// An event which is fired when a parameter is changed.
			// Should return true if parameter value has been updated after
			// validation. In case of doubt, just return true ;-)
			//
			virtual bool VTValidateParams( CParameter *lastChangedParam=0 ){ return false; }
		protected:
			friend class CParameter;
			#if defined(VTEP_USECRITICALSECTION)
				CCriticalSection effectCriticalSection;
			#endif
		};


		//
		// A parameter.
		// 
		// Each parameter has a number of representations, which can be converted
		// one to another using class methods:
		//
		// - Internal float value, which is usually equal to physical meaning of parameter.
		//   E.g.: for audio gain, it is a coefficient of voltage multiplication.
		//   The value can be stored either internally or externally (in this case,
		//   use AssociateWith() to link parameter to some existing variable).
		//
		// - Float value in range of [0..1]. Range of internal value is mapped to [0..1] 
		//   linearly by default, but a logarithmic scale or grid can be used.
		//   For example, audio gain is naturally controlled using logarihtmic 
		//   potentiometers.
		//
		// - "Visible" value. For example, audio gain may be displayed in dB.
		//
		// - Save value, in case it differs from internal representation 
		//   (the default functions which use unternal value are usually 
		//   not overloaded).
		//
		// - Text representation of visible value.
		//
		// Also, there are some type conversions:
		//
		// - To/from bool - available in CParameter and descendants
		// - To integer (getAsInt) - available in CParameter_int and descendants
		//
		class CParameter: public VT::CAssociatedValue<float>
		{
		public:
			static const size_t 
				kMaxLongNameSize = 64,
				kMaxCodeNameSize = kMaxParamStrLen;

			// A tag which can be used by param's owner to identify the parameter.
			// Initialized with -1 in constructor; not used anywhere in the class.
			int paramPos;

			// Initialized to 0, not used in the class.
			long m_Tag;

			// Constructor parameters:
			// nme - name of a paramter, e.g.: "Car Spd"
			// lbl - label, usually measurement unit, e.g.: "MPH"
			// min - minimal possible value of a parameter
			// max - maximal possible value of a parameter
			// def - initial / default value, can be restored using 
			//    resetToDefault()/resetToDefaultNE().
			// logH - use logarithmic handle (log scale for calculation between 
			//    internal value and "0..1" representation). See log handler
			//    functions which allow to tune the scale below.
			// gridH - use grid (discrete) handle. See grid handler
			//    functions which allow to tune the grid below.
			CParameter(CParameterValidator *ownr, const VTCHAR *nme, const VTCHAR *lbl, 
				float min, float max, float def, bool logH=false, bool gridH=false);

			~CParameter();

			VTINLINE static float validate01(float x) { return Limit01(x); } // Limit x to [0..1]
			VTINLINE float validate(float x) const { return Clip(x, fMin, fMax); } // Limit x to valid range
			VTINLINE float getMin() const { return fMin; } // Get min possible value
			VTINLINE float getMax() const { return fMax; } // Get max possible value
			VTINLINE float getMiddle() const { return float((getMax()+getMin())/2.0); } // Average of min & max values
			VTINLINE float getDefault() const { return fDefault; } // Default value
			
			#if defined(VTEP_GETVALUE_CLIPPING)
				VTINLINE float getValue() const { 
					return Clip( VT::CAssociatedValue<float>::getValue(), getMin(), getMax() ); } // Current value
			#else
				// Simple live with inherited getValue().
			#endif
			
			VTINLINE float isAtMax() const { return getValue()>=fMax; } // Check that current value is the max
			VTINLINE float isAtMin() const { return getValue()<=fMin; } // Check that current value is the min
			VTINLINE float isAtMax(float x) const { return x>=fMax; } // Check that x woule be the max
			VTINLINE float isAtMin(float x) const { return x<=fMin; } // Check that x woule be the min

			// Enable logarithmic handler movement
			VTINLINE void setLogHandler(bool val=true){ logHandler = val; }
			// Setup logarithmic handler, so for internal value 'x' the handler is located
			// at position 'y' (0..1). Precision and interval of search for result can be also 
			// passed as parameters.
			void adjustLogHandler(float x, float y, double prec=0.01, double smin=0.1, double smax=10.0);
			// Since calculation of the handler parameter isn't very fast, use this function
			// or copyHandleSettings() to copy it between similar parameters.
			VTINLINE void adjustLogHandler(CParameter *const from){ logMod = from->logMod; }

			// Enable snap-to-grid handler movement (only for handlers which work with 0..1 range). 
			VTINLINE void setGridHandler(bool val=true){ gridHandler=val; }
			VTINLINE void setGridStep(SDWORD val){ gridSteps=float(val); iGridSteps=val; }
			VTINLINE void setGridHandler(bool enbl, SDWORD val){ setGridStep(val); setGridHandler(enbl); }

			virtual bool canBeAutomated() const { return true; } 
			virtual const VTCHAR* getName() const { return name; }
			
			virtual const VTCHAR* getLongName() const { return longName; }
			virtual void setLongName( LPCVTCHAR newName );
			virtual const VTCHAR* getCodeName() const { return codeName; }
			virtual void setCodeName( LPCVTCHAR newName );

			virtual void setOwner(CParameterValidator *ownr){ owner = ownr; }
			virtual const VTCHAR *getLabel() const { return label; }

			// Convert value to [0..1]
			virtual float as01_raw(float x) const { return (x-fMin)/(fMax-fMin); } // Convert to 0..1 without using log scale
			virtual float as01(float x) const { return Limit01((logHandler)?float(ControlHandle_01_Log(as01_raw(x), logMod)):as01_raw(x)); }
			virtual float as01() const { 
				#if defined(VTEP_FAKE01)
					return (bReturnFakeValue)?fFake01Value:as01(getValue()); 
				#else
					return as01(getValue());
				#endif
			}
			virtual float getDefault_as01() const { return as01(getDefault()); }

			#if defined(VTEP_FAKE01)
				// This functions cases param to return previous 0..1 value 
				// set from outside instead of recalculating it from the real
				// value range. This is a bit faster and helps some hosts 
				// to understand that parameter automation should not be discrete
				// (even though internally it is).
				void setReturnFake01(bool x=true){ bReturnFakeValue=x; }

				void updateFake01(){ fFake01Value = as01(getValue()); }
			#endif

			VTINLINE float from01_raw(float x) const {
				return (fMax-fMin)*Limit01(x)+fMin;
			}

			// Convert [0..1] to value, not using lookup table or grid.
			VTINLINE float from01_NLTG(float x) const {
				if(logHandler) 
					x=float( ControlHandle_Log_01(x, logMod) ); 
				return from01_raw( x );
			}

			// Convert [0..1] to value, using lookup table or grid (if neccessary).	
			float from01(float x) const;

			// Creating a pre-calculated table of conversion between handle position
			// and internal value. This improves performance of the conversion for
			// logarithmic handles. The table can be also copied between objects.
			//
			// Warning: Use createHandleLookupTable() only in plugin constructor
			// and don't use deleteHandleLookupTable(), or lock the parameter 
			// into a critical section.
			// This function uses iGridSteps to determine lookup table size, 
			// so call setGridHandler() before it.
			virtual void createHandleLookupTable();
			// The same, but with gridSize initialization
			virtual void createHandleLookupTable(int gridSize){ setGridStep(gridSize); createHandleLookupTable(); }
			// Copies values of log handler, handler grid and lookup table from src param.
			virtual void copyHandleSettings(const CParameter *srcParam);
			// Very unlikely it may be useful outside the class, but...
			// No need to call this function before destroying the object (done automatically).
			void deleteHandleLookupTable();

			// This is the function which is used by VST setParameter().
			virtual void setFrom01(float x){
				#if defined(VTEP_FAKE01) 
					setValueNF01(from01(fFake01Value = validate01(x)));
				#else
					setValue(from01(validate01(x))); 
				#endif
			}
			
			#if defined(VTEP_FAKE01)
				// ...NF01 - the function doesn't update fFake01Value 
				// ...NE - the function doesn't fire "params changed" event on owner
				// With event, w/o fake value:
				virtual void  setValueNF01(float x); 
				// No event, no fake value; this function is called by all other setValue...()'s:
				virtual void  setValueNENF01(float x){ 
					// getValue()=validate(x); ???????????????????????
					CAssociatedValue<float>::setValueNE(validate(x));				
				} 
			#endif

			virtual void  setValueNE(float x){ // No event, set fake value
				#if defined(VTEP_FAKE01)
					setValueNENF01(x); 
					fFake01Value = as01(getValue()); 
				#else
					// fValue = validate(x);
					VT::CAssociatedValue<float>::setValueNE(validate(x));
				#endif
			} 	
			virtual void  setValue(float x); // Sets value, fake value and calls owner's "validate params" event

			virtual void  resetToDefault(){ setValue(getDefault()); } // Reset to default & call event
			virtual void  resetToDefaultNE(){ setValueNE(getDefault()); } // Reset w/o event

			virtual float getVisibleValue(float x) const { return x; }
			virtual float getVisibleValue() const { return getVisibleValue(getValue()); }
			virtual float getDefault_VisibleValue() const { return getVisibleValue(getDefault()); }
			virtual float fromVisibleValue(float x) const { return validate(x); }
			virtual void  setFromVisibleValue(float x){ setValue(fromVisibleValue(x)); } // Yep, double validation, sry
			virtual float getMinVisibleValue(){ return getVisibleValue(getMin()); }
			virtual float getMaxVisibleValue(){ return getVisibleValue(getMax()); }

			virtual void getAsText(float x, VTCHAR *out, size_t maxLen=kMaxParamStrLen) const { 
				float2string(getVisibleValue(x), out, maxLen, showPlusSign); }

			virtual void getAsText(VTCHAR *out) const { getAsText(getValue(), out); }
			virtual void getAsTextForVST(VTCHAR *out) const; // In "fake 01" mode, returns fake value
			virtual void textFromVisible(float x, VTCHAR *out, size_t maxLen=kMaxParamStrLen) const { float2string(x, out, maxLen, showPlusSign); }

			VTINLINE static bool asBool(float x){ return (x>=0.5)?true:false; }
			VTINLINE bool asBool() const { return asBool(getValue()); }
			virtual void setAsBool(bool x){ setValue( (x)?float(1.0): float(0.0) ); }
			virtual void setAsBoolNE(bool x){ setValueNE( (x)?float(1.0): float(0.0) ); }

			virtual void visibleText2Visible(VTCHAR *string, float &output) const { output = getValue(); VTSSCANF(string, _T("%f"), output); }
			virtual bool is_dB() const { return false; }

			// Note: these functions are not 100% safe, as they don't update handle stuff and etc.
			virtual void setMin(float m);
			virtual void setMax(float m);

			virtual bool CanIterateValues(){ return false; }
			virtual size_t GetNValues(); 
			virtual float GetValueByIdx(size_t idx);
			virtual size_t GetIdxOfValue(float value);

			// Some parameters may trick with saved/loaded value as well.
			virtual float getSaveValue() const { return getValue(); }
			virtual void setFromSaveValue( float value, bool notifyOwner = true );

			// Access to Range Change indicator. It is incremented each time
			// the parameter changes its min/max values, or set of labels
			// for combobox parameters. GUI control can use the variable
			// to update itself when necessary. Wrapping the variable around
			// on overflow (in case application was workig without restart
			// for many years) is not a problem, as GUI should check for a
			// CHANGE of the variable, not for its increment.
			unsigned long GetRangeChange() const { return m_uRangeChange; }

		protected:		
			bool showPlusSign;	// Should a "+" sign be displayed for positive values?
								// This flag should be honoured by all value-to-text
								// functions here and in all descendant classes.
			
			void RangeChanged(){ m_uRangeChange++; }

		private:
			VTCHAR 
				name[kMaxParamStrLen+1],	// Short parameter name, compatible with VST(i)
				label[kMaxParamStrLen+1],	// Parameter label (measurement units - "ms", "dB", etc.)
				longName[kMaxLongNameSize], // Long parameter name
				codeName[kMaxCodeNameSize];
			bool 
				logHandler,			// Is logarithmic handle used?
				gridHandler;		// Is grid (discrete) handle used?
			unsigned long 			
				m_uRangeChange;		// Incremented each time when range changes
			double logMod;			// Coefficient of logarihmic handle, usually
									// calculated so some certain internal value
									// was at some certain handle position (see adjustLogHandler()).
			float gridSteps;		// Grid step
			SDWORD iGridSteps;		// Grid step
			float *handleLookupTable; // Handle <-> internal value conversion table
			CParameterValidator *owner; // Receives events when parameter is chaged; can be NULL
			float 
				fMin,				// Minimal internal value
				fMax,				// Maximal internal value
				fDefault;			// Default internal value
			#if defined(VTEP_FAKE01)
				bool bReturnFakeValue;	// Return fake value? (precisely last assigned value)
				float fFake01Value;		// Last externally assigned value
			#endif
		};




		// 
		// An integer parameter. Though internal representation is still float ;-)
		//
		class CParameter_int: public CParameter
		{
		public:
			CParameter_int(CParameterValidator *ownr, const VTCHAR *nme, const VTCHAR *lbl, 
				SDWORD min, SDWORD max, SDWORD def):
				CParameter(ownr, nme, lbl, float(min), float(max), float(def), false)
			{}

			virtual void setValue(float x){ CParameter::setValue(float(SRound(x))); }
			virtual float getVisibleValue(float x) const { return float(validate( float(SRound(x)) )); }
			virtual float fromVisibleValue(float x) const { return validate(float( SRound(x) )); }	
			virtual void getAsText(float x, VTCHAR *out, size_t maxLen=kMaxParamStrLen) const { int2string(getVisibleValue(x), out, maxLen, showPlusSign); }
			virtual void textFromVisible(float x, VTCHAR *out, size_t maxLen=kMaxParamStrLen) const { int2string(x, out, maxLen, showPlusSign); }

			int getAsInt() const { return SRound( getValue() ); }

			virtual bool CanIterateValues(){ return true; }
			virtual size_t GetNValues(){ return size_t( SRound( getMax() )-SRound( getMin() ) + 1 ); }
			virtual float GetValueByIdx(size_t idx){ return float(SRound( getMin() )+int(idx)); }
			
			virtual size_t GetIdxOfValue(float value){ 
				int ret = SRound( validate(value) )-SRound( getMin() ); 
				VTASSERT( ret>=0 && ret<int(GetNValues()) );
				return size_t( ret );
			}
		};


		//
		// A scaled parameter which ranges from fMin to fMax inside 
		// and from fVisibleMin to fVisibleMax outside.
		// 
		class CParameter_scaled: public CParameter
		{
		public:
			CParameter_scaled(CParameterValidator *ownr, const VTCHAR *nme, const VTCHAR *lbl,
				float min, float max, float def, float visMin, float visMax ):
				CParameter( ownr, nme, lbl, min, max, def ),
				m_fVisMin( visMin ),
				m_fVisMax( visMax ),
				m_fVisRange( visMax-visMin ){}
			
			virtual float fromVisibleValue(float x) const { 
				return from01_raw( (x-m_fVisMin)/m_fVisRange );
			}
			
			virtual float getVisibleValue(float x) const { 
				return as01_raw(x)*m_fVisRange + m_fVisMin; 
			}

			// virtual void getAsText(float x, VTCHAR *out) const { int2string(getVisibleValue(x), out, kMaxParamStrLen, showPlusSign); }
			// virtual void textFromVisible(float x, VTCHAR *out) const { int2string(x, out, kMaxParamStrLen, showPlusSign); }

		protected:
			float m_fVisRange, m_fVisMin, m_fVisMax;
		};


		//
		// A scaled parameter which ranges from fMin to fMax inside 
		// and from iVisibleMin to iVisibleMax outside.
		// 
		class CParameter_scaled_int: public CParameter
		{
		public:
			CParameter_scaled_int(CParameterValidator *ownr, const VTCHAR *nme, const VTCHAR *lbl,
				float min, float max, float def, SDWORD visMin, SDWORD visMax ):
				CParameter( ownr, nme, lbl, min, max, def ),
				m_iVisMin( visMin ),
				m_iVisMax( visMax ),
				m_fVisRange( float(visMax-visMin) ){}
			
			virtual float fromVisibleValue(float x) const { 
				return from01_raw( (x-float(m_iVisMin))/m_fVisRange );
			}
			
			virtual float getVisibleValue(float x) const { 
				return as01_raw(x)*m_fVisRange + float(m_iVisMin); 
			}

			virtual void getAsText(float x, VTCHAR *out, size_t maxLen=kMaxParamStrLen) const { int2string(getVisibleValue(x), out, maxLen, showPlusSign); }
			virtual void textFromVisible(float x, VTCHAR *out, size_t maxLen=kMaxParamStrLen) const { int2string(x, out, maxLen, showPlusSign); }

		protected:
			SDWORD m_iVisMin, m_iVisMax;
			float m_fVisRange;
		};





		//
		// % parameter, ranging [0..100] or [1..100].
		//
		class CParameter_percent: public CParameter_int
		{
		public:
			CParameter_percent(CParameterValidator *ownr, const VTCHAR *nme, SDWORD def, SDWORD min=0):
				CParameter_int( ownr, nme, _T("%"), min, 100, def )
			{}
		};


		//
		// "Combobox" parameter - a list of strings which can be put into
		// pop-up menu, drop-down box, and etc. The value of the parameter is
		// the index of currently selected strings.
		//
		class CParameter_combo: public CParameter_int
		{
		public:		
			// "values" is a NULL-terminated array of pointers to string values.
			// Max value of the parameter is determined automatically, 
			// from length of the array.
			// Note that names of the entries are copied into internal storage.
			CParameter_combo(CParameterValidator *ownr, const VTCHAR *nme, 
				const VTCHAR *const *values, int def=0);
			~CParameter_combo();

			virtual void SetNewValues( const VTCHAR *const *values, bool notifyIfValueChanged );
			virtual void visibleText2Visible(VTCHAR *string, float &output) const;		
			virtual void getAsText(float x, VTCHAR *out, size_t maxLen=kMaxParamStrLen) const;

			virtual void textFromVisible(float x, VTCHAR *out, size_t maxLen=kMaxParamStrLen) const { getAsText(x, out, maxLen); }
		protected:	
			VTCHAR **labels;

			virtual void DeleteLabels();
		};


		//
		// "Combobox" parameter - a list of strings which can be put into
		// pop-up menu, drop-down box, and etc. The value of the parameter is
		// the number which is encoded in the selected string.
		//
		class CParameter_strval_combo: public CParameter
		{
		public:
			static const size_t kMaxLabel = 255;

			CParameter_strval_combo(CParameterValidator *ownr, const VTCHAR *nme, const VTCHAR *lbl,
				const VTCHAR *const *values, const float *realValues=NULL, float def=0);
			~CParameter_strval_combo();

			// virtual void visibleText2Visible(VTCHAR *string, float &output) const;
			virtual void getAsText(float x, VTCHAR *out, size_t maxLen=kMaxParamStrLen) const;
			virtual void textFromVisible(float x, VTCHAR *out, size_t maxLen=kMaxParamStrLen) const;

			virtual float getVisibleValue(float x) const;
			virtual float fromVisibleValue(float x) const;

			virtual bool CanIterateValues(){ return true; }
			virtual size_t GetNValues(){ return m_uEntries; }
			virtual float GetValueByIdx(size_t idx){ return m_pEntries[idx].m_fValue; }
			virtual size_t GetIdxOfValue(float value){ return FindClosestEntry( value ); }

		protected:
			size_t m_uEntries;

			struct TEntry{			
				VTCHAR m_Label[kMaxLabel+1];
				float m_fValue;
				TEntry(): m_fValue(0.0f){ m_Label[0]=0; }
			} *m_pEntries;

			// Find an entry which has a value closest to 'search'.
			size_t FindClosestEntry( float search ) const;
		};


		//
		// Float parameter which visually changes from -100 to +100, and [-1..1] internally.
		//
		class CParameter_pm100: public CParameter
		{
		public:	
			CParameter_pm100(CParameterValidator *ownr, const VTCHAR *nme, float def=0):
			  CParameter(ownr, nme, _T("%"), 
				  -1.005f, // fixes rounding error which causes min visible value to be -99
				  1.0f, def)
			{
				showPlusSign = true;
			}
			virtual float getVisibleValue(float x) const { return validate(x) * 100.0f; }
			virtual float fromVisibleValue(float x) const { return validate( x/100.0f ); }	
			virtual void getAsText(float x, VTCHAR *out, size_t maxLen=kMaxParamStrLen) const { int2string(getVisibleValue(x), out, maxLen, showPlusSign); }
			virtual void textFromVisible(float x, VTCHAR *out, size_t maxLen=kMaxParamStrLen) const { int2string(x, out, maxLen, showPlusSign); }
		};


		//
		// Panoram controller - [-1..1], visually represented as [-100%..100%].
		//
		// ???????????? Shouldn't it be just equal to CParameter_pm100?
		//
		class CParameter_panoram: public CParameter_pm100
		{
		public:	
			CParameter_panoram(CParameterValidator *ownr, const VTCHAR *nme, float def=0):
				CParameter_pm100(ownr, nme, def)
			{
				showPlusSign = false; // ??
			}
		};


		//
		// Integer parameter which visually changes from -100% to +100%, and [-1..1] internally.
		//
		class CParameter_int_pm100pc: public CParameter_int
		{
		public:	
			CParameter_int_pm100pc(CParameterValidator *ownr, const VTCHAR *nme, SDWORD def=0):
			  CParameter_int(ownr, nme, _T("%"), -100, 100, def)
			{
				showPlusSign = true;
			}
		};


		//
		// "Amplitude as Decibels" parameter: internally repesented as amplitude (voltage),
		// it is visually represented in decibels, with automatic conversion of course.
		// Logarithmic handle can be applied as well, to make the parameter even cooler.
		//
		class CParameter_dB: public CParameter
		{
		public:
			CParameter_dB(CParameterValidator *ownr, const VTCHAR *nme, float def, bool logH=false):
				CParameter(ownr, nme, LCT("dB"), 0, 1, def, logH),
				max_dB(0), min_dB(kMin_dB){}

			CParameter_dB(CParameterValidator *ownr, const VTCHAR *nme, float def, float maxdB, bool logH):
				CParameter(ownr, nme, LCT("dB"), 0, float(Convert_dBTo01_NoClip(maxdB)), def, logH),
				max_dB(maxdB), min_dB(kMin_dB){}	

			CParameter_dB(CParameterValidator *ownr, const VTCHAR *nme, float def, float mindB, float maxdB, bool logH):
				CParameter(ownr, nme, LCT("dB"), float(Convert_dBTo01_NoClip(mindB)), float(Convert_dBTo01_NoClip(maxdB)), def, logH),
				max_dB(maxdB), min_dB(mindB){}	

			virtual float getVisibleValue(float x) const { return float(Convert_01TodB(x, max_dB)); }
			virtual float fromVisibleValue(float x) const { return float(Convert_dBTo01(x, getMax())); }

			virtual void getAsText(float x, VTCHAR *out, size_t maxLen=kMaxParamStrLen) const { dB2string(x, out, maxLen, showPlusSign); }
			virtual void textFromVisible(float x, VTCHAR *out, size_t maxLen=kMaxParamStrLen) const { dB2stringNC(x, out, maxLen, showPlusSign); }
			virtual void visibleText2Visible(VTCHAR *string, float &output) const { VisibleTextToVisible_dB(string, output); }
			virtual bool is_dB() const { return true; }
		private:
			float max_dB, min_dB;
		};


		//
		// Millisecons. A feature of this parameter is that if value is over 10ms, then
		// it is displayed without fractions; otherwise, with tenths of msec.
		//
		class CParameter_ms: public CParameter
		{
		public:
			CParameter_ms(CParameterValidator *ownr, const VTCHAR *nme, float min, float max, float def, bool logH=false):
			  CParameter(ownr, nme, LCT("ms"), min, max, def, logH){}
			virtual void getAsText(float x, VTCHAR *out, size_t maxLen=kMaxParamStrLen) const { textFromVisible(getVisibleValue(x), out, maxLen); }
			virtual void textFromVisible(float x, VTCHAR *out, size_t maxLen=kMaxParamStrLen) const { VTSNPRINTFC(out, maxLen, (x<10)?"%0.1f":"%0.0f", x); }
		};

		//
		// Boolean parameter (ON/OFF).
		// asBool()/setAsBool() functions from CParameter can be used.
		//
		class CParameter_onoff: public CParameter_int
		{
		public:
			CParameter_onoff(CParameterValidator *ownr, const VTCHAR *nme, bool def=false):
			  CParameter_int(ownr, nme, _T(""), 0, 1, SDWORD(def)){}	
			virtual void getAsText(float x, VTCHAR *out, size_t maxLen=kMaxParamStrLen) const { textFromVisible(x, out, maxLen); }
			virtual void textFromVisible(float x, VTCHAR *out, size_t maxLen=kMaxParamStrLen) const { VTSTRNCPY(out, (x>=0.5)?"ON":"OFF", maxLen); }
		};

		
		//
		// A ratio of 2 numbers. When set to max value, it displays a string
		// maxValueLabel (usually, infinity symbol) instead of the number.
		// 
		class CParameter_ratio: public CParameter
		{
		public:
			CParameter_ratio(CParameterValidator *ownr, const VTCHAR *nme, float min, float max, 
				float def, bool logH):
			  CParameter(ownr, nme, _T("1/x"), min, max, def, logH)
			{ 
				strcpy(maxValueLabel, kInfinitySymbol); 
			}

			CParameter_ratio(CParameterValidator *ownr, const VTCHAR *nme, float min, float max, const VTCHAR *maxdisp, float def, bool logH):
			  CParameter(ownr, nme, _T("1/x"), min, max, def, logH){ strncpy(maxValueLabel, maxdisp, kMaxParamStrLen); }
			
			virtual void getAsText(float x, VTCHAR *out, size_t maxLen=kMaxParamStrLen) const { textFromVisible(x, out, maxLen); }
			
			virtual void textFromVisible(float x, VTCHAR *out, size_t maxLen=kMaxParamStrLen) const { 
				if(isAtMax(x) && maxValueLabel[0]) 
					VTSTRNCPY(out, maxValueLabel, maxLen); 
				else 
					CParameter::getAsText(x, out, maxLen); 
			}
		protected:
			VTCHAR maxValueLabel[ kMaxParamStrLen+1 ];
		};


		//
		// Frequency in Hertz. A feature of this parameter is that if value is over 10Hz, then
		// it is displayed without fractions; otherwise, with tenths of Hz.
		//
		class CParameter_Hz: public CParameter
		{
		public:
			CParameter_Hz(CParameterValidator *ownr, const VTCHAR *nme, float min, float max, float def, bool logH=true):
			  CParameter(ownr, nme, LCT("Hz"), min, max, def, logH){}
			virtual void getAsText(float x, VTCHAR *out, size_t maxLen=kMaxParamStrLen) const { textFromVisible(getVisibleValue(x), out, maxLen); }
			virtual void textFromVisible(float x, VTCHAR *out, size_t maxLen=kMaxParamStrLen) const { VTSNPRINTFC(out, maxLen, (x<10)?"%0.1f":"%0.0f", x); }
		};


		//
		// Audio filter order controller.
		// Has filter order in the internal representation, and dB/Octave display
		// i.e. 1 => 6 dB/Oct, 2 => 12 dB/Oct, and etc.
		//
		class CParameter_filterOrder: public CParameter_int
		{
		public:
			// min, max, def are filter orders
			CParameter_filterOrder(CParameterValidator *ownr, const VTCHAR *nme, 
				SDWORD min=1, SDWORD max=4, SDWORD def=1):
				CParameter_int(ownr, nme, LCT("dB/Oct"), min, max, def)
			{
				VTASSERT( min>=1 );
			}
			virtual float getVisibleValue(float x) const { return float(URound(CParameter_int::getVisibleValue(x)*6.0f)); }
			virtual float fromVisibleValue(float x) const { return CParameter_int::fromVisibleValue(x/6.0f); }

		};

	} // namespace Params
} // namespace VT


#endif


