//-----------------------------------------------------------------------------
//
// Parameter storage abstraction.
//
//-----------------------------------------------------------------------------

/*
	VT Framework License (BSL 1.0)

	Copyright (C) 2006 Sergey A. Galin 
	E-mail: sergey.galin@gmail.com

	Permission is hereby granted, free of charge, to any person or organization
	obtaining a copy of the software and accompanying documentation covered by
	this license (the "Software") to use, reproduce, display, distribute,
	execute, and transmit the Software, and to prepare derivative works of the
	Software, and to permit third-parties to whom the Software is furnished to
	do so, all subject to the following:

	The copyright notices in the Software and this entire statement, including
	the above license grant, this restriction and the following disclaimer,
	must be included in all copies of the Software, in whole or in part, and
	all derivative works of the Software, unless such copies or derivative
	works are solely in the form of machine-executable object code generated by
	a source language processor.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
	SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
	FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
	ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
	DEALINGS IN THE SOFTWARE.
*/

#ifndef VT_PARAMSTORAGE_H
#define VT_PARAMSTORAGE_H

#include "vt_assert.hpp"

namespace VT
{

	template<class T> class CStorage
	{
	public:
		virtual void set( T val )=0;
		virtual T get() const=0;		
	};

	template<class T> class CInternalStorage: public CStorage<T>
	{
	public:
		CInternalStorage<T>( T value=T(0) ): m_Value(value){}
		virtual void set( T val ){ m_Value = val; }
		virtual T get() const { return m_Value; }
	protected:
		T m_Value;
	};




	template<class T> class CPointerStorage: public CStorage<T>
	{
	public:
		CPointerStorage<T>( T* pvalue ): m_pValue(pvalue)
		{
			VTASSERT2T( m_pValue!=NULL, "NULL pointer in PointerStorage." );
			if( m_pValue==NULL )
				throw _T("NULL pointer in PointerStorage.");
		}
		
		virtual void set( T val ){ (*m_pValue) = val; }		
		virtual T get() const { return *m_pValue; }
	protected:
		T *m_pValue;
	};


	template<class T> class CFloatToBoolPointerStorage: public CStorage<T>
	{
	public:
		CFloatToBoolPointerStorage<T>( bool* pvalue ): m_pValue(pvalue)
		{
			VTASSERT2T( m_pValue!=NULL, "NULL pointer in FloatToBoolPointerStorage." );
			if( m_pValue==NULL )
				throw _T("NULL pointer in F2BPointerStorage.");
		}
		
		virtual void set( T val ){ (*m_pValue) = (val>=T(0.5))? true: false; }		
		virtual T get() const { return (*m_pValue)? T(1.0) : T(0.0); }
	protected:
		bool *m_pValue;
	};



	template<class T, class OWNER> class CClassPropertyStorage: public CStorage<T>
	{
	public:
		typedef T (OWNER::*TGetFunctionPtr)() const;
		typedef void (OWNER::*TSetFunctionPtr)(T);
		
		CClassPropertyStorage( OWNER *owner, TGetFunctionPtr get, TSetFunctionPtr set ):
			m_Owner( owner ),
			m_GetFunction( get ),
			m_SetFunction( set )
		{}

		virtual void set( T val ){ ((*m_Owner).*m_SetFunction)( val ); }		
		virtual T get() const { return ((*m_Owner).*m_GetFunction)(); }

	protected:
		OWNER* m_Owner;
		TGetFunctionPtr m_GetFunction;
		TSetFunctionPtr m_SetFunction;
	};





	//
	// Float-to-bool
	//
	template<class T, class OWNER> class CClassFloatBoolPropertyStorage: public CStorage<T>
	{
	public:
		typedef bool (OWNER::*TGetFunctionPtr)() const;
		typedef void (OWNER::*TSetFunctionPtr)(bool);
		
		CClassFloatBoolPropertyStorage( OWNER *owner, TGetFunctionPtr get, TSetFunctionPtr set ):
			m_Owner( owner ),
			m_GetFunction( get ),
			m_SetFunction( set )
		{}

		virtual void set( T val ){ ((*m_Owner).*m_SetFunction)( val>=T(0.5) ); }		
		virtual T get() const { return ((*m_Owner).*m_GetFunction)(); }

	protected:
		OWNER* m_Owner;
		TGetFunctionPtr m_GetFunction;
		TSetFunctionPtr m_SetFunction;
	};




	//
	// Float-to-int
	//
	template<class T, class OWNER> class CClassFloatIntPropertyStorage: public CStorage<T>
	{
	public:
		typedef int (OWNER::*TGetFunctionPtr)() const;
		typedef void (OWNER::*TSetFunctionPtr)(int);
		
		CClassFloatIntPropertyStorage( OWNER *owner, TGetFunctionPtr get, TSetFunctionPtr set ):
			m_Owner( owner ),
			m_GetFunction( get ),
			m_SetFunction( set )
		{}

		virtual void set( T val ){ ((*m_Owner).*m_SetFunction)( SRound(val) ); }
		virtual T get() const { return T(((*m_Owner).*m_GetFunction)()); }

	protected:
		OWNER* m_Owner;
		TGetFunctionPtr m_GetFunction;
		TSetFunctionPtr m_SetFunction;
	};





	//
	// CAssociatedValue is a "variable" which is stored either in itself,
	// in some other place (accessed via pointer) or is a class property.
	//
	template<class T> class CAssociatedValue
	{
	public:
		CAssociatedValue(T value=T(0)):
			m_Storage( new CInternalStorage<T>(value) )
		{}

		~CAssociatedValue(){ delete m_Storage; }

		// "Low-level" association.
		// Remember that the object will be deleted using 
		// operator delete in destructor.
		void Associate( CStorage<T> *storage, bool bCarryValue = false ){
			T oldValue;
			if( bCarryValue )
				oldValue = getValue();
			delete m_Storage;
			m_Storage = storage;
			if( bCarryValue )
				setValueNE( oldValue );
			OnAssociate();
		}
		
		void Deassociate(){
			T value = getValue();
			Associate( new CInternalStorage<T>(value) );
		}

		void AssociateWith( T* ptr, bool bCarryValue = false ){ 
			Associate( new CPointerStorage<T>(ptr), bCarryValue ); }

		void AssociateWithBoolean( bool *ptr, bool bCarryValue = false ){ 
			Associate( new CFloatToBoolPointerStorage<T>(ptr), bCarryValue ); }

		#if defined(_MSC_VER)
			#define VT_PARAMSTORAGE_TYPENAME typename
		#else
			#define VT_PARAMSTORAGE_TYPENAME
		#endif
		
		template<class OWNER> void AssociateWithProperty( 
			VT_PARAMSTORAGE_TYPENAME OWNER* owner,
			typename CClassPropertyStorage<T, OWNER>::TGetFunctionPtr get,
			typename CClassPropertyStorage<T, OWNER>::TSetFunctionPtr set,
			bool bCarryValue = false )
		{ 
			Associate( new CClassPropertyStorage<T, OWNER>(owner, get, set), bCarryValue );
		}

		template<class OWNER> void AssociateWithFloatBoolProperty( 
			VT_PARAMSTORAGE_TYPENAME OWNER* owner,
			typename CClassFloatBoolPropertyStorage<T, OWNER>::TGetFunctionPtr get,
			typename CClassFloatBoolPropertyStorage<T, OWNER>::TSetFunctionPtr set,
			bool bCarryValue = false )
		{ 
			Associate( new CClassFloatBoolPropertyStorage<T, OWNER>(owner, get, set), bCarryValue );
		}

		template<class OWNER> void AssociateWithFloatIntProperty( 
			VT_PARAMSTORAGE_TYPENAME OWNER* owner,
			typename CClassFloatIntPropertyStorage<T, OWNER>::TGetFunctionPtr get,
			typename CClassFloatIntPropertyStorage<T, OWNER>::TSetFunctionPtr set,
			bool bCarryValue = false )
		{ 
			Associate( new CClassFloatIntPropertyStorage<T, OWNER>(owner, get, set), bCarryValue );
		}

		// !!!!!!!!!! todo: somehow refactor member names
		// without breaking code compatibility.
		VTINLINE T getValue() const { return m_Storage->get(); }
		VTINLINE void setValueNE(T newval){ m_Storage->set(newval); }

	protected:
		CStorage<T> *m_Storage;
		virtual void OnAssociate(){;}
	};


} // namespace VT

#endif


