//-----------------------------------------------------------------------------
//
// C++ Cross-Compilation Portability layer
//
// Macros and type definitions to help writing portable applications, allowing
// to move source code without modifications between:
//
//   - GCC, MSC, C++ Builder
//   - Posix (Linux and, in future, Mac OS X) and Win32; DOS compatibility 
//     can also be implemented later if ever necessary
//   - UNICODE <-> Multi-Byte MSC and GCC builds
//
// Also:
//
//   - Configuration detection macros (detect Win32, Linux, wide strings...)
//   - Misc useful macros, like VTINLINE
//   - Ensures _DEBUG / _NDEBUG macros are always defined
//   - Ensures MAX_PATH is defined
//   - Ensures BYTE/WORD/DWORD/DWORD64/SWORD/SDWORD/SDWORD64 are defined
//   - Automatical switching to 64 bit file size, when available
//   - Simple built-in conversion between C strings and wchar_t strings 
//     (with built-in implementation for ASCII characters and Win API
//     functions in Windows VT_UNICODE builds)
//
// Not supported at the moment:
//
//   - Big-endian machines
//
// Types in this file are NOT NAMESPACED, because macros don't go into namespaces 
// anyway and there are typedefs which mimic :: types. However, functions are in
// VT:: namepace. VTFunctionName macros are provided for backward compatibility.
//
// Dependency: no dependencies, this is the lowest level of the food pyramid
// (right above standard system libraries).
//
//-----------------------------------------------------------------------------

/*
	VT Framework License (BSL 1.0)

	Copyright (C) 2006 Sergey A. Galin 
	E-mail: sergey.galin@gmail.com

	Permission is hereby granted, free of charge, to any person or organization
	obtaining a copy of the software and accompanying documentation covered by
	this license (the "Software") to use, reproduce, display, distribute,
	execute, and transmit the Software, and to prepare derivative works of the
	Software, and to permit third-parties to whom the Software is furnished to
	do so, all subject to the following:

	The copyright notices in the Software and this entire statement, including
	the above license grant, this restriction and the following disclaimer,
	must be included in all copies of the Software, in whole or in part, and
	all derivative works of the Software, unless such copies or derivative
	works are solely in the form of machine-executable object code generated by
	a source language processor.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
	SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
	FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
	ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
	DEALINGS IN THE SOFTWARE.
*/


#if !defined(VT_PORTABLE_H)
#define VT_PORTABLE_H

//-----------------------------------------------------------------------------
// Common includes & config macroses
//-----------------------------------------------------------------------------
//
// Fixing debug macros.
//
#if defined(DEBUG)
	#if !defined(_DEBUG)
		#define _DEBUG 1
	#endif
#endif

#if defined(_DEBUG)
	#if defined(NDEBUG)
		#error _DEBUG and NDEBUG cannot be defined at the same time!
	#endif
#else
	#if !defined(NDEBUG)
		#define NDEBUG 1
	#endif
#endif

#if !defined(NDEBUG) && !defined(_DEBUG)
	#warning Neither of _DEBUG and NDEBUG are defined.
#endif

//
// Adding system-dependent includes which are absolutely necessary.
//
#if defined(_MSC_VER) || defined(WINDOWS) || defined(WIN32) || defined(_WINDOWS) || defined(VT_WINDOWS)
	#if !defined(VT_WINDOWS)
		#define VT_WINDOWS
	#endif

	#include <windows.h>
	#include <tchar.h>
#else
	#define VT_NIXGCC

	#include <unistd.h>
	#include <sys/types.h>
	#include <dirent.h>
	#include <sys/stat.h>
	#include <fcntl.h>
#endif

#if defined(_MSC_VER)
	#define VT_64BIT_FILEOIO 1
#endif

#if defined(LINUX)
	#if LINUX!=0 
		#if !defined(VT_LINUX)
			#define VT_LINUX 1
		#endif
	#endif
#endif

#if defined(VT_LINUX) && !defined(LINUX)
	#define LINUX 1
#endif

//
// Adding standard includes which are necessary in the portability part.
//
#include <string.h>
#include <stdlib.h>
#include <stdio.h>


//-----------------------------------------------------------------------------
// Character types and string functions
//-----------------------------------------------------------------------------

#if defined(VT_WINDOWS) && defined(UNICODE)
	#define VTCHAR		wchar_t
	#define LPVTCHAR	LPWSTR
	#define LPCVTCHAR	LPCWSTR
#else
	#define VTCHAR		char
	typedef char* LPVTCHAR;
	typedef const char* LPCVTCHAR;
	#if !defined(_T)
		#define _T(x)	x
	#endif
#endif

//
// Some functions are named differently in GCC, MSC, BCC, in UNICODE/MBCS builds.
// We define macros for these functions; the macros all receive VTCHAR-based 
// arguments, so we can use VTCHAR strings and these macros and have a portable 
// code.
//
// Also, some useful string constants are defined here.
//
#if defined(VT_WINDOWS)
	#if defined(UNICODE)
		// Windows/UNICODE
		#define VTPRINTF	wprintf
		#define VTSTRNCPY	wcsncpy
		#define VTSTRNCAT	wcsncat
		#define VTSNPRINTF	wsprintf_s	
		#define VTSSCANF	swscanf		
		#define VTSTRLEN	wcslen
		#define VTUNLINK	_wunlink
		#define VTRMDIR		_wrmdir
		#define VTSTRCMP	wcscmp
		#define VTSTRICMP	wcsicmp
		#define VTSTRNCMP	wcsncmp
		#define VTSPRINTF	wsprintf
		#define VTFOPEN		_wfopen
		#define VTGETENV	_wgetenv
		#define VTISSPACE	iswspace
		#define VTISALPHA	iswalpha
		#define VTISDIGIT	iswdigit
		#define VTISXDIGIT	iswxdigit
		#define VTTOLOWER	towlower
		#define VTTOUPPER	towupper
	#else
		// Windows/MBCS
		#define VTPRINTF	printf
		#define VTSTRNCPY	strncpy
		#define VTSTRNCAT	strncat
		#define VTSNPRINTF	sprintf_s	
		#define VTSSCANF	sscanf
		#define VTSTRLEN	strlen
		#define VTUNLINK	_unlink
		#define VTRMDIR		_rmdir
		#define VTSTRCMP	strcmp
		#define VTSTRICMP	_stricmp
		#define VTSTRNCMP	strncmp
		#define VTSPRINTF	sprintf
		#define VTFOPEN		fopen
		#define VTGETENV	getenv	
		#define VTISSPACE	isspace
		#define VTISALPHA	isalpha
		#define VTISDIGIT	isdigit
		#define VTISXDIGIT	isxdigit
		#define VTTOLOWER	tolower
		#define VTTOUPPER	toupper
	#endif
	#define VTENDL "\r\n"
#else
	// Non-Windows (assuming ANSI/MBCS)
	#define VTSTRNCPY	strncpy
	#define VTSTRNCAT	strncat
	#if defined(_MSC_VER)
		#define VTSNPRINTF sprintf_s
	#else
		#define VTSNPRINTF snprintf
	#endif
	#define VTPRINTF	printf
	#define VTSSCANF	sscanf
	#define VTSTRLEN	strlen
	#define VTUNLINK	unlink
	#define VTRMDIR		rmdir
	#define VTSTRCMP	strcmp
	#define VTSTRICMP	VTStrICmp
	#define VTSTRNCMP	strncmp
	#define VTSPRINTF	sprintf
	#if defined(VT_64BIT_FILEOIO)
		#define VTFOPEN	fopen64
	#else
		#define VTFOPEN	fopen
	#endif
	#define VTGETENV	getenv
	#define VTISSPACE	isspace
	#define VTISALPHA	isalpha
	#define VTISDIGIT	isdigit
	#define VTISXDIGIT	isxdigit
	#define VTTOLOWER	tolower
	#define VTTOUPPER	toupper
	#define VTENDL "\n"
#endif

// snprintf function for char* (not for VTCHAR*).
#if defined(_MSC_VER)
	#define VTSNPRINTFC sprintf_s
#else
	#define VTSNPRINTFC snprintf
#endif

// Some compilers don't have this function in their run-time,
// so we simply link to our own replacement.
#define VTSTRNICMP VTStrNICmp

#define VTENDLT _T(VTENDL)

// 
// Simple char<->VTCHAR character conversion functions, for UNICODE builds work 
// correctly for ASCII (7-bit) characters only.
// Output buffer must be at least maxLen+1 characters size.
// If replaceChar is 0, it is not used (>7bit chars are just thrown away).
// These functions work the same for UNICODE and non-UNICODE, except that VTCHAR
// size changes, of course.
//
int VTASCII2VTCHAR(const char* src, LPVTCHAR dest, int maxLen, VTCHAR replaceChar=_T('?'), int *nReplacedChars=NULL);
int VTVTCHAR2ASCII(LPCVTCHAR src, char *dest, int maxLen, char replaceChar='?', int *nReplacedChars=NULL);

//
// stricmp/strcmpi and strnicmp implementation. 
// Some compilers don't have such functions in runtime.
//
int VTStrICmp(LPCVTCHAR s1, LPCVTCHAR s2);
int VTStrNICmp(LPCVTCHAR s1, LPCVTCHAR s2, size_t n);

//-----------------------------------------------------------------------------
// Function modifiers & compiler options
//-----------------------------------------------------------------------------

#if defined(_MSC_VER)
	#if !defined(_DEBUG)
		#define VTINLINE __forceinline
	#else
		#define VTINLINE inline
	#endif

	// Combatting that stupid warning which floods build log.
	// Says that C-style string functions are all evil, scary 
	// and dangerous and so on.
	#if !defined(_CRT_SECURE_NO_DEPRECATE)
		#define _CRT_SECURE_NO_DEPRECATE 1
	#endif
	#pragma warning( disable: 4996 )

#else
	#define VTINLINE inline
#endif

//-----------------------------------------------------------------------------
// C++ Exceptions
//-----------------------------------------------------------------------------
//
// Indicate that exceptions are supported by compiler.
//
#if defined(_MSC_VER) || defined(__GNUC__)
	#define VT_EXCEPTIONS
#endif


//-----------------------------------------------------------------------------
// Namespaces
//-----------------------------------------------------------------------------
//
// This section only needed if you want to make code which uses namespaces 
// compatible with an ancient compiler which doesn't understand them. Otherwise
// you can freely use the 'namespace' keyword.
//

#if defined(_MSC_VER) || defined(__GNUC__)
	#define VT_USE_NAMSESPACES
#endif


#if 0

// Namespace which can be simulated using a class:
//
// VTNAMESPACE nspMyNameSpace {
//	   VTNSMEMBERS;
//     VTNSM int x; // variable
//     VTNSM void blablablah(...); // function prototype
//     ...
// }
// nspMyNameSpace::blablablah(...){ ... } // implementation or call
// 
//
// Namespace which cannot be simulated:
//
// VT_OPNNAMESPACE_START( nspMyNameSpace )
//     int x; // variable
//     void blablablah(...);  // function prototype
//     ...
// VT_OPNNAMESPACE_END
//
// VTONSM( nspMyNameSpace, blablablah )(){ .... } // implementation or call
//


#if defined(VT_USE_NAMSESPACES)
	// For a namespace which can be simulated using a class with static functions
	#define VTNAMESPACE namespace
	#define VTNSM 
	#define VTUSENAMESPACE(x) using namespace x;
	#define VTNSMEMBERS
	
	// For a namespace which cannot be simulated using a class
	#define VT_OPNNAMESPACE_START(name)	namespace name{
	#define VT_OPNNAMESPACE_END			}; using namespace name;
	#define VTONSM(name, member)		name::member

#else
	// For a namespace which can be simulated using a class with static functions
	#define VTNAMESPACE class
	#define VTNSM static
	#define VTUSENAMESPACE(x)
	#define VTNSMEMBERS public:

	// For a namespace which cannot be simulated using a class
	#define VT_OPNNAMESPACE_START(name)	
	#define VT_OPNNAMESPACE_END
	#define VTONSM(name, member)		member
#endif

#endif


//-----------------------------------------------------------------------------
// Filesystem stuff
//-----------------------------------------------------------------------------

#if defined(VT_WINDOWS) || defined(VT_DOS)
	#define VTPATHSLASH		"\\"
	#define VTPATHSLASHCH	'\\'
	#define VT_FS_CASEINSENSITIVE	1
	#define VT_ALLFILESMASK "*.*"
#else
	#define VTPATHSLASH		"/"
	#define VTPATHSLASHCH	'/'	
	#define VT_FS_CASESENSITIVE		1
	#define VT_ALLFILESMASK "*"
#endif

// Also defined under non-UNIX systems, although most likely useless :)
#define VTDEFMKDIRMODE	0755 /* rwx-rx-rx */

#if defined(VT_WINDOWS) && defined(_MSC_VER)
	#define VTFSIZE __int64 
	#define VTFSEEK _fseeki64
	#define	VTFTELL _ftelli64
#else
	#if defined(VT_64BIT_FILEOIO)
		#define VTFSIZE off64_t 
		#define VTFSEEK fseek64
		#define	VTFTELL ftell64
	#else
		#define VTFSIZE off_t
		#define VTFSEEK fseek
		#define	VTFTELL ftell		
	#endif
#endif

#define VTPATHSLASHT		_T(VTPATHSLASH)
#define VTPATHSLASHCHT		_T(VTPATHSLASHCH)
#define VT_ALLFILESMASKT    _T(VT_ALLFILESMASK)

// 
// Make sure MAX_PATH is defined. All compilers have some max path macro,
// but it is named differently.
//
#if !defined(MAX_PATH)
	#if defined(MAXPATH)
		#define MAX_PATH MAXPATH
	#elif defined(_MAX_PATH)
		#define MAX_PATH _MAX_PATH
	#elif defined(_POSIX_PATH_MAX)
		#define MAX_PATH _POSIX_PATH_MAX
	#else
		#error "MAX_PATH can't be defined!"
	#endif
#endif

//-----------------------------------------------------------------------------
// Windows-like Data Types
//-----------------------------------------------------------------------------

#if !defined(VT_WINDOWS)
    typedef unsigned char	BYTE;
    typedef unsigned short	WORD;
	typedef signed short	SWORD;
    typedef unsigned int	DWORD; 
	typedef signed int		SDWORD; 
	typedef long long		DWORD64;	
    typedef BYTE BOOL;
    static const BOOL TRUE = 1;
    static const BOOL FALSE = 0;
    
    #define FOURCC DWORD
    #define MAKEFOURCC(a,b,c,d) VTMakeFOURCC((a),(b),(c),(d))
#else
	typedef signed int		SDWORD;
	typedef signed short	SWORD;
#endif

typedef signed short SIGNED_WORD;

// Portable FOURCC functions. Presuming that generating FOURCC is
// not a very frequent task, the functions were not made inline.
FOURCC VTMakeFOURCC(unsigned char a, unsigned char b=' ', unsigned char c=' ', unsigned char d=' ');
FOURCC VTMakeFOURCC(const char* str);
void VTFOURCC2Str(FOURCC code, char* str); // Convert FOURCC to text string
const char* VTFOURCC2Str(FOURCC code); // Convert FOURCC to text string


//-----------------------------------------------------------------------------
// Program arguments
//-----------------------------------------------------------------------------
//
// There should be some common way to allow library functions to access
// program command-line arguments.
// For VT lib, it is done by setting vars via VTSetGlobalArgs() at program
// entry point (e.g. 'main(argc, argv)') or somewhere during initialization.
//
// There's currently no support for UTF-16!!!!
//
extern int vt_argc;
extern const char* const* vt_argv; 

inline void VTSetGlobalArgs(int argc, const char* const* argv)
{
    vt_argc = argc;
    vt_argv = argv;
}

// Checks if global vars are set, and terminates program if not.
void VTCheckGlobalArgs();








#endif

