//-----------------------------------------------------------------------------
//
// A class for loading PCM Wav and raw audio which fits into RAM. 
// This is usually the case for sampler's stuff, small utilites...
// The classes also provides saving of audio to PCM WAV and raw and
// loading audio from executable's resources.
//
// NOTE: Full functionality is available only if basic data type is 
// floating-point (float / double / long double).
//
//-----------------------------------------------------------------------------

/*
	VT Framework License (BSL 1.0)

	Copyright (C) 2006 Sergey A. Galin 
	E-mail: sergey.galin@gmail.com

	Permission is hereby granted, free of charge, to any person or organization
	obtaining a copy of the software and accompanying documentation covered by
	this license (the "Software") to use, reproduce, display, distribute,
	execute, and transmit the Software, and to prepare derivative works of the
	Software, and to permit third-parties to whom the Software is furnished to
	do so, all subject to the following:

	The copyright notices in the Software and this entire statement, including
	the above license grant, this restriction and the following disclaimer,
	must be included in all copies of the Software, in whole or in part, and
	all derivative works of the Software, unless such copies or derivative
	works are solely in the form of machine-executable object code generated by
	a source language processor.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
	SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
	FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
	ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
	DEALINGS IN THE SOFTWARE.
*/

#ifndef VT_RAWAUDIO_HPP
#define VT_RAWAUDIO_HPP

#include <string.h>
#include <stdlib.h>
#include "vt_portable.hpp"
#include "vt_resampling.hpp"
#include "vt_dithering.hpp"
#include "vt_riff.hpp"
#include "vt_localization.hpp"
#include "vt_assert.hpp"
#include "vt_file.hpp"
#include "vt_assert.hpp"
#include "vt_audiotypes.hpp"

namespace VT {


	// This is a class which is intended for working with short audio files, 
	// which are fit into RAM completely. 
	// Audio data can be loaded from resources, PCM Wav file or raw file and 
	// saved into raw file. Also there's a resampling function which helps 
	// to match file sample rate to what you need in your application, and
	// functions to split/combine channels.
	//
	// T - sample type
	// NCH - number of channels
	template<class T, int NCH> class CInMemoryAudio
	{
	public:
		static const size_t kMinAllocSize = 16;	

		CInMemoryAudio(); // Use some Load...() or Allocate() to create empty buffer.
		#if defined(VT_WINDOWS)
			// Load raw audio stream from Win32 .exe file resources
			CInMemoryAudio(UINT iRes, LPCTSTR lpType=_T("RAWAUDIO"), HMODULE hModule=NULL);
		#endif
		~CInMemoryAudio();

		int GetSampleRate() const { return m_SampleRate; } // Get current sample rate
		void ForceSampleRate(int sr){ m_SampleRate = sr; } // Simply change sample rate setting without changing any data

		VTINLINE void Clear(); // Destroys data (but doesn't free memory buffer)
		// Check/allocate memory (don't initialize it).
		// It also does Clear() (i.e. cursor position is reset).
		// dataSize is set to 'nsamples'. 'allocate' is a suggested size (in samples 
		// of memory block to allocate. The function may ignore that value if more
		// memory is already allocated or nsamples is bigger.
		void Allocate(size_t nsamples, size_t allocate=0, bool bKeepData=false); 
		void Zero(); // Zero all samples
		void copyFrom( const CInMemoryAudio<T, NCH>& source ); // Copy everything from another mem audio
		template<class U> void ConvertCopyFrom( const CInMemoryAudio<U, NCH>& source ); // Copy with sample type conversion
		void Multiply( T k ); // Multiply all samples by 'k' (change volume).
		void Limit( T lim ); // Limit all samples by [-lim..lim]

		VTINLINE size_t getLength() const { return dataSize; } // Length in samples
		VTINLINE float getTimeLength() const { return float(dataSize)/float(m_SampleRate); } // Length in seconds
		VTINLINE T getSample(size_t pos, size_t channel) const { return data[pos].channels[channel]; }
		VTINLINE size_t getNChannels() const { return NCH; } // Number of channels
		VTINLINE size_t getSampleSize() const { return sizeof(T); } // Sample size in bytes
		VTINLINE size_t getFrameSize() const { return sizeof(TAudioFrame<T, NCH>); } // Frame size in bytes
		VTINLINE void getFrame(size_t idx, TAudioFrame<T, NCH> *out) const { (*out)=data[idx]; }
		VTINLINE TAudioFrame<T, NCH>& getFrameRef(size_t idx){ return data[idx]; }
		VTINLINE const TAudioFrame<T, NCH>& getConstFrame(size_t idx) const { return data[idx]; }

		// "Playback" functions, which can be helpful e.g. for mixing
		// the sound with your audio stream sample-by-sample.
		// toEnd() maybe used together with isAtEnd() to indicate that 
		// playback is stopped.
		VTINLINE void rewind(){ dataPointer = 0; } // Cursor to beginning of the sound
		VTINLINE void goTo(size_t pos){ dataPointer = pos; } // Move cursor to 'pos' sample
		VTINLINE void toEnd(){ dataPointer = dataSize; } // Cursor just after end of the sound 
		VTINLINE void next(); // Move to next sample
		VTINLINE bool isAtStart() const { return dataPointer==0; } // Is cursor at beginning?
		VTINLINE bool isAtEnd() const { return dataPointer>=dataSize; } // Is cursor after end of the sound?
		VTINLINE bool isLoaded() const { return bLoadSuccess; } // Is sound loaded successfully?
		VTINLINE bool IsEmpty() const { return !isLoaded() || dataSize==0 || data==NULL; }
		VTINLINE void SetLoaded(bool bValue){ bLoadSuccess=bValue; } // Force loaded flag
		VTINLINE T getCurrentSample(size_t channel); // Get sample under current postion. If cursor isAtEnd(), does rewind() first.
		VTINLINE const TAudioFrame<T, NCH>& getCurrentFrame(); // Get frame under current postion. If cursor isAtEnd(), does rewind() first.

		#if defined(VT_WINDOWS)
			// Load raw audio stream from Win32 .exe file resources
			bool LoadResource(UINT iRes, LPCTSTR lpType, HMODULE hModule=NULL);
			bool LoadResource(HRSRC hRes, HMODULE hModule);	
		#endif

		// Load/save data from/to raw audio file.
		// No conversions or check are done here.
		virtual bool LoadRaw(CFile *file, long size = -1); // If size is -1, then whole file is loaded
		virtual bool LoadRaw(LPCVTCHAR fileName); // Load raw audio stream (as is into memory)
		virtual bool SaveRaw(CFile *file, size_t *outSize = NULL); // Data size is saved in outSize (if it's not NULL)
		virtual bool SaveRaw(LPCVTCHAR fileName); // Save raw audio stream (as is from memory)

		void SplitChannelsTo(CInMemoryAudio<T, 1> *const outAudios[]) const; // Updates m_SampleRate
		void CombineChannelsFrom(CInMemoryAudio<T, 1> *const inAudios[]); // Does not update m_SampleRate
		virtual void ResampleFrom(const CInMemoryAudio<T, NCH>& srcAudio, CResampler<T>& resampler); // Updates m_SampleRate
		virtual void Resample( CResampler<T>& resampler );
		virtual void Resample( int newSampleRate, int stdMetaResamplerQuality = 0 );

		// Load data from PCM Wav file. Data is scaled to be within (-myRange..myRange).
		// 8, 16, 24, 32 bit is supported so far.
		// The file must have NCH channels.
		// m_SampleRate is also read and set.
		// WARNING: T must be a floating-point type, or scaling will be done improperly!
		virtual bool LoadPCMWav(CFile &fp, T myRange = T(1));	

		// Constructs CFILEFile and calls LoadPCMWav(CFile&, T).
		bool LoadPCMWav(FILE *fp, T myRange = T(1)); 
		
		// Constructs CDiskFile and calls LoadPCMWav(CFile&, T).
		bool LoadPCMWav(LPCVTCHAR fileName, T myRange = T(1)); 
		
		// Save data into 16-bit signed integer WAV.
		virtual bool SavePCMWav16SI(CFile &fp, T myRange = T(1), 
			TDitheringMethod method=kDitheringDefaultMethod) const;
		
		// Constructs CFILEFile and calls SavePCMWav16SI(CFile&, T).
		bool SavePCMWav16SI(FILE *fp, T myRange = T(1), 
			TDitheringMethod method=kDitheringDefaultMethod) const; 
		
		// Constructs CDiskFile and calls SavePCMWav16SI(CFile&, T).
		// Not 'const' because it may set szError.
		bool SavePCMWav16SI(LPCVTCHAR fileName, T myRange = T(1), 
			TDitheringMethod method=kDitheringDefaultMethod);

		// Delete samples before dataPointer by moving memory and updating dataSize and dataPointer.
		// bPadWithZeroes==true leaves old dataSize, padding the buffer with zeroes. 
		// bPadWithZeroes==false reduces dataSize to number of frames left.
		void RemovePastFrames(bool bPadWithZeroes);

		// Removes block of samples; length of the sound is reduced
		void DeleteFrames( size_t start, size_t nsamples );

		// Return last error message set by saving/loading function.
		// Returns LCT("No error") if no error occured.
		LPCVTCHAR GetLastErrorMessage() const { return szError; }

		VTINLINE const TAudioFrame<T, NCH> *GetData() const { return data; } // Low-level access to the buffer (read-only)
		VTINLINE TAudioFrame<T, NCH> *GetRawDataAccess() { return data; } // Low-level access to the buffer
		VTINLINE TAudioFrame<T, NCH> GetData(size_t idx) const { return data[idx]; } // Peek frame at position 'idx' 
		VTINLINE T GetData(size_t idx, size_t channel) const { return data[idx].channels[channel]; } // Peek sample at position 'idx'
		VTINLINE void SetData(size_t idx, const TAudioFrame<T, NCH>& sample){ data[idx]=sample; } // Set frame at position 'idx' 
		VTINLINE void SetData(size_t idx, size_t channel, T sample){ data[idx].channels[channel]=sample; } // Set sample at position 'idx', 'channel'

		// Compares data against data of another buffer.
		// Required for CArray<CInMemoryAudio>.
		//bool operator==(const CInMemoryAudio<T, NCH>& other) const;
	protected:
		// Load more information from RIFF Wav file.
		// This function is called from LoadPCMWav().
		// If this function returns 'false', loading is aborted.
		virtual bool OnLoadMoreInfo_Wav( VT::RIFF::CRIFFReader *reader ){ return true; } 
		
		// Called after sound is loaded, rebuilt or resampled.
		virtual void OnFinalize(){} 
		
		int m_SampleRate;	// Sample rate - used for information and by some resampling routines
		LPCVTCHAR szError;  // Pointer to the last error message 
		bool bLoadSuccess;  // Was the sound loaded/built successfully?
		size_t dataSize;    // Data size, in frames
		size_t allocSize;   // Allocated buffer size, in frames
		size_t dataPointer; // Data pointer, for playback and etc.
		TAudioFrame<T, NCH> *data; // Pointer to frame buffer
	};



	//////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////
	//
	// Implementation of the template class functions
	//

	template<class T, int NCH> 
		CInMemoryAudio<T, NCH>::CInMemoryAudio():
			dataSize(0),
			dataPointer(0),
			data(NULL),
			bLoadSuccess(false),
			m_SampleRate(44100),
			allocSize(0)
	{
		szError = LCT("No error.");
	}

	#if defined(VT_WINDOWS)
	template<class T, int NCH> 
		CInMemoryAudio<T, NCH>::CInMemoryAudio(UINT iRes, LPCTSTR lpType, HMODULE hModule):
			dataSize(0),
			dataPointer(0),
			data(NULL),
			bLoadSuccess(false),
			m_SampleRate(44100),
			allocSize(0)
	{ 
		szError = LCT("No error.");
		bLoadSuccess = LoadResource(iRes, lpType, hModule);

		#if defined(_DEBUG)
			if( !bLoadSuccess ){
				char buf[256];
				sprintf(buf, "Result: %d\r\nData Size: %u\r\nResource: %u", bLoadSuccess, dataSize, iRes);
				MessageBox(0, buf, __FUNCTION__, MB_OK);
			}
		#endif
	}
	#endif

	template<class T, int NCH> CInMemoryAudio<T, NCH>::~CInMemoryAudio()
	{
		if( data!= NULL){
			delete data;
			data = NULL;
		}
	}

	template<class T, int NCH> void CInMemoryAudio<T, NCH>::Clear()
	{
		dataSize = 0;
		dataPointer = 0;
		bLoadSuccess = false;
	}

	template<class T, int NCH> 
		void CInMemoryAudio<T, NCH>::Allocate(size_t nsamples, size_t allocate, bool bKeepData)
	{
		if( data==NULL || dataSize==0 || allocSize==0 )
			bKeepData = false;

		if( !bKeepData )
			Clear(); 
		else{
			// If keeping data, keep dataPointer at end at end :-)
			if( dataPointer>nsamples )
				dataPointer = nsamples;
		}
		
		// Change 'allocate' if neccessary
		if( allocate<kMinAllocSize )
			allocate=kMinAllocSize;
		if( allocate<allocSize )
			allocate=allocSize;
		if( nsamples>allocate ) // Increase allocation space to fit all data
			allocate = nsamples; // Now allocate>=nsamples
		
		// We already have enough memory allocated, so we can just update dataSize and return.
		if( allocSize>=allocate && data!=NULL ){
			// Oh, if we're in "keep data" mode, don't forget for this:
			if( bKeepData && nsamples>dataSize )
				memset( data+dataSize, 0, 
					sizeof(TAudioFrame<T, NCH>)*(nsamples-dataSize) );
			dataSize = nsamples;
			return;	
		}
		
		// *** Allocate new memory block ***

		TAudioFrame<T, NCH> *saveData;
		size_t saveDataSize;
		if( bKeepData ){
			saveData = data;
			saveDataSize = dataSize;
		}else{
			saveData = NULL;
			saveDataSize = 0;
			if( data!=NULL ){
				delete data;
				data = NULL;
			}
		}

		// Allocate
		dataSize = nsamples;
		allocSize = allocate;
		data = new TAudioFrame<T, NCH>[ allocSize ];
		if( data==NULL ){
			allocSize = 0;
			Clear();
			VTDEBUGMSGT("Out of memory while allocating raw audio storage.");
			throw LCT("Out of memory while allocating raw audio storage.");
		}

		// Restore data, if neccessary
		if( bKeepData && saveData!=NULL ){
			size_t olddatasize = Min( dataSize, saveDataSize );
			memcpy( data, saveData, sizeof(TAudioFrame<T, NCH>)*olddatasize );
			// When enlarging buffer with keeping all data, unfilled end 
			// of the new buffer must be zeroed.
			if( dataSize>saveDataSize ){
				// VTASSERT( olddatasize + (dataSize-saveDataSize) == dataSize );
				VTASSERT( olddatasize == saveDataSize );
				memset( data+saveDataSize, 0, sizeof(TAudioFrame<T, NCH>)*( dataSize-saveDataSize ) );
			}
			delete saveData;
		}
	}

	template<class T, int NCH> void CInMemoryAudio<T, NCH>::Zero()
	{
		if( data!=NULL && dataSize>0 )
			/*for(size_t ch=0; ch<NCH; ch++)
				for(size_t i=0; i<dataSize; i++)
					data[i].channels[ch] = T(0);*/
			memset(data, 0, dataSize * sizeof(TAudioFrame<T, NCH>));
	}

	template<class T, int NCH> void CInMemoryAudio<T, NCH>::copyFrom( const CInMemoryAudio<T, NCH>& source )
	{
		Allocate( source.dataSize );
		memcpy( data, source.data, dataSize*sizeof(TAudioFrame<T, NCH>) );
		this->m_SampleRate = source.m_SampleRate;
		bLoadSuccess = true;
	}

	template<class T, int NCH> template<class U>
		void CInMemoryAudio<T, NCH>::ConvertCopyFrom( const CInMemoryAudio<U, NCH>& source )
	{
		Allocate( source.dataSize );
		for( size_t i=0; i<source.getLength(); i++ )
			for( size_t ch=0; ch<NCH; ch++)
				this->data[i].channels[ch] = source->GetData(i, ch);
		this->m_SampleRate = source.m_SampleRate;
		bLoadSuccess = true;	
	}

	template<class T, int NCH> void CInMemoryAudio<T, NCH>::Multiply( T k )
	{
		if( data!=NULL )
			for(size_t ch=0; ch<NCH; ch++)
				for(size_t i=0; i<dataSize; i++)
					data[i].channels[ch] *= k;
	}

	template<class T, int NCH> void CInMemoryAudio<T, NCH>::Limit( T lim )
	{
		if( data!=NULL ){
			T nlim = -lim;
			for(size_t ch=0; ch<NCH; ch++)
				for(size_t i=0; i<dataSize; i++)
					if( data[i].channels[ch] < nlim )
						data[i].channels[ch] = nlim;
					else
						if( data[i].channels[ch] > lim )
							data[i].channels[ch] = lim;
		}
	}

	template<class T, int NCH> void CInMemoryAudio<T, NCH>::next()
	{ 
		if(dataPointer>=dataSize) 
			dataPointer=0;
		else
			dataPointer++;  
	}

	template<class T, int NCH> T CInMemoryAudio<T, NCH>::getCurrentSample(size_t channel)
	{ 
		if( isAtEnd() ) 
			rewind();
		return getSample(dataPointer, channel); 
	}

	template<class T, int NCH> const TAudioFrame<T, NCH>& CInMemoryAudio<T, NCH>::getCurrentFrame()
	{
		if( isAtEnd() ) 
			rewind();
		return getConstFrame(dataPointer); 
	}


	#if defined(VT_WINDOWS)

	template<class T, int NCH> 
		bool CInMemoryAudio<T, NCH>::LoadResource(UINT iRes, LPCTSTR lpType, HMODULE hModule)
	{
		rewind();
		LPCTSTR lpszResourceName = MAKEINTRESOURCE (iRes);
		if( lpszResourceName == NULL ){
			#if defined(_DEBUG)
				VTDEBUGMSGT("lpszResourceName == NULL");
			#endif
			return false;
		}
		HRSRC hRsrc = ::FindResource(hModule, lpszResourceName, lpType);
		return this->LoadResource(hRsrc, hModule);
	}

	template<class T, int NCH> 
		bool CInMemoryAudio<T, NCH>::LoadResource(HRSRC hRes, HMODULE hModule)
	{
		Clear();
		szError = LCT("No error.");
		DWORD rsize = SizeofResource(hModule, hRes);
		if( rsize==0 ){
			szError = LCT("Zero resource size.");
			VTDEBUGMSG( szError );
			return false;
		}
		if( (rsize % sizeof(TAudioFrame<T, NCH>)) !=0 ){
			szError = LCT("Size of the raw audio resource is not a multiplier "
				"of the size of the audio frame size.");
			VTDEBUGMSG( szError );
			return false;
		}
		HGLOBAL hMem = ::LoadResource(hModule, hRes);
		if( hMem ){		
			void* resPtr = LockResource(hMem);
			if( resPtr ){
				size_t newDataSize = size_t( rsize / sizeof(TAudioFrame<T, NCH>) );
				Allocate( newDataSize );
				memcpy( data, resPtr, newDataSize * sizeof(TAudioFrame<T, NCH>) );			
			}
			FreeResource(hMem);
			bLoadSuccess = true;
			OnFinalize();
			return true;
		}else{
			#if defined(_DEBUG)
				VTDEBUGMSGT("Failed to ::LoadResource().");
			#endif
			return false;
		}
	}

	#endif /*defined(VT_WINDOWS) */


		
	template<class T, int NCH> bool CInMemoryAudio<T, NCH>::LoadRaw(CFile *file, long size )
	{
		szError = LCT("No error.");
		if( !file->IsOK() ){
			szError = LCT("Raw audio file is not open for reading.");
			return false;
		}
		size_t fileSzToRead = 0;
		if( size>0 ){
			fileSzToRead = size; // Not verifying anything more here...
		}else{
			VTFSIZE longsz = file->getSize() - file->tell();
			if( longsz<=0 ){
				szError = LCT("Error determinging size of a raw audio file.");
				return false;
			}
			fileSzToRead = size_t( longsz );	
			// Check that file size is sane
			if( VTFSIZE(fileSzToRead)!=longsz ){
				szError = LCT("Raw audio file is too big.");
				return false;
			}
			if( (fileSzToRead % sizeof(TAudioFrame<T, NCH>)) !=0 ){
				szError = LCT("Size of raw audio file is not a multiplier of the audio frame size.");
				VTDEBUGMSG( szError );
				return false;
			}
		}	
		size_t szInSamples = fileSzToRead/sizeof(TAudioFrame<T, NCH>); // Size in frames
		Allocate( szInSamples ); 
		if( !file->read( data, szInSamples*sizeof(TAudioFrame<T, NCH>) ) ){
			szError = LCT("Raw audio file reading error.");
			return false;
		}
		OnFinalize(); // Finalization event
		bLoadSuccess = true;
		return true;
	}

	template<class T, int NCH> bool CInMemoryAudio<T, NCH>::LoadRaw(LPCVTCHAR fileName)
	{
		/*
		szError = LCT("No error.");
		rewind();
		FILE *f = VTFOPEN(fileName, _T("rb"));
		if( f==NULL ) 
			return false;
		VTFSEEK(f, 0, SEEK_END);
		size_t fsz = size_t( ftell(f) );
		size_t sz = fsz/sizeof(TAudioFrame<T, NCH>);
		VTFSEEK(f, 0, SEEK_SET);
		Allocate( sz, sz+1 );
		bool bRet = true;
		if( fread( data, fsz, 1, f )!=1 )
			bRet = false;
		fclose(f);
		OnFinalize();
		return bRet;
		*/

		CDiskFile file( fileName, _T("rb") );
		return LoadRaw( &file );
	}

	template<class T, int NCH> bool CInMemoryAudio<T, NCH>::SaveRaw(CFile *file, size_t *outSize )
	{
		szError = LCT("No error.");
		if( !file->IsOK() ){
			szError = LCT("Raw audio file is not open for writing.");
			return false;
		}
		size_t sz = dataSize*sizeof(TAudioFrame<T, NCH>);
		if( !file->write( data, sz ) ){
			szError = LCT("Raw audio file writing error.");
			return false;
		}
		if( outSize!=NULL )
			*outSize = sz;		
		return true;
	}

	template<class T, int NCH> bool CInMemoryAudio<T, NCH>::SaveRaw(LPCVTCHAR fileName)
	{
		/*
		szError = LCT("No error.");
		FILE *f = VTFOPEN(fileName, _T("wb"));
		if( f==NULL ) 
			return false; // Failed to open file
		bool bRet = true;
		if( fwrite( data, dataSize*sizeof(TAudioFrame<T, NCH>), 1, f )!=1 )
			bRet = false;
		fclose(f);
		return bRet;
		*/
		CDiskFile file( fileName, _T("wb") );
		return SaveRaw( &file );
	}

	template<class T, int NCH> 
		void CInMemoryAudio<T, NCH>::SplitChannelsTo(CInMemoryAudio<T, 1> *const outAudios[]) const
	{
		for(size_t ch=0; ch<NCH; ch++){
			outAudios[ch]->Allocate( dataSize );
			VTASSERT( outAudios[ch]->GetData()!=NULL && outAudios[ch]->getLength()==dataSize );		
			for(size_t s=0; s<dataSize; s++){
				outAudios[ch]->SetData(s, 0, this->data[s].channels[ch]);
			}
			outAudios[ch]->ForceSampleRate( GetSampleRate() );
			outAudios[ch]->SetLoaded( true );
		}
	}

	template<class T, int NCH> 
		void CInMemoryAudio<T, NCH>::CombineChannelsFrom(CInMemoryAudio<T, 1> *const inAudios[])
	{
		rewind();
		bLoadSuccess = false;
		size_t sz = inAudios[0]->getLength();
		for(size_t ch=1; ch<NCH; ch++)
			if( sz>inAudios[ch]->getLength() )
				sz = inAudios[ch]->getLength();
		Allocate( sz );			
		for(size_t s=0; s<dataSize; s++)
			for(size_t ch=0; ch<NCH; ch++)
				this->data[s].channels[ch] = inAudios[ch]->GetData(s, 0);
		bLoadSuccess = true;
		OnFinalize();
	}

	template<class T, int NCH> 
		void CInMemoryAudio<T, NCH>::ResampleFrom(const CInMemoryAudio<T, NCH>& srcAudio, CResampler<T>& resampler)
	{
		// *** NOTE ***
		// This function should work OK if srcAudio == *this!
		rewind();
		bLoadSuccess = false;
		m_SampleRate = URound(srcAudio.m_SampleRate * resampler.getRatio());
		if( NCH==1 && (srcAudio.data!=this->data) ){
			// One-channel resampling from one memory buffer to another. 
			// Everything is simple when working in-memory!
			resampler.restart();
			Allocate( resampler.outBufferSize( srcAudio.dataSize ) );
			dataSize = resampler.process( (T*)srcAudio.data, srcAudio.dataSize, (T*)this->data, true );
		}else{
			// Resampling using temporary one-channel buffers.
			// Split to channels, resample each channel, recombine channels
			CInMemoryAudio<T, 1> *splitchannels[ NCH ], *resampledChannels[ NCH ];
			// Allocate temporary buffers
			for(size_t ch=0; ch<NCH; ch++){
				splitchannels[ ch ] = new CInMemoryAudio<T, 1>;
				resampledChannels[ ch ] = new CInMemoryAudio<T, 1>;
			}
			// Copy source data into temporary buffers and split channels.
			srcAudio.SplitChannelsTo( splitchannels );
			// Resample each temporary buffer info another temporary buffer.
			for(int ch=0; ch<NCH; ch++){
				resampler.restart();
				resampledChannels[ch]->ResampleFrom( *(splitchannels[ch]), resampler );
			}
			// Recombine sound from output temporary buffers.
			this->CombineChannelsFrom( resampledChannels );
			// Free memory.
			for(size_t ch=0; ch<NCH; ch++){
				delete splitchannels[ ch ];
				delete resampledChannels[ ch ];
			}
		}
		bLoadSuccess = true;
		OnFinalize();
	}

	template<class T, int NCH> 
		void CInMemoryAudio<T, NCH>::Resample( CResampler<T>& resampler )
	{
		ResampleFrom( *this, resampler );
	}

	template<class T, int NCH> 
		void CInMemoryAudio<T, NCH>::Resample( int newSampleRate, int stdMetaResamplerQuality )
	{
		if( newSampleRate!=m_SampleRate ){
			CStdMetaResampler<T, T> resampler;
			resampler.init( double(newSampleRate)/double(m_SampleRate), stdMetaResamplerQuality );
			Resample( resampler );
			// Set new sample rate precisely (without rounding error 
			// probably added by Resample(  CResampler<T>& ).
			m_SampleRate = newSampleRate; 
		}
	}

	// Load data from PCM Wav file. Data is scaled to be within (-myRange..myRange).
	// Scale of the input data is determined from input WAV format (-32767..32767 for
	// 16-bit signed...)
	// BTW, 16-bit signed is the only supported format for now ;-)
	template<class T, int NCH> 
		bool CInMemoryAudio<T, NCH>::LoadPCMWav(CFile &fp, T myRange)
	{
		rewind();
		LPCVTCHAR noerror = LCT("No error.");
		szError = noerror;
		bLoadSuccess = false;

		VT::RIFF::CRIFFReader riffreader;
		riffreader.Open( &fp );
		if( !riffreader.IsOK() ){
			szError = LCT("RIFF format reading failed.");
			return false;
		}

		szError = LCT("Failed to load additional information.");
		if( !OnLoadMoreInfo_Wav( &riffreader ) )		
			return false;
		szError = noerror;

		RIFF::TFmt fmt; 
		DWORD dwDataSize=0;

		if( !riffreader.LoadFmt( &fmt ) ){
			szError = LCT("RIFF format chunk not found.");
			return false;
		}
		if( !fmt.IsValidPCMWav() ){
			szError = LCT("Not a PCM Wav.");
			return false;
		}
		if( fmt.wChannels!=NCH ){
			szError = LCT("Number of channels mismatch while loading a PCM Wav file.");
			return false;
		}
		
		if( fmt.wFormatTag==1 ){
			if(	fmt.wBitsPerSample!=16 && fmt.wBitsPerSample!=24 && 
				fmt.wBitsPerSample!=32 && fmt.wBitsPerSample!=8 ){
				szError = LCT("Unsupported PCM Wav bitrate for format 1.");
				return false;
			}
		}else if( fmt.wFormatTag==3 ){
			if( fmt.wBitsPerSample!=32 ){ // 32-bit IEEE float?
				szError = LCT("Unsupported PCM Wav bitrate for format 3.");
				return false;
			}
		}else{
			szError = LCT("Unsupported PCM Wav audio data format.");
			return false;
		}

		if( !riffreader.SeekToData('d', 'a', 't', 'a', &dwDataSize) ){
			szError = LCT("PCM data not found.");
			return false;
		}

		size_t nInputSamples = dwDataSize / (fmt.PCMBytesPerSample()*NCH);
		size_t nReadMonoSamples = nInputSamples * NCH;

		// Allocating temp buffer and reading WAV data into memory
		void *pcmdata = new char[dwDataSize];	
		if( !pcmdata ){
			szError = LCT("Out of memory.");
			return false;
		}
		if( !riffreader.Read(pcmdata, dwDataSize ) ){
			szError = LCT("Failed to read PCM data into memory.");
			delete pcmdata;
			return false;
		}

		// Allocating my own memory
		this->Allocate( nInputSamples );

		m_SampleRate = fmt.dwSamplesPerSec;
		
		// Converting samples
		if( fmt.wBitsPerSample==16 && fmt.wFormatTag==1 ){
			TSample16S *ptr = (TSample16S*)pcmdata;
			// Order of the cycles is determined by ordering of samples in PCM Wav
			// (left[0], right[0], left[1], right[1]...)
			T scale = myRange / T(32768);
			for(size_t i=0; i<nInputSamples; i++)
				for(size_t ch=0; ch<NCH; ch++)
					data[i].channels[ch] = T(*(ptr++)) * scale;
			bLoadSuccess = true;

		}else if( fmt.wBitsPerSample==32 && fmt.wFormatTag==1 ){
			SDWORD *ptr = (SDWORD*)pcmdata;
			T scale = myRange / T(2147483648.0); 
			for(size_t i=0; i<nInputSamples; i++)
				for(size_t ch=0; ch<NCH; ch++)
					data[i].channels[ch] = T(*(ptr++)) * scale;
			bLoadSuccess = true;

		}else if( fmt.wBitsPerSample==24 && fmt.wFormatTag==1 ){
			unsigned char *ptr = (unsigned char *)pcmdata;
			T scale = myRange / T(2147483648.0); 
			for(size_t i=0; i<nInputSamples; i++)
				for(size_t ch=0; ch<NCH; ch++){
					data[i].channels[ch] = T(ConvertSample24to32(ptr)) * scale;
					ptr+=3;
				}
			bLoadSuccess = true;

		}else if( fmt.wBitsPerSample==8 && fmt.wFormatTag==1 ){
			unsigned char *ptr = (unsigned char *)pcmdata;
			T scale = myRange / T(128.0); 
			for(size_t i=0; i<nInputSamples; i++)
				for(size_t ch=0; ch<NCH; ch++)
					data[i].channels[ch] = T(ConvertSample8Uto8S(*(ptr++))) * scale;
			bLoadSuccess = true;

		}else if( fmt.wBitsPerSample==32 && fmt.wFormatTag==3 ){ // 32-bit IEEE float
			float *ptr = (float *)pcmdata;
			for(size_t i=0; i<nInputSamples; i++)
				for(size_t ch=0; ch<NCH; ch++)
					data[i].channels[ch] = *(ptr++);
			bLoadSuccess = true;

		}else{
			szError = LCT("Failed to find a converter for PCM data.");
			bLoadSuccess = false;
		}	

		delete pcmdata;	
		OnFinalize();
		return true;
	}

	template<class T, int NCH> 
		bool CInMemoryAudio<T, NCH>::LoadPCMWav(FILE *fp, T myRange)
	{
		CFILEFile f(fp);
		return LoadPCMWav(f, myRange);
	}

	template<class T, int NCH> 
		bool CInMemoryAudio<T, NCH>::LoadPCMWav(LPCVTCHAR fileName, T myRange)
	{
		rewind();
		szError = LCT("No error.");
		bLoadSuccess = false;
		FILE *fp = VTFOPEN( fileName, _T("rb") );
		if( !fp ){
			szError = LCT("Could not open file for reading.");
			return false;
		}	
		// LoadPCMWav() sets this->bLoadSuccess, so we don't have to save the 
		// returned result in a local variable.
		this->LoadPCMWav( fp, myRange ); 
		// LoadPCMWav( fp, myRange ) sets szError on its own.
		// if( !bLoadSuccess )
		//	szError = LCT("Error while reading file data.");
		fclose( fp );
		return bLoadSuccess;
	}

	template<class T, int NCH> 
		bool CInMemoryAudio<T, NCH>::SavePCMWav16SI(CFile &fp, T myRange, TDitheringMethod method) const
	{
		// Create & write WAV header
		TRIFFWavHeader riff;
		// Initialize for standard CD-Audio format, i.e. Signed PCM 16-bit.
		// This saves us from setting some header fields, like format tag.
		riff.Init(); 
		riff.Channels = NCH;
		riff.SamplesPerSec = (m_SampleRate)?m_SampleRate:44100;
		riff.AvgBytesPerSec = riff.SamplesPerSec*NCH*2; // 2 bytes per sample
		riff.datalen = DWORD(dataSize*NCH*2); // 2 bytes per sample
		riff.WAVElen = riff.datalen+sizeof(riff);
		if( fp.write(&riff, sizeof(riff), 1)!=1 )
			return false;
		// Convert data into integer vector, using dithering algorithm
		CDitherer<T, TSample16S, -32767, 32767> ditherer;
		ditherer.SetMethod( method );
		CVector<TSample16S> out_data;
		out_data.resize(NCH*dataSize);	
		T scale = 32767.0f/myRange;
		for(size_t ch=0; ch<NCH; ch++){ // Interlacing channels
			ditherer.Restart();
			for(size_t sample=0, outidx=ch; sample<dataSize; sample++, outidx+=NCH)
				out_data.set(outidx, ditherer.Dither(data[sample].channels[ch]*scale));
		}
		
		// printf("dataSize=%u Channels=%u\n", dataSize, NCH);

		// Flush the happily converted PCM data to the file
		if( fp.write(out_data.getDataPtr(), sizeof(TSample16S), out_data.getSize())!=out_data.getSize() )
			return false;
		return true;
	}

	template<class T, int NCH> 
		bool CInMemoryAudio<T, NCH>::SavePCMWav16SI(FILE *fp, T myRange, TDitheringMethod method) const
	{
		CFILEFile f( fp );
		return SavePCMWav16SI( f, myRange, method );
	}


	template<class T, int NCH> 
		bool CInMemoryAudio<T, NCH>::SavePCMWav16SI(LPCVTCHAR fileName, T myRange, TDitheringMethod method) 
	{
		FILE *fp = VTFOPEN( fileName, _T("wb") );
		if( !fp ){
			szError = LCT("Could not open file for writing.");
			return false;
		}
		bool bRet = this->SavePCMWav16SI( fp, myRange, method );
		if( !bRet )
			szError = LCT("Error while writing file data.");
		fclose(fp);
		return bRet;
	}

	/*
	template<class T, int NCH> bool CInMemoryAudio<T, NCH>::operator==(const CInMemoryAudio<T, NCH>& other) const
	{
		if( data==NULL )
			return other.data==NULL;
		return memcmp( data, other.data, sizeof(TAudioFrame<T, NCH>)*dataSize )==0;	
	}*/

	// Delete samples before dataPointer by moving memory and updating dataSize and dataPointer.
	template<class T, int NCH> void CInMemoryAudio<T, NCH>::RemovePastFrames(bool bPadWithZeroes)
	{
		if( dataPointer>0 && dataSize>0 && data!=NULL ){ // (or don't have to do anything)
			if( dataPointer>=dataSize ){ // isAtEnd() => remove all
				// Remove all frames
				if( !bPadWithZeroes)
					dataSize = 0;
				else
					// Fill all with zeroes, leave dataSize as is.
					memset( data, 0, sizeof(TAudioFrame<T, NCH>)*dataSize );
			}else{
				// Remove some samples
				// Fix data pointer - will be used for calcs
				if( dataPointer>dataSize )
					dataPointer = dataSize; // >0
				size_t keptDataSize = dataSize-dataPointer; // dataPointer<dataSize here
				memmove( data, data+dataPointer, sizeof(TAudioFrame<T, NCH>)*keptDataSize );
				if( !bPadWithZeroes)
					dataSize = keptDataSize;
				else
					memset( data+keptDataSize, 0, sizeof(TAudioFrame<T, NCH>)*(dataSize-keptDataSize) );
			}
			dataPointer = 0;
		}
	}

	template<class T, int NCH> void CInMemoryAudio<T, NCH>::DeleteFrames( size_t start, size_t nsamples )
	{
		if( dataSize>0 && data!=NULL && start+1<dataSize && nsamples>0 ){ 
			// If the block to remove goes beyond data, then reduce the length to remove...
			if( start+nsamples > dataSize ){
				nsamples = dataSize - start;
				if( nsamples==0 )
					return;
			}
			// Move data after the removed block over it
			size_t tomove = dataSize - (start+nsamples-1);
			if( tomove )
				memmove( (void*)(data+start), (void*)(data+start+nsamples), tomove*sizeof(TAudioFrame<T, NCH>) );
			// Reduce data size
			dataSize -= nsamples;
		}
	}
} // namespace

#endif



