//-----------------------------------------------------------------------------
//
// Sound resampling library with a number of different resamplers
//
//-----------------------------------------------------------------------------

/*
	VT Framework License (BSL 1.0)

	Copyright (C) 2006 Sergey A. Galin 
	E-mail: sergey.galin@gmail.com

	Permission is hereby granted, free of charge, to any person or organization
	obtaining a copy of the software and accompanying documentation covered by
	this license (the "Software") to use, reproduce, display, distribute,
	execute, and transmit the Software, and to prepare derivative works of the
	Software, and to permit third-parties to whom the Software is furnished to
	do so, all subject to the following:

	The copyright notices in the Software and this entire statement, including
	the above license grant, this restriction and the following disclaimer,
	must be included in all copies of the Software, in whole or in part, and
	all derivative works of the Software, unless such copies or derivative
	works are solely in the form of machine-executable object code generated by
	a source language processor.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
	SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
	FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
	ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
	DEALINGS IN THE SOFTWARE.
*/

#ifndef VT_RESAMPLING_HPP
#define VT_RESAMPLING_HPP

#include "vt_audiomath.hpp"
#include "vt_buffers.hpp"
#include "vt_firs.hpp"
#include "vt_localization.hpp"

#if defined(USE_LSGUI) && !defined(USE_LSSTRING)
	#define USE_LSSTRING
#endif

#if defined(USE_LSSTRING)
	#include "LSString.h"
#endif

#if defined(DFEBUG) || defined(VT_CRTDEBUG)
	#include "vt_messageboxes.hpp" 
#else
	#if !defined(VTASSERT)
		#define VTASSERT(x)	{}
	#endif
#endif

namespace VT {

	//
	// Basic resampler interface.
	//
	// The workflow is:
	//
	// 1. Create & init() the resampler.
	// 2. Create output buffer. Use outBufferSize() to determine which minimal 
	//    memory size should be  allocated.
	// 3. Give process() chunks of data to be processed, and receive returned 
	//    number of output samples from outSamples. Set bStreamEnd to true when 
	//    processing the last chunk of data.
	//    You can check getMinMiddleInputSize() to make sure you're feeding it 
	//    with enough samples inside of the stream; using smaller blocks may 
	//    cause degradation of quality. If resampler works OK with any size 
	//    of input buffer, the function returns 0.
	// A resampler can be init() again or reset(), if you want to use it to 
	// resample another stream.
	//
	// NOTE 1: for best results, pass (input sample rate) samples per process(), this for sure
	// eliminates any problems at block edges which might appear with some simple imlementations.
	// E.g. for 44100->something, pass 44100 samples per block.
	//
	template<class T> class CResampler
	{
	public:
		CResampler(int quality=0): iQuality(quality), fRatio(1.0){}

		// Get name of the algorithm
		virtual void getName(LPVTCHAR out, int maxLen) const =0;

		// Restart resampling process.
		// Can be called at middle of the process, if it is terminated halfway.
		// This function is also always called by init().
		virtual void restart()=0;

		// Returns quality setting (0 - fastest quality)
		virtual int getBestQuality() const { return 0; }

		// Returns good quality setting 
		virtual int getGoodQuality() const { return this->getBestQuality(); }

		// Get description of the quality setting.
		virtual void getQualityDescription(int quality, LPVTCHAR out, int maxLen) const;
		virtual void getQualityDescription(LPVTCHAR out, int maxLen) const; // For current quality

		// Get description of the resampler, including the quality setting.
		virtual void getFullDescription(int quality, LPVTCHAR out, int maxLen) const;
		virtual void getFullDescription(LPVTCHAR out, int maxLen) const; // For current quality

		// Reconfigure the resampler for new sample rates & quality.	
		virtual void init(double ratio, int quality=0); // Redefine this function all in descendant classes
		virtual void init2(long sourceRate, long destRate, int quality=0);

		// Get minimum input block size in middle of the stream (the last block 
		// may be smaller, of course). If resampler can work correctly with any
		// size of input buffer, the function returns 0.
		virtual size_t getMinMiddleInputSize() const;

		// Get minumum output buffer size, in samples, for the current setup
		// and given max input buffer size.
		// Call this function after Init(). This functions suggests size with 
		// enough extra storage, to avoid buffer overruns.
		virtual size_t outBufferSize(size_t inBufferSize) const;

		// Buffer processing. The function receives nInSamples from input buffer and writes
		// something to outSamples. The number of samples written to outSamples is returned.
		virtual size_t process(const T* inSamples, size_t nInSamples, T* outSamples, bool bStreamEnd)=0;

		virtual double getRatio() const { return fRatio; }
		virtual double getInvRatio() const { return fInvRatio; }
		virtual int getQuality() const { return iQuality; }

		// Some handy functions for apps built with LString.
		// MSC is a bit stupid about overloading here, so all functions
		// have LS suffix.
		#if defined(USE_LSSTRING)
			LString getNameLS() const;
			LString getQualityDescriptionLS(int q) const;
			LString getQualityDescriptionLS() const;
			LString getFullDescriptionLS(int q) const;
			LString getFullDescriptionLS() const;
		#endif

	protected:
		int iQuality;
		double fRatio; // Output sample rate / Input sample rate
		double fInvRatio; // 1/fRatio
	};


	// CMetaResampler is a class which does not any resampling by itself,
	// but hosts sub-objects of various CResampler<T> descendants which
	// do the job using their algorithms.
	template<class T> class CMetaResampler: public CResampler<T>
	{
	public:
		// Override the constructor to create the resamplers
		CMetaResampler<T>(): pResamplers(NULL), numResamplers(0), iResampler(0){}
		~CMetaResampler<T>();
		virtual void getName(LPVTCHAR out, int maxLen) const;
		virtual void restart();
		virtual int getBestQuality() const;
		virtual void getQualityDescription(int quality, LPVTCHAR out, int maxLen) const;
		virtual void getFullDescription(int quality, LPVTCHAR out, int maxLen) const;
		virtual void init(double ratio, int quality=0); 
		virtual size_t getMinMiddleInputSize() const;
		virtual size_t outBufferSize(size_t inBufferSize) const;	
		virtual size_t process(const T* inSamples, size_t nInSamples, T* outSamples, bool bStreamEnd);
	protected:
		CResampler<T> **pResamplers;
		int numResamplers, iResampler;

		// From given CMetaResampler<T> quality, determine to which resampler
		// and what its quality it corresponds.
		void getResamplerIdxFromQuality(int q, int *out_resampler, int *out_quality) const;
		
	};


	// Very simple resampler, which decimates stream or duplicates samples.
	// Ignores quality setting.
	template<class T> class CDecimateDuplicateResampler: public CResampler<T>
	{
	public:
		CDecimateDuplicateResampler<T>();
		virtual void getName(LPVTCHAR out, int maxLen) const;
		virtual void restart();
		virtual void getFullDescription(int quality, LPVTCHAR out, int maxLen) const;
		virtual size_t process(const T* inSamples, size_t nInSamples, T* outSamples, bool bStreamEnd);
	};

	// Averaging resampler. 
	// Ignores quality setting.
	template<class T> class CAveragingResampler: public CResampler<T>
	{
	public:
		CAveragingResampler<T>();
		virtual void getName(LPVTCHAR out, int maxLen) const;
		virtual void restart();
		virtual void init(double ratio, int quality=0); 
		virtual size_t getMinMiddleInputSize() const;
		virtual void getFullDescription(int quality, LPVTCHAR out, int maxLen) const;
		virtual size_t process(const T* inSamples, size_t nInSamples, T* outSamples, bool bStreamEnd);
	protected:
		T tPrevSample;
		double fNumProcessedSamples, fNumOutSamples, fOutSampleHalfSrcSpan;
		int iOutSampleWidth;
		CVector<T> vBuffer;
	};



	// Polyphase Sinc interpolation resampler.
	// FIRT defines precision of calculations, use the same type or bigger
	// than T. Typical example would be CSincResampler<float, double>, so
	// a stream of float samples is processed via double-precision calculations.
	//
	// Define to enable buffer fill checks (you'll never need that):
	#if defined(_DEBUG) || defined(VT_CRTDEBUG)
		#define CVTSINC_CHECKBUFFERFILL
	#endif

	//
	// Define to enable calculation of real possible number of phases.
	// This should fasten initialization, but the code is not debugged yet and
	// so far it generates noise. So do not define it.
	// #define CVTSINC_PHASEOPTIMIZATION

	//
	// Define this to have phase dithering. Works only when 
	// CVTSINC_PHASEOPTIMIZATION is NOT defined.
	// This does not provide any visible effects when many enough phases are used,
	// so it has been switched off.
	// #define CVTSINC_PHASEDITHERING

	// Add "insane" quality setting into the list
	#if defined(VT_CRTDEBUG)
		#if !defined(CVTSINC_INSANE)
			#define CVTSINC_INSANE
		#endif
	#endif

	template<class T, class FIRT> class CSincResampler: public CResampler<T>
	{
	public:
		#if defined(CVTSINC_INSANE)
			static const int kMaxQuality = 3;
		#else
			static const int kMaxQuality = 2;
		#endif
		static const int kGoodQuality = 1;
		static const int kFIRSize[ kMaxQuality+1 ];
		static const int kMaxPhases[ kMaxQuality+1 ];

		typedef double TStreamPos; // long double didn't prove to increase precision.

		FIRT m_LowPassK; // 1.0 by default (keep all frequencies). Set this before init.

		CSincResampler<T, FIRT>();
		virtual void getName(LPVTCHAR out, int maxLen) const;
		virtual void restart();
		virtual int getBestQuality() const { return kMaxQuality; }
		virtual int getGoodQuality() const { return kGoodQuality; }
		virtual void getQualityDescription(int quality, LPVTCHAR out, int maxLen) const;
		virtual void getFullDescription(int quality, LPVTCHAR out, int maxLen) const;
		virtual void init(double ratio, int quality=0);
		// When CVTSINC_PHASEOPTIMIZATION is enabled, this version should be managing
		// to use smaller number of filter phases.
		virtual void init2(long sourceRate, long destRate, int quality=0);
		virtual void initCopy(const CSincResampler<T, FIRT>& src);
		virtual size_t getMinMiddleInputSize() const;
		virtual size_t outBufferSize(size_t inBufferSize) const;
		virtual size_t process(const T* inSamples, size_t nInSamples, T* outSamples, bool bStreamEnd);
	protected:
		void sincFinishInit();
		CCircularBuffer<FIRT> buffer; 
		CVector<FIRT> fir;
		CVector<T> inputBuffer;
		#if defined(CVTSINC_CHECKBUFFERFILL)
			size_t iBufferFill;
		#endif
		TStreamPos fPos;
		size_t iPhases;
		size_t iFIRSize;
		TAudioStreamSize iTotalInSamples, iTotalOutSamples;
	};

	// Sizes of FIR filters for various quality settings.
	// Max. resampling ratio in real life is: 192000/8000 = 24, so let's
	// have something bigger than that even in the fastest case.
	template<class T, class FIRT> const int 
		CSincResampler<T, FIRT>::kFIRSize[ CSincResampler<T, FIRT>::kMaxQuality+1 ] = {
			33, 159, 415
			#if defined(CVTSINC_INSANE)
				, 2047
			#endif
		};

	// Max possible number of phases for various quality settings.
	// When CVTSINC_PHASEOPTIMIZATION is disabled, this is the number
	// of phases which is always created.
	// Since we don't dither phase selection and since it doesn't have
	// big impact on speed, it's better to have plenty of phases at hand.
	template<class T, class FIRT> const int 
		CSincResampler<T, FIRT>::kMaxPhases[ CSincResampler<T, FIRT>::kMaxQuality+1 ] = {
			1024, 2048, 4096
			#if defined(CVTSINC_INSANE)
				, 4096 
			#endif
		};

	// Standard typedefs for single and double precision resamplers.
	typedef CSincResampler<float, double> CVTSincResamplerSP;
	typedef CSincResampler<double, double> CVTSincResamplerDP;




	// Standard resampler collection.
	template<class T, class FIRT> class CStdMetaResampler: public CMetaResampler<T>
	{
	public:
		CStdMetaResampler<T, FIRT>(){
			this->pResamplers = new CResampler<T>*[ 2 ];
			this->numResamplers = 2;
			this->iResampler = 1;
			//this->pResamplers[ 0 ] = new CDecimateDuplicateResampler<T>;
			this->pResamplers[ 0 ] = new CAveragingResampler<T>;
			this->pResamplers[ 1 ] = new CSincResampler<T, FIRT>;
		}

	};

	// Standard resampler collections for single and double precision resamplers.
	typedef CStdMetaResampler<float, double> CVTStdMetaResamplerSP;
	typedef CStdMetaResampler<double, double> CVTStdMetaResamplerDP;











	//////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////
	//
	// Implementation of template & inline functions
	//

	//****************************************************************************
	//*** CResampler<T>
	//****************************************************************************

	template<class T> void CResampler<T>::getQualityDescription(int quality, LPVTCHAR out, int maxLen) const
	{
		VTSTRNCPY(out, LCT("Normal"), maxLen);
	}

	template<class T> void CResampler<T>::getQualityDescription(LPVTCHAR out, int maxLen) const
	{
		getQualityDescription(this->iQuality, out, maxLen);
	}

	template<class T> void CResampler<T>::getFullDescription(int quality, LPVTCHAR out, int maxLen) const
	{
		this->getName(out, maxLen);
		maxLen -= int(VTSTRLEN(out));
		if( maxLen<=3 )
			return;
		VTSTRNCAT(out, _T(": "), maxLen);
		maxLen -= 2;
		if( maxLen<=0 )
			return;
		this->getQualityDescription(quality, out, maxLen);
	}

	template<class T> void CResampler<T>::getFullDescription(LPVTCHAR out, int maxLen) const
	{
		getFullDescription(this->iQuality, out, maxLen);
	}

	// Reconfigure the resampler for new sample rates & quality.
	template<class T> void CResampler<T>::init2(long sourceRate, long destRate, int quality)
	{
		if( sourceRate<1 || destRate<1 ){
			VTDEBUGMSGT("Invalid sample rate.");
			throw LCT("Invalid sample rate.");
		}
		init(double(destRate) / double(sourceRate), quality);
	}

	template<class T> void CResampler<T>::init(double ratio, int quality)
	{
		if( ratio < 0.01 ){
			VTDEBUGMSGT("Resampling ratio is too small.");
			throw LCT("Resampling ratio is too small.");
		}
		this->iQuality = quality;
		this->fRatio = ratio;
		this->fInvRatio = 1.0/ratio;
		restart();
	}

	template<class T> size_t CResampler<T>::getMinMiddleInputSize() const
	{
		return 0;
	}

	template<class T> size_t CResampler<T>::outBufferSize(size_t inBufferSize) const
	{
		// Request some additional space to avoid buffer overrun.
		return size_t((double(inBufferSize) * this->fRatio) * 1.1 + 128.0);
	}

	#if defined(USE_LSSTRING)
	template<class T> LString CResampler<T>::getNameLS() const
	{
		VTCHAR buf[256];
		this->getName(buf, sizeof(buf)-1);
		return LString(buf);
	}
			
	template<class T> LString CResampler<T>::getQualityDescriptionLS(int q) const
	{
		VTCHAR buf[256];
		this->getQualityDescription(q, buf, sizeof(buf)-1);
		return LString(buf);
	}

	template<class T> LString CResampler<T>::getQualityDescriptionLS() const
	{
		VTCHAR buf[256];
		this->getQualityDescription(buf, sizeof(buf)-1);
		return LString(buf);
	}
			
	template<class T> LString CResampler<T>::getFullDescriptionLS(int q) const
	{
		VTCHAR buf[256];
		this->getFullDescription(q, buf, sizeof(buf)-1);
		return LString(buf);
	}

	template<class T> LString CResampler<T>::getFullDescriptionLS() const
	{
		VTCHAR buf[256];
		this->getFullDescription(buf, sizeof(buf)-1);
		return LString(buf);
	}

	#endif



	//****************************************************************************
	//*** CMetaResampler<T>
	//****************************************************************************

	template<class T> CMetaResampler<T>::~CMetaResampler()
	{
		for(int i=0; i<this->numResamplers; i++)
			delete this->pResamplers[i];
		delete this->pResamplers;
	}

	template<class T> void CMetaResampler<T>::getName(LPVTCHAR out, int maxLen) const
	{
		VTSTRNCPY(out, LCT("Resampler Collection"), maxLen);
	}

	template<class T> void CMetaResampler<T>::restart()
	{
		this->pResamplers[this->iResampler]->restart();
	}

	template<class T> int CMetaResampler<T>::getBestQuality() const
	{
		int q=0;
		for(int i=0; i<this->numResamplers; i++){
			q += this->pResamplers[i]->getBestQuality();
			q++;
		}
		return --q;
	}

	template<class T>
		void CMetaResampler<T>::getQualityDescription(int quality, LPVTCHAR out, int maxLen) const
	{	int res, q;
		this->getResamplerIdxFromQuality(quality, &res, &q);
		return this->pResamplers[res]->getQualityDescription(q, out, maxLen);
	}
		
	template<class T> void CMetaResampler<T>::getFullDescription(int quality, LPVTCHAR out, int maxLen) const
	{
		int res, q;
		this->getResamplerIdxFromQuality(quality, &res, &q);
		return this->pResamplers[res]->getFullDescription(q, out, maxLen);
	}

	template<class T> void CMetaResampler<T>::init(double ratio, int quality)
	{
		int res, q;
		this->getResamplerIdxFromQuality(quality, &res, &q);
		this->iResampler = res;
		this->pResamplers[res]->init(ratio, q); // Init the scapegoat
		this->CResampler<T>::init(ratio, quality); // Also init my own parent
	}

	template<class T> size_t CMetaResampler<T>::getMinMiddleInputSize() const
	{
		int res, q;
		this->getResamplerIdxFromQuality(this->iQuality, &res, &q);
		return this->pResamplers[res]->getMinMiddleInputSize();
	}

	template<class T> size_t CMetaResampler<T>::outBufferSize(size_t inBufferSize) const
	{
		int res, q;
		this->getResamplerIdxFromQuality(this->iQuality, &res, &q);
		return this->pResamplers[res]->outBufferSize(inBufferSize);
	}

	template<class T>
		size_t CMetaResampler<T>::process(const T* inSamples, size_t nInSamples, T* outSamples, bool bStreamEnd)
	{
		return this->pResamplers[this->iResampler]->process(inSamples, nInSamples, outSamples, bStreamEnd);
	}

	template<class T>
		void CMetaResampler<T>::getResamplerIdxFromQuality(int q, int *out_resampler, int *out_quality) const
	{
		int qx=0;
		for(int i=0; i<this->numResamplers; i++){
			int bq = this->pResamplers[i]->getBestQuality();
			if( qx+bq >= q){
				*out_resampler = i;
				*out_quality = q-qx;
				return;
			}
			qx += bq;
			qx++;
		}
		*out_resampler = 0;
		*out_quality = 0;
	}





	//****************************************************************************
	//*** CDecimateDuplicateResampler<T>
	//****************************************************************************

	template<class T> CDecimateDuplicateResampler<T>::CDecimateDuplicateResampler()
	{
	}

	template<class T> void CDecimateDuplicateResampler<T>::getName(LPVTCHAR out, int maxLen) const
	{
		VTSTRNCPY(out, LCT("Decimate/Duplicate"), maxLen);
	}

	template<class T> void CDecimateDuplicateResampler<T>::restart()
	{
	}

	template<class T> void CDecimateDuplicateResampler<T>::getFullDescription(int quality, LPVTCHAR out, int maxLen) const
	{
		getName(out, maxLen);
	}

	template<class T> size_t CDecimateDuplicateResampler<T>::process(const T* inSamples, size_t nInSamples, T* outSamples, bool bStreamEnd)
	{
		#if defined(MSCFASTMATH)
			#pragma float_control(push)
			#pragma float_control(fast, on)
		#endif
		size_t nOutSamples = size_t(URound(double(nInSamples)*this->fRatio));
		long nInSamplesM1 = long(nInSamples)-1;
		for(size_t outi = 0; outi<nOutSamples; outi++){		
			long idx = long((double(outi)/this->fRatio+0.499999)); // Calculate index of closest input sample		
			idx = Clip<long>(idx, 0, nInSamplesM1); // Fix it, if it's out of range somehow
			outSamples[ outi ] = inSamples[ idx ]; // And then just copy the sample
		}
		return nOutSamples;
		#if defined(MSCFASTMATH)
			#pragma float_control(pop)
		#endif
	}


	//****************************************************************************
	//*** CDecimateDuplicateResampler<T>
	//****************************************************************************

	template<class T> CAveragingResampler<T>::CAveragingResampler():
		tPrevSample(T(0)),
		fNumProcessedSamples(0.0),
		fNumOutSamples(0.0),
		fOutSampleHalfSrcSpan(1.0),
		iOutSampleWidth(1),
		vBuffer(64) // Should be enough for everyone - (C) Bill Gates  (but will be increased if ever needed)
	{
	}

	template<class T> void CAveragingResampler<T>::getName(LPVTCHAR out, int maxLen) const
	{
		VTSTRNCPY(out, LCT("Linear Interpolation/Averaging"), maxLen);
	}

	template<class T> void CAveragingResampler<T>::init(double ratio, int quality)
	{
		CResampler<T>::init(ratio, quality);
		fOutSampleHalfSrcSpan = this->fInvRatio / 2.0; 
		iOutSampleWidth = URound(ceil(fOutSampleHalfSrcSpan))+1;
		vBuffer.resize( iOutSampleWidth, false );
	}

	template<class T> void CAveragingResampler<T>::restart()
	{
		tPrevSample = T(0);
		fNumProcessedSamples = 0.0;
		fNumOutSamples = 0.0;
		vBuffer.zero();
	}

	template<class T> size_t CAveragingResampler<T>::getMinMiddleInputSize() const
	{
		return vBuffer.getSize();
	}

	template<class T> void CAveragingResampler<T>::getFullDescription(int quality, LPVTCHAR out, int maxLen) const
	{
		getName(out, maxLen);
	}

	#if defined(MSCFASTMATH)
		#pragma float_control(push)
		#pragma float_control(precise, off)
		#pragma float_control(except, off)
	#endif
	template<class T> size_t CAveragingResampler<T>::process(const T* inSamples, size_t nInSamples, T* outSamples, bool bStreamEnd)
	{
		size_t nOutSamples = 0;
		if( !nInSamples ) 
			return 0;
		long nInSamplesM1 = long(nInSamples-1);
		if( this->fRatio>1.0 ){ // Upsampling
			//printf("UP");
			double fPos = (this->fNumOutSamples * this->fInvRatio)-this->fNumProcessedSamples;
			double fOutSamples = floor(-fPos + double(nInSamples) * this->fRatio);
			register size_t outi = 0;
			if( fOutSamples >= 0.9999999999 ){
				nOutSamples = URound(fOutSamples);
				while( fPos<0.0 && outi<nOutSamples ){
					outSamples[ outi++ ] = T(inSamples[ 0 ]*(1.0+fPos) - tPrevSample*fPos);
					fPos += this->fInvRatio;
				}
				for(; outi<nOutSamples; outi++, fPos += this->fInvRatio){
					int ipos1 = Clip<long>(URound(floor(fPos)), 0, nInSamplesM1), // Left closest source sample
						ipos2 = Clip<long>(URound(ceil(fPos)), 0, nInSamplesM1); // Right closest source sample
					if( ipos1==ipos2 ) 
						outSamples[ outi ] = inSamples[ ipos1 ]; 
					else 
						outSamples[ outi ] = T(inSamples[ ipos1 ]*(double(ipos2)-fPos) + inSamples[ ipos2 ]*(fPos-double(ipos1)));
				}		
			} // else do nothing (no output samples)

		}else{ // Downsampling (this->fRatio < 1.0, output ratio < input ratio)
			//printf("DOWN");
			double fLeftPos = (this->fNumOutSamples * this->fInvRatio)-this->fNumProcessedSamples-0.5;
			double fLastPos = double(nInSamples-0.5); 
			VTASSERT(fLastPos>fLeftPos);
			double fOutSamples = ( fLastPos - fLeftPos ) * this->fRatio;		
			if( fOutSamples >= 0.9999999999 ){
				nOutSamples = URound(floor(fOutSamples));
				//printf("*** Left Pos: %f Last Pos: %f fOut: %f nOut: %d\n", float(fLeftPos), float(fLastPos), float(fOutSamples), int(nOutSamples));
				double fRightPos = fLeftPos + this->fInvRatio; //fOutSampleHalfSrcSpan;			
				for(size_t outi = 0; outi<nOutSamples; outi++, 
					fLeftPos+=this->fInvRatio, fRightPos+=this->fInvRatio){
					//printf("*** Left: %f Right: %f\n", float(fLeftPos), float(fRightPos));
					// Step trough samples inside fLeftPos to fRightPos, weighting and summing them.
					double sum = 0.0;
					long iLeftPos = SRound( ceil(fLeftPos) );
					long iRightPos = SRound( floor(fRightPos) );
					VTASSERT(iRightPos>=iLeftPos);
					for(long i=iLeftPos; i<=iRightPos; i++){
						double sample, weight;
						if( i<0 ){
							VTASSERT( -i < long(vBuffer.getSize()) ); 
							sample = vBuffer.get(-i);
						}else{
							VTASSERT( i < long(nInSamples) );
							sample = inSamples[ i ];
						}		
						if( iLeftPos==iRightPos )
							weight = fRightPos-fLeftPos;
						else if( i==iLeftPos )
							weight = double(iLeftPos)-fLeftPos + 0.5; 
						else if( i==iRightPos )
							weight = fRightPos-double(iRightPos) + 0.5; 
						else // i is between iLeftPos and iRightPos
							weight = 1.0; //this->fInvRatio;
						VTASSERT(weight>0.0);
						sum += sample * weight;
					}
					outSamples[ outi ] = T(sum * this->fRatio);
				}

			} // else do nothing (no output samples)
			// Copy last samples into buffer
			int cnt = 0;
			for(long i=long(nInSamples)-1; i>=0 && cnt<int(vBuffer.getSize()); cnt++, i--)
				vBuffer.set(cnt, inSamples[i]);
			for( ; cnt<int(vBuffer.getSize()); cnt++)
				vBuffer.set(cnt, 0);
		};
		
		
		this->tPrevSample = inSamples[ nInSamplesM1 ];
		this->fNumProcessedSamples += double(nInSamples);
		this->fNumOutSamples += double(nOutSamples);

		return nOutSamples;
	}
	#if defined(MSCFASTMATH)
		#pragma float_control(pop)
	#endif




	//****************************************************************************
	//*** CSincResampler<T>
	//****************************************************************************

	template<class T, class FIRT> CSincResampler<T, FIRT>::CSincResampler():
		#if defined(CVTSINC_CHECKBUFFERFILL)
			iBufferFill(0),
		#endif
		iPhases(0),
		fPos(0),
		inputBuffer(0, 32),
		buffer(kFIRSize[kMaxQuality]+2),
		fir( kFIRSize[kMaxQuality]*(kMaxPhases[kMaxQuality]+1) ),
		m_LowPassK(FIRT(1.0))
	{

	}

	template<class T, class FIRT> void CSincResampler<T, FIRT>::getName(LPVTCHAR out, int maxLen) const
	{
		VTSTRNCPY(out, LCT("Sinc Interpolation"), maxLen);
	}

	template<class T, class FIRT> void CSincResampler<T, FIRT>::getQualityDescription(int quality, LPVTCHAR out, int maxLen) const
	{
		switch( quality ){
			case 0: VTSTRNCPY(out, LCT("Fast"), maxLen); return;
			case 1: VTSTRNCPY(out, LCT("Normal"), maxLen); return;
			case 2: VTSTRNCPY(out, LCT("High Quality"), maxLen); return;
			#if defined(CVTSINC_INSANE)
				case 3: VTSTRNCPY(out, LCT("Insane"), maxLen); return;
			#endif
		}
		VTSTRNCPY(out, LCT("[Bad value]"), maxLen);
	}

	template<class T, class FIRT> void CSincResampler<T, FIRT>::getFullDescription(int quality, LPVTCHAR out, int maxLen) const
	{
		switch( quality ){
			case 0: VTSTRNCPY(out, LCT("Fast Sinc Interpolation"), maxLen); return;
			case 1: VTSTRNCPY(out, LCT("Normal Sinc Interpolation"), maxLen); return;
			case 2: VTSTRNCPY(out, LCT("High Quality Sinc Interpolation"), maxLen); return;
			#if defined(CVTSINC_INSANE)
				case 3: VTSTRNCPY(out, LCT("Insane Quality Sinc Interpolation"), maxLen); return;
			#endif
		}
		VTSTRNCPY(out, LCT("[Bad value]"), maxLen);
	}

	template<class T, class FIRT> void CSincResampler<T, FIRT>::restart()
	{
		#if defined(CVTSINC_CHECKBUFFERFILL)
			this->iBufferFill = 0;
		#endif
		this->iTotalInSamples = 0;
		this->iTotalOutSamples = 0;
		this->buffer.zero();
		this->buffer.reset();
		this->fPos = 0.0;
		this->inputBuffer.zero();
		this->inputBuffer.resize(0);	
	}

	template<class T, class FIRT> void CSincResampler<T, FIRT>::init(double ratio, int quality)
	{
		CResampler<T>::init(ratio, quality);
		if( quality<0 || quality>this->kMaxQuality ){
			VTDEBUGMSGT("Bad quality setting.");
			throw LCT("Bad quality setting.");	
		}

		#if defined(CVTSINC_PHASEOPTIMIZATION)
			// If ratio is an integer value, we're increasing sample rate
			// few times, so we need only zero phase shift
			if( ratio>1.0 && fmod( ratio, 1.0 )<0.00001 ){
				#if defined(VT_CRTDEBUG)
					printf("*** Integer upsample detected => one phase shift.\n");
				#endif
				iPhases = 1;
				sincFinishInit();
		
			// Downsampling stream with integer ratio
			}else if( ratio<1.0 && fmod( 1.0/ratio, 1.0 )<0.00001 ){
				#if defined(VT_CRTDEBUG)
					printf("*** Integer downsample detected => one phase shift.\n");
				#endif
				iPhases = 1;
				sincFinishInit();
		
			}else{
				#if defined(VT_CRTDEBUG)
					printf("*** Non-integer resampling detected => max phase shifts.\n");
				#endif
				iPhases = kMaxPhases[ this->iQuality ];
				sincFinishInit();
			}
		#else
			iPhases = kMaxPhases[ this->iQuality ];
			sincFinishInit();
		#endif
	}

	template<class T, class FIRT>
		void CSincResampler<T, FIRT>::init2(long sourceRate, long destRate, int quality)
	{
		if( quality<0 || quality>this->kMaxQuality ){
			VTDEBUGMSGT("Bad quality setting.");
			throw LCT("Bad quality setting.");
		}
		if( sourceRate<1 || destRate<1 ){
			VTDEBUGMSGT("Invalid sample rate.");
			throw LCT("Invalid sample rate.");
		}
		// Cannot use inherited init2(), because in the base class it will 
		// then call our own virtual init().
		CResampler<T>::init(double(destRate)/double(sourceRate), quality);
		iPhases = kMaxPhases[ this->iQuality ];

		#if defined(CVTSINC_PHASEOPTIMIZATION)
			SDWORD lcm, gcd;
			if( LCM_and_GCD( sourceRate, destRate, &lcm, &gcd ) )
				if( lcm>=1 && gcd>=1 ){
					#if defined(VT_CRTDEBUG)
						printf("*** %ld=>%ld, GCD=%d, LCM=%d\n",
							sourceRate, destRate, int(gcd), int(lcm) );
					#endif
					if( gcd==destRate ){ // So sourceRate is a divisor of destRate
						iPhases = 1;
						#if defined(VT_CRTDEBUG)
							printf("*** Source rate is a divisor of destRate, so one phase.\n");
						#endif				
					}else{
						iPhases = (lcm/sourceRate) * (lcm/destRate);
						#if defined(VT_CRTDEBUG)
							printf("*** Number of phases possible: %d.\n", phases);
						#endif								
						if( iPhases>this->kMaxPhases[ this->iQuality ] ){
							iPhases = this->kMaxPhases[ this->iQuality ];
							#if defined(VT_CRTDEBUG)
								printf("*** Too many phases, limiting to the max allowed: %d.\n",
									phases);
							#endif									
						}
					}
					this->sincFinishInit();
					return;
				}
		
			#if defined(VT_CRTDEBUG)
				printf("*** Failed to find LCT of the rates, using max phases.\n");
			#endif
			this->sincFinishInit();
		#else
			this->sincFinishInit(); // iPhases is already set above
		#endif
	}

	#if defined(MSCFASTMATH)
		#pragma float_control(push)
		#pragma float_control(precise, off)
		#pragma float_control(except, off)
	#endif
	template<class T, class FIRT> void CSincResampler<T, FIRT>::sincFinishInit()
	{
		VTASSERT( iPhases>=1 && int(iPhases)<=this->kMaxPhases[ this->iQuality ] );
		VTASSERT( this->iQuality>=0 && this->iQuality<=this->kMaxQuality );
		iFIRSize = kFIRSize[ this->iQuality ];
		#if defined(CVTSINC_PHASEDITHERING)
			fir.resize( iFIRSize * (iPhases+1) ); 
		#else
			fir.resize( iFIRSize * iPhases ); 
		#endif
		CVector<FIRT> windowVector( iFIRSize );
		windowVector.fill(FIRT(1.0));
		// Hamming window works here better than Blackman, Kaiser even better.
		// HammingWindow( windowVector );	

		FIR::KaiserWindow<FIRT, long double>( windowVector, 5.6533, 1e-40l );
			
		FIRT *firDataPtr = fir.getDataPtr();
		FIRT fc;
		if( this->fRatio < 1.0 ) 
			fc = FIRT(0.5*m_LowPassK*this->fRatio); // Downsampling - throw away high frequencies
		else 
			fc = FIRT(0.5*m_LowPassK); // Upsampling - leave all frequencies	
		#if defined(VT_CRTDEBUG)
			printf("*** Initializing %d-sample sinc filters with %d phase offsets, %d values.\n",
				iFIRSize, iPhases, int(fir.getSize()));
			printf("*** Low-pass Frequency: %.3f (%.3f of input Nyquist)\n",
				float(fc), float(fc*2.0));
		#endif	
		#if defined(CVTSINC_PHASEDITHERING)
			for( int phasei=0; phasei<=int(iPhases); phasei++ ){
		#else
			for( int phasei=0; phasei<int(iPhases); phasei++ ){
		#endif
			FIRT phaseShift = FIRT(phasei) / FIRT(iPhases);
			FIR::LowPass( firDataPtr, iFIRSize, fc, phaseShift, windowVector.getDataPtr() );
			firDataPtr += iFIRSize;
		}
		buffer.resize( int(iFIRSize) );
		this->restart();
	}
	#if defined(MSCFASTMATH)
		#pragma float_control(pop)
	#endif

	template<class T, class FIRT> void CSincResampler<T, FIRT>::initCopy(const CSincResampler<T, FIRT>& src)
	{
		CResampler<T>::init(src.fRatio, src.iQuality);
		this->m_LowPassK = src.m_LowPassK;
		this->fir.copyFrom( src.fir );
		this->buffer.resize( int(src.buffer.getSize()) );
		this->iPhases = src.iPhases;
		this->iFIRSize = src.iFIRSize;
		this->restart();	
	}

	template<class T, class FIRT> size_t CSincResampler<T, FIRT>::getMinMiddleInputSize() const
	{
		return 0;
	}

	template<class T, class FIRT> size_t CSincResampler<T, FIRT>::outBufferSize(size_t inBufferSize) const
	{
		// Request some additional space to avoid buffer overrun.
		return size_t((double(inBufferSize+buffer.getSize()+64) * this->fRatio) * 1.1 + 128.0);
	}


	/*template<class T, class FIRT> size_t CSincResampler<T, FIRT>::outBufferSize(size_t inBufferSize) const
	{
		// Give plenty of extra room
		return (inBufferSize + this->buffer.getSize()) * URound( ceil(this->fRatio * 1.1) );
	}*/

	#if defined(MSCFASTMATH)
		#pragma float_control(push)
		#pragma float_control(precise, off)
		#pragma float_control(except, off)
	#endif

	template<class T, class FIRT> size_t
		CSincResampler<T, FIRT>::process(const T* inSamples, size_t nInSamples, 
		T* outSamples, bool bStreamEnd)
	{
		size_t retOutSamples = 0, bufferHalf = buffer.getSize()/2;
		#if defined(VT_CRTDEBUG) || defined(_DEBUG)
			size_t nInSamples_Save = nInSamples;
		#endif

		while( iTotalInSamples<=TAudioStreamSize(bufferHalf) && nInSamples ){	
			buffer.quickInsert( *(inSamples++) );
			nInSamples--;
			iTotalInSamples++;
			#if defined(CVTSINC_CHECKBUFFERFILL)
				iBufferFill++;
			#endif
		}

		if( !nInSamples && !bStreamEnd )
			return 0;

		if( iTotalInSamples>TAudioStreamSize(bufferHalf) || inputBuffer.getSize() ||
			bStreamEnd ){

			//double mustBeTotalOutSamples =
			// (double(iTotalInSamples+nInSamples)+0.5)*this->fRatio - double(iTotalOutSamples);
			// Caclulate total number of samples which should be in output stream 
			// after this run. This number is then rounded down (should take that
			// into account in calculations).
			double mustBeTotalOutSamples =
				(double(iTotalInSamples+nInSamples)+inputBuffer.getSize() )*
					this->fRatio -
				double(iTotalOutSamples);

			if( this->fRatio>=1.0 )
				mustBeTotalOutSamples += 0.5*this->fRatio;

			// If we are not finishing yet, keep half of buffer unused
			if( !bStreamEnd ) 
				mustBeTotalOutSamples -= (double(bufferHalf)+0.5)*this->fRatio;

			if( mustBeTotalOutSamples>=1.0 ){
				retOutSamples = size_t( URound(floor(mustBeTotalOutSamples)) );
		
				// printf("In: %d TotalIn: %ld Must be out: %f Now: %ld\n",
				// 	int(nInSamples), long(iTotalInSamples), float(mustBeTotalOutSamples),
				// 	long(retOutSamples));
				
				for(size_t outSample = 0; outSample<retOutSamples; outSample++){
					// FIR filtering
					{
						#if !defined(CVTSINC_PHASEDITHERING)
							// No dithering - way faster and produces good result with
							// many enough phases.
							SDWORD phaseIdx = URound( (fmod(fPos, TStreamPos(1.0))) * 
								double(iPhases-1));
							VTASSERT( phaseIdx<SDWORD(iPhases) );
						#else
							// Rectangular dithering
							SDWORD phaseIdx;
							double fPhaseIdx = fmod(fPos, TStreamPos(1.0)) * double(iPhases);
							if( fmod(fPhaseIdx, 1.0)>double(rand())/(double(RAND_MAX)*2.0) )
								phaseIdx = SDWORD( fPhaseIdx+0.5000001 ); 
							else
								phaseIdx = SDWORD( fPhaseIdx);
							if( phaseIdx>SDWORD(iPhases) ){
								#if defined(VT_CRTDEBUG)
									printf("HOW HOW HOW!!!\n");
								#endif
								phaseIdx = SDWORD(iPhases);
							}
							VTASSERT( phaseIdx<=SDWORD(iPhases) );
						#endif
						VTASSERT( phaseIdx>=0 );

						#if 0
							register int firi = int(iFIRSize*phaseIdx);
							register int firend = int(firi+iFIRSize);
							register size_t bufferIdx=0;
							for(; firi<firend; firi++)
								sum+=buffer.sampleFO( bufferIdx++ ) * fir.get(firi);
						#else
							// Bi-directional summing has 2 aims: tad better speed
							// and, more important, higher precision, because
							// numbers are summed from least to greatest.
							register int
								firi1 = int(iFIRSize*phaseIdx),
								firi2 = firi1+int(buffer.getSize())-1;
							register size_t
								bufferIdx1=0,
								bufferIdx2=buffer.getSize()-1;
							register FIRT sum = 0;
							while( bufferIdx2 > bufferIdx1 )
								sum += buffer.sampleFO( bufferIdx1++ ) * fir.get(firi1++) +
									buffer.sampleFO( bufferIdx2-- ) * fir.get(firi2--);
							VTASSERT( bufferIdx1==bufferIdx2 );
							VTASSERT( firi1==firi2 );
							sum += buffer.sampleFO( bufferIdx1 ) * fir.get(firi1);
						#endif
						outSamples[ outSample ] = Clip(T(sum), T(-0.999999), T(0.999999));
					}
					
					iTotalOutSamples++;
					int pos1 = int(fPos); // URound(floor(fPos));
					fPos += TStreamPos(this->fInvRatio);
					int pos2 = int(fPos); // URound(floor(fPos));
		
					// We don't care about integer part of fPos, so we just strip it out
					// to keep the precision. This is a neccessary measure.
					fPos = fmod(fPos, TStreamPos(1.0));

					//while( fPos>TStreamPos(1.0) )
						//fPos-=TStreamPos(1.0);
		
					// Eating input samples
					for(int i=pos1; i<pos2; i++){ // i<pos2
						if( inputBuffer.getSize() ){ // Using up samples from prev call
							// Buffer shifting - looks ineffective, but we
							// actually do that 1-2 times per block usually, so WTH.
							buffer.quickInsert( inputBuffer.get(0) );
							#if defined(CVTSINC_CHECKBUFFERFILL)
								if( iBufferFill<buffer.getSize() )
									iBufferFill++;
							#endif						
							for(int i=0; i<int(inputBuffer.getSize())-1; i++)
								inputBuffer.set(i, inputBuffer.get(i+1));
							inputBuffer.resize( inputBuffer.getSize()-1, true );
							iTotalInSamples++;
						}else if( nInSamples ){ // Using samples given in this call
							buffer.quickInsert( *(inSamples++) );
							nInSamples--;
							#if defined(CVTSINC_CHECKBUFFERFILL)
								if( iBufferFill<buffer.getSize() )
									iBufferFill++;
							#endif
							iTotalInSamples++;
						}else{ // Stream must be ending. Use samples from buffer.
							if( !bStreamEnd ){ // This should never happen, actually!
								#if defined(VT_CRTDEBUG)
									printf("*** Wanting zeroes before stream end.\n");
									printf("*** outSample# = %d of %d, fPos = %f of %d\n",
										int(outSample),	int(retOutSamples), float(fPos),
										int(nInSamples_Save));
								#endif
								return outSample+1; // All in&out here is done, so just return
							}else{ // We still can use some samples from the buffer.
								buffer.quickInsert( 0 );
								#if defined(CVTSINC_CHECKBUFFERFILL)
									if( !iBufferFill )
										throw _T("Attempt to move beyond buffer.");
									iBufferFill--;
								#endif
								iTotalInSamples++;
							}
						}
					}
				}
			}else{
				#if defined(VT_CRTDEBUG)
					printf("\r*** No samples to output!\n");
				#endif
			}
			if( nInSamples ){ // Feed all unused yet input samples into the buffer.
				/*#if defined(VT_CRTDEBUG)
					printf("*** Saving %d unprocessed samples (BTW, the ratio is %.3f).\n",
						int(nInSamples), float(this->fRatio));
					printf("*** fPos = %f of %d, buffer size = %d, out samples = %d\n",
						float(fPos), int(nInSamples_Save), int(buffer.getSize()),
						int(retOutSamples));
				#endif*/
				int oldsz = int(inputBuffer.getSize());
				inputBuffer.resize(nInSamples+oldsz, true);
				for( int i=oldsz; i<int(inputBuffer.getSize()); i++ )
					inputBuffer.set(i, *(inSamples++));
			}
			return retOutSamples;
		}else
			return 0;
	}
	#if defined(MSCFASTMATH)
		#pragma float_control(pop)
	#endif


} // namespace

#endif


