//-----------------------------------------------------------------------------
//
// RIFF file format definition and simple tools
//
//-----------------------------------------------------------------------------

/*
	VT Framework License (BSL 1.0)

	Copyright (C) 2006 Sergey A. Galin 
	E-mail: sergey.galin@gmail.com

	Permission is hereby granted, free of charge, to any person or organization
	obtaining a copy of the software and accompanying documentation covered by
	this license (the "Software") to use, reproduce, display, distribute,
	execute, and transmit the Software, and to prepare derivative works of the
	Software, and to permit third-parties to whom the Software is furnished to
	do so, all subject to the following:

	The copyright notices in the Software and this entire statement, including
	the above license grant, this restriction and the following disclaimer,
	must be included in all copies of the Software, in whole or in part, and
	all derivative works of the Software, unless such copies or derivative
	works are solely in the form of machine-executable object code generated by
	a source language processor.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
	SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
	FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
	ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
	DEALINGS IN THE SOFTWARE.
*/

// Some useful resources:
//		http://mediasrv.ns.ac.yu/extra/fileformat/sound/wav/wave.htm

#include "vt_riff.hpp"
#include "vt_assert.hpp"

using namespace VT;

// **************************************************************************
// *** Legacy code
// **************************************************************************

const struct TRIFFWavHeader kDefaultRIFF = {
	MAKEFOURCC('R', 'I', 'F', 'F'),
	0,
	MAKEFOURCC('W', 'A', 'V', 'E'),
	MAKEFOURCC('f', 'm', 't', ' '),
	0x10,
	1,
	2,
	44100,
	44100*4,
	0,
	16,
	MAKEFOURCC('d', 'a', 't', 'a'),
	0
};	


void TRIFFWavHeader::Init()
{
	(*this) = kDefaultRIFF;
}

bool TRIFFWavHeader::VerifySignatures() const
{
	return 
		riffheader	== kDefaultRIFF.riffheader	&&
		WAVEsig		== kDefaultRIFF.WAVEsig		&&
		fmtsig		== kDefaultRIFF.fmtsig		&&
		dataheader	== kDefaultRIFF.dataheader;
}

bool TRIFFWavHeader::IsValidPCMWav(bool strict) const
{
	WORD bypf = RIFF::TFmt::PCMBytesPerFrame(BitsPerSample, Channels);
	return 
		VerifySignatures() &&
		fmtlen == 0x10 &&
		FormatTag == 1 && 
		Channels >= 1 &&			
		(!strict || 
			(
				SamplesPerSec*bypf==AvgBytesPerSec &&
				(BlockAlign == 0 || BlockAlign==bypf)
			)
		);
}

const TRIFFWavHeader* TRIFFWavHeader::GetDefaultHeader()
{	
	return &kDefaultRIFF;
}




// **************************************************************************
// *** RIFF
// **************************************************************************

using namespace RIFF;

const FOURCC
	RIFFsig = MAKEFOURCC('R', 'I', 'F', 'F'),
	FMT		= MAKEFOURCC('f', 'm', 't', ' '),
	DATA	= MAKEFOURCC('d', 'a', 't', 'a'),
	LIST	= MAKEFOURCC('L', 'I', 'S', 'T'), 
	LISTlc  = MAKEFOURCC('l', 'i', 's', 't'), 
	ADTL	= MAKEFOURCC('a', 'd', 't', 'l'),
	LABL	= MAKEFOURCC('l', 'a', 'b', 'l'),
	CUE		= MAKEFOURCC('c', 'u', 'e', ' ');



bool TFmt::VerifySignature() const
{
	return chunkID == FMT && chunkSize>=16;
}

bool TFmt::IsValidPCMWav(bool strict) const
{
	WORD bytesperf = PCMBytesPerFrame();
	return 
		VerifySignature() && 		
		// chunkSize == 0x10 - not neccessary for smart RIFF reader; 
		// And chunkSize >= 0x10 is verified by VerifySignature().
		wFormatTag == 1 && 
		wChannels >=1 &&	
		(!strict ||
			(dwSamplesPerSec*bytesperf == dwAvgBytesPerSec && 
			(wBlockAlign == 0 || wBlockAlign==bytesperf))
		);
}

void TFmt::SetupPCM(DWORD srate, WORD channels, WORD bits)
{
	chunkID = FMT;
	chunkSize = 16;
	wFormatTag = 1;
	wChannels = channels;
	dwSamplesPerSec = srate;
	wBitsPerSample = bits;
	wBlockAlign = PCMBytesPerFrame( bits, channels ); // wBlockAlign is a size of frame in bytes
	dwAvgBytesPerSec = dwSamplesPerSec*wBlockAlign;
}


CRIFFReader::CRIFFReader(): 
	m_File(NULL), 
	chunks(NULL), 
	nChunks(0), 
	fileSize(0), 
	RIFFType(0),
	m_CuePoints(NULL),
	m_nCuePoints(0),
	m_Labels(NULL),
	m_nLabels(0)
{
}

CRIFFReader::~CRIFFReader()
{ 
	Close(); 
	FreeChunks(); 
}

bool CRIFFReader::ContinueOpen()
{
	TChunkHeaderHeader hdr;
	TIFFHeader filehdr;
	//Close();
	// m_File = VTFOPEN( fileName, _T("rb") );
	//if( !m_File )
	//	return false;
	ReallocChunks( 32 ); // 99.999% of files contain fewer chunks...
	VTASSERT( chunks );
	
	// Reading & verifying RIFF header
	if( m_File->read( &filehdr, sizeof(filehdr), 1 )!=1 ){
		#if defined(VT_CRTDEBUG)
			printf("Error: failed to read file header.\n");
		#endif
		return false;
	}
	if( filehdr.id != RIFFsig ){
		#if defined(VT_CRTDEBUG)
			printf("Error: RIFF signature not found.\n");
		#endif
		return false;
	}
	RIFFType = filehdr.type;

	// chunks[ nChunks++ ].Set( hdr, 0 );

	//VTFSEEK( m_File, 0, SEEK_END );
	//fileSize = VTFTELL( m_File );
	fileSize = m_File->getSize();
	
	// Now, reading all chunks
	VTFSIZE nextchunkpos = sizeof( filehdr ); 
	while( nextchunkpos<fileSize ){
		//VTFSEEK( m_File, nextchunkpos, SEEK_SET );
		m_File->seek( nextchunkpos );
		if( m_File->read( &hdr, sizeof(hdr), 1 )!=1 ){
			#if defined(VT_CRTDEBUG)
				printf("Error: chunk header read error.\n");
			#endif
			break;
		}
		if( nChunks>=allocChunks )
			ReallocChunks( allocChunks+100 );
		chunks[ nChunks++ ].Set( hdr, nextchunkpos );
		if( hdr.size & 1 )
			hdr.size++;
		nextchunkpos += hdr.size + 8;
	}

	return true;
}

bool CRIFFReader::Open( CFile *file )
{
	Close();
	bCloseFile = false;
	m_File = file;
	return ContinueOpen();
}

bool CRIFFReader::Open( LPCVTCHAR fileName )
{
	Close();
	m_File = new CDiskFile( fileName, _T("rb") );
	if( !m_File->IsOK() ){
		delete m_File;
		m_File = NULL;
		return false;
	}
	bCloseFile = true;
	return ContinueOpen();
}

void CRIFFReader::Close()
{
	//if( m_File!=NULL ){
	//	fclose( m_File );
	//	m_File = NULL;
	//}

	if( bCloseFile && m_File!=NULL )
		delete m_File;
	bCloseFile = false;
	m_File = NULL;

	RIFFType = 0;
	fileSize = 0;
	nChunks = 0;

	DeleteAllInfo();
}

long CRIFFReader::FindChunk( FOURCC id, size_t startpos ) const
{
	for(; startpos<nChunks; startpos++)
		if( chunks[startpos].id == id )
			return long(startpos);
	return -1;
}

bool CRIFFReader::SeekToPos(VTFSIZE pos)
{
	if( !IsOK() )
		return false;
	if( m_File->seek( pos )!=0 )
		return false;
	return m_File->tell() == pos; 
}

bool CRIFFReader::SeekToChunk(FOURCC signature, DWORD *size, size_t startpos)
{
	if( !IsOK() )
		return false;
	long idx = FindChunk( signature, startpos );
	if( idx<0 )
		return false;
	if( size!=NULL )
		*size = chunks[idx].size;
	return SeekToChunk( chunks[idx] );
}

bool CRIFFReader::SeekToData(FOURCC signature, DWORD *size, size_t startpos)
{
	if( !IsOK() )
		return false;
	long idx = FindChunk( signature, startpos );
	if( idx<0 )
		return false;
	if( size!=NULL )
		*size = chunks[idx].size;
	return SeekToData( chunks[idx] );
}

bool CRIFFReader::Read(void *out, size_t size)
{
	if( !IsOK() )
		return false;
	if( m_File->read( out, 1, size )!=size )
		return false;
	else
		return true;
}

void CRIFFReader::FreeChunks()
{
	if( chunks!=NULL ){
		delete chunks;
		chunks = NULL;
		nChunks = 0;
		allocChunks = 0;
	}
}

void CRIFFReader::ReallocChunks(size_t newsize)
{
	if( newsize<32 )
		newsize = 32;
	if( newsize<=allocChunks && chunks!=NULL )
		return;
	TChunkInfo *newchunks = new TChunkInfo[ newsize ];
	if( chunks!=NULL ){
		if( nChunks>0 )
			memcpy( newchunks, chunks, sizeof(TChunkInfo)*nChunks );
		delete chunks; 
	}
	chunks = newchunks;
	allocChunks = newsize;
}

// Read CUE points. Delete the array using 'operator delete' after use.
// Returns NULL if no CUE points found.
// Only one CUE chunk can be present in a file.
TCuePoint* CRIFFReader::LoadCuePoints(size_t *outSize)
{
	*outSize = 0;
	if( !IsOK() )
		return NULL;
	const DWORD sign = CUE;
	long idx = FindChunk( sign );
	if( idx<0 )
		return NULL;
	m_File->seek( chunks[idx].position );
	TCueChunkHeader hdr;
	if( m_File->read( &hdr, sizeof(hdr), 1 )!=1 )
		return NULL;
	if( hdr.chunkID!=sign ) // No one ever died from double-check
		return NULL;
	if( !hdr.dwCuePoints )
		return NULL;
	TCuePoint *points = new TCuePoint[ hdr.dwCuePoints ];
	if( !points )
		return NULL;
	if( m_File->read( points, sizeof(TCuePoint), hdr.dwCuePoints )!=hdr.dwCuePoints ){
		delete points;
		return NULL;
	}
	*outSize = size_t( hdr.dwCuePoints );
	return points;
}

TLabel** CRIFFReader::LoadLabels(size_t *outSize)
{
	*outSize = 0;
	if( !IsOK() )
		return NULL;
	long idx = FindChunk( LIST );
	if( idx<0 )
		idx = FindChunk( LISTlc );
	if( idx<0 )
		return NULL;
	m_File->seek( chunks[idx].position );
	TIFFHeader hdr;
	if( m_File->read( &hdr, sizeof(hdr), 1 )!=1 )
		return NULL;
	if( (hdr.id!=LIST && hdr.id!=LISTlc) || hdr.type!=ADTL )
		return NULL;
	// Now read all sub-chunks
	VTFSIZE 
		// Current position in file, used to determine if we're 
		// at the end of the chunk or not.
		pos = chunks[idx].position+sizeof(hdr), 
		// End of the LIST chunk
		rightpos = chunks[idx].position+8+hdr.size; 

	size_t labelsSize = 32, labelsPos = 0;
	TLabel **labels = new TLabel*[labelsSize];
	while( pos<rightpos ){
		TLabelChunk subchunkhdr;
		if( m_File->read( &subchunkhdr, sizeof(TChunkHeaderHeader), 1 )!=1 )
			break;
		pos += sizeof(TChunkHeaderHeader);
		if( subchunkhdr.chunkID==LABL ){ // Found a label
			// Reading the rest of the structure
			if( m_File->read( ((char*)(&subchunkhdr))+sizeof(TChunkHeaderHeader),
				sizeof(TLabelChunk)-sizeof(TChunkHeaderHeader), 1 )!=1 )
				break;
			pos += sizeof(TLabelChunk)-sizeof(TChunkHeaderHeader);
			
			if( subchunkhdr.chunkSize & 1 )
				subchunkhdr.chunkSize++;

			// Reading string
			size_t strlength = size_t( subchunkhdr.chunkSize-sizeof(subchunkhdr.dwIdentifier) );	
			size_t structsize = size_t( sizeof(TLabel)+strlength+1 ); // +1 is for fun
			TLabel *label = (TLabel*) (new char[structsize]);
			label->dwIdentifier = subchunkhdr.dwIdentifier;
			memset( label->dwText, 0, strlength+1 );
			if( m_File->read( label->dwText, strlength, 1)!=1 ){
				delete label;
				break;
			}
			pos += strlength;
			
			// Adding label into return array
			if( labelsPos>=labelsSize-1 ){
				labelsSize += 100;
				TLabel **newLabels = new TLabel*[labelsSize];
				memcpy( newLabels, labels, sizeof(TLabel*)*labelsPos );
				delete labels;
				labels = newLabels;
			}
			labels[ labelsPos++ ] = label;

		}else{
			// Skipping the entry
			if( subchunkhdr.chunkSize & 1 )
				subchunkhdr.chunkSize++;
			pos += subchunkhdr.chunkSize;
			m_File->seek( subchunkhdr.chunkSize, SEEK_CUR );
		}			

		VTASSERT( pos==m_File->tell() );
	}
	if( labelsPos ){
		*outSize = size_t( labelsPos );
		return labels;
	}else{
		delete labels;
		return NULL;
	}
}


bool CRIFFReader::LoadFmt(TFmt* out)
{
	if( !IsOK() )
		return false;
	long idx = FindChunk( FMT );
	if( idx<0 )
		return false;
	if( !SeekToPos( chunks[idx].position ) )
		return false;
	if( !Read( out, sizeof(TFmt) ) )
		return false;
	if( !out->VerifySignature() )
		return false;
	return true;
}

bool CRIFFReader::LoadAllLists()
{
	if( !IsOK() )
		return false;
	DeleteAllInfo();
	m_CuePoints = LoadCuePoints( &m_nCuePoints );
	m_Labels = LoadLabels( &m_nLabels );
	return true;
}

void CRIFFReader::DeleteAllInfo()
{
	// Clearing additional information
	if( m_CuePoints!=NULL ){
		delete m_CuePoints;
		m_CuePoints = NULL;
		m_nCuePoints = 0;
	}
	if( m_Labels!=NULL ){
		for( size_t i=0; i<m_nLabels; i++ )
			delete m_Labels[i];
		delete m_Labels;
		m_Labels = NULL;
		m_nLabels = 0;
	}
}

const TCuePoint* CRIFFReader::GetCuePointById(DWORD uid) const
{
	if( m_CuePoints==NULL || m_nCuePoints==0 )
		return NULL;
	for( size_t i=0; i<m_nCuePoints; i++ )
		if( m_CuePoints[i].dwIdentifier == uid )
			return m_CuePoints+i;
	return NULL;
}

const char* CRIFFReader::GetLabelById(DWORD uid) const
{
	if( m_Labels==NULL || m_nLabels==0 )
		return "";
	for( size_t i=0; i<m_nLabels; i++ )
		if( m_Labels[i]->dwIdentifier == uid )
			return m_Labels[i]->dwText;
	return "";
}

bool CRIFFReader::GetIdByLabel(const char *label, DWORD *result) const
{
	*result = 0;
	if( m_Labels==NULL || m_nLabels==0 )
		return false;
	for( size_t i=0; i<m_nLabels; i++ )
		if( !strcmp(label, m_Labels[i]->dwText) ){
			*result = m_Labels[i]->dwIdentifier;
			return true;
		}
	return false;	
}

const TCuePoint* CRIFFReader::GetCueByLabel(const char *label) const
{
	DWORD uid;
	if( GetIdByLabel(label, &uid) )
		return GetCuePointById( uid );
	else
		return NULL;
}





