//-----------------------------------------------------------------------------
//
// RIFF file format definition and simple tools
//
//-----------------------------------------------------------------------------

/*
	VT Framework License (BSL 1.0)

	Copyright (C) 2006 Sergey A. Galin 
	E-mail: sergey.galin@gmail.com

	Permission is hereby granted, free of charge, to any person or organization
	obtaining a copy of the software and accompanying documentation covered by
	this license (the "Software") to use, reproduce, display, distribute,
	execute, and transmit the Software, and to prepare derivative works of the
	Software, and to permit third-parties to whom the Software is furnished to
	do so, all subject to the following:

	The copyright notices in the Software and this entire statement, including
	the above license grant, this restriction and the following disclaimer,
	must be included in all copies of the Software, in whole or in part, and
	all derivative works of the Software, unless such copies or derivative
	works are solely in the form of machine-executable object code generated by
	a source language processor.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
	SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
	FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
	ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
	DEALINGS IN THE SOFTWARE.
*/

#if !defined(VT_RIFF_H)
#define VT_RIFF_H

#include <stdio.h>
#include <string.h>
#include "vt_portable.hpp"
#include "vt_file.hpp"

// **************************************************************************
// *** Legacy code
// **************************************************************************
//
// A VERY basic structure which is suitable for reading and writing typical,
// non-compressed WAV files without any custom tags ONLY.
// Do not use it, it is here for bakward compatibility only.
//
// IMPORTANT:
// Default initialization done by Init() sets all fields as for CD Audio format:
// 16-bit signed PCM, stereo. This behaviour should not be changed.
struct TRIFFWavHeader
{
	// Some comments from http://www.ringthis.com/dev/wave_format.htm
	FOURCC riffheader;		// "RIFF"
	DWORD  WAVElen;			// Size of the file, DWORD
							// The file size LESS the size of the "RIFF" description (4 bytes) 
							// and the size of file description (4 bytes). This is usually file size - 8.
	FOURCC WAVEsig;         // "WAVE"
	FOURCC fmtsig;          // "fmt "
	DWORD  fmtlen;			// DWORD, The size of the WAVE type format (2 bytes) 
							// + mono/stereo flag (2 bytes) + sample rate (4 bytes) + 
							// bytes/sec (4 bytes) + block alignment (2 bytes) + bits/sample (2 bytes). 
							// This is usually 16 (or 0x10).
	WORD   FormatTag;		// 1 == PCM
	WORD   Channels;		// Mono == 1, Stereo == 2
	DWORD  SamplesPerSec;	// Sample rate, DWORD
	DWORD  AvgBytesPerSec;	// bps, DWORD
	WORD   BlockAlign;		// Block alignment
	WORD   BitsPerSample;	/* format specific for PCM */
	FOURCC dataheader;		// "data"
	DWORD  datalen;			// number of bytes in data

	void Init();  // Sets all fields by default	
	bool VerifySignatures() const; // Checks that all text signatures are OK
	// Checks that the header is valid for simple PCM Wav.
	// 'strict' enables checking of bps and align fields
	bool IsValidPCMWav(bool strict=false) const; 
	static const TRIFFWavHeader* GetDefaultHeader();
};



// **************************************************************************
// *** RIFF - a collection of classes for reading any RIFF files.
// **************************************************************************

namespace VT {
	namespace RIFF {

		// IFF header
		struct TIFFHeader
		{
			FOURCC id; 
			DWORD  size; // Size of file/chunk minus 8 bytes
			FOURCC type; // Type of the file, e.g. 'WAVE'
		};

		// Chunk header, as it is in file
		struct TChunkHeaderHeader
		{
			FOURCC  id;   // Signature		
			DWORD	size; // Size of the chunk minus 8 bytes
		};

		// Chunk information, as it is stored in memory
		struct TChunkInfo: public TChunkHeaderHeader
		{
			VTFSIZE position; // Position of chunk (staring from id) in file.
			
			VTINLINE void Set( const TChunkHeaderHeader& fci, VTFSIZE pos ){
				id = fci.id;
				size = fci.size;
				position = pos;
			}
		};

		// ***************************************************************
		// *** Chunk-specific structures
		// ***************************************************************

		struct TFmt{
			FOURCC chunkID;				// "fmt "
			DWORD  chunkSize;			// DWORD, The size of the WAVE type format (2 bytes) 
										// This is usually 16 (or 0x10). Can be bigger for 
										// 
			WORD   wFormatTag;			// 1 == PCM
			WORD   wChannels;			// Mono == 1, Stereo == 2
			DWORD  dwSamplesPerSec;		// Sample rate, DWORD
			DWORD  dwAvgBytesPerSec;	// bps, DWORD
			WORD   wBlockAlign;			// Block alignment
			WORD   wBitsPerSample;		// format specific for PCM 

			static WORD PCMBytesPerSample(WORD bitsps){ return (bitsps+7)>>3; }
			WORD PCMBytesPerSample() const { return PCMBytesPerSample(wBitsPerSample); }		
			static WORD PCMBytesPerFrame(WORD bitps, WORD chan) { return PCMBytesPerSample(bitps)*chan; }
			WORD PCMBytesPerFrame() const { return PCMBytesPerFrame(wBitsPerSample, wChannels); }

			bool VerifySignature() const;
			bool IsValidPCMWav(bool strict=false) const; // 'strict' enables checking of bps and align fields
			void SetupPCM(DWORD srate=44100, WORD channels=2, WORD bits=16);
		};

		// CUE Sheet	
		struct TCueChunkHeader
		{
			FOURCC chunkID;       // 'cue '
			DWORD  chunkSize;
			DWORD  dwCuePoints;
		};
		
		// Array of TCuePoint follows TCueChunkHeader
		struct TCuePoint
		{
		  DWORD  dwIdentifier;	  // Unique number 
		  DWORD  dwPosition;	  // Position within play order
		  FOURCC fccChunk;        // Chunk ID of the Data or Wave List
		  // 0 for uncompressed WAV, used for files with multiple 'data' 
		  // and 'slnt' chunks; specifies the byte offset of the start of 
		  // the 'data' or 'slnt' chunk which actually contains the waveform 
		  // data to which this CuePoint refers.
		  DWORD  dwChunkStart;    
		  // 0 for uncompressed WAV, used for files with multiple 'data' 
		  // and 'slnt' chunks or compressed files. The field specifies 
		  // the byte offset of the start of the block containing the position. 
		  // This offset is relative to the start of the waveform data within 
		  // the 'data' or 'slnt' chunk.
		  DWORD  dwBlockStart;	  
		  DWORD  dwSampleOffset; // Sample offset of the cue point relative to the start of the block.			
		};

		// Labels, notes and labeled text chunks are contained inside 'LIST' ('list') /'adtl' 
		// (Associated Data List) chunk.
		struct TLabelChunk
		{
			FOURCC chunkID; // 'labl'
			DWORD  chunkSize;
			DWORD  dwIdentifier; 
			// Followed by char dwText[].
		};

		// Label as it is represended in memory
		struct TLabel
		{
			DWORD dwIdentifier;
			char  dwText[1]; // Variable
		};

		struct TNoteChunk
		{
			FOURCC chunkID; // 'note'
			DWORD  chunkSize;
			DWORD  dwIdentifier; 
			// Followed by char dwText[].
		};

		struct TLabelTextChunk
		{
			FOURCC chunkID; // 'ltxt'
			DWORD  chunkSize;
			DWORD  dwIdentifier;
			DWORD  dwSampleLength;
			DWORD  dwPurpose;
			WORD   wCountry;
			WORD   wLanguage;
			WORD   wDialect;
			WORD   wCodePage;
			// Followed by char dwText[].
		};

		//
		// A flexible class for reading RIFF files
		//
		class CRIFFReader
		{
		public:
			CRIFFReader();
			~CRIFFReader();
			
			bool Open( CFile *file ); // Opens file and loads all chunk headers (not subchunks)
			bool Open( LPCVTCHAR fileName ); // Opens file and loads all chunk headers (not subchunks)
			void Close();
			bool IsOK() const { if( m_File==NULL) return false; return m_File->IsOK() && nChunks; }
			
			bool Read(void *out, size_t size);
			bool SeekToPos(VTFSIZE pos); // Seek to raw position

			//******************************************************************
			// File seeking utiltes.
			// 'size' parameter to all these functions is a pointer to buffer 
			// where chunk data size is returned. Pass NULL if you don't need
			// that information.
			//******************************************************************
			// Seek to chunk start.		
			bool SeekToChunk(const TChunkInfo& chunk){ return SeekToPos(chunk.position); } 
			bool SeekToChunk(FOURCC signature, DWORD *size=NULL, size_t startpos=0);
			bool SeekToChunk(unsigned char a, unsigned char b, unsigned char c, unsigned char d, DWORD *size=NULL, size_t startpos=0){
				return SeekToChunk(VTMakeFOURCC(a, b, c, d), size, startpos); }
			bool SeekToChunk(const char* riff, DWORD *size=NULL, size_t startpos=0){ 
				return SeekToChunk(VTMakeFOURCC(riff), size, startpos); }
			// Seek to chunk data
			bool SeekToData(const TChunkInfo& chunk){ return SeekToPos(chunk.position+8); } 
			bool SeekToData(FOURCC signature, DWORD *size=NULL, size_t startpos=0);
			bool SeekToData(unsigned char a, unsigned char b, unsigned char c, unsigned char d, 
				DWORD *size=NULL, size_t startpos=0){
				return SeekToData(VTMakeFOURCC(a, b, c, d), size, startpos); }
			bool SeekToData(const char* riff, DWORD *size=NULL, size_t startpos=0){ 
				return SeekToData(VTMakeFOURCC(riff), size, startpos); }

			// Access to variables
			const size_t		GetChunkCount() const { return nChunks; }
			const TChunkInfo*	GetChunk(size_t i) const { return chunks+i; }
			const VTFSIZE		GetFileSize() const { return fileSize; }
			
			// Return value is chunk index, -1 if not found.
			// startpos is the first chunk index checked.
			long FindChunk( FOURCC id, size_t startpos=0 ) const;
			long FindChunk( unsigned char a, unsigned char b, unsigned char c, unsigned char d, size_t startpos=0 ) const{ 
				return FindChunk( VTMakeFOURCC(a, b, c, d), startpos ); }
			long FindChunk( const char* str, size_t startpos=0 ) const{ return FindChunk( VTMakeFOURCC(str), startpos ); }

			// ***************************************************************
			// *** Chunk type-specific utilites
			// ***************************************************************

			// Read CUE points. Delete the array using 'operator delete' after use.
			// Returns NULL if no CUE points found.
			TCuePoint* LoadCuePoints(size_t *outSize); 

			// Read text labels. Delete the array items & the array of pointers itself after use.
			TLabel** LoadLabels(size_t *outSize); 

			// Load TFmt structure
			bool LoadFmt(TFmt* out);


			// ***************************************************************
			// *** High-level chunk utilites
			// ***************************************************************
			//
			// The functions in this section load chunk data into object's 
			// own memory, and provide various high-level stuff based upon
			// the information.
			//

			// Load all additional information supported by the class.
			// Returns false on error (though error checking is pretty weak).
			// Memory used by the information is freed using DeleteAllInfo()
			// automatically by Close().
			virtual bool LoadAllLists(); 
			
			//
			// Working with data loaded via LoadAllLists()
			//
			const TCuePoint* GetCuePoint(size_t idx) const { return m_CuePoints+idx; }
			size_t			 GetCuePointCount() const { return m_nCuePoints; }
			const TCuePoint* GetCuePointById(DWORD uid) const; // NULL if not found
			const TLabel*	 GetLabel(size_t idx) const { return m_Labels[idx]; }
			size_t			 GetLabelCount() const { return m_nLabels; }			
			const char*		 GetLabelById(DWORD uid) const; // Empty string if not found
			bool			 GetIdByLabel(const char *label, DWORD *result) const; // If not found, returns false and 0 in *result
			const TCuePoint* GetCueByLabel(const char *label) const; // NULL if not found
			

		protected:
			bool ContinueOpen();
			void FreeChunks(); // Delete all chunk data		
			void ReallocChunks(size_t newsize); // Reallocate (resize) chunk array
			virtual void DeleteAllInfo(); // Delete all additional information from memory

			// FILE *m_File;
			CFile *m_File;
			bool bCloseFile;

			TChunkInfo *chunks;
			size_t nChunks, allocChunks;
			VTFSIZE fileSize;
			FOURCC RIFFType;
			
			// Additional information (allocated & filled by LoadAllLists()):
			TCuePoint* m_CuePoints;
			size_t m_nCuePoints;
			TLabel** m_Labels;
			size_t m_nLabels;


		};
	}; // namespace
}; // namespace

#endif


