//-----------------------------------------------------------------------------
//
// Yet one more C++ string class, which is mostly based on array template.
//
//-----------------------------------------------------------------------------

/*
	VT Framework License (BSL 1.0)

	Copyright (C) 2006 Sergey A. Galin 
	E-mail: sergey.galin@gmail.com

	Permission is hereby granted, free of charge, to any person or organization
	obtaining a copy of the software and accompanying documentation covered by
	this license (the "Software") to use, reproduce, display, distribute,
	execute, and transmit the Software, and to prepare derivative works of the
	Software, and to permit third-parties to whom the Software is furnished to
	do so, all subject to the following:

	The copyright notices in the Software and this entire statement, including
	the above license grant, this restriction and the following disclaimer,
	must be included in all copies of the Software, in whole or in part, and
	all derivative works of the Software, unless such copies or derivative
	works are solely in the form of machine-executable object code generated by
	a source language processor.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
	SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
	FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
	ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
	DEALINGS IN THE SOFTWARE.
*/

#include <ctype.h>
#include "vt_string.hpp"

using namespace VT;

bool CString::TruncateBy0()
{
	FixZero();
	size_t len = VTSTRLEN( m_pData );
	VTASSERT( m_uDataSize>=1 );
	if( len < m_uDataSize-1 ){
		m_uDataSize = len+1;
		FixZero();
		return true;
	}
	return false;
}

bool CString::Truncate( size_t newLength )
{
	if( newLength<Length() ){
		m_uDataSize = newLength+1;
		FixZero();
		return true;
	}
	return false;
}

// This function minds that strings take at least 1 (zero) character of space.
// Also it sets the zero character.
void CString::resize(size_t newDataSize, bool saveData, size_t newAllocSize)
{
	if( newDataSize<1 )
		newDataSize = 1;
	if( m_uAllocSize>0 ){
		VTASSERT( m_uDataSize<=m_uAllocSize );
		FixZero(); // Extra safety...
	}
	CArray<VTCHAR>::resize(newDataSize, saveData, newAllocSize);
	VTASSERT( m_uAllocSize>0 );
	FixZero();
}

CString& CString::CopyFrom(LPCVTCHAR str)
{
	if( str==NULL )
		Empty();
	else
		rawCopyFromSameType( str, VTSTRLEN(str)+1 );
	return *this;
}

CString& CString::CopyFrom(LPCVTCHAR str, size_t nChars)
{
	if( str==NULL )
		Empty();
	else{
		size_t len=VTSTRLEN(str);
		if( len<nChars )
			nChars = len;
		rawCopyFromSameType( str, nChars+1 );
		FixZero();
	}
	return *this;
}

CString& CString::ConcatStr( LPCVTCHAR str )
{
	if( str!=NULL ){
		size_t len=VTSTRLEN(str), myoldlen = Length();
		// printf("'%s'.ConcatStr( '%s' ): myoldlen=%u\n", m_pData, str, myoldlen );
		resize( myoldlen+len+1, true ); // Resize saving m_pData
		memcpy( m_pData+myoldlen, str, (len+1)*sizeof(VTCHAR) );
	}
	return *this;
}

CString& CString::ConcatStr( LPCVTCHAR str, size_t nChars )
{
	if( str!=NULL ){
		size_t len=VTSTRLEN(str), myoldlen = Length();
		if( len<nChars )
			nChars = len;
		resize( myoldlen+nChars+1, true ); // Resize saving m_pData
		memcpy( m_pData+myoldlen, str, nChars*sizeof(VTCHAR) );
		FixZero(); // Set zero byte at m_pData[ string length ]
	}
	return *this;
}

CString& CString::ConcatVTStr( const CString& str )
{
	size_t len=str.Length(), myoldlen = Length();
	resize( myoldlen+len+1, true ); // Resize saving m_pData
	memcpy( m_pData+myoldlen, str.Str(), (len+1)*sizeof(VTCHAR) );
	return *this;
}

CString& CString::ConcatVTStr( const CString& str, size_t nChars )
{
	size_t len=str.Length(), myoldlen = Length();
	if( len<nChars )
		nChars = len;
	resize( myoldlen+nChars+1, true ); // Resize saving m_pData
	memcpy( m_pData+myoldlen, str.Str(), nChars*sizeof(VTCHAR) );
	FixZero(); // Set zero byte at m_pData[ string length ]
	return *this;
}


CString& CString::ConcatChar( VTCHAR ch )
{
	if( ch ){
		resize( m_uDataSize+1, true );
		VTASSERT( m_uDataSize>=2 );
		m_pData[ m_uDataSize-2 ] = ch;
		FixZero();
	}
	return *this;
}


CString& CString::ConcatHex( const BYTE *ptr, size_t nBytes, size_t nBytesPerLine )
{
	const VTCHAR hexes[17] = _T("0123456789ABCDEF");
	size_t col = 0;
	for( size_t i=0; i<nBytes; i++ ){
		BYTE byt = *(ptr++);
		ConcatChar( hexes[ byt>>4 ] );
		ConcatChar( hexes[ byt & 0x0F ] );		
		if( i<nBytes-1 ){
			bool newline = false;
			if( nBytesPerLine > 0 ){
				col++;
				if( col>=nBytesPerLine ){
					newline = true;
					ConcatStr( VTENDLT );
					col = 0;
				}
			}
			if( !newline )
				ConcatChar( _T(' ') );
		}		
	}
	return *this;
}


CString& CString::ConcatLong( long x, LPCVTCHAR format )
{
	VTCHAR buffer[ 128 ];
	VTSPRINTF( buffer, format, x );
	ConcatStr( buffer );
	return *this;
}

CString& CString::ConcatDouble( double x, LPCVTCHAR format )
{
	VTCHAR buffer[ 128 ];
	VTSPRINTF( buffer, format, x );
	ConcatStr( buffer );
	return *this;
}

int CString::CompareCaseSensitive(LPCVTCHAR str, int maxChars) const
{
	if( str==NULL )
		return !IsEmpty();
	if( maxChars<0 )
		return VTSTRCMP( Str(), str );
	else
		return VTSTRNCMP( Str(), str, size_t(maxChars) );
}

int CString::CompareCaseInsensitive(LPCVTCHAR str, int maxChars) const
{
	if( str==NULL )
		return !IsEmpty();
	if( maxChars<0 )
		return VTSTRICMP( Str(), str );
	else
		return VTSTRNICMP( Str(), str, size_t(maxChars) );
}

void CString::TrimLeft()
{	
	if( m_uDataSize>1 && VTISSPACE( m_pData[0] ) ){
		size_t nSpaces=1; // Already found one space
		while( VTISSPACE( m_pData[nSpaces] ) ){
			if( nSpaces >= Length() || m_pData[nSpaces]==0 ){ // All characters in string are spaces
				Empty();
				return;
			}
			nSpaces++;
		}
		m_uDataSize = Length()-nSpaces+1;
		memmove( m_pData, m_pData+sizeof(VTCHAR)*nSpaces, sizeof(VTCHAR)*m_uDataSize );
	}
}

void CString::TrimRight()
{
	if( m_uDataSize>1 ){
		while( VTISSPACE( m_pData[m_uDataSize-2] ) ){ // Last index is size-1, and also step back over zero 
			if( m_uDataSize<=2 ){ // All characters in string are spaces
				Empty();
				return;
			}
			m_uDataSize--;
		}
		FixZero();
	}
}

void CString::TrimEndlChars()
{
	if( m_uDataSize>1 ){
		// Last index is size-1, and also step back over zero 
		while( m_pData[m_uDataSize-2]==_T('\r') || m_pData[m_uDataSize-2]==_T('\n') ){ 
			if( m_uDataSize<=2 ){ // All characters in string are spaces
				Empty();
				return;
			}
			m_uDataSize--;
		}
		FixZero();
	}
}

CString& CString::ToLower()
{
	for( size_t i=0, e=m_uDataSize-1; i<e; i++ )
		m_pData[i] = VTTOLOWER( m_pData[i] );
	return *this;
}

CString& CString::ToUpper()
{
	for( size_t i=0, e=m_uDataSize-1; i<e; i++ )
		m_pData[i] = VTTOUPPER( m_pData[i] );
	return *this;
}


// Copy fragment from src to character 'delimiter' or to end of the string.
// Delimiter is not copied. 
// Returns pointer set either to zero or to the character after delimiter.
LPCVTCHAR CString::ExtractUntil( LPCVTCHAR src, VTCHAR delimiter, VTCHAR noDelimPfx )
{
	//printf("SRC: [%s] delim=[%c] pfx=[%c] *** ", src, delimiter, noDelimPfx);
	Empty();
	if( src==NULL )
		return _T("");
	if( !(*src) )
		return src;
	size_t nChars=0;
	for(; src[nChars]; nChars++)
		if( noDelimPfx && src[nChars]==noDelimPfx && src[nChars+1]==delimiter )
			nChars++;
		else if( src[nChars]==delimiter )
			break;			 
	LPCVTCHAR ret;
	// src is not empty, so nChars is >= 1 here.
	if( !src[nChars] ) // Ended due end of the string
		ret = &src[nChars]; // Point to the zero char found
	else // Ended due src[nChars] == delimiter
		ret = src+nChars+1; // Point to symbol after the delimiter
	resize( nChars+1, false );
	// Copy characters skipping noDelimPfx in (noDelimPfx, delimiter) sequences
	size_t i2=0;
	for(size_t i=0; i<nChars; i++){
		if( src[i]==noDelimPfx && src[i+1]==delimiter )
			i++;
		m_pData[i2++] = src[i];
	}
	m_pData[i2]=0;
	resize( ++i2, true ); // This may reduce size, if some noDelimPfx's were skipped
	//printf("RESULT: [%s] RET: [%s]\n", m_pData, ret);
	return ret;
}

CString& CString::Implode( const CStringArray& arr, VTCHAR separator, bool separatorAfter )
{
	Empty();
	for( size_t i=0; i<arr.getSize(); i++ ){
		ConcatStr( arr.getConst(i) );
		if( i+1<arr.getSize() )
			ConcatChar( separator );
	}
	if( separatorAfter )
		ConcatChar( separator );
	return *this;
}

CString& CString::ImplodeS( const CStringArray& arr, const CString& separator, bool separatorAfter )
{	
	size_t arrsz = arr.getSize();
	if( arrsz>0 ){
		size_t gsm1 = arrsz-1; // getSize()>0
		// Pre-allocating string buffer
		size_t sz = separator.Length()*arrsz + arr.GetTotalLength();
		resize( 0, false, sz+1 ); 
		Empty(); // Repairing m_pData after resize
		for( size_t i=0; i<arrsz; i++ ){
			ConcatVTStr( arr.getConst(i) );
			if( i<gsm1 )
				ConcatVTStr( separator );
		}
	}else
		Empty();
	if( separatorAfter )
		ConcatVTStr( separator );
	return *this;
}

CString& CString::Implode( const LPCVTCHAR *nullTermList, const CString& separator, bool separatorAfter )
{	
	if( nullTermList!=NULL ){
		size_t sz = 0;
		for( const LPCVTCHAR *ptr = nullTermList; *ptr!=NULL; ptr++ )
			sz += VTSTRLEN( *ptr ) + separator.Length();
		resize( 0, false, sz+1 );
		Empty(); // Repairing m_pData after resize
		for( const LPCVTCHAR *ptr = nullTermList; *ptr!=NULL; ptr++ ){
			ConcatStr( *ptr );			
			if( *(ptr+1)!=NULL )
				ConcatVTStr( separator );
		}
	}else
		Empty();
	if( separatorAfter )
		ConcatVTStr( separator );
	return *this;
}

void CString::ConcatAll( const CStringArray& arr ) 
{
	size_t arrsz = arr.getSize();
	if( arrsz>0 ){
		// Pre-allocating string buffer
		resize( 0, false, arr.GetTotalLength()+1 ); 
		Empty(); // Repairing m_pData after resize
		for( size_t i=0; i<arrsz; i++ )
			ConcatVTStr( arr.getConst(i) );
	}else
		Empty();
}




bool CString::Load( CFile *file, long maxChunkSizeChars )
{
	Empty();
	if( !file->IsOK() )
		return false;
	size_t sz = size_t(file->getSize() - file->tell());
	if( maxChunkSizeChars>=0 )
		sz = Min<size_t>( size_t(maxChunkSizeChars), sz );
	resize( sz+1 );
	zero();
	if( sz > 0 ){
		bool ret = file->read( m_pData, sz*sizeof(VTCHAR) );
		TruncateBy0();
		return ret;
	}else
		// (The string is already emptied by resize(1) and zero() above.)
		return true;
}

bool CString::Save( CFile *file ) const
{
	if( !file->IsOK() )
		return false;
	return file->write( m_pData, Length()*sizeof(VTCHAR) );
}

bool CString::Load( LPCVTCHAR fileName )
{
	CDiskFile file( fileName, _T("rb") );
	return Load( &file );
}

bool CString::Save( LPCVTCHAR fileName ) const
{
	CDiskFile file( fileName, _T("wb") );
	return Save( &file );
}

VTCHAR CString::FirstNonspaceChar() const
{
	for( VTCHAR *ptr=m_pData; *ptr; ptr++ ){
		if( !(*ptr) )
			return 0;
		if( !VTISSPACE(*ptr) )
			return *ptr;
	}
	return 0;
}

VTCHAR CString::LastNonspaceChar() const
{
	if( m_uDataSize<2 )
		return 0;
	for( VTCHAR *ptr=m_pData+m_uDataSize-2; ptr>=m_pData; ptr-- )
		if( !VTISSPACE(*ptr) )
			return *ptr;
	return 0;
}


long CString::ToLong(bool *conversionOK, long defValue ) const
{
	long ret=defValue;
	bool ok = VTSSCANF( Str(), _T("%ld"), &ret )==1;
	if( conversionOK!=NULL )
		*conversionOK = ok;
	if( !ok ) 
		return defValue;
	return ret;
}

double CString::ToDouble(bool *conversionOK, long defValue ) const
{
	double ret=defValue;
	bool ok = VTSSCANF( Str(), _T("%lf"), &ret )==1;
	if( conversionOK!=NULL )
		*conversionOK = ok;
	if( !ok ) 
		return defValue;
	return ret;
}



CStringArray::CStringArray( size_t size, size_t m_uAllocSize ):
	CArray<CString>( size, m_uAllocSize )
{
}



int CStringArray::Find(LPCVTCHAR str, bool bCaseSensitive, size_t startIndex) const
{
	for(; startIndex<m_uDataSize; startIndex++)
		if( m_pData[startIndex].Compare( str, bCaseSensitive )==0 )
			return int(startIndex);
	return -1;
}

size_t CStringArray::AddUnique( LPCVTCHAR str, bool bCaseSensitive )
{
	int idx = this->Find( str, bCaseSensitive, 0 );
	if( idx>=0 )
		return size_t(idx);
	else{
		size_t newidx = this->getSize();
		this->resize( newidx+1, true );
		this->m_pData[ newidx ].CopyFrom( str );
		return newidx;
	}	
}

void CStringArray::Explode(LPCVTCHAR str, VTCHAR separator, int maxItems)
{
	size_t nParts=0;
	resize( 0, false );
	if( str!=NULL )
		for(;;){
			nParts++;
			if( m_uDataSize<nParts )
				resize( nParts, true );
			
			if( nParts == size_t(maxItems) ){
				m_pData[ nParts-1 ].CopyFrom( str );
				break;

			}else{
				size_t len = 0;
				for(; str[len] && str[len]!=separator; len++);
				m_pData[ nParts-1 ].CopyFrom( str, len );
				if( str[len]==separator )
					str += len+1;
				else // Found 0 character
					break;
			}
		}
	if( nParts==1 && m_pData[0].IsEmpty() )
		resize( 0, false );
}

size_t CStringArray::GetTotalLength() const
{
	size_t ret=0;
	for( size_t i=0; i<getSize(); i++ )
		ret += m_pData[i].Length();
	return ret;
}

void CStringArray::TrimAllRight()
{
	for( size_t i=0; i<getSize(); i++ )
		m_pData[i].TrimRight();
}

void CStringArray::TrimAllLeft()
{
	for( size_t i=0; i<getSize(); i++ )
		m_pData[i].TrimLeft();
}

void CStringArray::TrimAll()
{
	for( size_t i=0; i<getSize(); i++ )
		m_pData[i].Trim();
}

void CStringArray::UppercaseAll()
{
	for( size_t i=0; i<getSize(); i++ )
		m_pData[i].ToUpper();
}

void CStringArray::LowercaseAll()
{
	for( size_t i=0; i<getSize(); i++ )
		m_pData[i].ToLower();
}

bool CStringArray::Load( CFile *file )
{
	resize(0);
	CString str;
	if( str.Load( file ) ){
		Explode( str.Str(), _T('\n') );
		for( size_t i=0; i<getSize(); i++ )
			m_pData[i].TrimEndlChars();
		return true;
	}else
		return false;
}	

CString* CStringArray::InsertAt( size_t pos, LPCVTCHAR str )
{
	size_t oldSize = getSize();
	if( pos>=oldSize ){
		// Inserting after end of array (probably way after, 
		// so we have to resize it using pos and clean the tail).		
		resize( pos+1, true );
		for( size_t i=oldSize; i<pos; i++ )
			m_pData[i].Empty();	
	}else{
		// Inserting between existing elements, have to 
		// move all elements after pos 1 position down.
		resize( oldSize+1, true );
		if( oldSize>0 )
			for( size_t i=getSize()-1, j=i-1; i>pos; i--, j-- ){
				VTASSERT( j>=pos );
				m_pData[j].Swap( &m_pData[i] ); 
			}
	}
	VTASSERT( getSize()>pos );
	m_pData[ pos ].CopyFrom( str );
	return &m_pData[ pos ];
}


void CStringArray::RemoveFrom( size_t pos )
{
	if( pos<getSize() ){ 
		// pos>=0 and pos<getSize() => getSize()>0
		size_t newSize = getSize()-1;
		for( size_t i=pos, j=pos+1; i<newSize; i++, j++ )
			m_pData[i].Swap( &m_pData[ j ] );
		resize( newSize, true ); 
	}else
		throw LCT("String array index is out of bounds.");
}

void CStringArray::Sort( bool reverse, bool caseSensitive )
{
	// !!! Use a better sorting method here ;-)
	if( !getSize() )
		return;
	size_t imaxm1 = getSize()-1;
	for( size_t i=0; i<imaxm1; i++ )
		for( size_t j=i+1; j<getSize(); j++ ){
			int compare = m_pData[i].Compare( m_pData[j], caseSensitive );			
			if( !compare )
				continue;
			if( !reverse ){
				if( compare>0 ) // upper is greater than lower
					m_pData[i].Swap( &(m_pData[j]) );
			}else{
				if( compare<0 ) // upper is smaller than lower
					m_pData[i].Swap( &(m_pData[j]) );
			}
		}
}


