//-----------------------------------------------------------------------------
//
// Yet one more C++ string class, which is mostly based on array template.
//
// - Fully portable.
// - Uses wchar_t when it's available, otherwise uses char.
// - Has conversion from/to ASCII, 8-bit, system locale.
// - Has conversion from/to other string classes.
// - Functionality related to VT framework.
// - Stream-like operators.
//
// Also includes string array class.
//
//-----------------------------------------------------------------------------


/*
	VT Framework License (BSL 1.0)

	Copyright (C) 2006 Sergey A. Galin 
	E-mail: sergey.galin@gmail.com

	Permission is hereby granted, free of charge, to any person or organization
	obtaining a copy of the software and accompanying documentation covered by
	this license (the "Software") to use, reproduce, display, distribute,
	execute, and transmit the Software, and to prepare derivative works of the
	Software, and to permit third-parties to whom the Software is furnished to
	do so, all subject to the following:

	The copyright notices in the Software and this entire statement, including
	the above license grant, this restriction and the following disclaimer,
	must be included in all copies of the Software, in whole or in part, and
	all derivative works of the Software, unless such copies or derivative
	works are solely in the form of machine-executable object code generated by
	a source language processor.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
	SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
	FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
	ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
	DEALINGS IN THE SOFTWARE.
*/

#ifndef VT_STRING_H
#define VT_STRING_H

#include "vt_vectors.hpp"
#include "vt_file.hpp"

namespace VT {

	class CString;
	class CStringArray;
	class CFile;

	//
	// This class is something like Pascal ANSI strings, i.e. it contains length
	// information field (m_uDataSize is always length of string +1) 
	// and on the other hand it keeps string buffer 0-terminated,
	// so it can be easily converted to C string.
	// On the other hand, there's no shared m_pData with reference counting like in 
	// ANSI strings.
	//
	// Note that address of the buffer is never NULL, which saves us a lot of
	// checks. Functions which receive LPCVTCHAR parameters, however, all check
	// if input parameter is NULL (which is equal to passing an empty C string).
	//
	class CString: protected CVector<VTCHAR>
	{
	public:
		virtual LPCVTCHAR GetClassName() const { return _T("CString"); }
		static const size_t kMinAllocSize = 32;
		static const size_t kMinGrowSize = 32;
		virtual size_t GetMinAllocSize() const { return kMinAllocSize; }
		virtual size_t GetGrowSize() const { return kMinGrowSize; } 

		CString(): CVector<VTCHAR>(1, kMinAllocSize){ Empty(); };
		CString(LPCVTCHAR str): CVector<VTCHAR>(VTSTRLEN(str)+1, kMinAllocSize){ CopyFrom(str); };
		CString(double x): CVector<VTCHAR>(1, kMinAllocSize){ Empty(); ConcatDouble(x); };
		CString(long x): CVector<VTCHAR>(1, kMinAllocSize){ Empty(); ConcatLong(x); };
		CString(const CString& str): CVector<VTCHAR>(str.Length()+1, kMinAllocSize){ CopyFrom(str); };

		/* ~CString(){ printf("STRING_DESTRUCTOR [%ld]\n", long(this)); } */

		// Often used as str.Empty()<<something<<else;
		CString& Empty(){ resize(1, false); FixZero(); return *this; } 
		size_t Length() const { return m_uDataSize-1; } 
		bool IsEmpty() const { return Length()==0; }
		
		bool TruncateBy0(); // Trim string by zero character, returns true if trimmed
		bool Fix(){ return TruncateBy0(); }
		bool Truncate( size_t newLength );
		void FixZero(){ m_pData[m_uDataSize-1] = VTCHAR(0); }
		
		void TrimLeft();
		void TrimRight();
		void Trim(){ TrimRight(); TrimLeft(); }
		void TrimEndlChars();
		CString& ToLower();
		CString& ToUpper();

		LPCVTCHAR Str() const {	return m_pData; }
		operator LPCVTCHAR() const { return m_pData; }

		CString& CopyFrom(LPCVTCHAR str);
		CString& CopyFrom(LPCVTCHAR str, size_t nChars);
		VTINLINE CString& CopyFrom(const CString& str); // Faster, because doesn't have to use strlen()
		CString& operator=( LPCVTCHAR str ){ CopyFrom(str); return *this; }
		CString& operator=( CString& str ){ CopyFrom(str); return *this; }	

		// Swap contents of strings (fast operation which can be used for sorting and etc.)
		VTINLINE void Swap( CString *str ){ CArray<VTCHAR>::Swap( (CArray<VTCHAR>*)str ); }

		// All functions return reference to self
		CString& ConcatStr( LPCVTCHAR str );
		CString& ConcatStr( LPCVTCHAR str, size_t nChars );
		CString& ConcatVTStr( const CString& str );
		CString& ConcatVTStr( const CString& str, size_t nChars );
		CString& ConcatEndl(){ ConcatStr(VTENDLT); }
		CString& ConcatLong( long x, LPCVTCHAR format=_T("%ld") );
		CString& ConcatDouble( double x, LPCVTCHAR format=_T("%lf") );
		CString& ConcatChar( VTCHAR ch );
		CString& ConcatHex( const BYTE *ptr, size_t nBytes, size_t nBytesPerLine=0 );
		CString& operator<<( LPCVTCHAR str ){ return ConcatStr(str); }
		CString& operator<<( long x ){ return ConcatLong(x);  }
		// CString& operator<<( double x ){ return ConcatDouble(x);  }	

		CString& operator=( long x ){ Empty(); return ConcatLong(x); }
		//CString& operator=( double x ){ Empty(); return ConcatDouble(x); }

		int CompareCaseSensitive(LPCVTCHAR str, int maxChars=-1) const;
		int CompareCaseInsensitive(LPCVTCHAR str, int maxChars=-1) const;
		inline int Compare(LPCVTCHAR str, bool bCaseSensitive, int maxChars=-1) const;
		
		int operator==(LPCVTCHAR str) const { return CompareCaseSensitive(str)==0; }
		int operator<(LPCVTCHAR str) const { return CompareCaseSensitive(str)>0; }
		int operator>(LPCVTCHAR str) const { return CompareCaseSensitive(str)<0; }
		int operator<=(LPCVTCHAR str) const { return CompareCaseSensitive(str)>=0; }
		int operator>=(LPCVTCHAR str) const { return CompareCaseSensitive(str)<=0; }

		VTCHAR FirstChar() const { return (m_uDataSize>1)? m_pData[0]: 0; }
		VTCHAR LastChar() const { return (m_uDataSize>2)? m_pData[m_uDataSize-2]: 0; }
		VTCHAR FirstNonspaceChar() const;
		VTCHAR LastNonspaceChar() const;

		// Copy fragment from src to character 'delimiter' or to end of the string.
		// Delimiter is not copied. 
		// noDelimPfx allows to ignore delimiters which follow the symbol. 
		// (noDelimPfx|delimiter) sequences are replaced with delimiter on output.
		// Returns pointer set either to zero or to the character after delimiter.
		LPCVTCHAR ExtractUntil( LPCVTCHAR src, VTCHAR delimiter, VTCHAR noDelimPfx=0 );

		LPVTCHAR getDataPtr(){ return m_pData; }
		// LPCVTCHAR getRODataPtr() const { return m_pData; } - use Str()

		// This function minds zero-termination of strings.
		// To keep behaviour of CArray's resize, it always sets m_uDataSize to ns,
		// so length of the string is not correct after resizing.
		// So remember to update string contents or call Fix() or Empty() after resize().
		virtual void resize(size_t newDataSize, bool saveData=false, size_t newAllocSize=0); 
		
		void fill(VTCHAR f){ CArray<VTCHAR>::fill(f); FixZero(); }
		
		// Join lines from 'arr' via 'separator'. If separatorAfter==false, then
		// the functions work as a reverse of CStringArray::Implode() (separator
		// is inserted only between lines). If separatorAfter==true, then separator
		// is also added at end of the combined string.
		CString& Implode( const CStringArray& arr, VTCHAR separator, bool separatorAfter=false ); 
		CString& ImplodeS( const CStringArray& arr, const CString& separator, bool separatorAfter=false ); 
		CString& Implode( const LPCVTCHAR *nullTermList, const CString& separator, bool separatorAfter=false ); 
		
		void ConcatAll( const CStringArray& arr ); // Join lines from 'arr' (without separators)

		long ToLong(bool *conversionOK=NULL, long defValue=0 ) const;
		double ToDouble(bool *conversionOK=NULL, long defValue=0 ) const;

		// Read text file from current file position till EOF or maxChunkSizeChars into a string
		virtual bool Load( CFile *file, long maxChunkSizeChars = -1 ); 
		// Write string into file, from current file position
		virtual bool Save( CFile *file ) const;

		// Read text file (via CDiskFile) from current file position till EOF into a string
		virtual bool Load( LPCVTCHAR fileName ); 
		// Write string into a file (via CDiskFile)
		virtual bool Save( LPCVTCHAR fileName ) const; 
	};	

	// Array of strings adds some additional functionality to CArray<CString>.
	class CStringArray: public CArray<CString>
	{
	public:
		CStringArray( size_t size = 0, size_t allocSize = 16 );

		virtual int Find(LPCVTCHAR str, bool bCaseSensitive, size_t startIndex=0) const; // Returns -1 if not found
		
		// Split 'str' to string array by separators.
		// The reverse function (Implode) is in CString.
		void Explode(LPCVTCHAR str, VTCHAR separator, int maxItems=0); 	
		
		// Get sum of lengths of all strings
		size_t GetTotalLength() const;	
		
		// Many m_pData type files ignore space characters at beginning and
		// end of lines. So this is it.
		void TrimAllRight();
		void TrimAllLeft();
		void TrimAll();
		void UppercaseAll();
		void LowercaseAll();

		// Insert string at position, returns ref to string.
		// If pos > getSize(), empty strings are added at [getSize()..pos)
		CString* InsertAt( size_t pos, const LPCVTCHAR str ); 

		CString* Insert( const LPCVTCHAR str ){ return InsertAt( getSize(), str ); }

		virtual size_t AddUnique( LPCVTCHAR str, bool bCaseSensitive = true );

		// Remove string from position pos. If pos is out of range,
		// an exception is thrown.
		void RemoveFrom( size_t pos ); 
		
		// Loads file as a string, splits it by '\n' character, removes
		// '\r' characters at end of lines. 
		virtual bool Load( CFile *file );

		virtual void Sort( bool reverse = false, bool caseSensitive = true );
	};





	//////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////
	//
	// Implementation of template & inline functions
	//

	/*template<> VTINLINE void Swap( CString *a, CString *b )
	{
		a->Swap( b );
	}*/


	CString& CString::CopyFrom(const CString& str)
	{
		rawCopyFromSameType( str );
		return *this;
	}

	int CString::Compare(LPCVTCHAR str, bool bCaseSensitive, int maxChars) const
	{
		return bCaseSensitive? 
			CompareCaseSensitive(str, maxChars): 
			CompareCaseInsensitive(str, maxChars);
	}

} // namespace

#endif




