//-----------------------------------------------------------------------------
//
// CVector<T> expands CArray<T> with functionality related to 
// numbers, like math operations, resampling vectors and etc.
//
//-----------------------------------------------------------------------------

/*
	VT Framework License (BSL 1.0)

	Copyright (C) 2006 Sergey A. Galin 
	E-mail: sergey.galin@gmail.com

	Permission is hereby granted, free of charge, to any person or organization
	obtaining a copy of the software and accompanying documentation covered by
	this license (the "Software") to use, reproduce, display, distribute,
	execute, and transmit the Software, and to prepare derivative works of the
	Software, and to permit third-parties to whom the Software is furnished to
	do so, all subject to the following:

	The copyright notices in the Software and this entire statement, including
	the above license grant, this restriction and the following disclaimer,
	must be included in all copies of the Software, in whole or in part, and
	all derivative works of the Software, unless such copies or derivative
	works are solely in the form of machine-executable object code generated by
	a source language processor.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
	SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
	FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
	ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
	DEALINGS IN THE SOFTWARE.
*/


#ifndef VT_VECTORS_H
#define VT_VECTORS_H

#include <stdio.h>
#include "vt_arrays.hpp"

namespace VT {

	// This class expands Array class by adding functionality which
	// is specific to scalar types (numbers and characters).
	// The memory is never reallocated unless vector is resized over allocSize.
	// This default behaviour should never be changed, because the objects are 
	// used in sound processing routines which should not do any OS calls.
	// For example, VST's processXXX under Mac OS X VST's can't do that.
	// shrink() can be used to reduce memory size occupied by the vector.
	template<class T> class CVector: public CArray<T>
	{
	public:
		static const size_t kDefaultAllocSize = 32;

		virtual LPCVTCHAR GetClassName() const { return _T("CVector"); }

		// Note that size can be 0.
		CVector<T>(size_t size=kDefaultAllocSize, size_t alloc=0): 
		  CArray<T>(size, alloc)
		{}

		// Symmetically copy left side of the array to the right side, 
		// or vice versa, if useright==true.
		void symmet(bool useright=false); 
		
		VTINLINE T get(size_t x) const { return this->m_pData[x]; } // Copy item from array
			
		template<class U> VTINLINE void multiply_points(const CVector<U>& src);
		template<class U> VTINLINE void add_points(const CVector<U>& src);
		
		// void zero(){ memset( m_pData, 0, sizeof(T)*getSize() ); } // fill(T(0)); 
		VTINLINE void zero(){ this->ZeroMemoryBytes(); } 
		
		T getMax() const;
		T getMin() const;
		virtual int Find(const T& what, int startIdx=0) const; // Find an item in array using T::operator==(); returns index; -1 if not found
		virtual size_t Count(const T& what) const;

		bool operator==( const CArray<T>& otherarray ) const; 

		// Resample (change length) of vector 'src' by 'k', k>1.
		// Length of resulting vector is src.getLength()/k.
		// Algotithm for calculating elements of resulting vector depends on the function.	
		// These functions are designed for resizing stuff like frequency response graphs.
		template<class U> VTINLINE void shorten_averaging(const CVector<U>& src, DWORD k);
		template<class U> VTINLINE void shorten_maxmodulus(const CVector<U>& src, DWORD k);

		// A couple of debugging functions for dumping
		// vector into a CSV file (one array item in a row).
		// Set bIndex to true to have array indexes in the first column.
		void write_txt(FILE *f, bool bIndex=false) const; 
		void write_txt(LPCVTCHAR fn, bool bIndex=false) const; 

		// Inserts new element into array, if it was not there yet. Returns index 
		// of the value in the array (either newly added or old).
		virtual size_t AddUnique( const T& newValue );
	};

	typedef CVector<float> CVTSPVector; // "Single precision" vector
	typedef CVector<double> CVTDPVector; // "Double precision" vector

	// Scalar multiplication
	template<class T> T multiply(const CVector<T>& a, const CVector<T>& b);

	template<class T> VTINLINE T operator*(const CVector<T>& a, const CVector<T>& b){
		return multiply( a, b );
	}






	//////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////
	//
	// Implementation of the template class functions
	//



	template<class T> T CVector<T>::getMax() const
	{
		T ret=this->m_pData[0];
		for(size_t i=1; i<this->getSize(); i++)
			if( this->m_pData[i] > ret ) 
				ret = this->m_pData[i];
		return ret;
	}

	template<class T> T CVector<T>::getMin() const
	{
		T ret=this->m_pData[0];
		for(size_t i=1; i<this->getSize(); i++)
			if( this->m_pData[i] < ret )
				ret = this->m_pData[i];
		return ret;
	}

	template<class T> int CVector<T>::Find(const T& what, int startIdx) const
	{
		for(int i=startIdx; i<int(this->m_uDataSize); i++)
			if( this->m_pData[i]==what )
				return i;
		return -1;
	}

	template<class T> size_t CVector<T>::Count(const T& what) const
	{
		size_t ret = 0;
		for( size_t i=0; i<this->m_uDataSize; i++ )
			if( this->m_pData[i]==what )
				ret++;
		return ret;
	}

	template<class T> bool CVector<T>::operator==( const CArray<T>& otherarray ) const
	{
		if( otherarray.getSize()!=this->m_uDataSize )
			return false;
		for( size_t i=0; i<this->m_uDataSize; i++ )
			if( !(this->m_pData[i]==otherarray.m_pData[i]) )
				return false;
		return true;
	}


	// Non-class function (scalar multiplication)
	template<class T> T multiply(const CVector<T>& a, const CVector<T>& b)
	{
		size_t minsz = Min( a.getSize(), b.getSize() );
		T sum = 0;
		for(size_t i=0; i<minsz; i++)
			sum+=a.get(i)*b.get(i);
		return sum;
	}

	template<class T> template<class U> void CVector<T>::multiply_points(const CVector<U>& src)
	{
		size_t minsz = Min( this->getSize(), src.getSize() );	
		for(size_t i=0; i<minsz; i++)
			this->m_pData[i] *= T(src.m_pData[i]);
	}

	template<class T> template<class U> void CVector<T>::add_points(const CVector<U>& src)
	{
		size_t minsz = Min( this->getSize(), src.getSize() );	
		for(size_t i=0; i<minsz; i++)
			this->m_pData[i] += T(src.m_pData[i]);
	}

	template<class T> void CVector<T>::symmet(bool useright)
	{
		register size_t i, j=this->m_uDataSize, half = this->m_uDataSize/2;
		if( !useright )
			for(i=0; i<half; i++)
				this->m_pData[--j] = this->m_pData[i];
		else
			for(i=0; i<half; i++)
				this->m_pData[i] = this->m_pData[--j];
	}

	template<class T> void CVector<T>::write_txt(FILE *f, bool bIndex) const
	{
		if( !bIndex )
			for(size_t i=0; i<this->getSize(); i++)
				fprintf(f, "%.20le\n", (double)this->get(i));
		else
			for(size_t i=0; i<this->getSize(); i++)
				fprintf(f, "%u,%.20le\n", unsigned(i), (double)this->get(i));
	}

	template<class T> void CVector<T>::write_txt(LPCVTCHAR fn, bool bIndex) const
	{
		FILE *fo=fopen(fn, _T("wt"));
		if(!fo){
			VTDEBUGMSGT("File open error.");
			throw LCT("File open error.");
		}
		this->write_txt(fo, bIndex);
		fclose(fo);
	}

	template<class T> template<class U> void CVector<T>::shorten_averaging(const CVector<U>& src, DWORD k)
	{
		size_t i, j;
		resize(src.getSize()/k);
		for(i=0, j=0; i<this->getSize(); i++){
			// Sum the chunk
			register U sum=src.get(j++);
			for( register DWORD l=1; l<k; l++)
				sum+=src.get(j++);
			this->set(i, T(sum/U(k)));
		}
	}

	template<class T> template<class U> void CVector<T>::shorten_maxmodulus(const CVector<U>& src, DWORD k)
	{
		size_t i, j;
		resize(src.getSize()/k);
		for(i=0, j=0; i<this->getSize(); i++){
			register U max=src.get(j++), maxmod=Abs(max), x, modx;
			// Find max value in the chunk
			for( register DWORD l=1; l<k; l++){
				x = src.get(j++);
				modx = Abs(x);
				if(modx>maxmod){
					maxmod = modx;
					max = x;
				}
			}
			this->set(i, T(max));
		}
	}

	template<class T> size_t CVector<T>::AddUnique( const T& newValue )
	{
		int idx = this->Find( newValue );
		if( idx>=0 )
			return size_t(idx);
		else{
			size_t newidx = this->getSize();
			this->resize( newidx+1, true );
			this->m_pData[ newidx ] = newValue;
			return newidx;
		}
	}

} // namespace

#endif


