//-----------------------------------------------------------------------------
//
// Zlib wrapper.
//
//-----------------------------------------------------------------------------

/*
	VT Framework License (BSL 1.0)

	Copyright (C) 2006 Sergey A. Galin 
	E-mail: sergey.galin@gmail.com

	Permission is hereby granted, free of charge, to any person or organization
	obtaining a copy of the software and accompanying documentation covered by
	this license (the "Software") to use, reproduce, display, distribute,
	execute, and transmit the Software, and to prepare derivative works of the
	Software, and to permit third-parties to whom the Software is furnished to
	do so, all subject to the following:

	The copyright notices in the Software and this entire statement, including
	the above license grant, this restriction and the following disclaimer,
	must be included in all copies of the Software, in whole or in part, and
	all derivative works of the Software, unless such copies or derivative
	works are solely in the form of machine-executable object code generated by
	a source language processor.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
	SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
	FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
	ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
	DEALINGS IN THE SOFTWARE.
*/

#include <stdlib.h>
#include "zlib.h"
#include "vt_zlib.hpp"

//#define VTZDEBUG

#define VTZTLEN DWORD
#define VTZMEMTOLERANCE 32 // Do not re-allocate if it will save less than that bytes
#define VTZ_MINSIZE 17

using namespace VT;

const VTZTLEN VTZSign = 'ZcD!';

void *ZLib::Compress(const void *in, size_t in_size, size_t *out_size, TCompression lev)
{
	#ifdef VTZDEBUG
		printf("=== ZLib::Compress() ===\n");
	#endif

	//const int chunk_size = 1024*128;
	*out_size = 0;

	if(!in_size || !in)
		return NULL;

	int level,		// compression level, for zlib
		ret,		// values returned by zlib 
		m_alloc_size,	// size of allocated out buffer
		have, m_out_size, m_avail_out_size;
	z_stream strm;
	char *out;
	
	switch(lev){
		case kBest: level = Z_BEST_COMPRESSION; break;
		case kFast: level = Z_BEST_SPEED; break;
		default: level = Z_DEFAULT_COMPRESSION; 
	}
    
    strm.zalloc = Z_NULL;
    strm.zfree = Z_NULL;
    strm.opaque = Z_NULL;
    ret = deflateInit(&strm, level);    
	if (ret != Z_OK){
		#if defined(VT_EXCEPTIONS)
			throw(_T("Failed to initialize decompressor."));
		#endif
		return NULL;
	}
	strm.avail_in = (int)in_size;
    strm.next_in = (Byte*)in;

	m_avail_out_size = (m_alloc_size = ((int)in_size*10)/9+sizeof(VTZTLEN)*2+16);
	m_out_size = 0;
	out = (char*)malloc(m_alloc_size);		
	if(!out){
		deflateEnd(&strm);
		#if defined(VT_EXCEPTIONS)
		throw(_T("Compressor memory allocation error #1."));
		#endif
		return NULL;	
	}

	// Recording signature & uncompressed data length into out
	*((VTZTLEN*)out) = VTZSign;
	*(((VTZTLEN*)out)+1) = VTZTLEN(in_size);
	m_avail_out_size -= sizeof(VTZTLEN)*2;
	m_out_size += sizeof(VTZTLEN)*2;

	// Doing Zlib compression cycle
	do{	
		strm.avail_out = m_avail_out_size;
		strm.next_out = (Byte*) (out + m_out_size);

		#ifdef VTZDEBUG
			printf(
				"*** Next compression stage ***\n"
				"avail_out: %d\n"
				"m_out_size: %d\n"
				"m_avail_out_size: %d\n", strm.avail_out, m_out_size, m_avail_out_size);
		#endif

		ret = deflate(&strm, Z_FINISH);  

		// Counting output
		have = m_avail_out_size - strm.avail_out; // how many bytes we have in out
		m_avail_out_size = strm.avail_out;

		#ifdef VTZDEBUG
			printf("Survived deflate().\n");
		#endif
		if( strm.avail_out==0 ){
			// Increase size of the buffer
			int chunk_size = (strm.avail_in*5)/4+1024*32; // +32K

			#ifdef VTZDEBUG
				printf(
					"ADDING MEMORY!\n"
					"Available input: %d\n"
					"Available output: %d\n"
					"Chunk to add: %d\n", strm.avail_in, strm.avail_out, chunk_size);
			#endif

			m_alloc_size += chunk_size;
			m_avail_out_size += chunk_size;
			char *out2 = (char*)realloc(out, m_alloc_size);
			if( out2 == NULL ){
				deflateEnd(&strm);
				free(out);
				#if defined(VT_EXCEPTIONS)
					throw(_T("Compressor memory allocation error #2."));
				#endif
				return NULL;
			}
			out = out2;
		}
		m_out_size += have;

	}while( strm.avail_out == 0 );// ( strm.avail_in!=0 );
    deflateEnd(&strm);

	*out_size = m_out_size;

	if(!m_out_size){
		free(out);
		return NULL;
	}

	if(m_alloc_size-(*out_size)<=VTZMEMTOLERANCE)
		return out;

	char* out2 = (char*)realloc(out, *out_size);
	if( out2 == NULL ){
		delete out;
		*out_size = 0;
		#if defined(VT_EXCEPTIONS)
			throw(_T("Compressor memory allocation error #3."));
		#endif
		return NULL;
	}
	return out2;
}

void *ZLib::Decompress(const void *in, size_t in_size, size_t *out_size)
{
	#ifdef VTZDEBUG
		printf("=== ZLib::Decompress() ===\n");
	#endif

	const int chunk_size = 1024*128;
	*out_size = 0;

	if(!in_size || !in)
		return NULL;

	// If data appears to be not compressed, return it as is
	bool bSendAsIs = false;
	if(in_size<sizeof(VTZTLEN)*2)
		bSendAsIs = true;
	else
		if( *((VTZTLEN*)in)!=VTZSign )
			bSendAsIs = true;
	if( bSendAsIs ){
		#ifdef VTZDEBUG
			printf("Returning data AS IS!\n");
		#endif
		char *copy = (char*)malloc(in_size);
		if(!copy)
			return NULL;
		memcpy(copy, in, in_size);
		*out_size = in_size;
		return copy;
	}

	if(*(((VTZTLEN*)in)+1)<=0){
		#if defined(VT_EXCEPTIONS)
			throw(_T("Compressed data header is invalid."));
		#endif
		return NULL;
	}

	if( in_size<VTZ_MINSIZE ){
		#if defined(VT_EXCEPTIONS)
			throw(_T("Data block is to small to contain valid compressed data."));
		#endif
		return NULL;	
	}

    int ret;
	char *out;
	int m_out_size, m_avail_out_size, m_alloc_size, have;
    z_stream strm;
    strm.zalloc = Z_NULL;
    strm.zfree = Z_NULL;
    strm.opaque = Z_NULL;
    strm.avail_in = 0;
    strm.next_in = Z_NULL;
    ret = inflateInit(&strm);
	if (ret != Z_OK){
		#if defined(VT_EXCEPTIONS)
			throw(_T("Failed to initialize decompressor."));
		#endif
		return NULL;
	}
	
	strm.avail_in = (int)in_size - sizeof(VTZTLEN)*2;
	strm.next_in = (Byte*)((char*)in + sizeof(VTZTLEN)*2);

	//m_avail_out_size = (m_alloc_size = (int)in_size * 3);

	m_avail_out_size = (m_alloc_size = *(((VTZTLEN*)in)+1)+4);

	out = (char*)malloc( m_alloc_size );
	if( out == NULL ){
		inflateEnd(&strm);
		#if defined(VT_EXCEPTIONS)
			throw(_T("Decompressor memory allocation error #1."));
		#endif
		return NULL;
	}

	m_out_size = 0;

	// Doing Zlib decompression cycle.
	do{
		strm.avail_out = m_avail_out_size;
		strm.next_out = (Byte*) (out + m_out_size);

		#ifdef VTZDEBUG
			printf(
				"*** Next decompression stage ***\n"
				"avail_out: %d\n"
				"m_out_size: %d\n"
				"m_avail_out_size: %d\n", strm.avail_out, m_out_size, m_avail_out_size);
		#endif

		ret = inflate(&strm, Z_NO_FLUSH);

		// Decompressor is bo-bo
		if( ret == Z_STREAM_ERROR || ret ==  Z_NEED_DICT || ret == Z_DATA_ERROR || ret == Z_MEM_ERROR){
			inflateEnd(&strm);
			free(out);
			#if defined(VT_EXCEPTIONS)
				throw(_T("Decompressor: corrupted data stream."));
			#endif
			return NULL;
		}

		have = m_avail_out_size - strm.avail_out; // how many bytes we have in out
		m_avail_out_size = strm.avail_out;

		if( strm.avail_out==0 ){
			#ifdef VTZDEBUG
				printf("Adding memory for decompression!\n");
			#endif

			// Reallocate out buffer
			m_alloc_size += chunk_size;
			m_avail_out_size += chunk_size;
			char *out2 = (char*)realloc(out, m_alloc_size);
			if( out2 == NULL ){
				inflateEnd(&strm);
				free(out);
				#if defined(VT_EXCEPTIONS)
					throw(_T("Decompressor memory allocation error #3."));
				#endif
				return NULL;
			}
			out = out2;
		}
		m_out_size += have;

		#ifdef VTZDEBUG
			printf("==> m_out_size = %d\n", m_out_size);
		#endif
	} while (strm.avail_out == 0); // In zlib example: (strm.avail_out == 0)

	inflateEnd(&strm);
	
	if(ret != Z_STREAM_END /*&& ret!=Z_OK*/){
		free(out);
		#if defined(VT_EXCEPTIONS)
			throw(_T("Broken compressed file."));
		#endif
		return NULL;
	}

	*out_size = m_out_size;
	
	if(!m_out_size){
		free(out);
		return NULL;
	}

	if(m_alloc_size-(*out_size)<=VTZMEMTOLERANCE)
		return out;

	char* out2 = (char*)realloc(out, *out_size);
	if( out2 == NULL ){
		delete out;
		*out_size = 0;
		#if defined(VT_EXCEPTIONS)
			throw(_T("Memory allocation error."));
		#endif
		return NULL;
	}
	return out2;
}


bool ZLib::IsCompressedData(const void *in, size_t in_size)
{
	if( in_size<VTZ_MINSIZE )
		return false;
	if( *((VTZTLEN*)in)!=VTZSign ||
		*(((VTZTLEN*)in)+1)<=0 )
		return false;
	return true;
}


